unit ForevalZ_Command;
{******************************************************************************}
{                                                                              }
{                                 ver 1.3.5.229                                }
{----------------------------------------------------------------------------- }
{  :    ,                            }
{       :                    }
{  FWAIT.                                                           }
{  : Pentium Pro                                                 }
{      FAST ZDIV                             }
{ -----------------------------------------------------------------------------}
{ Command module: contains executable functions,                               }
{ Only use processors with AutoSync : missing FWAIT.                           }
{ min proc.: Pentium Pro                                                       }
{ in functions is used fast division FAST ZDIV                                 }
{------------------------------------------------------------------------------}
{                  SOREL   (C)CopyRight 2000. Russia, S.-Petersburg.           }
{******************************************************************************}
{
                  Random numbers generator  algorithm( taus113 (Pierre L'Ecuyer),mt19937 (Mersenne Twister), xor4096) ,
                  spherical harmonic,
                  Gamma function,
                  Bessel functions
                  Legendre polynoms

              (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/
}



// {$RangeChecks On}
// {$OverFlowChecks On}
 //{$ZEROBASEDSTRINGS Off}

{$R-}
{$Q-}

{$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}


interface

uses
  math, ForevalZ_SpecFunc, Windows, SysUtils;



{spec. func. }


 const  Max2Power: integer = 16383;  //2^Max2Power
 const  Max2PowerN: integer = -16383;  //2^Max2Power
 const  Power2Overflow: integer =17000; //Any > Max2Power
 const  C1: double = 1;
 const  C05: double = 0.5;
 const  C2: integer = 2;
 const  C3: double = 3;
 const  C025: double = 0.25;





type
TArrayD = array of Double;
type
TArrayE = array of Extended;
type
TArrayI = array of Integer;
type
PArrayD = ^TArrayD;
type
PArrayE = ^TArrayE;
type
PArrayI = ^TArrayI;







const _Double = 2;
const _Extended = 3;
const _Pointer = 5;
const _Integer = 4;

const NumberFact = 1754;
const NumberFact2 = 1604;

//const S_c05: double = 0.5;
//const S_c2: double = 2;
//const S_c025: double = 0.25;
//const S_cs2: double = -2;
//const E_CPId2: extended  = 3.1415926535897932384626433832795/2;
//const E_2PI: extended  = 3.1415926535897932384626433832795*2;
//const c3: double = 3;
//const c6: double = 6;
//const c4: double = 4;
//const c1: double = 1;

const
FCW1: Word = $1f32;
FCW2: Word = $1332;




    procedure Z_MUL;
    procedure Z_RZDIV;
    procedure Z_DIV;
    procedure Z_RZDIV_FAST;
    procedure Z_DIV_FAST;
    procedure Z_SIN;
    procedure Z_COS;
    procedure Z_TAN;
    procedure Z_COSEC;
    procedure Z_SEC;
    procedure FPWR;
    procedure ZPWR;
    procedure Z_ReCxFPWR;
    procedure Z_CxReFPWR;
    procedure CreateFactMass;
    procedure InitMEM(Addr: Cardinal);
    procedure R_FACT;
    procedure R_FACT2;
    procedure Z_LN;
    procedure Z_SQRT;
    procedure Z_SQR;
    procedure Z_CxCxLOG;
    procedure R_LogN;
    procedure Z_ReCxLOG;
    procedure Z_CxReLOG;
    procedure Z_COTAN;
    procedure R_ARCSIN;
    procedure R_ARCCOS;
    procedure R_ARCCOTAN;
    procedure Z_ARCCOTAN;
    procedure Z_ARCSIN;
    procedure Z_ARCCOS;
    procedure Z_ARCTAN;
    procedure R_SINH;
    procedure R_COSH;
    procedure R_TANH;
    procedure R_COTANH;
    procedure R_ARCSINH;
    procedure R_ARCCOSH;
    procedure R_ARCTANH;
    procedure R_ARCCOTANH;
    procedure R_ARCCOSECH;
    procedure R_ARCSECH;
    procedure R_COSECH;
    procedure R_SECH;
    procedure R_ARCCOSEC;
    procedure R_ARCSEC;
    procedure Z_ARCSEC;
    procedure Z_ARCCOSEC;
    procedure Z_SINH;
    procedure Z_COSH;
    procedure Z_COSECH;
    procedure Z_SECH;
    procedure Z_TANH;
    procedure Z_COTANH;
    procedure Z_ARCSINH;
    procedure Z_ARCCOSH;
    procedure Z_ARCTANH;
    procedure Z_ARCCOTANH;
    procedure Z_ARCCOSECH;
    procedure Z_ARCSECH;
    procedure Z_EXP;
    procedure R_EXP;
    procedure Z_CxIROOT;
    procedure R_IROOT;
    procedure R_FROOT;
    procedure Z_IPWR3;
    procedure Z_IPWR4;
    procedure Z_IPWRN1;
    procedure Z_IPWRN2;
    procedure Z_IPWRN3;
    procedure Z_IPWRN4;
    procedure Z_IPWRN5;
    procedure Z_IPWRN6;
    procedure R_IPWRSGN;
    procedure Z_IPWRSGN; {189}
    procedure RI_NORM;
    procedure RD_NORM;
    procedure RE_NORM;
    procedure RI_SUM;
    procedure RD_SUM;
    procedure RE_SUM;
    procedure RI_AVR;
    procedure RD_AVR;
    procedure RE_AVR;
    procedure RI_SUMQ;
    procedure RD_SUMQ;
    procedure RE_SUMQ;
    procedure R_LEN;
    procedure RI_MAX;
    procedure RD_MAX;
    procedure RE_MAX;
    procedure RI_MIN;
    procedure RD_MIN;
    procedure RE_MIN;
    procedure RI_PROD;
    procedure RD_PROD;
    procedure RE_PROD;
    procedure R_SGN;
    procedure R_F2PWR;
    procedure R_MAX(adr,len: Cardinal);  stdcall;
    procedure R_MIN(adr,len: Cardinal);  stdcall;
    procedure R_AVR(adr,len: Cardinal);  stdcall;
    procedure R_SUM(adr,len: Cardinal);  stdcall;
    procedure R_SUMQ(adr,len: Cardinal);  stdcall;
    procedure R_NORM(adr,len: Cardinal);  stdcall;
    procedure R_PROD(adr,len: Cardinal);  stdcall;
    procedure R_DEV(adr,len: Cardinal); stdcall;
    procedure R_DEVS(adr,len: Cardinal); stdcall;
    procedure RI_VEC;
    procedure RD_VEC;
    procedure RE_VEC;
    procedure RI_VECS;
    procedure RD_VECS;
    procedure RE_VECS;
    procedure R_IPWR;
    procedure Z_IPWR; {v177}
    procedure Z_CxIROOTA;
    procedure R_Decart;
    procedure RI_DEV;
    procedure RI_DEVS;
    procedure RD_DEV;
    procedure RD_DEVS;
    procedure RE_DEV;
    procedure RE_DEVS;
    procedure RE_POLY;
    procedure RD_POLY;
    procedure RI_POLY;
    procedure CxE_POLY;
    procedure CxD_POLY;
    procedure CxI_POLY;
    procedure _MAX2;
    procedure _MIN2;
    procedure SetLengthArray(AdrV: Pointer; TypeV: Cardinal; Len: Cardinal);  stdcall;
    procedure R_GCD(adr,len: Cardinal); stdcall;
    procedure V_GCDD(VD: TArrayD); stdcall;
    procedure V_GCDE(VD: TArrayE); stdcall;
    procedure V_GCDI(VD: TArrayI); stdcall;
    procedure R_LCM(adr,len: Cardinal); stdcall;
    procedure V_LCMD(VD: TArrayD); stdcall;
    procedure V_LCME(VD: TArrayE); stdcall;
    procedure V_LCMI(VD: TArrayI); stdcall;
    procedure RD_GCD; assembler;
    procedure RE_GCD; assembler;
    procedure RI_GCD; assembler;
    procedure RD_LCM; assembler;
    procedure RE_LCM; assembler;
    procedure RI_LCM; assembler;
    //procedure _MaskFPUException;  assembler;
    //procedure _IsFPUException; assembler;
    procedure _MaskFPUExceptionA;  assembler;
    procedure _IsFPUExceptionA; assembler;

   // procedure _SafeCall(Adr: Cardinal); {.219}






    function _FACT1(n: integer): extended;//N -> EAX
    function std_rand2(x1,x2:PExtended): extended;  stdcall;
    procedure _Randomize;  stdcall;
    procedure InitRandomGenerator;


var
Factorial,Factorial1,Factorial2,Factorial2F: array of extended;
F_FastSpec: Boolean;


implementation
var
MEM: Cardinal;
CWMEM: Cardinal;




procedure InitMEM(Addr: Cardinal);
begin
  MEM:=Addr;
end;



procedure CreateFactMass;
var
i,j,k1,k2: Integer;
S: extended;
begin                   //  

SetLength(Factorial,NumberFact+1);

S:=1;
for j:=1 to NumberFact do
begin
 S:=S*j;
 Factorial[j]:=S;
end;
 Factorial[0]:=1;


k1:=1; k2:=1;
SetLength(Factorial1,NumberFact2+1);  SetLength(Factorial2,NumberFact2+1);
Factorial1[0]:=1; Factorial2[0]:=1;
 for i := 1 to NumberFact2*2 do
 begin
  if i mod 2 = 0 then  begin Factorial2[k2]:=Factorial2[k2-1]*i; inc(k2);  end
  else
    begin Factorial1[k1]:=Factorial1[k1-1]*i;  inc(k1); end;
 end;

 k1:=1; k2:=1;
 SetLength(Factorial2F,NumberFact2*2+1);   Factorial2F[0]:=1;
 for i := 1 to NumberFact2*2 do
 begin
  if i mod 2 = 0 then  begin Factorial2F[i]:=Factorial2[k2-1]*i; inc(k2);  end
  else
    begin Factorial2F[i]:=Factorial1[k1-1]*i;  inc(k1); end;
 end;

end;



procedure Z_MUL; assembler;
//z1*z2; z1.RE-ST,z1.IM-ST(1),z2.RE-ST(2),z2.IM-ST(3);
asm
 fld  st(0)
 fmul st(0),st(3)
 fxch st(1)
 fmul st(0),st(4)
 fxch st(4)
 fmul st(0),st(2)
 fsubp
 fxch st(2)
 fmulp
 faddp st(2),st(0)
end;


procedure Z_RZDIV_FAST; assembler;
//r/z1; r-ST,z1.RE-ST(1),z1.IM-ST(2);
//FAST ZDIV!!!
asm
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
end;



procedure Z_RZDIV; assembler;
//r/z1; r-ST,z1.RE-ST(1),z1.IM-ST(2);
asm
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fdiv  st(1),st
  fdivp st(2),st
end;



procedure Z_DIV_FAST; assembler;
//z1/z2; z1.RE-ST,z1.IM-ST(1),z2.RE-ST(2),z2.IM-ST(3);
//FAST ZDIV!!!
asm
 fxch  st(2)
 fld   st
 fmul  st,st
 fxch  st(4)
 fld   st
 fmul  st,st
 faddp st(5),st
 fxch  st(4)
 fld1
 fdivr

 fxch  st(4)
 fld   st
 fmul  st,st(3)
 fxch
 fmul  st,st(4)
 fxch  st(3)
 fmul  st,st(2)
 fxch  st(4)
 fmulp st(2),st
 faddp
 fmul  st,st(3)
 fxch  st(2)
 fsubrp
 fmulp st(2),st
end;




procedure Z_DIV; assembler;
//z1/z2; z1.RE-ST,z1.IM-ST(1),z2.RE-ST(2),z2.IM-ST(3);
asm
 fxch  st(2)
 fld   st
 fmul  st,st
 fxch  st(4)
 fld   st
 fmul  st,st
 faddp st(5),st
 fxch  st(4)

 fxch  st(4)
 fld   st
 fmul  st,st(3)
 fxch
 fmul  st,st(4)
 fxch  st(3)
 fmul  st,st(2)
 fxch  st(4)
 fmulp st(2),st
 faddp
 fdiv  st,st(3)
 fxch  st(2)
 fsubrp
 fdivrp st(2),st
end;



procedure  Z_SIN;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm
   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF

        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
 @@exp:
   fld   st
   fld1              //FAST ZDIV
   fdivrp
   //fdivr S_c1
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(3),st
   fmulp

@@end:
end;


procedure  Z_COSEC;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm
   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF

        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   fld1              //FAST ZDIV
   fdivrp
   //fdivr S_c1
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(3),st
   fmulp
   //FAST ZDIV (1/rez)
  fld1
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
@@end:
end;




procedure  Z_SEC;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm

   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   fld1                    //FAST ZDIV
   fdivrp
   //fdivr S_c1
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(2),st
   fchs
   fmulp st(2),st
     //FAST ZDIV (1/rez)
  fld1
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
 @@end:
end;

{
RE(sqrt) = sqrt((sqrt(re^2+im^2)+re)*0.5)
IM(sqrt) = sign(im)*((sqrt(re^2+im^2)-re)*0.5)
}
procedure Z_SQRT;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
asm
 //push   ax
 fxch
 ftst
 fstsw  ax
 sahf
 fmul   st,st
 fxch
 fld    st
 fmul   st,st
 faddp st(2),st
 fxch
 fsqrt
 fld   st
 fsub  st,st(2)
 fmul  c05
 fsqrt
 jnb   @@1
 fchs
@@1:
 fxch  st(2)
 faddp
 fmul  c05
 fsqrt
 //pop   ax

end;




procedure Z_SQR;
const c2: double=2;
//RE-ST(0); IM-ST(1);
asm
 fld   st
 fmul  st,st
 fxch
 fmul  st,st(2)
 fmul  c2
 fxch  st(2)
 fmul  st,st
 fsubp
end;



procedure Z_LN;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
asm
 fld  st
 fxch st(2)
 fld  st
 fxch st(3)
 fpatan
 fxch st(2)
 fmul st,st
 fxch
 fmul st,st
 faddp
 fldln2
 fxch
 fyl2x
 fmul  c05
end;






procedure R_F2PWR;   //2^X
const F2: double =2.0;
var
M: Cardinal;
MX: Extended;
asm
//2^X
//ST(0) <- X




     FLD       ST
     //FRNDINT
     FISTP M
     FILD  M
     FCOMIP   ST(0),ST(1)
     JZ        @@int
     JMP       @@flow

@@int:


       //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxI
        //call Overflow Error:
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxI:  //ipower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2i
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2i:
        FISTP M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
    //**********************

     JMP     @@end


@@flow:

     // X=0 ??
     FLD F2

     FLDZ            {.}
     FCOMP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@NoZero     //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??
     //FXCH
     //FLDZ
     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz
     JB    @@end              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)
     fld1
     fchs
     fsqrt
     JMP     @@end      {.}

@@NoZero:
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP  ST(1)
     }
      //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)

        //fpower >  Max2Power ??
        FXCH ST(1)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2f
        fstp st(0)
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)

    //**********************

@@end:

end;





procedure FPWR;
const
MaxInt: double= 2147483647;
var
M: Cardinal;
MX: Extended;
asm
//X^Y
//ST(0) <- X
//ST(1) <- Y



     FXCH

     FLD       ST
     FABS
     FLD      MaxInt
     fcompp
     fstsw   ax
     sahf
     jb      @@flow


     FIST      M
     FICOM     M
     {fld      st
     frndint
     fcomp}


     FNSTSW    AX
     SAHF;
     JZ        @@int
     JMP       @@flow

@@int:

     FXCH
     FSTP     ST(1)
     //FISTP    M


     MOV      EAX, M



     PUSH     ECX
     PUSH     EDX

     fld1
     fxch
     mov     ecx, eax
     cdq
     xor     eax, edx
     sub     eax, edx
     jnz     @@5
     fstp    ST(0)

     jmp     @@endInt
@@4: fmul    ST, ST
@@5: shr     eax,1
     jnc     @@4
     fmul    ST(1),ST
     jnz     @@4
     fstp    st
     cmp     ecx, 0



     jge     @@endInt
     //fdivr   S_c1
     fld1
     fdivrp

@@endInt:
     POP     EDX
     POP     ECX
     JMP     @@end

@@flow:

     // X=0 ??
     FXCH

     FLDZ            {.}
     FCOMP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@NoZero           //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??

     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz
     JB    @@end              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)
     fld1
     fchs
     fsqrt
     JMP     @@end      {.}

@@NoZero:
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP  ST(1)
     }
      //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)

        //fpower >  Max2Power ??
        FXCH ST(1)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2f
        fstp st(0)
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2f:
        FISTP M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)

@@end:

end;





procedure ZPWR;
//z1^z2; z1.RE-ST,z1.IM-ST(1),z2.RE-ST(2),z2.IM-ST(3);
var
M: Cardinal;
MX: Extended;
asm


 fld   st
 fmul  st,st
 fxch  st(2)
 fld   st
 fmul  st,st
 faddp st(3),st
 fxch  st(2)
 fsqrt
 //LN:
 fldln2
 fxch
 fyl2x

 fxch  st(2)
 fxch
 fpatan
 fld   st
 fmul  st,st(4)
 fxch
 fmul  st,st(3)
 fxch  st(2)
 fmul  st(3),st
 fmulp st(4),st
 fsubp st(2),st
 faddp st(2),st

 //EXP:
 FLDL2E
 FMUL
 FLD     ST(0)
 //FRNDINT
 FISTP M
 FILD  M
 FSUB    ST(1), ST
 FXCH    ST(1)
 F2XM1
 FLD1
 FADD
 {
 FSCALE
 FSTP    ST(1)
 }
 //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
         //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************

@@exp:
 fxch
 fsincos
 fmul   st,st(2)
 fxch   st(2)
 fmulp
 fxch

@@end:
end;





procedure Z_ReCxFPWR;
//Re^z2; Re-ST,z2.RE-ST(1),z2.IM-ST(2);
var
M: Cardinal;
MX: Extended;
asm


 ftst
 fstsw  ax
 sahf


 fabs
 //LN:
 fldln2
 fxch
 fyl2x

 fld   st
 jnb    @@1

 fmul  st,st(2)
 fldpi
 fmul  st,st(4)
 fsubp
 fxch  st(3)
 fmulp
 fldpi
 fmulp st(2),st
 faddp
 fxch
 jmp   @@2


@@1:
 fmulp st(2),st
 fmulp st(2),st
@@2:
  //EXP:
 FLDL2E
 FMUL
 FLD     ST(0)
 //FRNDINT
 FISTP M
 FILD  M
 FSUB    ST(1), ST
 FXCH    ST(1)
 F2XM1
 FLD1
 FADD
 {
 FSCALE
 FSTP    ST(1)
 }
 //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
 fxch
 fsincos
 fmul   st,st(2)
 fxch
 fmulp  st(2),st
@@end:
end;




procedure Z_CxReFPWR;
//z1^Re; z1.RE-ST,z1.IM-ST(1),Re-ST(2);
const c2: double=2;
var
M: Cardinal;
MX: Extended;
asm

        FXCH      ST(2)
        FIST      M//MEM
        FICOM     M//MEM
        FNSTSW    AX
        SAHF;
        JZ        @@int
        FXCH      ST(2)
        JMP       @@float

@@int:
        FSTP     ST
        FXCH
        MOV      EAX, M//MEM

        fldz
        fxch    ST(2)
        fld1
        fxch    ST(2)
        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jz      @@3
        jmp     @@2

   @@1: fld   st          //^2
        fmul  st,st
        fxch
        fmul  st,st(2)
        fmul  c2
        fxch  st(2)
        fmul  st,st
        fsubp

   @@2: shr     eax,1         //*z
        jnc     @@1

        fld   ST(0)
        fmul  ST(0),ST(3)
        fxch  ST(3)
        fmul  ST(0),ST(2)
        fxch  ST(4)
        fld   ST(0)
        fmul  ST(0),ST(3)
        fxch  ST(1)
        fmul  ST(0),ST(2)
        faddp ST(5),ST(0)
        fsubp ST(3),ST(0)
        jnz   @@1

@@3:
        fstp   st
        fstp   st

        cmp    ecx, 0
        jge    @@end

        //FAST_ZDIV:
        fld    st
        fmul   st,st
        fxch   st(2)
        fld    st
        fmul   st,st
        faddp  st(3),st
        fld1                    //FAST ZDIV
        fdivrp st(3),st
        fmul   st,st(2)
        fchs
        fxch   st(2)
        fmulp  st(1),st

        jmp    @@end

@@float:
        //ZPUSHST
        fld   st
        fxch  st(2)
        fld   st
        fxch  st(3)

        fmul  st,st
        fxch
        fmul  st,st
        faddp
        fsqrt

        //LN:
        fldln2
        fxch
        fyl2x

        fmul  st,st(3)

        //EXP:
        FLDL2E
        FMUL
        FLD     ST(0)
        //FRNDINT
        FISTP M
        FILD  M
        FSUB    ST(1), ST
        FXCH    ST(1)
        F2XM1
        FLD1
        FADD
        {
        FSCALE
        FSTP    ST(1)
         }
         //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
    @@exp:
        fxch  st(2)
        fxch
        fpatan
        fmulp st(2),st
        fxch
        fsincos
        fmul   st,st(2)
        fxch
        fmulp  st(2),st

@@end:
end;





{v177}
procedure Z_IPWR;
const c2:double=2;
//z1^Int; z1.RE-ST,z1.IM-ST(1),Int-EAX;
asm

       { FXCH      ST(2)
        FIST      M//MEM
        FICOM     M//MEM
        FNSTSW    AX
        SAHF;
        JZ        @@int
        FXCH      ST(2)
        JMP       @@float }

@@int:
        {FSTP     ST
        FXCH
        MOV      EAX, M//MEM   }

        fldz
        fxch    ST(2)
        fld1
        fxch    ST(2)
        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jz      @@3
        jmp     @@2

   @@1: fld   st          //^2
        fmul  st,st
        fxch
        fmul  st,st(2)
        //fmul  S_c2
        fmul c2             {.217}
        fxch  st(2)
        fmul  st,st
        fsubp

   @@2: shr     eax,1         //*z
        jnc     @@1

        fld   ST(0)
        fmul  ST(0),ST(3)
        fxch  ST(3)
        fmul  ST(0),ST(2)
        fxch  ST(4)
        fld   ST(0)
        fmul  ST(0),ST(3)
        fxch  ST(1)
        fmul  ST(0),ST(2)
        faddp ST(5),ST(0)
        fsubp ST(3),ST(0)
        jnz   @@1

@@3:
        fstp   st
        fstp   st

        cmp    ecx, 0
        jge    @@end

        //fl_FAST_ZDIV:
        fld    st
        fmul   st,st
        fxch   st(2)
        fld    st
        fmul   st,st
        faddp  st(3),st
        fld1                     //FAST ZDIV
        fdivrp st(3),st
        fmul   st,st(2)
        fchs
        fxch   st(2)
        fmulp  st(1),st

@@end :
end;




procedure R_FACT; assembler;
//N -> EAX
asm
 push  ecx
 mov   ecx, dword ptr [Factorial]
 lea   ecx, [ecx+eax*8]
 lea   eax, [ecx+eax*2]
 fld   tbyte ptr [eax]
 pop   ecx
 //ret
end;





procedure R_FACT2; assembler;
asm
      push  ecx
      mov   ecx, dword ptr [Factorial2F]
      lea   ecx, [ecx+eax*8]
      lea   eax, [ecx+eax*2]
      fld   tbyte ptr [eax]
      pop   ecx
      //push eax
     (* push ecx
      mov  ecx,eax
      and  ecx,1
      test ecx,ecx
      jnz  @_fct1
      shr  eax,1
      mov  ecx,dword ptr [Factorial2]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
      jmp  @end
  @_fct1:
      add  eax,1
      shr  eax,1
      mov  ecx,dword ptr [Factorial1]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
  @end:
      pop  ecx
   *)
      //pop  eax
end;





procedure  Z_COS;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm

   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   fld1                    //FAST ZDIV
   fdivrp
   //fdivr S_c1
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(2),st
   fchs
   fmulp st(2),st
@@end:
end;



procedure Z_ARCSINH;
const c2: double=2;
const c05: double=0.5;
//RE-ST(0); IM-ST(1);
asm

 //ZPUSHST
 fld   st
 fxch  st(2)
 fld   st
 fxch  st(3)

 //ZSQR
 fld   st
 fmul  st,st
 fxch
 fmul  st,st(2)
 fmul  c2
 fxch  st(2)
 fmul  st,st
 fsubp

 //fadd  S_c1
 fld1
 faddp

 //ZSQRT:
 push   ax
 fxch
 ftst
 fstsw  ax
 sahf
 fmul   st,st
 fxch
 fld    st
 fmul   st,st
 faddp st(2),st
 fxch
 fsqrt
 fld   st
 fsub  st,st(2)
 fmul  c05
 fsqrt
 jnb   @@1
 fchs
@@1:
 fxch  st(2)
 faddp
 fmul  c05
 fsqrt
 pop   ax

 faddp st(2),st
 faddp st(2),st

 //ZLN:
 fld  st
 fxch st(2)
 fld  st
 fxch st(3)
 fpatan
 fxch st(2)
 fmul st,st
 fxch
 fmul st,st
 faddp
 fldln2
 fxch
 fyl2x
 fmul  c05
end;




procedure Z_ARCCOSH;
const c2: double=2;
const c05: double=0.5;
//RE-ST(0); IM-ST(1);
//  use in form: arccosh=ln(z+sqrt(z^2-1)) - no always correct , best: "ln(z+sqrt(z+1)*sqrt(z-1))"
asm

 //ZPUSHST
 fld   st
 fxch  st(2)
 fld   st
 fxch  st(3)

 //ZSQR
 fld   st
 fmul  st,st
 fxch
 fmul  st,st(2)
 fmul  c2
 fxch  st(2)
 fmul  st,st
 fsubp

 //fsub  S_c1
 fld1
 fsubp

 //ZSQRT:
 push   ax
 fxch
 ftst
 fstsw  ax
 sahf
 fmul   st,st
 fxch
 fld    st
 fmul   st,st
 faddp st(2),st
 fxch
 fsqrt
 fld   st
 fsub  st,st(2)
 fmul  c05
 fsqrt
 jnb   @@1
 fchs
@@1:
 fxch  st(2)
 faddp
 fmul  c05
 fsqrt
 pop   ax

 faddp st(2),st
 faddp st(2),st

 //ZLN:
 fld  st
 fxch st(2)
 fld  st
 fxch st(3)
 fpatan
 fxch st(2)
 fmul st,st
 fxch
 fmul st,st
 faddp
 fldln2
 fxch
 fyl2x
 fmul  c05
end;




procedure Z_ARCTANH;
const c2: double=2;
const c05: double=0.5;
const c025: double=0.25;
//RE-ST(0); IM-ST(1);
asm

  //fsubr  S_c1
  fld1
  fsubrp

  fld    st
  fsubr  c2
  fmul   st,st(1)
  fxch
  fmul   st,st
  fxch   st(2)
  fld    st
  fmul   c2
  fxch
  fmul   st,st
  fsub   st(2),st
  faddp  st(3),st
  fxch   st(2)
  //fdivr  S_c1
  fld1                    //FAST ZDIV
  fdivrp
  fmul   st(1),st
  fmulp  st(2),st

//ZLN*0.5
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fmul  c05
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c025
end;




procedure Z_ARCCOTANH;
const c2: double=2;
const c05: double=0.5;
const c025: double=0.25;
const cs2: double=-2;
//RE-ST(0); IM-ST(1);
asm

  //fsub   c1
  fld1
  fsubp

  fld    st
  fadd   c2
  fmul   st,st(1)
  fxch
  fmul   st,st
  fxch   st(2)
  fld    st
  fmul   cs2
  fxch
  fmul   st,st
  fadd   st(2),st
  faddp  st(3),st
  fxch   st(2)
  //fdivr  c1
  fld1                       //FAST ZDIV
  fdivrp
  fmul   st(1),st
  fmulp  st(2),st

//ZLN*0.5
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fmul  c05
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c025
end;









procedure Z_TANH;
const c2: double=2;
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
//RE-ST(0); IM-ST(1);
asm

   fmul   c2
   fxch
   fmul   c2
   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld    st
   fmul   st,st
   fmul   c05
   fld    st
   fsub   c05
   fxch
   fadd   c05
   fxch   st(3)
   fsincos
   fmul   st,st(3)
   faddp  st(4),st
   fxch   st(3)
   //fdivr  S_c1
   fld1                         //FAST ZDIV
   fdivrp
   fxch
   fmul   st,st(1)
   fxch   st(3)
   fmulp  st(2),st
   fmulp
   fxch
@@end:
end;




procedure Z_COTANH;
const c2: double=2;
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
//RE-ST(0); IM-ST(1);
asm

   fmul   c2
   fxch
   fmul   c2
   fxch
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld    st
   fmul   st,st
   fmul   c05
   fld    st
   fsub   c05
   fxch
   fadd   c05
   fxch   st(3)
   fsincos
   fmul   st,st(3)
   faddp  st(4),st
   fxch   st(3)
   //fdivr  S_c1
   fld1                              //FAST ZDIV
   fdivrp
   fxch
   fmul   st,st(1)
   fxch   st(3)
   fmulp  st(2),st
   fmulp
   fxch

     //r/z1; r-ST,z1.RE-ST(1),z1.IM-ST(2);
  fld1

  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  //fdivr S_c1
  fld1                          //FAST ZDIV
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
@@end:
end;




procedure Z_TAN;
const c2: double=2;
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
//RE-ST(0); IM-ST(1);
asm

   fmul   c2
   fxch
   fmul   c2
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
    }
    //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld    st
   fmul   st,st
   fmul   c05
   fld    st
   fsub   c05
   fxch
   fadd   c05
   fxch   st(3)
   fsincos
   fmul   st,st(3)
   faddp  st(4),st
   fxch   st(3)
   //fdivr  S_c1
   fld1                            //FAST ZDIV
   fdivrp
   fxch
   fmul   st,st(1)
   fxch   st(3)
   fmulp  st(2),st
   fmulp
@@end:
end;





procedure Z_COTAN;
const c2: double=2;
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
//RE-ST(0); IM-ST(1);
asm

   fmul   c2
   fxch
   fmul   c2
   //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld    st
   fmul   st,st
   fmul   c05
   fld    st
   fsub   c05
   fxch
   fadd   c05
   fxch   st(3)
   fsincos
   fmul   st,st(3)
   faddp  st(4),st
   fxch   st(3)
   //fdivr  S_c1
   fld1                                 //FAST ZDIV
   fdivrp
   fxch
   fmul   st,st(1)
   fxch   st(3)
   fmulp  st(2),st
   fmulp

   //r/z1; r-ST,z1.RE-ST(1),z1.IM-ST(2);
  fld1

  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  //fdivr S_c1
  fld1                               //FAST ZDIV
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
@@end:
end;



procedure Z_EXP;
//RE-ST(0); IM-ST(1);
var
M: Cardinal;
MX: Extended;
asm
  { push ax
   fldz
   fcomp
   fstsw ax
   sahf
   jnz @@zexp
   fstp st
   fsincos
   jmp @@endp}

   fldz                        {.217}
   fcomip    st(0),st(1)
   jnz @@zexp
   fstp st(0)
   fsincos
   jmp @@end

   @@zexp:
 //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
    }
    //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF


        //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow

        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fxch
   fsincos
   fmul  st,st(2)
   fxch
   fmulp st(2),st

@@end:
   //pop  ax
end;



procedure Z_SINH;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm

  //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   //fdivr S_c1
   fld1
   fdivrp
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(3),st
   fmulp
   fxch
 @@end:
end;




procedure Z_COSECH;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm

  //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   //fdivr S_c1
   fld1
   fdivrp
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(3),st
   fmulp
   fxch
    //FAST ZDIV (1/rez)
  fld1
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
@@end:
end;



procedure Z_COSH;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm


 //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
    }
    //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   //fdivr S_c1
   fld1
   fdivrp
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(2),st
   fmulp st(2),st
@@end:
end;


procedure Z_SECH;
//RE-ST(0); IM-ST(1);
const c05: double=0.5;
var
M: Cardinal;
MX: Extended;
asm


 //EXP:
   FLDL2E
   FMUL
   FLD     ST(0)
   //FRNDINT
   FISTP M
   FILD  M
   FSUB    ST(1), ST
   FXCH    ST(1)
   F2XM1
   FLD1
   FADD
   {
   FSCALE
   FSTP    ST(1)
   }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
   fld   st
   //fdivr S_c1
   fld1
   fdivrp
   fld   st
   fadd  st,st(2)
   fmul  c05
   fxch  st(2)
   fsubrp
   fmul  c05
   fxch  st(2)
   fsincos
   fmulp st(2),st
   fmulp st(2),st
     //FAST ZDIV (1/rez)
  fld1
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp

  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st
@@end:
end;




procedure Z_ARCTAN;
const c2: double=2;
const c05: double=0.5;
//RE-ST(0); IM-ST(1);
asm

  fxch
  //fadd   S_c1
  fld1
  faddp
  fld    st
  fsubr  c2
  fmul   st,st(1)
  fxch
  fmul   st,st
  fxch   st(2)
  fld    st
  fmul   c2
  fxch
  fmul   st,st
  fsub   st(2),st
  faddp  st(3),st
  fxch   st(2)
  //fdivr  S_c1
  fld1                         //FAST ZDIV
  fdivrp
  fmul   st(1),st
  fmulp  st(2),st
   //ZLN:
   fld  st
   fxch st(2)
   fld  st
   fxch st(3)
   fpatan
   fxch st(2)
   fmul st,st
   fxch
   fmul st,st
   faddp
   fldln2
   fxch
   fyl2x
   fmul  c05

   fmul  c05
   fchs
   fxch
   fmul  c05
end;





procedure Z_ARCCOTAN;
//RE-ST(0); IM-ST(1);
const c2: double=2;
const c05: double=0.5;
const CPId2: extended  = 3.1415926535897932384626433832795/2;

asm

 (*
  fxch
  //fsub   S_c1
  fld1
  fsubp
  fld    st
  fadd   S_c2
  fmul   st,st(1)
  fxch
  fmul   st,st
  fxch   st(2)
  fld    st
  fmul   S_c2
  fxch
  fmul   st,st
  fadd   st(2),st
  faddp  st(3),st
  fxch   st(2)
  //fdivr  S_c1
  fld1                                //FAST ZDIV
  fdivrp
  fmul   st(1),st
  fmulp  st(2),st
   //ZLN:
   fld  st
   fxch st(2)
   fld  st
   fxch st(3)
   fpatan
   fxch st(2)
   fmul st,st
   fxch
   fmul st,st
   faddp
   fldln2
   fxch
   fyl2x
   fmul  S_c05

   fmul  S_c05
   fchs
   fxch
   fmul  S_c05

   *)
                   {.217}
   fxch
  //fadd   S_c1
  fld1
  faddp
  fld    st
  fsubr  c2
  fmul   st,st(1)
  fxch
  fmul   st,st
  fxch   st(2)
  fld    st
  fmul   c2
  fxch
  fmul   st,st
  fsub   st(2),st
  faddp  st(3),st
  fxch   st(2)
  //fdivr  S_c1
  fld1                         //FAST ZDIV
  fdivrp
  fmul   st(1),st
  fmulp  st(2),st
   //ZLN:
   fld  st
   fxch st(2)
   fld  st
   fxch st(3)
   fpatan
   fxch st(2)
   fmul st,st
   fxch
   fmul st,st
   faddp
   fldln2
   fxch
   fyl2x
   fmul  c05

   fmul  c05
   //fchs
   fxch
   fmul  c05

   fld  CPId2
   fsubrp

end;



procedure Z_ARCSIN;
//RE-ST(0); IM-ST(1);
const c2: double=2;
const c05: double=0.5;
asm


    //ZPUSHST:
   fld   st
   fxch  st(2)
   fld   st
   fxch  st(3)
   //ZSQR:
   fld   st
   fmul  st,st
   fxch
   fmul  st,st(2)
   fmul  c2
   fxch  st(2)
   fmul  st,st
   fsubp

   //fsubr S_c1
   fld1
   fsubrp
   fxch
   fchs
   //ZSQRT':
   ftst
   fstsw  ax
   sahf
   fmul   st,st
   fxch
   fld    st
   fmul   st,st
   faddp st(2),st
   fxch
   fsqrt
   fld   st
   fsub  st,st(2)
   fmul  c05
   fsqrt
   jnb   @@1
   fchs
  @@1:
   fxch  st(2)
   faddp
   fmul  c05
   fsqrt

   fsubrp st(3),st
   faddp
   fxch
   //ZLN:
   fld  st
   fxch st(2)
   fld  st
   fxch st(3)
   fpatan
   fxch st(2)
   fmul st,st
   fxch
   fmul st,st
   faddp
   fldln2
   fxch
   fyl2x
   fmul  c05

   fchs
   fxch


end;





procedure Z_ARCCOS;
const c2: double=2;
const c05: double=0.5;
//RE-ST(0); IM-ST(1);
asm


   //ZPUSHST:
   fld   st
   fxch  st(2)
   fld   st
   fxch  st(3)
   //ZSQR:
   fld   st
   fmul  st,st
   fxch
   fmul  st,st(2)
   fmul  c2
   fxch  st(2)
   fmul  st,st
   fsubp

   //fsubr S_c1
   fld1
   fsubrp
   fxch
   fchs
   //ZSQRT':
   ftst
   fstsw  ax
   sahf
   fmul   st,st
   fxch
   fld    st
   fmul   st,st
   faddp st(2),st
   fxch
   fsqrt
   fld   st
   fsub  st,st(2)
   fmul  c05
   fsqrt
   jnb   @@1
   fchs
  @@1:
   fxch  st(2)
   faddp
   fmul  c05
   fsqrt

   faddp  st(3),st
   fsubp  st(1),st
   //ZLN:
   fld  st
   fxch st(2)
   fld  st
   fxch st(3)
   fpatan
   fxch st(2)
   fmul st,st
   fxch
   fmul st,st
   faddp
   fldln2
   fxch
   fyl2x
   fmul  c05

   fchs
   fxch

end;




procedure Z_ARCSEC;
const c2: double=2;
const c05: double=0.5;
const CPId2: extended  = 3.1415926535897932384626433832795/2;

//RE-ST(0); IM-ST(1);
asm
//arcsec=pi/2+i*ln(sqrt(1-(1/z)^2)+i/z)

  fld1

  // 1/z: FAST ZDIV
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st

   // ZPUSH    1/z, 1/z
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)

  // ZSQR  1/z^2
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp


  // 1-1/z^2
  fchs
  fld1
  faddp
  fxch st(1)
  fchs
  fxch  st(1)

 // ZSQRT: sqrt(1-1/z^2)
  push   ax
  fxch
  ftst
  fstsw  ax
  sahf
  fmul   st,st
  fxch
  fld    st
  fmul   st,st
  faddp st(2),st
  fxch
  fsqrt
  fld   st
  fsub  st,st(2)
  fmul  c05
  fsqrt
  jnb   @@1
  fchs
 @@1:
  fxch  st(2)
  faddp
  fmul  c05
  fsqrt
  pop   ax

 // sqrt+i/z
  fsubrp st(3),st(0)
  faddp
  fxch

  // LN
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c05

  // *i
  fxch
  fchs

  //+pi/2
  fld tbyte ptr [CPId2]
  faddp

end;




procedure Z_ARCCOSEC;
const c2: double=2;
const c05: double=0.5;
//const CPId2: extended  = 3.1415926535897932384626433832795/2;

//RE-ST(0); IM-ST(1);
asm
//arccosec=-i*ln(sqrt(1-(1/z)^2)+i/z)

  fld1

  // 1/z: FAST ZDIV
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st

   // ZPUSH    1/z, 1/z
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)

  // ZSQR  1/z^2
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp


  // 1-1/z^2
  fchs
  fld1
  faddp
  fxch st(1)
  fchs
  fxch  st(1)

 // ZSQRT: sqrt(1-1/z^2)
  push   ax
  fxch
  ftst
  fstsw  ax
  sahf
  fmul   st,st
  fxch
  fld    st
  fmul   st,st
  faddp st(2),st
  fxch
  fsqrt
  fld   st
  fsub  st,st(2)
  fmul  c05
  fsqrt
  jnb   @@1
  fchs
 @@1:
  fxch  st(2)
  faddp
  fmul  c05
  fsqrt
  pop   ax

 // zsqrt+i/z
  fsubrp st(3),st(0)
  faddp
  fxch

  // LN
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c05

  // *(-i)
  fchs
  fxch

end;




procedure Z_ARCSECH;
const c1: double=1;
const c05: double=0.5;
const c2: double=2;
//const CPId2: extended  = 3.1415926535897932384626433832795/2;

//RE-ST(0); IM-ST(1);
//  use in form: arcsech=ln(1/z+sqrt((1/z)^2-1)) - no always correct , best: "ln(1/z+sqrt(1/z+1)*sqrt(1/z-1))"
asm
fld1

  // 1/z: FAST ZDIV
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st

   // ZPUSH    1/z, 1/z
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)

  // ZSQR  1/z^2
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp

   // 1/z^2-1
  fsub qword ptr [c1]

   // SQRT: sqrt(1/z^2-1)
  push   ax
  fxch
  ftst
  fstsw  ax
  sahf
  fmul   st,st
  fxch
  fld    st
  fmul   st,st
  faddp st(2),st
  fxch
  fsqrt
  fld   st
  fsub  st,st(2)
  fmul  c05
  fsqrt
  jnb   @@1
  fchs
 @@1:
  fxch  st(2)
  faddp
  fmul  c05
  fsqrt
  pop   ax

  //sqrt+1/z
  faddp st(2),st(0)
  faddp st(2),st(0)

   // LN
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c05

end;



procedure Z_ARCCOSECH;
const c1: double=1;
const c05: double=0.5;
const c2: double=2;
//const CPId2: extended  = 3.1415926535897932384626433832795/2;

//RE-ST(0); IM-ST(1);
//  arccosech=ln(1/z+sqrt((1/z)^2+1))
asm
fld1

  // 1/z: FAST ZDIV
  fxch st(2)
  fld  st
  fmul st,st
  fxch
  fmul st,st(3)
  fchs
  fxch st(3)
  fmul st,st(2)
  fxch st(2)
  fmul st,st
  faddp
  fld1
  fdivrp
  fmul  st(1),st
  fmulp st(2),st

   // ZPUSH    1/z, 1/z
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)

  // ZSQR  1/z^2
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp

   // 1/z^2+1
  fadd qword ptr [c1]

   // SQRT: sqrt(1/z^2-1)
  push   ax
  fxch
  ftst
  fstsw  ax
  sahf
  fmul   st,st
  fxch
  fld    st
  fmul   st,st
  faddp st(2),st
  fxch
  fsqrt
  fld   st
  fsub  st,st(2)
  fmul  c05
  fsqrt
  jnb   @@1
  fchs
 @@1:
  fxch  st(2)
  faddp
  fmul  c05
  fsqrt
  pop   ax

  //sqrt+1/z
  faddp st(2),st(0)
  faddp st(2),st(0)

   // LN
  fld  st
  fxch st(2)
  fld  st
  fxch st(3)
  fpatan
  fxch st(2)
  fmul st,st
  fxch
  fmul st,st
  faddp
  fldln2
  fxch
  fyl2x
  fmul  c05

end;




procedure Z_CxCxLOG;
//log(z1,z2); z1.Re-ST,z1.Im-ST(1),z2.Re-ST(2),z2.Im-ST(3);
const c05: double=0.5;
const c025: double=0.25;
asm

  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch
  fpatan
  fxch
  //LN:
  fldln2
  fxch
  fyl2x

  fxch  st(3)
  fld   st
  fmul  st,st
  fxch  st(3)
  fld   st
  fmul  st,st
  faddp st(4),st
  fpatan
  fxch  st(2)
  //LN:
  fldln2
  fxch
  fyl2x

  fxch
  fld   st
  fmul  st,st
  fxch  st(4)
  fld   st
  fmul  st,st
  fmul  c025

  faddp st(5),st
  fxch  st(4)
  fld1
  fdivrp
  //fdivr S_c1
  fxch  st(4)
  fld   st
  fmul  st,st(3)
  fxch
  fmul  st,st(4)
  fxch  st(4)
  fmul  st,st(2)
  fxch  st(3)
  fmulp st(2),st
  fmul  c025
  faddp st(2),st
  fsubp st(2),st
  fmul  st,st(2)
  fxch
  fmul  c05
  fmulp st(2),st

end;







procedure R_LogN; assembler;
//ST(0) <- Y
//ST(1) <- X
asm
        FXCH
        FLD1
        FXCH
        FYL2X
        FXCH
        FLD1
        FXCH
        FYL2X
        FDIV
end;







procedure Z_ReCxLOG;
const c05: double=0.5;
const c025: double=0.25;
//log(r,z); Re-ST,z2.RE-ST(1),z2.IM-ST(2);
asm
  ftst
  fstsw  ax
  sahf
  jb     @@pi
  fldz
  jmp    @@1
@@pi:
  fldpi
@@1:
  fxch
  fmul  st,st
  //LN:
  fldln2
  fxch
  fyl2x

  fxch  st(3)
  fld   st
  fmul  st,st
  fxch  st(3)
  fld   st
  fmul  st,st
  faddp st(4),st
  fpatan
  fxch  st(2)
  //LN:
  fldln2
  fxch
  fyl2x

  fxch
  fld   st
  fmul  st,st
  fxch  st(4)
  fld   st
  fmul  st,st
  fmul  c025

  faddp st(5),st
  fxch  st(4)
  //fdivr S_c1
  fld1
  fdivrp
  fxch  st(4)
  fld   st
  fmul  st,st(3)
  fxch
  fmul  st,st(4)
  fxch  st(4)
  fmul  st,st(2)
  fxch  st(3)
  fmulp st(2),st
  fmul  c025
  faddp st(2),st
  fsubp st(2),st
  fmul  st,st(2)
  fxch
  fmul  c05
  fmulp st(2),st
end;






procedure Z_CxReLOG;
const c05: double=0.5;
const c025: double=0.25;
//log(z,r);
//z1.Re-ST(0); z1.Im-ST(1); r-ST(2);

asm
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch
  fpatan
  fxch
  //LN:
  fldln2
  fxch
  fyl2x

  fxch  st(2)
  ftst
  fstsw  ax
  sahf
  jb     @@pi
  fldz
  jmp    @@1
@@pi:
  fldpi
@@1:
  fxch
  fmul  st,st
  //LN:
  fldln2
  fxch
  fyl2x

  fxch
  fxch  st(2)
  fld   st
  fmul  st,st
  fxch  st(4)
  fld   st
  fmul  st,st
  fmul  c025

  faddp st(5),st
  fxch  st(4)
  //fdivr S_c1
  fld1
  fdivrp
  fxch  st(4)
  fld   st
  fmul  st,st(3)
  fxch
  fmul  st,st(4)
  fxch  st(4)
  fmul  st,st(2)
  fxch  st(3)
  fmulp st(2),st
  fmul  c025
  faddp st(2),st
  fsubp st(2),st
  fmul  st,st(2)
  fxch
  fmul  c05
  fmulp st(2),st

end;




procedure Z_CxIROOT; assembler;
//root(Z,N); Re-ST, Im-ST(1), N-EAX;
var
M: Cardinal;
asm
 mov   M,eax
 //ZPUSHST
 fld   st
 fxch  st(2)
 fld   st
 fxch  st(3)

 fpatan
 fild  M
 fdivp
 fxch  st(2)
 fmul  st,st
 fxch
 fmul  st,st
 faddp
 fsqrt

 //IROOT:
 call R_IROOT

  fxch
  fsincos
  fmul   st,st(2)
  fxch
  fmulp  st(2),st
end;






{.215}
procedure  R_IROOT; assembler;
//EAX - ; ST(0) - ;
var
M: Cardinal;
MX: Extended;
asm
     //push    bx

     xor     bx,bx    //  '+'
     mov     M,eax  //
     fldz
     fcomp
     fnstsw  ax
     sahf
     mov     eax,M //
     ja      @@1  //  < 0
     jb      @@2  //  > 0
                  //  = 0:    root(0,P)
     fstp    ST(0)
     fldz

     // P = ??
     fild    M
     fcomp
     fnstsw  ax
     sahf
     ja      @@END    // P > 0

     // P <=0: call error
     fld1
     fchs
     fsqrt
     jmp     @@END

@@1:
     bt      eax,0
     jc      @@3      //  

     //xor     eax,eax    //    
     //div     eax
     fsqrt     {.203}
@@3:
     or      bx,1b     // '-'
     fabs
@@2:

     fld1
     fild    dword ptr [M]
     fdivp
     fxch
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD

     {
     FSCALE
     FSTP    ST(1)
     }

     //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
     (*
        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //call error
        fld1
        fchs
        fsqrt
        JMP       @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP       @@end

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        *)


        FXCH    st(1)
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)


    //**********************

     bt      bx,0
     jnc     @@END
     fchs

@@END:
     //pop     bx

end;




procedure  R_FROOT; assembler;
//ST(1) - ; ST(0) - ;
var
M: Cardinal;
MX: Extended;
asm
     //push      eax

     FXCH
     FIST      M
     FICOM     M
     FNSTSW    AX
     SAHF;
     JNZ        @@F   // 

     FXCH
     FSTP     ST(1)
     MOV      EAX, M
     CALL     R_IROOT

     JMP      @@END



@@F:
       // X=0 ??
     FXCH

     FLDZ            {.}
     FCOMP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@ROOT   //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??
     //FXCH            {.215}
     //FLDZ            {.215}
     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz            {.215}
     JB    @@END              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)               {.215}
     fld1
     fchs
     fsqrt
     JMP     @@END      {.}

@@ROOT:
     FXCH
     fld1
     fdivrp
     fxch
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP    ST(1)
     }
     //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        //PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        //POP   EAX
    //**********************
@@END:
     //pop      eax
end;






procedure R_ARCSIN;
 asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FPATAN
end;



procedure R_ARCCOS;
 asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FXCH
  FPATAN
end;




procedure R_ARCCOTAN;
const E_CPId2: extended  = 3.1415926535897932384626433832795/2;
 asm
  FLD1
  FPATAN
  FLD  E_CPid2
  fsubrp
end;




procedure R_SINH; assembler;
var
MX: Extended;
M:  Cardinal;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
  //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
      {  PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }

        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD   ST(0)
  FLD1
  FDIVR
  FSUB
  FMUL   C05

@@end:
end;



procedure R_COSH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }
         //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        //PUSH    EAX
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        //POP     EAX
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FDIVR
  FADD
  FMUL  C05
@@end:
end;




procedure R_TANH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FIMUL   C2
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FSUB
  FLD1
  FADD    ST,ST(2)
  FDIV
  FSTP    ST(1)
@@end:
end;




procedure R_COTANH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FIMUL   C2
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP   ST(1)
   }
    //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FSUB
  FLD1
  FADD    ST,ST(2)
  FDIVR
  FSTP    ST(1)
@@end:
end;




procedure R_ARCSINH;
asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FADD
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;



procedure R_ARCCOSH;
asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUB
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;



procedure R_ARCTANH;
const c05: double=0.5;
asm
  FLD  ST
  FLD1
  FADD
  FXCH
  FLD1
  FSUBR
  FDIV
  FLDLN2
  FXCH  ST(1)
  FYL2X
  FMUL  C05
end;



procedure R_ARCCOTANH;
const c05: double=0.5;
asm
  FLD  ST
  FLD1
  FADD
  FXCH
  FLD1
  FSUB
  FDIV
  FLDLN2
  FXCH  ST(1)
  FYL2X
  FMUL  C05
end;




procedure R_ARCSEC; assembler;
const c1: double=1;
asm      //arccos(1/x)
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FXCH
  FPATAN
end;


procedure R_ARCCOSEC; assembler;
const c1: double=1;
asm            //arcsin(1/x)
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FPATAN
end;



procedure R_SECH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FDIVR
  FADD
  FMUL  C05
  FDIVR C1
@@end:
end;




procedure R_COSECH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
       { PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }

        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        fld1
        fchs
        fsqrt
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD   ST(0)
  FLD1
  FDIVR
  FSUB
  FMUL   C05
  FDIVR  C1
@@end:
end;



procedure R_ARCSECH; assembler;
const c1: double=1;
asm
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUB
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;



procedure R_ARCCOSECH; assembler;
const c1: double=1;
asm
  push  ecx
  push  eax
  xor   ecx,ecx

  fldz
  fcomp
  fstsw   ax
  sahf
  jb      @@bz
  mov ecx,1

@@bz:
  fabs

  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FADD
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X

  cmp     ecx, 0
  je     @@end
  fchs
  @@end:

  pop  eax
  pop  ecx
end;



procedure R_EXP; assembler;
var
M: Cardinal;
MX: Extended;
asm
        FLDL2E
        FMUL
        FLD     ST(0)
        //FRNDINT
        FISTP M
        FILD  M
        FSUB    ST(1), ST
        FXCH    ST(1)
        F2XM1
        FLD1
        FADD
       {
        FSCALE
        FSTP ST(1)
       }
        //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX    }

        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@end

@@pwr2f:
        //PUSH    EAX
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        //POP     EAX
    //**********************
@@end:
end;





procedure Z_IPWR3; assembler;
//z1^3; z1.RE-ST,z1.IM-ST(1);
const c3: double=3;
asm
  fld   st
  fmul  st,st
  fmul  st,st(1)
  fxch  st(2)
  fld   st
  fmul  st,st
  fmul  st,st(1)
  fxch
  fld   st
  fmul  st,st(3)
  fmul  c3
  fxch
  fmul  st,st(1)
  fsubp st(4),st
  fmulp st(2),st
  fsubp
  fxch
end;




procedure Z_IPWR4; assembler;
//z1^4; z1.RE-ST,z1.IM-ST(1);
//: ZSQR+ZSQR;
const c4: double=4;
const c6: double=6;
asm
  fld   st
  fmul  st,st(2)
  fxch
  fmul  st,st
  fxch  st(2)
  fmul  st,st
  fxch  st(2)
  fld   st
  fsub  st,st(3)
  fxch  st(3)
  fmul  st,st
  fxch
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  c4
  fmulp st(4),st
  fmul  st,st
  fxch
  faddp st(2),st
  fmul  c6
  fsubp
end;




procedure Z_IPWRN1; assembler;
//z1^(-1); z1.RE-ST,z1.IM-ST(1);
asm
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st
end;



procedure Z_IPWRN2; assembler;
//z1^(-2); z1.RE-ST,z1.IM-ST(1);
const c2: double=2;
asm
    //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp

  //1/Z
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st

end;



procedure Z_IPWRN3; assembler;
//z1^(-2); z1.RE-ST,z1.IM-ST(1);
const c3: double=3;
asm
  // Z^3
  fld   st
  fmul  st,st
  fmul  st,st(1)
  fxch  st(2)
  fld   st
  fmul  st,st
  fmul  st,st(1)
  fxch
  fld   st
  fmul  st,st(3)
  fmul  c3
  fxch
  fmul  st,st(1)
  fsubp st(4),st
  fmulp st(2),st
  fsubp
  fxch
  //1/Z
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st

end;




procedure Z_IPWRN4; assembler;
//z1^(-2); z1.RE-ST,z1.IM-ST(1);
const c2: double=2;
asm
  // Z^4
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp

  //1/Z
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st


end;





procedure Z_IPWRN5; assembler;
//Z^(-5); z1.RE-ST,z1.IM-ST(1);
const c2: double=2;
asm
  //Z^5
  //ZPUSHST
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp
  //ZMULP
  fld  st(0)
  fmul st(0),st(3)
  fxch st(1)
  fmul st(0),st(4)
  fxch st(4)
  fmul st(0),st(2)
  fsubp
  fxch st(2)
  fmulp
  faddp st(2),st(0)

  //1/Z
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st

end;




procedure Z_IPWRN6; assembler;
//Z^(-6); z1.RE-ST,z1.IM-ST(1);
const c2: double=2;
asm
  //Z^6
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp
   //ZPUSHST
  fld   st
  fxch  st(2)
  fld   st
  fxch  st(3)
  //ZSQR
  fld   st
  fmul  st,st
  fxch
  fmul  st,st(2)
  fmul  c2
  fxch  st(2)
  fmul  st,st
  fsubp
  //ZMULP
  fld  st(0)
  fmul st(0),st(3)
  fxch st(1)
  fmul st(0),st(4)
  fxch st(4)
  fmul st(0),st(2)
  fsubp
  fxch st(2)
  fmulp
  faddp st(2),st(0)

  //1/Z
  fld   st
  fmul  st,st
  fxch  st(2)
  fld   st
  fmul  st,st
  faddp st(3),st
  fxch  st(2)
  fld1
  fdivrp
  fmul  st(1),st
  fchs
  fmulp st(2),st

end;



procedure R_IPWR;assembler;
const c1: double=1;
asm


        FLD1
        FXCH


        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jnz     @@2
        fstp    ST(0)
        jmp     @@3
   @@1: fmul    ST, ST
   @@2: shr     eax,1
        jnc     @@1
        fmul    ST(1),ST
        jnz     @@1
        fstp    st
        cmp     ecx, 0
        jge     @@3
        fdivr   c1

   @@3:
end;



procedure R_IPWRSGN; assembler;
asm

       fld1
       and  eax,1
       jz   @end
       fchs
@end:
end;




//(-1)^z1=exp(-pi*z1.im)*(cossin(pi*z1.re))
procedure Z_IPWRSGN; assembler;      {189}
var
M: Cardinal;
MX: Extended;
asm
{  ftst
 fstsw  ax
 sahf}
 //push ax

 fxch
 fldpi
 fmulp
 fchs

 //EXP:
 FLDL2E
 FMUL
 FLD     ST(0)
 //FRNDINT
 FISTP M
 FILD  M
 FSUB    ST(1), ST
 FXCH    ST(1)
 F2XM1
 FLD1
 FADD
 {
 FSCALE
 FSTP    ST(1)
 }
 //******ISCALE********* {.219}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
           //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:

 fxch
 fld st(0)

 (*
 sub esp,4    {.227}
 fnstcw [esp];{.227}
 fnclex {.227}
 fldcw fcw1
 //FRNDINT
 FISTP M
 FILD  M
 //fldcw fcw2
 fldcw [esp];  {.227}
 add esp, 4     {.227}

 fcomp
 FNSTSW    AX
 SAHF;
 JZ        @@int
 *)
 FISTP M
 FILD  M
 FCOMIP  ST(0),ST(1)
 JZ        @@int


 fldpi
 fmulp
 fsincos
 fmul   st,st(2)
 fxch
 fmulp  st(2),st
 jmp @@end


@@int:
 fldpi
 fmulp
 fcos
 fmulp
 fldz
 fxch

@@end:
  //pop ax
end;




procedure R_Decart;
//R->ST(0); PHI->ST(1): Re->ST(0); Im->ST(1)
asm
  fxch
  fsincos
  fmul st(0),st(2)
  fxch st(2)
  fmulp
  fxch
end;






//***********************************************************{begin}**************************************************************************************







procedure RI_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fild  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ecx
end;


procedure RD_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fld  qword ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ecx
end;



procedure RE_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fld  tbyte ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ebx
       pop  ecx
end;




procedure RI_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@end:
       pop  ecx
end;


procedure RD_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       faddp
       loop @@L
       @@end:
       pop  ecx
end;





procedure RE_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       faddp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;






procedure RI_PROD; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmulp
       loop @@L
       @@end:
       pop  ecx
end;






procedure RD_PROD; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmulp
       loop @@L
       @@end:
       pop  ecx
end;




procedure RE_PROD;  assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmulp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;






procedure RI_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@res
       dec  ecx
       fild  dword ptr [eax]
       faddp
       jecxz @@res
       @@L:
       fild  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ecx
end;



procedure RD_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@res
       dec  ecx
       fld  qword ptr [eax]
       faddp
       jecxz @@res
       @@L:
       fld  qword ptr [eax+8*ecx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ecx
end;




procedure RE_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@res
       dec  ecx
       fld  tbyte ptr [eax]
       faddp
       jecxz @@res
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ebx
       pop  ecx
end;




procedure RI_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ecx
end;


procedure RD_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RE_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;







procedure R_LEN; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       mov  ecx,eax
       test ecx,ecx
       jz @@nil
       mov  ecx,[ecx-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       @@nil:
       push ecx
       fild dword ptr [esp]
       pop  ecx
       pop  ecx

      { push ecx
       mov  eax, [eax]
       mov  ecx,[eax-4]
       push ecx
       fild [esp]
       pop  ecx
       pop  ecx}
end;




procedure RI_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fild   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fild   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fild   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;


procedure RD_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   qword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,8
 @@cycl:
     fld   qword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fld   qword ptr [ebx]
 @@next:
     add   ebx,8
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;





procedure RE_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   tbyte ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,10
 @@cycl:
     fld   tbyte ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fld   tbyte ptr [ebx]
 @@next:
     add   ebx,10
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RI_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fild   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fild   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fild   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RD_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   qword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,8
 @@cycl:
     fld   qword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fld   qword ptr [ebx]
 @@next:
     add   ebx,8
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;




procedure RE_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   tbyte ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,10
 @@cycl:
     fld   tbyte ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fld   tbyte ptr [ebx]
 @@next:
     add   ebx,10
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;





procedure RI_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call RI_AVR
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;



procedure RI_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RI_AVR
       //mov  edi,ecx
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;





procedure RD_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RD_AVR
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;





procedure RD_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RD_AVR
       //mov  edi,ecx
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;






procedure RE_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RE_AVR
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  edx
       pop  ecx
end;





procedure RE_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RE_AVR
       //mov  edi,ecx
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  edx
       pop  ecx
end;









procedure RE_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  qword ptr [eax+8*ecx]  //vd[n-1]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       //fld  qword ptr [eax+8*ecx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  edx
       pop  ecx
end;




procedure RD_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       fld  qword ptr [eax+8*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;





procedure RI_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       fild  dword ptr [eax+4*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;





procedure CxE_POLY; assembler;      //polynom(vec,z)=vec[n]*z^(n)+...+vec[1]*z+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)   ;   z.re->st(0),z.im->st(1)
       push ecx
       push edx
       fldz
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       fstp st
       dec  ecx
       fldz
       //fld  qword ptr [eax+8*ecx]  //vd[n-1]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end
       dec  ecx
       @@L:

       fld   st(0)
       fmul  st(0),st(3)
       fxch  st(1)
       fmul  st(0),st(4)
       fxch  st(2)
       fld   st(0)
       fmul  st(0),st(5)
       fsubp st(2),st(0)
       fmul  st(0),st(3)
       faddp st(2),st(0)

       //fld  qword ptr [eax+8*ecx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(2)
       fstp st(2)
       pop  edx
       pop  ecx
end;



procedure CxD_POLY; assembler;      //polynom(vec,z)=vec[n]*z^(n)+...+vec[1]*z+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)   ;   z.re->st(0),z.im->st(1)
       push ecx
       fldz
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       fstp st
       dec  ecx
       fldz
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:

       fld   st(0)
       fmul  st(0),st(3)
       fxch  st(1)
       fmul  st(0),st(4)
       fxch  st(2)
       fld   st(0)
       fmul  st(0),st(5)
       fsubp st(2),st(0)
       fmul  st(0),st(3)
       faddp st(2),st(0)

       fld  qword ptr [eax+8*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(2)
       fstp st(2)
       pop  ecx
end;





procedure CxI_POLY; assembler;      //polynom(vec,z)=vec[n]*z^(n)+...+vec[1]*z+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)   ;   z.re->st(0),z.im->st(1)
       push ecx
       fldz
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       fstp st
       dec  ecx
       fldz
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:

       fld   st(0)
       fmul  st(0),st(3)
       fxch  st(1)
       fmul  st(0),st(4)
       fxch  st(2)
       fld   st(0)
       fmul  st(0),st(5)
       fsubp st(2),st(0)
       fmul  st(0),st(3)
       faddp st(2),st(0)

       fild  dword ptr [eax+4*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(2)
       fstp st(2)
       pop  ecx
end;



//***********************************************************{end}**************************************************************************************




procedure R_SGN; assembler;
asm
     push    ax
     ftst
     fstp  st(0)
     fstsw   ax
     sahf
     jnz     @@nz
     fldz
     jmp     @@end
 @@nz:
     fld1
     jnb      @@end
     fchs
 @@end:
     pop     ax
end;




procedure R_MAX(adr,len: Cardinal); stdcall;
var
i:Integer;
MAX: Extended;
begin
MAX:=PExtended(Adr)^;
for I := 1 to len - 1 do
begin
  if PExtended(Adr+16*i)^ > MAX then  MAX:=PExtended(Adr+16*i)^;
end;

 asm
  fld MAX
 end
end;




procedure R_MIN(adr,len: Cardinal); stdcall;
var
i:Integer;
MIN: Extended;
begin
MIN:=PExtended(Adr)^;
for I := 1 to len - 1 do
begin
  if PExtended(Adr+16*i)^ < MIN then  MIN:=PExtended(Adr+16*i)^;
end;

 asm
  fld MIN
 end
end;



procedure R_AVR(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+PExtended(Adr+16*i)^;
end;

S:=S/len;
 asm
  fld S
 end
end;



procedure R_SUM(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+PExtended(Adr+16*i)^;
end;

 asm
  fld S
 end
end;




procedure R_SUMQ(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^);
end;

 asm
  fld S
 end
end;




procedure R_NORM(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^);
end;

 asm
  fld S
  fsqrt
 end
end;




procedure R_PROD(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=1;
for I := 0 to len - 1 do
begin
  S:=S*PExtended(Adr+16*i)^;
end;

 asm
  fld S
 end
end;





procedure R_DEV(adr,len: Cardinal); stdcall;
var
i:Integer;
S,avg: Extended;
begin

avg:=0;
for I := 0 to len - 1 do
begin
  avg:=avg+PExtended(Adr+16*i)^;
end;

avg:=avg/len;

S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^-avg);
end;

S:=sqrt(S/len);

 asm
  fld S
 end
end;



procedure R_DEVS(adr,len: Cardinal); stdcall;
var
i:Integer;
S,avg: Extended;
begin

avg:=0;
for I := 0 to len - 1 do
begin
  avg:=avg+PExtended(Adr+16*i)^;
end;

avg:=avg/len;

S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^-avg);
end;

S:=sqrt(S/(len-1));

 asm
  fld S
 end
end;




procedure RI_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   fild  dword ptr [ebx+4*ecx]
end;




procedure RD_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  ECX<-num
   mov  ebx, [ebx]
   fld  qword ptr [ebx+8*ecx]
end;


procedure RE_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  ECX<-num
   mov  ebx, [ebx]
   //fld  tbyte ptr [ebx+10*ecx]
   lea   ecx,[ecx+4*ecx]
   fld   tbyte ptr [ebx+2*ecx]
end;


procedure RI_VECS; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   fistp  dword ptr [ebx+4*ecx]
end;

procedure RD_VECS; assembler;
asm              //load vec[num]: EBX<-addr(vec);  ECX<-num
   mov  ebx, [ebx]
   fstp  qword ptr [ebx+8*ecx]
end;


procedure RE_VECS; assembler;
asm              //load vec[num]: EBX<-addr(vec);  ECX<-num
   mov  ebx, [ebx]
   //fld  tbyte ptr [ebx+10*ecx]
   lea   ecx,[ecx+4*ecx]
   fstp   tbyte ptr [ebx+2*ecx]
end;


procedure _MAX2; assembler;
asm
     //push eax
     fcom
     fstsw ax
     sahf
     jb    @@end
     fxch st(1)
@@end:
     fstp  st(0)
     //pop eax
end;



procedure _MIN2; assembler;
asm
     //push eax
     fcom
     fstsw ax
     sahf
     jnb    @@end
     fxch st(1)
@@end:
     fstp  st(0)
     //pop eax
end;


procedure Z_CxIROOTA; assembler;
const C2PI: extended  = 3.1415926535897932384626433832795*2;
//root(Z,N,K);
var
M,N,K: Cardinal;
asm
 fld   tbyte ptr [eax+16]   //im
 fld   tbyte ptr [eax]      //re
 fld   tbyte ptr [eax+32]   //N
 fistp N
 fld   tbyte ptr [eax+48]   //K
 fistp K

 //mov   M,eax
 //ZPUSHST
 fld   st
 fxch  st(2)
 fld   st
 fxch  st(3)

 fpatan
 fild  K
 fld tbyte ptr [C2PI]
 fmulp
 faddp
 fild  N
 fdivp
 fxch  st(2)
 fmul  st,st
 fxch
 fmul  st,st
 faddp
 fsqrt

 //IROOT:
 mov eax, N
 call R_IROOT        {.215}


  fxch
  fsincos
  fmul   st,st(2)
  fxch
  fmulp  st(2),st
end;




procedure SetLengthArray(AdrV: Pointer; TypeV: Cardinal; Len: Cardinal);  stdcall;
var
L1: integer;
begin

if TypeV = _Double then
begin
  SetLength(PArrayD(AdrV)^,Len);
end
else
if TypeV = _Extended then
begin
  SetLength(PArrayE(AdrV)^,Len);
end
else
if TypeV = _Integer then
begin
  SetLength(PArrayI(AdrV)^,Len);
end;
end;




function std_rand2(x1,x2:PExtended): extended;  stdcall;
begin
 std_rand2:=(x1^-x2^)*random+x2^;
end;





//randomize()
procedure _Randomize;
var
ht,lt,ISeed: Cardinal;
b1,b2: Integer;
begin
InitRandomGenerator;
{
  asm
   push   eax
   push   edx
   RDTSC  //eax:edx
   shr    eax,10
   mov    lt,eax
   mov    ht,edx
   pop    edx
   pop    eax
 end;

  b1:=1; b2:=1;
  if lt  mod 2 <> 0 then b1:=-1;
  if ht  mod 2 <> 0 then b2:=-1;

  ISeed:=b1*lt+b2*ht;

  mt19937_init(ctxMT,ISeed);
  taus113_init(ctxPLE,ISeed);
  xor4096_init(ctxXOR,ISeed);
  mt19937_init(ctxMTG,ISeed); //only for RandG
}
end;



procedure InitRandomGenerator;
var

T1,i,nrnd,x1,x2,ht,lt,ISeed: Cardinal;

DT,DT2,Seed,b1,b2,b3,b4,b5,b6,b7,r: double;
DY,DM,DD,DHr,DMn,DSc,DMs: WORD;
begin
  {.203}
 DT:=Now;
 //DecodeDate( DT, DY, DM, DD);
 DecodeTime( DT, DHr, DMn, DSc, DMs);

 T1:=GetTickCount;

 asm
   push   eax
   push   edx
   RDTSC  //eax:edx
   mov    lt,eax
   mov    ht,edx
   pop    edx
   pop    eax
end;

 //_GetTickProc(ht,lt);
 {
 b1:=1; b2:=1; b3:=1;
 if Dms mod 2 <> 0 then b1:=-1;
 if T1  mod 2 <> 0 then b2:=-1;
 if lt  mod 2 <> 0 then b3:=-1;
 }
 //Seed:=(DY+DM+DD+Dhr+Dmn+Dsc)*DMs;
 //r:=lt/100000000;
 //Seed:=(DY/100+DM+DD+Dhr+Dmn+b1*Dsc+b2*Dms/10+b3*lt/100000000)*(T1/100);

 //ISeed:=Round(abs(Seed));

 b1:=1; b2:=1; b3:=1; b4:=1; b5:=1; b6:=1;b7:=1;
 if DHr mod 2 <> 0 then b1:=-1;
 if DMn mod 2 <> 0 then b2:=-1;
 if DSc mod 2 <> 0 then b3:=-1;
 if DMs mod 2 <> 0 then b4:=-1;
 if T1  mod 2 <> 0 then b5:=-1;
 if lt  mod 2 <> 0 then b6:=-1;
 if ht  mod 2 <> 0 then b7:=-1;

 //ISeed:=Trunc(DHr*b1+DMs*b2+DSc*b3+DMs*b4+T1*b5+lt*b6+ht*b7);
 //ISeed:=DHr*DMs*DSc*DMs*T1*lt*ht;
 ISeed:=(DHr+1)*(DMs+1)*(DSc+1)*(DMs+1)*(T1+1)*(lt+1)*(ht+1);

 mt19937_init(ctxMT,ISeed);
 taus113_init(ctxPLE,ISeed);
 xor4096_init(ctxXOR,ISeed);
 mt19937_init(ctxMTG,ISeed); //only for RandG

 mt19937_next(ctxMT);
 taus113_next(ctxPLE);
 xor4096_next(ctxXOR);
 mt19937_next(ctxMTG);  //only for RandG




end;



          //***************
//randomize()
//randmt(), randple(), randxor()  in (0,1)
//intrandmt(Num), intrandple(Num), intrandxor(Num)  in (0,Num) Num - int
//rand(array)






function _FACT1(n: integer): extended;//N -> EAX
asm
 push eax
 mov  eax,n
 push  ecx
 mov   ecx, dword ptr [Factorial]
 lea   ecx, [ecx+eax*8]
 lea   eax, [ecx+eax*2]
 fld   tbyte ptr [eax]
 pop   ecx
 pop   eax
 //ret
end;





function _FACT2(n: integer): extended;
asm
      push eax
      mov  eax,n
      push ecx
      mov  ecx,eax
      and  ecx,1
      test ecx,ecx
      jnz  @_fct1
      shr  eax,1
      mov  ecx,dword ptr [Factorial2]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
      jmp  @end
  @_fct1:
      add  eax,1
      shr  eax,1
      mov  ecx,dword ptr [Factorial1]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
  @end:
      pop  ecx

      pop  eax
end;






procedure R_GCD(adr,len: Cardinal); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r: longint;
begin

if len = 0  then S:=0 else
if len = 1  then S:=PExtended(Adr)^ else
begin
  n:=1;
  u:=Trunc(PExtended(Adr)^);
  v:=Trunc(PExtended(Adr+16)^);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(PExtended(Adr+16*n)^);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_GCDD(VD: TArrayD); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,len: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;



procedure V_GCDE(VD: TArrayE); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,len: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_GCDI(VD: TArrayI); stdcall;
label lp,endp;
var
i,n:Integer;
Res: Extended;
u,v,t,r,len,S: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=(VD[0]);
  v:=(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=(S);
    u:=(VD[n]);
    goto lp;
  end;

end;

endp:
Res:=S;
 asm
  fld Res
 end
end;






procedure R_LCM(adr,len: Cardinal); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1: longint;
begin


if len = 0  then S:=0 else
if len = 1  then S:=PExtended(Adr)^ else
begin
  n:=1;
  u:=Trunc(PExtended(Adr)^);
  v:=Trunc(PExtended(Adr+16)^);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(PExtended(Adr+16*n)^);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_LCMD(VD: TArrayD); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1,len: longint;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_LCME(VD: TArrayE); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1,len: longint;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;






procedure V_LCMI(VD: TArrayI); stdcall;
label lp,endp;
var
i,n:Integer;
S: longint;
u,v,t,r,u1,v1,len: longint;
Res: Extended ;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=(VD[0]);
  v:=(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=(S);
    u:=(VD[n]);
    goto lp;
  end;

end;

endp:
Res:=S;
 asm
  fld Res
 end
end;





procedure RD_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDD
end;


procedure RE_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDE
end;


procedure RI_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDI
end;






procedure RD_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCMD
end;


procedure RE_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCME
end;


procedure RI_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCMI
end;


(*
procedure _MaskFPUException;    {.227}
asm
    fnstcw word ptr [CWMEM];       //->store
    sub esp,4
    fnstcw [esp];
    or word ptr [esp], 003Fh;
    fldcw [esp];
    add esp, 4
end;

  {.227}
procedure _IsFPUException; //function: st(0)
asm
    push eax
    xor  eax,eax
    fnstsw ax
    sahf
    test ax, 001Fh
    fldz
    jz @@Z
    fstp st(0)
    fld1
    @@Z:
    pop eax
    fnclex
    fldcw word ptr [CWMEM]   //<-load
end;
*)


procedure _MaskFPUExceptionA;
asm
    fnstcw word ptr [ECX];       //->store
    sub esp,4
    fnstcw [esp];
    or word ptr [esp], 003Fh;
    fldcw [esp];
    add esp, 4
 end;




procedure _IsFPUExceptionA;
asm
    xor  eax,eax
    fnstsw ax
    sahf
    and ax, 001Fh
    push eax
    fild [esp]
    pop eax
    fnclex
    fldcw word ptr [ECX]   //<-load
end;



(*
procedure _SafeCall(Adr: Cardinal);       {.219}
var
M: Cardinal;
begin
  try
     asm
       //push eax
       //mov  Adr, eax
       call Adr
       //pop  eax
     end;
  except

  end;
end;
 *)



initialization
begin
 //init random numbers generator
 InitRandomGenerator;
 F_FastSpec:=True;
 CWMEM:=$1372;   {.227}
end;


end.
