unit Foreval_Command;

{******************************************************************************}
{                                                                              }
{                                 ver 8.4.17.249                               }
{----------------------------------------------------------------------------- }
{  :    ,                            }
{ +      .                         }
{       :  FWAIT  }
{  : Pentium Pro                                                 }
{ FSCALE                                                               }
{ -----------------------------------------------------------------------------}
{ Command module: contains executable functions,                               }
{ + initialize an array of functions and table of replacements.                }
{ Only use processors with AutoSync : missing FWAIT.                           }
{ min. proc.: Pentium Pro                                                      }
{ FSCALE is changed                                                            }
{------------------------------------------------------------------------------}
{                  SOREL   (C)CopyRight 2000. Russia, S.-Petersburg.           }
{******************************************************************************}


    {
              Random numbers generator  algorithm( taus113 (Pierre L'Ecuyer),mt19937 (Mersenne Twister), xor4096)
              Gamma function,
              Bessel   functions,
              Legendre, Chebyshev polynoms,
              Zeta functions, and other spec. functions are from
                  (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/
 }

 {$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

//{$OverFlowChecks Off}
//{$RangeChecks Off}
//{$ZEROBASEDSTRINGS Off}


{$R-}
{$Q-}

interface
uses
Foreval_Main, ForevalDiff_Lib, Foreval_SpecFunc, MATH, SysUtils, Windows;



const


 MaxInt: double= 2147483647;
 Max2Power: integer = 16383;  //2^Max2Power
 Max2PowerN: integer = -16383;  //2^Max2Power
 Power2OverFlow: integer = 17000; //Any > Max2Power

 C1: double = 1;
 C05: double = 0.5;
 //C1dpi: extended = 0.3183098861838; {1/Pi;}
 //C2dpi: extended = 0.6366197723676; {2/Pi;}
 C2: integer = 2;
 C3: double = 3;
 C025: double = 0.25;
 E_CPId2: extended = 3.141592653589793238462643/2; //1.57079632679489662;{Pi/2;}
 FCW1: Word = $1f32;    {Trunc}
 FCW2: Word = $1332;
 FCW3: Word = $1232;

 const  _Yes: Integer = 1;
 const  _No: Integer = 2;

 const f_Jordan = 1;
 const f_Gauss =  2;

 const rPI: extended = 3.14159265358979323846264/180;
 const dPI: extended = 180/3.14159265358979323846264;



type
TInteg2 = record
           I1: Integer;
           I2: Integer;
          end;
type
TArrayF_2 = array of TInteg2;


type
TSplain = record
             A0: TArrayF;
             A1: TArrayF;
             A2: TArrayF;
             A3: TArrayF;
             PS: TArrayF;
             SL: Integer;
             DS: TFloatType;
            end;



type
TArrayD = array of Double;
type
TArrayE = array of Extended;
type
TArrayS = array of Single;
type
TArrayI = array of Integer;
type
TArrayC = array of Cardinal;
type
PArrayD = ^TArrayD;
type
PArrayE = ^TArrayE;
type
PArrayI = ^TArrayI;
type
PArrayS = ^TArrayS;
type
TArray2E = array of array of extended;

const
_ARGF = 12345;




const
//     1    (+1) !!!
// C_IPWR3-C_IPWR16;C_IPWR3N-C_IPWR16N   (+1  )!!!
//{Internal Function: identeficator - IDF}
 C_LDSC      =   1;
 C_LDSF      =   2;
 C_LDSI      =   3;
 C_FAVB      =   4;
 C_IAVB      =   5;
 C_AIFB      =   6;
 C_AAI       =   7;
 C_AAF       =   8;
 C_AFI       =   9;
 C_MAF       =   10;
 C_MUL       =   11;
 C_ADD       =   12;
 C_NOP       =   13;
 C_PACKAGE =     14;
 C_SQR     =     15;
 C_SQRT =        16;
 C_LN =          17;
 C_EXP =         18;
 C_LOG2 =        19;
 C_LOG10 =       20;
 C_IPWR3 =       21;
 C_IPWR4 =       22;
 C_IPWR5 =       23;
 C_IPWR6 =       24;
 C_IPWR7 =       25;
 C_IPWR8 =       26;
 C_IPWR9 =       27;
 C_IPWR10 =      28;
 C_IPWR11 =      29;
 C_IPWR12 =      30;
 C_IPWR13 =      31;
 C_IPWR14 =      32;
 C_IPWR15 =      33;
 C_IPWR16 =      34;
 C_SIN =         35;
 C_COS =         36;
 C_TAN =         37;
 C_COTAN =       38;
 C_ARCSIN =      39;
 C_ARCCOS =      40;
 C_ARCTAN =      41;
 C_ARCCOTAN =    42;
 C_SINH =        43;
 C_COSH =        44;
 C_TANH =        45;
 C_COTANH =      46;
 C_ARCSINH =     47;
 C_ARCCOSH =     48;
 C_ARCTANH =     49;
 C_ARCCOTANH =   50;
 C_ABS =         51;
 C_TRUNC =       52;
 C_LOGN =        53;
 C_FPWR =        54;
 C_FACT =        55;
 C_IPWR =        56;
 C_IPWRSGN =     57;
 C_I2PWR =       58;
 C_POWER =       59;
 C_ROOT =        60;
 C_FROOT =       61;
 C_IROOT =       62;
 C_IPWR3N =      63;
 C_IPWR4N =      64;
 C_IPWR5N =      65;
 C_IPWR6N =      66;
 C_IPWR7N =      67;
 C_IPWR8N =      68;
 C_IPWR9N =      69;
 C_IPWR10N =     70;
 C_IPWR11N =     71;
 C_IPWR12N =     72;
 C_IPWR13N =     73;
 C_IPWR14N =     74;
 C_IPWR15N =     75;
 C_IPWR16N =     76;
 C_FPWRDVR =     77;
 C_SQRN =        78;
 C_SIGN =        79;
 C_FRAC =        80;
 C_LDSV =        81;
 C_SUMV =        82;
 C_SUMQV =       83;
 C_AVRV   =      84;
 C_LENV   =      85;
 C_MAXV  =       86;
 C_MINV =        87;
 C_NORMV =       88;
 C_LDSD =        89;
 C_SUMD =        90;
 C_SUMQD =       91;
 C_AVRD   =      92;
 C_LEND   =      93;
 C_MAXD  =       94;
 C_MIND =        95;
 C_NORMD =       96;
 C_SUM =         97;
 C_SUMQ =        98;
 C_AVR   =       99;
 C_LEN   =       100;
 C_MAX  =        101;
 C_MIN =         102;
 C_NORM =        103;
 C_IF   =        104;
 C_IFNOP =       105;
 C_JMP =         106;
 C_FLD =         107;
 C_LDA =         108;
 C_COMPILED =    109;
 C_PROD =        110;
 C_PRODV =       111;
 C_PRODD =       112;
 C_FACT2 =       113;
 C_MAX2 =        114;
 C_MIN2 =        115;
 C_SQRT2 =       116;
 C_GOTO =        117;
 C_AVR2 =        118;
 C_INC =         119;
 C_DEC =         120;
 C_RAND =        121;    //algorithm mt19937 (Mersenne Twister)
 C_RANDPLE =     122;    //algorithm taus113 (Pierre L'Ecuyer)
 C_SAVEVX =      123;
 C_SAVEV =       124;
 C_SETLENX =     125;
 C_SETLEN  =     126;
 C_SEC  =        127;
 C_COSEC  =      128;
 C_ARCSEC   =    129;
 C_ARCCOSEC   =  130;
 C_SECH  =       131;
 C_COSECH  =     132;
 C_ARCSECH   =   133;
 C_ARCCOSECH   = 134;
 C_SUMP        = 135;
 C_SUMPS       = 136;
 C_SUMPX       = 137;
 C_SUMPSX      = 138;
 C_DEV         = 139;
 C_DEVV        = 140;
 C_DEVS        = 141;
 C_DEVSV       = 142;
 C_GAMMA       = 143;
 C_BESSELJ0    = 144;
 C_BESSELJ1    = 145;
 C_BESSELJN    = 146;
 C_RAD         = 147;
 C_DEG         = 148;
 C_DEVD        = 149;
 C_DEVSD       = 150;
 C_LEGENDRE    = 151;
 C_LEGENDREP   = 152;
 C_LEGENDREPA  = 153;
 C_BESSELY0    = 154;
 C_BESSELY1    = 155;
 C_BESSELYN    = 156;
 C_MOD         = 157;
 C_POLY        = 158;
 C_POLYDIFF    = 159;
 C_POLYDIFFN   = 160;
 C_SPHBESSELJN = 161;
 C_ARCTAN2 =     162;         {.209}
 C_RANDXOR =     163;    //algorithm xor4096()     {.209}
 C_RANDG  =      164;          {.209}
 C_INTRANDMT =   165;
 C_RAND01MT  =   166;
 C_RAND01PLE  =  167;
 C_INTRANDPLE =  168;
 C_RAND01XOR  =  169;
 C_INTRANDXOR =  170;
 C_RANDOMIZE  =  171;
 C_RANDARRAYD =  172;
 C_RANDMT     =  173;
 C_PSI        =  174;
 C_PSID1      =  175;
 C_PSID2      =  176;
 C_PSID3      =  177;
 C_ZETA       =  178;
 C_CHEBYSHEV_T = 179;
 C_CHEBYSHEV_U = 180;
 C_GCDD        = 181;
 C_GCD         = 182;
 C_LCMD        = 183;
 C_LCM         = 184;
 C_GCDV        = 185;
 C_LCMV        = 186;
 C_ERF         = 187;
 C_ERFC        = 188;
 C_IPWRSGNF    = 189;       {.219}
 C_HERMITE     = 190;
 C_LAGUERRE    = 191;
 C_LAGUERREP   = 192;
 C_LAGUERREPA  = 193;
 C_MAX3 =        194;
 C_MIN3 =        195;
 C_NORM2 =       196;
 C_NORM3 =       197;
 C_AVR3 =        198;
 C_F2PWR =       199;
 C_LDSVA =       200;
 C_TRY   =       201;   {.248}
 C_EXCEPTION  =  202;    {.248}




  LFuncList: Integer = 203;



 {const
 c05dln2,c1dln2,c05dln10,c1dln10: extended;
 const c2: single = 2;
 const c05: single = 0.5;
 const c1: single = 1;
 const c025: single = 0.25;
 const c3: single = 3;
 const c4: single = 4;
 const c6: single = 6;
 const cs2: single = -2;

 c05dln2:=0.5/ln(2);
 c05dln10:=0.5/ln(10);
 c1dln2:=1/ln(2);
 c1dln10:=1/ln(10);
 }







 
//procedure CreateInternalFuncList(var InternalFuncList: TArrayInternalFunc);
//procedure CreateExternalFuncList(var ExternalFuncList: TArrayExternalFunc);
procedure CreateExchTable(FuncList: TArrayFuncList; var XchTable: TArray2I);
procedure CreateFuncList(var FuncList: TArrayFuncList);
//procedure AddSpecialFunction;



procedure R_FPWR;
procedure R_SINH;
procedure R_EXP;
procedure R_LogN;
procedure R_ARCSIN;
procedure R_ARCCOS;
procedure R_ARCSEC;
procedure R_ARCCOSEC;
procedure R_COSH;
procedure R_SECH;
procedure R_COSECH;
procedure R_TANH;
procedure R_COTANH;
procedure R_ARCSINH;
procedure R_ARCCOSH;
procedure R_ARCTANH;
procedure R_ARCCOTANH;
procedure R_ARCSECH;
procedure R_ARCCOSECH;
procedure R_IPWR;
procedure R_IPWR_Correct;
procedure R_IPWRSGN;
procedure R_IPWRSGNF;
procedure R_FPWR2F;
procedure R_FPWR2I;
procedure R_IROOT;
procedure R_FROOT;
procedure R_FACT;
procedure R_FACT2;
procedure R_SIGN;
procedure RD_VEC;
procedure RE_VEC;
procedure RS_VEC;
procedure RI_VEC;
procedure RD_SUM;
procedure RE_SUM;
procedure RS_SUM;
procedure RI_SUM;
procedure RD_SUMQ;
procedure RE_SUMQ;
procedure RS_SUMQ;
procedure RI_SUMQ;
procedure RI_NORM;
procedure RD_NORM;
procedure RS_NORM;
procedure RE_NORM;
procedure RI_AVR;
procedure RS_AVR;
procedure RD_AVR;
procedure RE_AVR;
procedure R_LEN;
procedure RS_MIN;
procedure RI_MIN;
procedure RD_MIN;
procedure RE_MIN;
procedure RS_MAX;
procedure RI_MAX;
procedure RD_MAX;
procedure RE_MAX;
procedure RD_PROD;
procedure RE_PROD;
procedure RS_PROD;
procedure RI_PROD;
procedure RD_SUMP;
procedure RD_SUMPS;
procedure RE_SUMP;
procedure RE_SUMPS;
procedure RD_DEV;
procedure RD_DEVS;
procedure RE_DEV;
procedure RE_DEVS;
procedure RS_DEV;
procedure RS_DEVS;
procedure RI_DEV;
procedure RI_DEVS;
procedure RS_SUMP;
procedure RS_SUMPS;
procedure RI_SUMP;
procedure RI_SUMPS;
procedure RD_POLY;
procedure RS_POLY;
procedure RE_POLY;
procedure RI_POLY;
procedure RD_POLYDIFF1;
procedure RD_POLYDIFF2;
procedure RD_POLYDIFF3;
procedure RE_POLYDIFF1;
procedure RE_POLYDIFF2;
procedure RE_POLYDIFF3;
procedure RS_POLYDIFF1;
procedure RS_POLYDIFF2;
procedure RS_POLYDIFF3;
procedure RI_POLYDIFF1;
procedure RI_POLYDIFF2;
procedure RI_POLYDIFF3;

procedure R_AVR(adr,len: Cardinal); stdcall;
procedure R_MAX(adr,len: Cardinal); stdcall;
procedure R_MIN(adr,len: Cardinal); stdcall;
procedure R_SUM(adr,len: Cardinal); stdcall;
procedure R_SUMQ(adr,len: Cardinal); stdcall;
procedure R_NORM(adr,len: Cardinal); stdcall;
procedure R_PROD(adr,len: Cardinal); stdcall;
procedure R_DEV(adr,len: Cardinal); stdcall;
procedure R_DEVS(adr,len: Cardinal); stdcall;
procedure R_GCD(adr,len: Cardinal); stdcall;
procedure R_LCM(adr,len: Cardinal); stdcall;
procedure RD_GCD; assembler;
procedure V_GCDD(VD: TArrayD); stdcall;
procedure RE_GCD; assembler;
procedure V_GCDE(VD: TArrayE); stdcall;
procedure RI_GCD; assembler;
procedure V_GCDI(VD: TArrayI); stdcall;
procedure RS_GCD; assembler;
procedure V_GCDS(VD: TArrayS); stdcall;

procedure RD_LCM; assembler;
procedure V_LCMD(VD: TArrayD); stdcall;
procedure RE_LCM; assembler;
procedure V_LCME(VD: TArrayE); stdcall;
procedure RI_LCM; assembler;
procedure V_LCMI(VD: TArrayI); stdcall;
procedure RS_LCM; assembler;
procedure V_LCMS(VD: TArrayS); stdcall;

{.199}
{procedure _MAX2;
procedure _MIN2; }

function std_rand2(x1,x2:PExtended): extended;  stdcall;
procedure RAND_ARRAYD(vd: TArrayD); stdcall;
procedure _Randomize;  stdcall;
function _RandG(Mean, StdDev: PExtended): Extended;  stdcall;

procedure InitRandomGenerator;


procedure SetLengthArray(AdrV: Pointer; TypeV: Cardinal; Len: Cardinal);  stdcall;
function  PolyDiffD(va: PArrayD; x: PExtended;  p: Pinteger): extended;    stdcall;
function  PolyDiffE(va: PArrayE; x: PExtended;  p: Pinteger): extended;    stdcall;
function  PolyDiffS(va: PArrayS; x: PExtended;  p: Pinteger): extended;    stdcall;
function  PolyDiffI(va: PArrayI; x: PExtended;  p: Pinteger): extended;    stdcall;
function PolyDiffN(pvec: Pointer; px: PExtended;  p: Pinteger; VType: PInteger): extended;  stdcall;






procedure _LSE(A: TArray2E;  B: TArrayE; var B1: TArrayE; var D: TFloatType);
procedure _Gausse(A: TArray2E; B: TArrayE; var B1: TArrayE; var D: TFloatType);
procedure _Jordan(A: TArray2E; B: TArrayE; var B1: TArrayE; var D: TFloatType);
procedure _CreateSplain3(MX: TArrayE; MF: TArrayE; var idf: cardinal; var  E: integer);
procedure _XCHSC(i,j: TIntegType; M: TArray2E; B: TArrayE; var E: Integer);



function _FSPL3(x: PExtended; id: integer): TFloatType;   stdcall;
function _FSPL3D1(x: PExtended; id: integer): TFloatType; stdcall;
function _FSPL3D2(x: PExtended; id: integer): TFloatType; stdcall;
function _FSPL3D3(x: PExtended; id: integer): TFloatType; stdcall;
function _FSPL3D4(x: PExtended; id: integer): TFloatType; stdcall;

{
procedure MaskFPUException; assembler;
procedure IsFPUException; assembler;
}
procedure MaskFPUExceptionA; assembler;
procedure IsFPUExceptionA; assembler;


var
G_TS,G_TC: TArrayF_2; //   / (-  Matrix )
F_Clm,F_Str: Integer; // 
_LSEType: Integer;
GSPL3: array of TSplain;
S_Error: integer;
F_FastSpec: Boolean;
//A3,A2,A1,A0,PS,Pk0,Pk: TArrayE;



implementation
var
MEM: Integer;
MEMX: Extended;
CWMEM: Cardinal;






{************************************begin********************************************************}

procedure AddCode1(B1: Byte; FL: TArrayFuncList; Code: Integer);
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+1);
FL[Code].inlCode[High(FL[Code].inlCode)]:=B1;
end;

procedure AddCode2(B1,B2: Byte; FL: TArrayFuncList; Code: Integer);
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+2);
FL[Code].inlCode[High(FL[Code].inlCode)-1]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)]:=B2;
end;


procedure AddCode3(B1,B2,B3: Byte; FL: TArrayFuncList; Code: Integer);
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+3);
FL[Code].inlCode[High(FL[Code].inlCode)-2]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)-1]:=B2;
FL[Code].inlCode[High(FL[Code].inlCode)]:=B3;
end;

procedure AddCode2A(B1,B2: Byte; Adr: TAddress; FL: TArrayFuncList; Code: Integer);
var
P: ^TAddress;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+6);
FL[Code].inlCode[High(FL[Code].inlCode)-5]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)-4]:=B2;
P:=@FL[Code].inlCode[High(FL[Code].inlCode)-3];
P^:=Adr;
end;


procedure FSCALE(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$FD,FL,Code);
end;


procedure FCHS(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$E0,FL,Code);
end;


procedure FILD(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode2A($DB,$05,Adr,FL,Code);
end;


procedure FDIVR1(Code: Integer; FL: TArrayFuncList);
begin
AddCode2A($DC,$3D,TAddress(@C1),FL,Code);
end;


procedure FLDLN2(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$ED,FL,Code);
end;

procedure FLDLG2(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$EC,FL,Code);
end;

procedure FXCH1(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$C9,FL,Code);
end;

procedure FYL2X(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$F1,FL,Code);
end;


procedure FMUL0(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D8,$C8,FL,Code);
end;

procedure FLD0(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$C0,FL,Code);
end;


procedure FMULP1(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DE,$C9,FL,Code);
end;


procedure FABS(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$E1,FL,Code);
end;

procedure FPTAN(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$F2,FL,Code);
end;

procedure FPATAN(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$F3,FL,Code);
end;

procedure FDIVRP(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DE,$F1,FL,Code);
end;

procedure FLD1(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$E8,FL,Code);
end;


procedure FRNDINT(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($D9,$FC,FL,Code);
end;

procedure FSTP0(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DD,$D8,FL,Code);
end;

procedure FSUBRP(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DE,$E1,FL,Code);
end;


procedure FSUBP(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DE,$E9,FL,Code);
end;

procedure LEA_CC8A(Code: Integer; FL: TArrayFuncList);
begin
AddCode3($8D,$0C,$C1,FL,Code);
end;


procedure LEA_AC2A(Code: Integer; FL: TArrayFuncList);
begin
AddCode3($8D,$04,$41,FL,Code);
end;


procedure FLD(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode2A($DB,$2D,Adr,FL,Code);
end;


procedure FLDA_E(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DB,$28,FL,Code);
end;



procedure FLDCW(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode2A($D9,$2D,Adr,FL,Code);
end;



procedure FNCLEX(Code: Integer; FL: TArrayFuncList);
begin
AddCode2($DB,$E2,FL,Code);
end;


procedure PUSHCW(Code: Integer; FL: TArrayFuncList);
begin
//sub esp,4
AddCode3($83,$EC,$04,FL,Code);
//fnstcw [esp];
AddCode3($D9,$3C,$24,FL,Code);
end;



procedure POPCW(Code: Integer; FL: TArrayFuncList);
begin
//fldcw [esp];
AddCode3($D9,$2C,$24,FL,Code);
//add esp, 4
AddCode3($83,$C4,$04,FL,Code);
end;


procedure MOVNC(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
var
P: ^Integer;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+5);
FL[Code].inlCode[High(FL[Code].inlCode)-4]:=$B9;
P:=@FL[Code].inlCode[High(FL[Code].inlCode)-3];
P^:=Adr;
end;

procedure MOVAM(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode2A($89,$05,Adr,FL,Code);
end;


{*******************************end*************************************************************}

(*
for DELPHI:

procedure AddCode(B1: Byte; FL: TArrayFuncList; Code: Integer); overload;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+1);
FL[Code].inlCode[High(FL[Code].inlCode)]:=B1;
end;

procedure AddCode(B1,B2: Byte; FL: TArrayFuncList; Code: Integer); overload;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+2);
FL[Code].inlCode[High(FL[Code].inlCode)-1]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)]:=B2;
end;


procedure AddCode(B1,B2,B3: Byte; FL: TArrayFuncList; Code: Integer);  overload;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+3);
FL[Code].inlCode[High(FL[Code].inlCode)-2]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)-1]:=B2;
FL[Code].inlCode[High(FL[Code].inlCode)]:=B3;
end;

procedure AddCode(B1,B2: Byte; Adr: TAddress; FL: TArrayFuncList; Code: Integer); overload;
var
P: ^TAddress;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+6);
FL[Code].inlCode[High(FL[Code].inlCode)-5]:=B1;
FL[Code].inlCode[High(FL[Code].inlCode)-4]:=B2;
P:=@FL[Code].inlCode[High(FL[Code].inlCode)-3];
P^:=Adr;
end;


procedure FSCALE(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$FD,FL,Code);
end;


procedure FCHS(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$E0,FL,Code);
end;


procedure FILD(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode($DB,$05,Adr,FL,Code);
end;


procedure FDIVR1(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DC,$3D,TAddress(@C1),FL,Code);
end;


procedure FLDLN2(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$ED,FL,Code);
end;

procedure FLDLG2(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$EC,FL,Code);
end;

procedure FXCH1(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$C9,FL,Code);
end;

procedure FYL2X(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$F1,FL,Code);
end;


procedure FMUL0(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D8,$C8,FL,Code);
end;

procedure FLD0(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$C0,FL,Code);
end;


procedure FMULP1(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DE,$C9,FL,Code);
end;


procedure FABS(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$E1,FL,Code);
end;

procedure FPTAN(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$F2,FL,Code);
end;

procedure FPATAN(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$F3,FL,Code);
end;

procedure FDIVRP(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DE,$F1,FL,Code);
end;

procedure FLD1(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$E8,FL,Code);
end;


procedure FRNDINT(Code: Integer; FL: TArrayFuncList);
begin
AddCode($D9,$FC,FL,Code);
end;

procedure FSTP0(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DD,$D8,FL,Code);
end;

procedure FSUBRP(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DE,$E1,FL,Code);
end;


procedure FSUBP(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DE,$E9,FL,Code);
end;

procedure LEA_CC8A(Code: Integer; FL: TArrayFuncList);
begin
AddCode($8D,$0C,$C1,FL,Code);
end;


procedure LEA_AC2A(Code: Integer; FL: TArrayFuncList);
begin
AddCode($8D,$04,$41,FL,Code);
end;


procedure FLD(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode($DB,$2D,Adr,FL,Code);
end;


procedure FLDA_E(Code: Integer; FL: TArrayFuncList);
begin
AddCode($DB,$28,FL,Code);
end;



procedure FLDCW(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode($D9,$2D,Adr,FL,Code);
end;


procedure MOVNC(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
var
P: ^Integer;
begin
SetLength(FL[Code].inlCode,Length(FL[Code].inlCode)+5);
FL[Code].inlCode[High(FL[Code].inlCode)-4]:=$B9;
P:=@FL[Code].inlCode[High(FL[Code].inlCode)-3];
P^:=Adr;
end;

procedure MOVAM(Code: Integer; FL: TArrayFuncList; Adr: TAddress);
begin
AddCode($89,$05,Adr,FL,Code);
end;
*)

procedure _LN(FL: TArrayFuncList);
begin
FL[C_LN].Name[1]:='ln';
FL[C_LN].arg:=1;
FL[C_LN].FType:=2;
FL[C_LN].NST:=1;
FL[C_LN].inl:=1;
FLDLN2(C_LN,FL);
FXCH1(C_LN,FL);
FYL2X(C_LN,FL);
end;


procedure _LOG10(FL: TArrayFuncList);
begin
SetLength(FL[C_LOG10].Name,3);
FL[C_LOG10].Name[1]:='log10';
FL[C_LOG10].Name[2]:='lg';
FL[C_LOG10].arg:=1;
FL[C_LOG10].FType:=2;
FL[C_LOG10].NST:=1;
FL[C_LOG10].inl:=1;
FLDLG2(C_LOG10,FL);
FXCH1(C_LOG10,FL);
FYL2X(C_LOG10,FL);
end;


procedure _LOG2(FL: TArrayFuncList);
begin
SetLength(FL[C_LOG2].Name,3);
FL[C_LOG2].Name[1]:='log2';
FL[C_LOG2].Name[2]:='lg2';
FL[C_LOG2].arg:=1;
FL[C_LOG2].FType:=2;
FL[C_LOG2].NST:=1;
FL[C_LOG2].inl:=1;
FLD1(C_LOG2,FL);
FXCH1(C_LOG2,FL);
FYL2X(C_LOG2,FL);
end;


procedure _SIN(FL: TArrayFuncList);
begin
FL[C_SIN].Name[1]:='sin';
FL[C_SIN].arg:=1;
FL[C_SIN].FType:=2;
FL[C_SIN].NST:=0;
FL[C_SIN].inl:=1;
SetLength(FL[C_SIN].inlCode,2);
FL[C_SIN].inlCode[0]:=$D9;  FL[C_SIN].inlCode[1]:=$FE;
end;

procedure _COS(FL: TArrayFuncList);
begin
FL[C_COS].Name[1]:='cos';
FL[C_COS].arg:=1;
FL[C_COS].FType:=2;
FL[C_COS].NST:=0;
FL[C_COS].inl:=1;
SetLength(FL[C_COS].inlCode,2);
FL[C_COS].inlCode[0]:=$D9; FL[C_COS].inlCode[1]:=$FF;
end;


procedure _COSEC(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
SetLength(FL[C_COSEC].Name,3);
FL[C_COSEC].Name[1]:='cosec';
FL[C_COSEC].Name[2]:='csc';
FL[C_COSEC].arg:=1;
FL[C_COSEC].FType:=2;
FL[C_COSEC].NST:=0;
FL[C_COSEC].inl:=1;
SetLength(FL[C_COSEC].inlCode,8);
//fsin:
FL[C_COSEC].inlCode[0]:=$D9;  FL[C_COSEC].inlCode[1]:=$FE;
//fdivr
FL[C_COSEC].inlCode[2]:=$DC; FL[C_COSEC].inlCode[3]:=$3D;
//addr(C1)
P_C:=@FL[C_COSEC].inlCode[4]; P_C^:=Cardinal(@C1);
end;


procedure _SEC(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_SEC].Name[1]:='sec';
FL[C_SEC].arg:=1;
FL[C_SEC].FType:=2;
FL[C_SEC].NST:=0;
FL[C_SEC].inl:=1;
SetLength(FL[C_SEC].inlCode,8);
//fsin:
FL[C_SEC].inlCode[0]:=$D9;  FL[C_SEC].inlCode[1]:=$FF;
//fdivr
FL[C_SEC].inlCode[2]:=$DC; FL[C_SEC].inlCode[3]:=$3D;
//addr(C1)
P_C:=@FL[C_SEC].inlCode[4]; P_C^:=Cardinal(@C1);
end;



procedure _RAD(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_RAD].Name[1]:='rad';
FL[C_RAD].arg:=1;
FL[C_RAD].FType:=2;
FL[C_RAD].NST:=1;
FL[C_RAD].inl:=1;
SetLength(FL[C_RAD].inlCode,8);
//fld tbyte ptr []
FL[C_RAD].inlCode[0]:=$DB;  FL[C_RAD].inlCode[1]:=$2D;
//addr(rPI)
P_C:=@FL[C_RAD].inlCode[2]; P_C^:=Cardinal(@rPI);
//fmulp
FL[C_RAD].inlCode[6]:=$DE; FL[C_RAD].inlCode[7]:=$C9;
end;



procedure _DEG(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_DEG].Name[1]:='deg';
FL[C_DEG].arg:=1;
FL[C_DEG].FType:=2;
FL[C_DEG].NST:=1;
FL[C_DEG].inl:=1;
SetLength(FL[C_DEG].inlCode,8);
//fld tbyte ptr []
FL[C_DEG].inlCode[0]:=$DB;  FL[C_DEG].inlCode[1]:=$2D;
//addr(rPI)
P_C:=@FL[C_DEG].inlCode[2]; P_C^:=Cardinal(@dPI);
//fmulp
FL[C_DEG].inlCode[6]:=$DE; FL[C_DEG].inlCode[7]:=$C9;
end;




procedure _SQR(FL: TArrayFuncList);
begin
FL[C_SQR].Name[1]:='sqr';
FL[C_SQR].arg:=1;
FL[C_SQR].FType:=2;
FL[C_SQR].NST:=0;
FL[C_SQR].inl:=1;
FMUL0(C_SQR,FL);
end;

procedure _SQRN(FL: TArrayFuncList);
begin
FL[C_SQRN].Name[1]:='@@sqrn';
FL[C_SQRN].arg:=1;
FL[C_SQRN].FType:=2;
FL[C_SQRN].NST:=1;
FL[C_SQRN].inl:=1;
FMUL0(C_SQRN,FL);

FLD1(C_SQRN,FL);
FDIVRP(C_SQRN,FL);
end;

procedure _SQRT(FL: TArrayFuncList);
begin
FL[C_SQRT].Name[1]:='sqrt';
FL[C_SQRT].arg:=1;
FL[C_SQRT].FType:=2;
FL[C_SQRT].NST:=0;
FL[C_SQRT].inl:=1;
SetLength(FL[C_SQRT].inlCode,2);
FL[C_SQRT].inlCode[0]:=$D9;   FL[C_SQRT].inlCode[1]:=$FA;
end;


procedure _SQRT2(FL: TArrayFuncList);
begin
FL[C_SQRT2].Name[1]:='sqrt2';
FL[C_SQRT2].arg:=1;
FL[C_SQRT2].FType:=2;
FL[C_SQRT2].NST:=0;
FL[C_SQRT2].inl:=1;
SetLength(FL[C_SQRT2].inlCode,4);
FL[C_SQRT2].inlCode[0]:=$D9;   FL[C_SQRT2].inlCode[1]:=$FA;
FL[C_SQRT2].inlCode[2]:=$D9;   FL[C_SQRT2].inlCode[3]:=$FA;
end;


procedure _COTAN(FL: TArrayFuncList);
begin
SetLength(FL[C_COTAN].Name,4);
FL[C_COTAN].Name[1]:='cotan';
FL[C_COTAN].Name[2]:='ctg';
FL[C_COTAN].Name[3]:='cot';
FL[C_COTAN].arg:=1;
FL[C_COTAN].FType:=2;
FL[C_COTAN].NST:=1;
FL[C_COTAN].inl:=1;
FPTAN(C_COTAN,FL);
FDIVRP(C_COTAN,FL);
end;


procedure _TAN(FL: TArrayFuncList);
begin
SetLength(FL[C_TAN].Name,3);
FL[C_TAN].Name[1]:='tan';
FL[C_TAN].Name[2]:='tg';
FL[C_TAN].arg:=1;
FL[C_TAN].FType:=2;
FL[C_TAN].NST:=1;
FL[C_TAN].inl:=1;
FPTAN(C_TAN,FL);
FSTP0(C_TAN,FL);
end;




procedure _ARCTAN(FL: TArrayFuncList);
begin
SetLength(FL[C_ARCTAN].Name,5);
FL[C_ARCTAN].Name[1]:='arctan';
FL[C_ARCTAN].Name[2]:='arctg';
FL[C_ARCTAN].Name[3]:='atg';
FL[C_ARCTAN].Name[4]:='atan';
FL[C_ARCTAN].arg:=1;
FL[C_ARCTAN].FType:=2;
FL[C_ARCTAN].NST:=1;
FL[C_ARCTAN].inl:=1;
FLD1(C_ARCTAN,FL);
FPATAN(C_ARCTAN,FL);
end;



procedure _ARCTAN2(FL: TArrayFuncList);
begin
SetLength(FL[C_ARCTAN2].Name,5);
FL[C_ARCTAN2].Name[1]:='arctan2';
FL[C_ARCTAN2].Name[2]:='arctg2';
FL[C_ARCTAN2].Name[3]:='atg2';
FL[C_ARCTAN2].Name[4]:='atan2';
FL[C_ARCTAN2].arg:=2;
FL[C_ARCTAN2].FType:=4;
FL[C_ARCTAN2].NST:=1;
FL[C_ARCTAN2].inl:=1;
//FLD1(C_ARCTAN2,FL);
FXCH1(C_ARCTAN2,FL);
FPATAN(C_ARCTAN2,FL);
end;



procedure _ARCCOTAN(FL: TArrayFuncList);
begin
SetLength(FL[C_ARCCOTAN].Name,6);
FL[C_ARCCOTAN].Name[1]:='arccotan';
FL[C_ARCCOTAN].Name[2]:='arcctg';
FL[C_ARCCOTAN].Name[3]:='actg';
FL[C_ARCCOTAN].Name[4]:='arccot';
FL[C_ARCCOTAN].Name[5]:='acot';
FL[C_ARCCOTAN].arg:=1;
FL[C_ARCCOTAN].FType:=2;
FL[C_ARCCOTAN].NST:=1;
FL[C_ARCCOTAN].inl:=1;
FLD1(C_ARCCOTAN,FL);
FPATAN(C_ARCCOTAN,FL);
FLD(C_ARCCOTAN,FL,TAddress(@E_CPid2));
FSUBRP(C_ARCCOTAN,FL);
end;


procedure _ABS(FL: TArrayFuncList);
begin
FL[C_ABS].Name[1]:='abs';
FL[C_ABS].arg:=1;
FL[C_ABS].FType:=2;
FL[C_ABS].inl:=1;
FABS(C_ABS,FL);
end;


procedure _TRUNC(FL: TArrayFuncList);
begin
FL[C_TRUNC].Name[1]:='trunc';
FL[C_TRUNC].arg:=1;
FL[C_TRUNC].FType:=2;
FL[C_TRUNC].inl:=1;
PUSHCW(C_TRUNC,FL);      {.248}
FNCLEX(C_TRUNC,FL);      {.248}
FLDCW(C_TRUNC,FL,TAddress(@FCW1));
FRNDINT(C_TRUNC,FL);
POPCW(C_TRUNC,FL);{.248}//FLDCW(C_TRUNC,FL,TAddress(@FCW2));
end;


procedure _FRAC(FL: TArrayFuncList);
begin
FL[C_FRAC].Name[1]:='frac';
FL[C_FRAC].arg:=1;
FL[C_FRAC].FType:=2;
FL[C_FRAC].inl:=1;
FL[C_FRAC].NST:=1;
FLD0(C_FRAC,FL);
PUSHCW(C_FRAC,FL);      {.248}
FNCLEX(C_FRAC,FL);      {.248}
FLDCW(C_FRAC,FL,TAddress(@FCW1));
FRNDINT(C_FRAC,FL);
POPCW(C_FRAC,FL);      {.248}//FLDCW(C_FRAC,FL,TAddress(@FCW2));
FSUBP(C_FRAC,FL);
end;



procedure _FACT(FL: TArrayFuncList);
begin
FL[C_FACT].Name[1]:='fact';
FL[C_FACT].arg:=1;
FL[C_FACT].FType:=1;
FL[C_FACT].inl:=1;
FL[C_FACT].NST:=1;
FL[C_FACT].SaveReg:=5; //EAX+ECX
MOVNC(C_FACT,FL,TAddress(@Factorial[0]));
LEA_CC8A(C_FACT,FL);
LEA_AC2A(C_FACT,FL);
FLDA_E(C_FACT,FL);

FL[C_FACT].Addr:=@R_FACT;//  flGet
end;



procedure _FACT2(FL: TArrayFuncList);
begin
FL[C_FACT2].Name[1]:='fact2';
FL[C_FACT2].arg:=1;
FL[C_FACT2].FType:=1;
FL[C_FACT2].inl:=1;
FL[C_FACT2].NST:=1;
FL[C_FACT2].SaveReg:=5; //EAX+ECX
MOVNC(C_FACT2,FL,TAddress(@Factorial2F[0]));
LEA_CC8A(C_FACT2,FL);
LEA_AC2A(C_FACT2,FL);
FLDA_E(C_FACT2,FL);

FL[C_FACT2].Addr:=@R_FACT2;//  flGet
(*
FL[C_FACT2].Name[1]:='fact2';
FL[C_FACT2].arg:=1;
FL[C_FACT2].FType:=1;
FL[C_FACT2].inl:=0;
FL[C_FACT2].NST:=1;
FL[C_FACT2].SaveReg:=5; //EAX+ECX
FL[C_FACT2].Addr:=@R_FACT2;
*)
end;


procedure _AVR2(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_AVR2].Name[1]:='avr2';
FL[C_AVR2].arg:=2;
FL[C_AVR2].FType:=4;
FL[C_AVR2].NST:=0;
FL[C_AVR2].inl:=1;
SetLength(FL[C_AVR2].inlCode,8);
//faddp:
FL[C_AVR2].inlCode[0]:=$DE; FL[C_AVR2].inlCode[1]:=$C1;
//fmul
FL[C_AVR2].inlCode[2]:=$DC; FL[C_AVR2].inlCode[3]:=$0D;
//addr(C05)
P_C:=@FL[C_AVR2].inlCode[4]; P_C^:=Cardinal(@C05);
end;


{.199}
procedure _MAX2(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_MAX2].Name[1]:='max2';
FL[C_MAX2].arg:=2;
FL[C_MAX2].FType:=4;
FL[C_MAX2].NST:=0;
FL[C_MAX2].inl:=1;


{.243}
SetLength(FL[C_MAX2].inlCode,6);
//fcomi st(0), st(1)
FL[C_MAX2].inlCode[0]:=$DB; FL[C_MAX2].inlCode[1]:=$F1;
//fcmovb st(0), st(1)
FL[C_MAX2].inlCode[2]:=$DA; FL[C_MAX2].inlCode[3]:=$C1;
//fstp  st(1)
FL[C_MAX2].inlCode[4]:=$DD; FL[C_MAX2].inlCode[5]:=$D9;


(*
FL[C_MAX2].SaveReg:=1;

SetLength(FL[C_MAX2].inlCode,11);
//fcom st(1)
FL[C_MAX2].inlCode[0]:=$D8; FL[C_MAX2].inlCode[1]:=$D1;
//fstsw ax
FL[C_MAX2].inlCode[2]:=$DF; FL[C_MAX2].inlCode[3]:=$E0;
//sahf
FL[C_MAX2].inlCode[4]:=$9E;
//jb $2
FL[C_MAX2].inlCode[5]:=$72; FL[C_MAX2].inlCode[6]:=$02;
//fxch st(1)
FL[C_MAX2].inlCode[7]:=$D9; FL[C_MAX2].inlCode[8]:=$C9;
//fstp st(0)
FL[C_MAX2].inlCode[9]:=$DD; FL[C_MAX2].inlCode[10]:=$D8;
*)
end;



{.199}
procedure _MIN2(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_MIN2].Name[1]:='min2';
FL[C_MIN2].arg:=2;
FL[C_MIN2].FType:=4;
FL[C_MIN2].NST:=0;
FL[C_MIN2].inl:=1;


{.243}
SetLength(FL[C_MIN2].inlCode,6);
//fcomi st(0), st(1)
FL[C_MIN2].inlCode[0]:=$DB; FL[C_MIN2].inlCode[1]:=$F1;
//fcmovnb st(0), st(1)
FL[C_MIN2].inlCode[2]:=$DB; FL[C_MIN2].inlCode[3]:=$C1;
//fstp  st(1)
FL[C_MIN2].inlCode[4]:=$DD; FL[C_MIN2].inlCode[5]:=$D9;


{
FL[C_MIN2].SaveReg:=1;

SetLength(FL[C_MIN2].inlCode,11);
//fcom st(1)
FL[C_MIN2].inlCode[0]:=$D8; FL[C_MIN2].inlCode[1]:=$D1;
//fstsw ax
FL[C_MIN2].inlCode[2]:=$DF; FL[C_MIN2].inlCode[3]:=$E0;
//sahf
FL[C_MIN2].inlCode[4]:=$9E;
//jb $2
FL[C_MIN2].inlCode[5]:=$73; FL[C_MIN2].inlCode[6]:=$02;
//fxch st(1)
FL[C_MIN2].inlCode[7]:=$D9; FL[C_MIN2].inlCode[8]:=$C9;
//fstp st(0)
FL[C_MIN2].inlCode[9]:=$DD; FL[C_MIN2].inlCode[10]:=$D8;
}
end;



procedure _NORM2(FL: TArrayFuncList);
begin
FL[C_NORM2].Name[1]:='norm2';
FL[C_NORM2].arg:=2;
FL[C_NORM2].FType:=4;
FL[C_NORM2].NST:=0;
FL[C_NORM2].inl:=1;
SetLength(FL[C_NORM2].inlCode,10);
//fmul st(0),st(0):
FL[C_NORM2].inlCode[0]:=$D8; FL[C_NORM2].inlCode[1]:=$C8;
//fxch st(0),st(1)
FL[C_NORM2].inlCode[2]:=$D9; FL[C_NORM2].inlCode[3]:=$C9;
//fmul st(0),st(0):
FL[C_NORM2].inlCode[4]:=$D8; FL[C_NORM2].inlCode[5]:=$C8;
//faddp st(1)
FL[C_NORM2].inlCode[6]:=$DE; FL[C_NORM2].inlCode[7]:=$C1;
//fsqrt
FL[C_NORM2].inlCode[8]:=$D9; FL[C_NORM2].inlCode[9]:=$FA;
end;


{.243}
procedure _MAX3(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_MAX3].Name[1]:='max3';
FL[C_MAX3].arg:=3;
FL[C_MAX3].FType:=-8;
FL[C_MAX3].ArgType:=_Extended;
FL[C_MAX3].NST:=0;
FL[C_MAX3].inl:=1;


SetLength(FL[C_MAX3].inlCode,12);
//fcomi st(0), st(1)
FL[C_MAX3].inlCode[0]:=$DB; FL[C_MAX3].inlCode[1]:=$F1;
//fcmovb st(0), st(1)
FL[C_MAX3].inlCode[2]:=$DA; FL[C_MAX3].inlCode[3]:=$C1;
//fstp  st(1)
FL[C_MAX3].inlCode[4]:=$DD; FL[C_MAX3].inlCode[5]:=$D9;
//fcomi st(0), st(1)
FL[C_MAX3].inlCode[6]:=$DB; FL[C_MAX3].inlCode[7]:=$F1;
//fcmovb st(0), st(1)
FL[C_MAX3].inlCode[8]:=$DA; FL[C_MAX3].inlCode[9]:=$C1;
//fstp  st(1)
FL[C_MAX3].inlCode[10]:=$DD; FL[C_MAX3].inlCode[11]:=$D9;

end;




{.243}
procedure _MIN3(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_MIN3].Name[1]:='min3';
FL[C_MIN3].arg:=3;
FL[C_MIN3].FType:=-8;
FL[C_MIN3].ArgType:=_Extended;
FL[C_MIN3].NST:=0;
FL[C_MIN3].inl:=1;


SetLength(FL[C_MIN3].inlCode,12);
//fcomi st(0), st(1)
FL[C_MIN3].inlCode[0]:=$DB; FL[C_MIN3].inlCode[1]:=$F1;
//fcmovnb st(0), st(1)
FL[C_MIN3].inlCode[2]:=$DB; FL[C_MIN3].inlCode[3]:=$C1;
//fstp  st(1)
FL[C_MIN3].inlCode[4]:=$DD; FL[C_MIN3].inlCode[5]:=$D9;
//fcomi st(0), st(1)
FL[C_MIN3].inlCode[6]:=$DB; FL[C_MIN3].inlCode[7]:=$F1;
//fcmovnb st(0), st(1)
FL[C_MIN3].inlCode[8]:=$DB; FL[C_MIN3].inlCode[9]:=$C1;
//fstp  st(1)
FL[C_MIN3].inlCode[10]:=$DD; FL[C_MIN3].inlCode[11]:=$D9;

end;



procedure _NORM3(FL: TArrayFuncList);
begin
FL[C_NORM3].Name[1]:='norm3';
FL[C_NORM3].arg:=3;
FL[C_NORM3].FType:=-8;
FL[C_NORM3].ArgType:=_Extended;
FL[C_NORM3].NST:=0;
FL[C_NORM3].inl:=1;


SetLength(FL[C_NORM3].inlCode,16);
//fmul st(0),st(0):
FL[C_NORM3].inlCode[0]:=$D8; FL[C_NORM3].inlCode[1]:=$C8;
//fxch st(0),st(1)
FL[C_NORM3].inlCode[2]:=$D9; FL[C_NORM3].inlCode[3]:=$C9;
//fmul st(0),st(0):
FL[C_NORM3].inlCode[4]:=$D8; FL[C_NORM3].inlCode[5]:=$C8;
//fxch st(0),st(2)
FL[C_NORM3].inlCode[6]:=$D9; FL[C_NORM3].inlCode[7]:=$CA;
//fmul st(0),st(0):
FL[C_NORM3].inlCode[8]:=$D8; FL[C_NORM3].inlCode[9]:=$C8;
//faddp st(1)
FL[C_NORM3].inlCode[10]:=$DE; FL[C_NORM3].inlCode[11]:=$C1;
//faddp st(1)
FL[C_NORM3].inlCode[12]:=$DE; FL[C_NORM3].inlCode[13]:=$C1;
//fsqrt
FL[C_NORM3].inlCode[14]:=$D9; FL[C_NORM3].inlCode[15]:=$FA;
end;




procedure _AVR3(FL: TArrayFuncList);
var
P_C:^Cardinal;
begin
FL[C_AVR3].Name[1]:='avr3';
FL[C_AVR3].arg:=3;
FL[C_AVR3].FType:=-8;
FL[C_AVR3].ArgType:=_Extended;
FL[C_AVR3].NST:=0;
FL[C_AVR3].inl:=1;



SetLength(FL[C_AVR3].inlCode,10);
//faddp:
FL[C_AVR3].inlCode[0]:=$DE; FL[C_AVR3].inlCode[1]:=$C1;
//faddp:
FL[C_AVR3].inlCode[2]:=$DE; FL[C_AVR3].inlCode[3]:=$C1;
//fdiv
FL[C_AVR3].inlCode[4]:=$DC; FL[C_AVR3].inlCode[5]:=$35;
//addr(C3)
P_C:=@FL[C_AVR3].inlCode[6]; P_C^:=Cardinal(@C3);
end;




procedure _IPWRSGN(FL: TArrayFuncList);
begin
FL[C_IPWRSGN].Name[1]:='@@IPWRSGN';
FL[C_IPWRSGN].arg:=1;
FL[C_IPWRSGN].FType:=1;
FL[C_IPWRSGN].inl:=1;
FL[C_IPWRSGN].NST:=1;
FL[C_IPWRSGN].SaveReg:=1;

FLD1(C_IPWRSGN,FL);
//and eax,1
SetLength(FL[C_IPWRSGN].inlCode,Length(FL[C_IPWRSGN].inlCode)+3);
FL[C_IPWRSGN].inlCode[High(FL[C_IPWRSGN].inlCode)-2]:=$83;
FL[C_IPWRSGN].inlCode[High(FL[C_IPWRSGN].inlCode)-1]:=$E0;
FL[C_IPWRSGN].inlCode[High(FL[C_IPWRSGN].inlCode)]:=$01;
//jz +$2
SetLength(FL[C_IPWRSGN].inlCode,Length(FL[C_IPWRSGN].inlCode)+2);
FL[C_IPWRSGN].inlCode[High(FL[C_IPWRSGN].inlCode)-1]:=$74;
FL[C_IPWRSGN].inlCode[High(FL[C_IPWRSGN].inlCode)]:=$02;

FCHS(C_IPWRSGN,FL);
end;


procedure _FPWRDVR(FL: TArrayFuncList);
begin
FL[C_FPWRDVR].Name[1]:='@@FPWRDVR';
FL[C_FPWRDVR].arg:=1;
FL[C_FPWRDVR].FType:=2;
FL[C_FPWRDVR].NST:=1;
FL[C_FPWRDVR].inl:=1;

FLD1(C_FPWRDVR,FL);
FDIVRP(C_FPWRDVR,FL);
end;


procedure _I2PWR(FL: TArrayFuncList);  //2^integer
var
P_C:^Cardinal;
begin
FL[C_I2PWR].Name[1]:='@@I2PWR';
FL[C_I2PWR].arg:=1;
FL[C_I2PWR].FType:=1;

FL[C_I2PWR].inl:=1;
FL[C_I2PWR].NST:=1;//fscale=2
FL[C_I2PWR].SaveReg:=1;

{
        SUB     ESP,16
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [ESP],   0
        MOV     DWORD PTR [ESP+4], 80000000H
        MOV     DWORD PTR [ESP+8], EAX
        FLD     TBYTE PTR [ESP]
        ADD     ESP,16
}
SetLength(FL[C_I2PWR].inlCode,33);

// SUB     ESP, 16
FL[C_I2PWR].inlCode[0]:=$83; FL[C_I2PWR].inlCode[1]:=$EC; FL[C_I2PWR].inlCode[2]:=$10;

// ADD     EAX, 00003FFFH
FL[C_I2PWR].inlCode[3]:=$05; FL[C_I2PWR].inlCode[4]:=$FF;
FL[C_I2PWR].inlCode[5]:=$3F; FL[C_I2PWR].inlCode[6]:=$00; FL[C_I2PWR].inlCode[7]:=$00;

// MOV     DWORD PTR [ESP],  0
FL[C_I2PWR].inlCode[8]:=$C7;  FL[C_I2PWR].inlCode[9]:=$04;   FL[C_I2PWR].inlCode[10]:=$24;
FL[C_I2PWR].inlCode[11]:=$00; FL[C_I2PWR].inlCode[12]:=$00;
FL[C_I2PWR].inlCode[13]:=$00; FL[C_I2PWR].inlCode[14]:=$00;

// MOV     DWORD PTR [ESP+4], 0
FL[C_I2PWR].inlCode[15]:=$C7; FL[C_I2PWR].inlCode[16]:=$44;  FL[C_I2PWR].inlCode[17]:=$24;  FL[C_I2PWR].inlCode[18]:=$04;
FL[C_I2PWR].inlCode[19]:=$00; FL[C_I2PWR].inlCode[20]:=$00;
FL[C_I2PWR].inlCode[21]:=$00; FL[C_I2PWR].inlCode[22]:=$80;

// MOV     DWORD PTR [ESP+8], EAX
FL[C_I2PWR].inlCode[23]:=$89; FL[C_I2PWR].inlCode[24]:=$44;  FL[C_I2PWR].inlCode[25]:=$24; FL[C_I2PWR].inlCode[26]:=$08;

//  FLD     TBYTE PTR [ESP]
FL[C_I2PWR].inlCode[27]:=$DB; FL[C_I2PWR].inlCode[28]:=$2C;   FL[C_I2PWR].inlCode[29]:=$24;

// ADD     ESP, 16
FL[C_I2PWR].inlCode[30]:=$83; FL[C_I2PWR].inlCode[31]:=$C4; FL[C_I2PWR].inlCode[32]:=$10;


{
// MOV     DWORD PTR [MEMX+4], 0
FL[C_I2PWR].inlCode[15]:=$C7; FL[C_I2PWR].inlCode[16]:=$05;
P_C:=@FL[C_I2PWR].inlCode[17]; P_C^:=Cardinal(@MEMX)+4;
FL[C_I2PWR].inlCode[21]:=$00; FL[C_I2PWR].inlCode[22]:=$00;
FL[C_I2PWR].inlCode[23]:=$00; FL[C_I2PWR].inlCode[24]:=$80;
}

{
MOVAM(C_I2PWR,FL,TAddress(@MEM));
FILD(C_I2PWR,FL,TAddress(@MEM));
FLD1(C_I2PWR,FL);
FSCALE(C_I2PWR,FL);
FXCH1(C_I2PWR,FL);
FSTP0(C_I2PWR,FL);
}
end;





procedure _IPWR3(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR3
else Code:=C_IPWR3N;

FL[Code].NST:=1;

FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;


procedure _IPWR4(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR4
else Code:=C_IPWR4N;

FL[Code].NST:=0;

FMUL0(Code,FL);
FMUL0(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;


procedure _IPWR5(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR5
else Code:=C_IPWR5N;

FL[Code].NST:=1;

FLD0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;


procedure _IPWR6(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR6
else Code:=C_IPWR6N;

FL[Code].NST:=1;

FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR7(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR7
else Code:=C_IPWR7N;

FL[Code].NST:=2;

FLD0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR8(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR8
else Code:=C_IPWR8N;

FL[Code].NST:=0;

FMUL0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR9(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR9
else Code:=C_IPWR9N;

FL[Code].NST:=1;

FLD0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR10(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR10
else Code:=C_IPWR10N;

FL[Code].NST:=1;

FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR11(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR11
else Code:=C_IPWR11N;

FL[Code].NST:=2;

FLD0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR12(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR12
else Code:=C_IPWR12N;

FL[Code].NST:=1;

FMUL0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);


if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR13(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR13
else Code:=C_IPWR13N;

FL[Code].NST:=2;

FLD0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);
FMULP1(Code,FL);

if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR14(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR14
else Code:=C_IPWR14N;

FL[Code].NST:=2;

FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);
FMULP1(Code,FL);


if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR15(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR15
else Code:=C_IPWR15N;

FL[Code].NST:=3;

FLD0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FLD0(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);
FMULP1(Code,FL);
FMUL0(Code,FL);
FMULP1(Code,FL);


if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWR16(FL: TArrayFuncList; N: Integer);
var
Code: Integer;
begin
if N = 0 then Code:=C_IPWR16
else Code:=C_IPWR16N;

FL[Code].NST:=0;


FMUL0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);
FMUL0(Code,FL);


if N = 1 then FDIVR1(Code,FL);
end;



procedure _IPWRC(FL: TArrayFuncList);
var
Code,CodeN,i: Integer;
S: String;
begin
S:='@@IPWR'; Code:=C_IPWR3-1; CodeN:=C_IPWR3N-1;
for i:=3 to 16 do
begin
 Code:=Code+1;
 FL[Code].Name[1]:=S+IntToStr(i);
 FL[Code].arg:=1;
 FL[Code].FType:=2;
 FL[Code].inl:=1;
 CodeN:=CodeN+1;
 FL[CodeN].Name[1]:=S+IntToStr(i)+'N';
 FL[CodeN].arg:=1;
 FL[CodeN].FType:=2;
 FL[CodeN].inl:=1;
end;

_IPWR3(FL,0);  _IPWR3(FL,1);
_IPWR4(FL,0);  _IPWR4(FL,1);
_IPWR5(FL,0);  _IPWR5(FL,1);
_IPWR6(FL,0);  _IPWR6(FL,1);
_IPWR7(FL,0);  _IPWR7(FL,1);
_IPWR8(FL,0);  _IPWR8(FL,1);
_IPWR9(FL,0);  _IPWR9(FL,1);
_IPWR10(FL,0); _IPWR10(FL,1);
_IPWR11(FL,0); _IPWR11(FL,1);
_IPWR12(FL,0); _IPWR12(FL,1);
_IPWR13(FL,0); _IPWR13(FL,1);
_IPWR14(FL,0); _IPWR14(FL,1);
_IPWR15(FL,0); _IPWR15(FL,1);
_IPWR16(FL,0); _IPWR16(FL,1);
end;




procedure CreateFuncList(var FuncList: TArrayFuncList);
var
i: Integer;
begin
//  ()  ;  ,  @@    

SetLength(FuncList,LFuncList+1);
for i:=1 to  High(FuncList) do
begin
 FuncList[i].NST:=0;
 SetLength(FuncList[i].Name,2);
 FuncList[i].arg:=0;
 FuncList[i].Addr:=nil;
 FuncList[i].inl:=0;
 FuncList[i].inlCode:=nil;
 FuncList[i].FType:=0;
 FuncList[i].Id_Func:=0;
 FuncList[i].CLST:=0;
 FuncList[i].Arrow:=0;
 FuncList[i].Set_ID:=0;
 FuncList[i].ArgType:=0;
 FuncList[i].CalcConst:=_Yes;
 FuncList[i].Enabled:=_Yes;
 FuncList[i].RetNum:=_Yes;
 FuncList[i].ReplFunc:=_No;
 FuncList[i].SaveReg:=0;      {!!!}
 FuncList[i].XchArg:=_No;
end;

FuncList[C_LDSC].Name[1]:='@@LDSC';
FuncList[C_LDSC].FType:=0;

FuncList[C_LDSF].Name[1]:='@@LDSF';
FuncList[C_LDSF].FType:=0;

FuncList[C_LDSI].Name[1]:='@@LDSI';
FuncList[C_LDSI].FType:=0;

FuncList[C_LDSV].Name[1]:='@@LDSV';
FuncList[C_LDSV].FType:=0;
FuncList[C_LDSV].Arg:=1;

FuncList[C_LDSVA].Name[1]:='@@LDSVA';    {arraycpu}{.245} //address of array
FuncList[C_LDSVA].FType:=0;
FuncList[C_LDSVA].RetNum:=_No;
FuncList[C_LDSVA].Arg:=1;

FuncList[C_FAVB].Name[1]:='@@FAVB';
FuncList[C_FAVB].FType:=0;

FuncList[C_IAVB].Name[1]:='@@IAVB';
FuncList[C_IAVB].FType:=0;

FuncList[C_AIFB].Name[1]:='@@AIFB';
FuncList[C_AIFB].FType:=0;

FuncList[C_AAI].Name[1]:='@@AAI';
FuncList[C_AAI].FType:=0;

FuncList[C_AAF].Name[1]:='@@AAF';
FuncList[C_AAF].FType:=0;

FuncList[C_AFI].Name[1]:='@@AFI';
FuncList[C_AFI].FType:=0;

FuncList[C_MAF].Name[1]:='@@MAF';
FuncList[C_MAF].FType:=0;

FuncList[C_MUL].Name[1]:='@@MUL';
FuncList[C_MUL].FType:=0;

FuncList[C_ADD].Name[1]:='@@ADD';
FuncList[C_ADD].FType:=0;

FuncList[C_NOP].Name[1]:='@@NOP';
FuncList[C_NOP].FType:=0;
FuncList[C_NOP].RetNum:=_No;

FuncList[C_PACKAGE].Name[1]:='@@PACKAGE';
FuncList[C_PACKAGE].FType:=0;
FuncList[C_PACKAGE].RetNum:=_NO;   {.203}

FuncList[C_INC].Name[1]:='inc';
FuncList[C_INC].FType:=-17;
FuncList[C_INC].arg:=1;
FuncList[C_INC].RetNum:=_NO;

FuncList[C_DEC].Name[1]:='dec';
FuncList[C_DEC].FType:=-17;
FuncList[C_DEC].arg:=1;
FuncList[C_DEC].RetNum:=_NO;

FuncList[C_SAVEVX].Name[1]:='savev';     {.199}
FuncList[C_SAVEVX].FType:=-17;
FuncList[C_SAVEVX].arg:=3;
FuncList[C_SAVEVX].RetNum:=_NO;

FuncList[C_SAVEV].Name[1]:='@@SAVEV';
FuncList[C_SAVEV].FType:=-17;
FuncList[C_SAVEV].arg:=1;
FuncList[C_SAVEV].RetNum:=_No;






FuncList[C_SUMV].Name[1]:='@@sumv';
FuncList[C_SUMV].FType:=5;
FuncList[C_SUMV].arg:=1;
FuncList[C_SUMV].NST:=2;
FuncList[C_SUMV].ReplFunc:=_YES;

FuncList[C_SUMQV].Name[1]:='@@sumqv';
FuncList[C_SUMQV].FType:=5;
FuncList[C_SUMQV].arg:=1;
FuncList[C_SUMQV].NST:=2;
FuncList[C_SUMQV].ReplFunc:=_YES;

FuncList[C_AVRV].Name[1]:='@@avrv';
FuncList[C_AVRV].FType:=5;
FuncList[C_AVRV].arg:=1;
FuncList[C_AVRV].NST:=2;
FuncList[C_AVRV].ReplFunc:=_YES;

FuncList[C_NORMV].Name[1]:='@@normv';
FuncList[C_NORMV].FType:=5;
FuncList[C_NORMV].arg:=1;
FuncList[C_NORMV].NST:=2;
FuncList[C_NORMV].ReplFunc:=_YES;

FuncList[C_MAXV].Name[1]:='@@maxv';
FuncList[C_MAXV].FType:=5;
FuncList[C_MAXV].arg:=1;
FuncList[C_MAXV].NST:=2;
FuncList[C_MAXV].ReplFunc:=_YES;

FuncList[C_MINV].Name[1]:='@@minv';
FuncList[C_MINV].FType:=5;
FuncList[C_MINV].arg:=1;
FuncList[C_MINV].NST:=2;
FuncList[C_MINV].ReplFunc:=_YES;

FuncList[C_LENV].Name[1]:='@@lenv';
FuncList[C_LENV].FType:=5;
FuncList[C_LENV].arg:=1;
FuncList[C_LENV].NST:=2;

FuncList[C_PRODV].Name[1]:='@@prodv';
FuncList[C_PRODV].FType:=5;
FuncList[C_PRODV].arg:=1;
FuncList[C_PRODV].NST:=2;
FuncList[C_PRODV].ReplFunc:=_YES;

FuncList[C_DEVV].Name[1]:='@@devv';
FuncList[C_DEVV].FType:=5;
FuncList[C_DEVV].arg:=1;
FuncList[C_DEVV].NST:=3;
FuncList[C_DEVV].ReplFunc:=_YES;

FuncList[C_DEVSV].Name[1]:='@@devsv';
FuncList[C_DEVSV].FType:=5;
FuncList[C_DEVSV].arg:=1;
FuncList[C_DEVSV].NST:=3;
FuncList[C_DEVSV].ReplFunc:=_YES;

FuncList[C_GCDV].Name[1]:='@@gcdv';
FuncList[C_GCDV].FType:=5;
FuncList[C_GCDV].arg:=1;
FuncList[C_GCDV].NST:=3;
FuncList[C_GCDV].ReplFunc:=_YES;

FuncList[C_LCMV].Name[1]:='@@lcmv';
FuncList[C_LCMV].FType:=5;
FuncList[C_LCMV].arg:=1;
FuncList[C_LCMV].NST:=3;
FuncList[C_LCMV].ReplFunc:=_YES;




FuncList[C_SUMD].Name[1]:='@@sumd';
FuncList[C_SUMD].FType:=-14;
FuncList[C_SUMD].ArgType:=_Extended;
FuncList[C_SUMD].NST:=2;
FuncList[C_SUMD].FType:=-14;
FuncList[C_SUMD].SaveReg:=15;             {195}
FuncList[C_SUMD].Addr:=@R_SUM;

FuncList[C_SUMQD].Name[1]:='@@sumqd';
FuncList[C_SUMQD].FType:=-14;
FuncList[C_SUMQD].ArgType:=_Extended;
FuncList[C_SUMQD].NST:=2;
FuncList[C_SUMQD].SaveReg:=15;
FuncList[C_SUMQD].Addr:=@R_SUMQ;

FuncList[C_AVRD].Name[1]:='@@avrd';
FuncList[C_AVRD].FType:=-14;
FuncList[C_AVRD].ArgType:=_Extended;
FuncList[C_AVRD].NST:=2;
FuncList[C_AVRD].SaveReg:=15;
FuncList[C_AVRD].Addr:=@R_AVR;

FuncList[C_NORMD].Name[1]:='@@normd';
FuncList[C_NORMD].FType:=-14;
FuncList[C_NORMD].ArgType:=_Extended;
FuncList[C_NORMD].NST:=2;
FuncList[C_NORMD].SaveReg:=15;
FuncList[C_NORMD].Addr:=@R_NORM;

FuncList[C_MAXD].Name[1]:='@@maxd';
FuncList[C_MAXD].FType:=-14;
FuncList[C_MAXD].ArgType:=_Extended;
FuncList[C_MAXD].NST:=2;
FuncList[C_MAXD].SaveReg:=15;
FuncList[C_MAXD].Addr:=@R_MAX;

FuncList[C_MIND].Name[1]:='@@mind';
FuncList[C_MIND].FType:=-14;
FuncList[C_MIND].ArgType:=_Extended;
FuncList[C_MIND].NST:=2;
FuncList[C_MIND].SaveReg:=15;
FuncList[C_MIND].Addr:=@R_MIN;

FuncList[C_PRODD].Name[1]:='@@prodd';
FuncList[C_PRODD].FType:=-14;
FuncList[C_PRODD].ArgType:=_Extended;
FuncList[C_PRODD].NST:=2;
FuncList[C_PRODD].SaveReg:=15;
FuncList[C_PRODD].Addr:=@R_PROD;

FuncList[C_DEVD].Name[1]:='@@devd';
FuncList[C_DEVD].FType:=-14;
FuncList[C_DEVD].ArgType:=_Extended;
FuncList[C_DEVD].NST:=2;
FuncList[C_DEVD].SaveReg:=15;
FuncList[C_DEVD].Addr:=@R_DEV;

FuncList[C_DEVSD].Name[1]:='@@devsd';
FuncList[C_DEVSD].FType:=-14;
FuncList[C_DEVSD].ArgType:=_Extended;
FuncList[C_DEVSD].NST:=2;
FuncList[C_DEVSD].SaveReg:=15;
FuncList[C_DEVSD].Addr:=@R_DEVS;


FuncList[C_GCDD].Name[1]:='@@GCD';
FuncList[C_GCDD].FType:=-14;
FuncList[C_GCDD].ArgType:=_Extended;
FuncList[C_GCDD].NST:=2;
FuncList[C_GCDD].SaveReg:=15;
FuncList[C_GCDD].Addr:=@R_GCD;

FuncList[C_LCMD].Name[1]:='@@LCM';
FuncList[C_LCMD].FType:=-14;
FuncList[C_LCMD].ArgType:=_Extended;
FuncList[C_LCMD].NST:=2;
FuncList[C_LCMD].SaveReg:=15;
FuncList[C_LCMD].Addr:=@R_LCM;




FuncList[C_SUM].Name[1]:='sum';
FuncList[C_SUM].FType:=-17;

FuncList[C_SUMQ].Name[1]:='sumq';
FuncList[C_SUMQ].FType:=-17;

FuncList[C_DEV].Name[1]:='dev';
FuncList[C_DEV].FType:=-17;


FuncList[C_DEVS].Name[1]:='devs';
FuncList[C_DEVS].FType:=-17;

FuncList[C_GCD].Name[1]:='gcd';
FuncList[C_GCD].FType:=-17;

FuncList[C_LCM].Name[1]:='lcm';
FuncList[C_LCM].FType:=-17;



SetLength(FuncList[C_AVR].Name,3);
FuncList[C_AVR].Name[1]:='avr';
FuncList[C_AVR].Name[2]:='avg';
FuncList[C_AVR].FType:=-17;

FuncList[C_NORM].Name[1]:='norm';
FuncList[C_NORM].FType:=-17;


FuncList[C_MAX].Name[1]:='max';
FuncList[C_MAX].FType:=-17;


FuncList[C_MIN].Name[1]:='min';
FuncList[C_MIN].FType:=-17;

FuncList[C_LEN].Name[1]:='len';
FuncList[C_LEN].FType:=-17;

FuncList[C_PROD].Name[1]:='prod';
FuncList[C_PROD].FType:=-17;

FuncList[C_SUMP].Name[1]:='@@sump';
FuncList[C_SUMP].FType:=-17;
FuncList[C_SUMP].NST:=3;
FuncList[C_SUMP].ReplFunc:=_YES;

FuncList[C_SUMPS].Name[1]:='@@sumps';
FuncList[C_SUMPS].NST:=4;
FuncList[C_SUMPS].FType:=-17;
FuncList[C_SUMPS].ReplFunc:=_YES;

FuncList[C_SUMPX].Name[1]:='sump';
FuncList[C_SUMPX].Arg:=2;
FuncList[C_SUMPX].FType:=-17;

FuncList[C_SUMPSX].Name[1]:='sumps';
FuncList[C_SUMPSX].Arg:=3;
FuncList[C_SUMPSX].FType:=-17;

FuncList[C_MOD].Name[1]:='mod';
FuncList[C_MOD].FType:=-17;
FuncList[C_MOD].arg:=2;
FuncList[C_MOD].ReplFunc:=_YES;

SetLength(FuncList[C_POLY].Name,3);
FuncList[C_POLY].Name[1]:='polynom';
FuncList[C_POLY].Name[2]:='poly';
FuncList[C_POLY].FType:=-17;
FuncList[C_POLY].arg:=2;
FuncList[C_POLY].ReplFunc:=_YES;


FuncList[C_POLYDIFF].Name[1]:='polynomdiff';
FuncList[C_POLYDIFF].FType:=-17;
FuncList[C_POLYDIFF].arg:=3;
FuncList[C_POLYDIFF].ReplFunc:=_YES;


FuncList[C_POLYDIFFN].Name[1]:='@@polydiffn';
FuncList[C_POLYDIFFN].Addr:=@PolyDiffN;
FuncList[C_POLYDIFFN].FType:=-15;
FuncList[C_POLYDIFFN].arg:=4;
FuncList[C_POLYDIFFN].ReplFunc:=_YES;
FuncList[C_POLYDIFFN].CLST:=0;
FuncList[C_POLYDIFFN].Arrow:=0;
SetLength(FuncList[C_POLYDIFFN].ArgTypeList,4);
FuncList[C_POLYDIFFN].ArgTypeList[0]:=_ArrayE{_Integer};     {.245}    {arraycpu}
FuncList[C_POLYDIFFN].ArgTypeList[1]:=_Extended;
FuncList[C_POLYDIFFN].ArgTypeList[2]:=_Integer;
FuncList[C_POLYDIFFN].ArgTypeList[3]:=_Integer;
FuncList[C_POLYDIFFN].ArgType:=_differ;
FuncList[C_POLYDIFFN].NST:=3;
FuncList[C_POLYDIFFN].SaveReg:=15;


(*
FuncList[C_SETLEN].Name[1]:='@@setlen';
FuncList[C_SETLEN].FType:=-10;
FuncList[C_SETLEN].ArgType:=_INTEGER;
FuncList[C_SETLEN].SaveReg:=255;            {!!!}
FuncList[C_SETLEN].NST:=1;
FuncList[C_SETLEN].arg:=3;
FuncList[C_SETLEN].Addr:=@SetLengthArray;
FuncList[C_SETLEN].RetNum:=_No;
FuncList[C_SETLEN].CalcConst:=_No;
FuncList[C_SETLEN].CLST:=0;
FuncList[C_SETLEN].Arrow:=0;
*)


_SIN(FuncList);
_COS(FuncList);
_LN(FuncList);
_SQR(FuncList);
_SQRT(FuncList);
_COTAN(FuncList);
_TAN(FuncList);
_ARCTAN(FuncList);
_ARCCOTAN(FuncList);
_LOG10(FuncList);
_LOG2(FuncList);
_ABS(FuncList);
_TRUNC(FuncList);
_FRAC(FuncList);
_FACT(FuncList);
_FACT2(FuncList);
_IPWRSGN(FuncList);
_FPWRDVR(FuncList);
//_I2PWR(FuncList);
_IPWRC(FuncList);
_SQRN(FuncList);
_SQRT2(FuncList);
_AVR2(FuncList);
_COSEC(FuncList);
_SEC(FuncList);
_DEG(FuncList);
_RAD(FuncList);
_MAX2(FuncList);     {.199}
_MIN2(FuncList);     {.199}
_ARCTAN2(FuncList); {.209}
_MAX3(FuncList);     {.243}
_MIN3(FuncList);     {.243}
_NORM2(FuncList);     {.243}
_NORM3(FuncList);     {.243}
_AVR3(FuncList);     {.243}


FuncList[C_EXP].Name[1]:='exp';
FuncList[C_EXP].Addr:=@R_EXP;
FuncList[C_EXP].arg:=1;
FuncList[C_EXP].FType:=2;
FuncList[C_EXP].NST:=2;



SetLength(FuncList[C_ARCSIN].Name,3);
FuncList[C_ARCSIN].Name[1]:='arcsin';
FuncList[C_ARCSIN].Name[2]:='asin';
FuncList[C_ARCSIN].Addr:=@R_ARCSIN;
FuncList[C_ARCSIN].arg:=1;
FuncList[C_ARCSIN].FType:=2;
FuncList[C_ARCSIN].NST:=2;



SetLength(FuncList[C_ARCCOS].Name,3);
FuncList[C_ARCCOS].Name[1]:='arccos';
FuncList[C_ARCCOS].Name[2]:='acos';
FuncList[C_ARCCOS].Addr:=@R_ARCCOS;
FuncList[C_ARCCOS].arg:=1;
FuncList[C_ARCCOS].FType:=2;
FuncList[C_ARCCOS].NST:=2;


SetLength(FuncList[C_ARCSEC].Name,3);
FuncList[C_ARCSEC].Name[1]:='arcsec';
FuncList[C_ARCSEC].Name[2]:='asec';
FuncList[C_ARCSEC].Addr:=@R_ARCSEC;
FuncList[C_ARCSEC].arg:=1;
FuncList[C_ARCSEC].FType:=2;
FuncList[C_ARCSEC].NST:=2;


SetLength(FuncList[C_ARCCOSEC].Name,4);
FuncList[C_ARCCOSEC].Name[1]:='arccosec';
FuncList[C_ARCCOSEC].Name[2]:='acosec';
FuncList[C_ARCCOSEC].Name[3]:='acsc';
FuncList[C_ARCCOSEC].Addr:=@R_ARCCOSEC;
FuncList[C_ARCCOSEC].arg:=1;
FuncList[C_ARCCOSEC].FType:=2;
FuncList[C_ARCCOSEC].NST:=2;


SetLength(FuncList[C_SINH].Name,3);
FuncList[C_SINH].Name[1]:='sinh';
FuncList[C_SINH].Name[2]:='sh';
FuncList[C_SINH].Addr:=@R_SINH;
FuncList[C_SINH].arg:=1;
FuncList[C_SINH].FType:=2;
FuncList[C_SINH].NST:=2;


SetLength(FuncList[C_COSH].Name,3);
FuncList[C_COSH].Name[1]:='cosh';
FuncList[C_COSH].Name[2]:='ch';
FuncList[C_COSH].Addr:=@R_COSH;
FuncList[C_COSH].arg:=1;
FuncList[C_COSH].FType:=2;
FuncList[C_COSH].NST:=2;


SetLength(FuncList[C_TANH].Name,3);
FuncList[C_TANH].Name[1]:='tanh';
FuncList[C_TANH].Name[2]:='th';
FuncList[C_TANH].Addr:=@R_TANH;
FuncList[C_TANH].arg:=1;
FuncList[C_TANH].FType:=2;
FuncList[C_TANH].NST:=2;


SetLength(FuncList[C_COTANH].Name,4);
FuncList[C_COTANH].Name[1]:='cotanh';
FuncList[C_COTANH].Name[2]:='cth';
FuncList[C_COTANH].Name[3]:='coth';
FuncList[C_COTANH].Addr:=@R_COTANH;
FuncList[C_COTANH].arg:=1;
FuncList[C_COTANH].FType:=2;
FuncList[C_COTANH].NST:=2;


SetLength(FuncList[C_ARCSINH].Name,5);
FuncList[C_ARCSINH].Name[1]:='arcsinh';
FuncList[C_ARCSINH].Name[2]:='arcsh';
FuncList[C_ARCSINH].Name[3]:='asinh';
FuncList[C_ARCSINH].Name[4]:='ash';
FuncList[C_ARCSINH].Addr:=@R_ARCSINH;
FuncList[C_ARCSINH].arg:=1;
FuncList[C_ARCSINH].FType:=2;
FuncList[C_ARCSINH].NST:=2;


SetLength(FuncList[C_ARCCOSH].Name,5);
FuncList[C_ARCCOSH].Name[1]:='arccosh';
FuncList[C_ARCCOSH].Name[2]:='arcch';
FuncList[C_ARCCOSH].Name[3]:='acosh';
FuncList[C_ARCCOSH].Name[4]:='ach';
FuncList[C_ARCCOSH].Addr:=@R_ARCCOSH;
FuncList[C_ARCCOSH].arg:=1;
FuncList[C_ARCCOSH].FType:=2;
FuncList[C_ARCCOSH].NST:=2;



SetLength(FuncList[C_ARCTANH].Name,5);
FuncList[C_ARCTANH].Name[1]:='arctanh';
FuncList[C_ARCTANH].Name[2]:='arcth';
FuncList[C_ARCTANH].Name[3]:='ath';
FuncList[C_ARCTANH].Name[4]:='atanh';
FuncList[C_ARCTANH].Addr:=@R_ARCTANH;
FuncList[C_ARCTANH].arg:=1;
FuncList[C_ARCTANH].FType:=2;
FuncList[C_ARCTANH].NST:=2;



SetLength(FuncList[C_ARCCOTANH].Name,5);
FuncList[C_ARCCOTANH].Name[1]:='arccotanh';
FuncList[C_ARCCOTANH].Name[2]:='arccth';
FuncList[C_ARCCOTANH].Name[3]:='acth';
FuncList[C_ARCCOTANH].Name[4]:='acoth';
FuncList[C_ARCCOTANH].Addr:=@R_ARCCOTANH;
FuncList[C_ARCCOTANH].arg:=1;
FuncList[C_ARCCOTANH].FType:=2;
FuncList[C_ARCCOTANH].NST:=2;


FuncList[C_SECH].Name[1]:='sech';
FuncList[C_SECH].Addr:=@R_SECH;
FuncList[C_SECH].arg:=1;
FuncList[C_SECH].FType:=2;
FuncList[C_SECH].NST:=2;

SetLength(FuncList[C_COSECH].Name,3);
FuncList[C_COSECH].Name[1]:='cosech';
FuncList[C_COSECH].Name[2]:='csch';
FuncList[C_COSECH].Addr:=@R_COSECH;
FuncList[C_COSECH].arg:=1;
FuncList[C_COSECH].FType:=2;
FuncList[C_COSECH].NST:=2;


SetLength(FuncList[C_ARCCOSECH].Name,4);
FuncList[C_ARCCOSECH].Name[1]:='arccosech';
FuncList[C_ARCCOSECH].Name[2]:='arccsch';
FuncList[C_ARCCOSECH].Name[3]:='acsch';
FuncList[C_ARCCOSECH].Addr:=@R_ARCCOSECH;
FuncList[C_ARCCOSECH].arg:=1;
FuncList[C_ARCCOSECH].FType:=2;
FuncList[C_ARCCOSECH].NST:=2;


SetLength(FuncList[C_ARCSECH].Name,3);
FuncList[C_ARCSECH].Name[1]:='arcsech';
FuncList[C_ARCSECH].Name[2]:='asech';
FuncList[C_ARCSECH].Addr:=@R_ARCSECH;
FuncList[C_ARCSECH].arg:=1;
FuncList[C_ARCSECH].FType:=2;
FuncList[C_ARCSECH].NST:=2;


SetLength(FuncList[C_RAND].Name,4);    //     random01mt, intrandommt,  randommt
FuncList[C_RAND].Name[1]:='random';
FuncList[C_RAND].Name[2]:='rand';
FuncList[C_RAND].Name[3]:='rnd';
FuncList[C_RAND].FType:=-11;


SetLength(FuncList[C_RANDMT].Name,4);
FuncList[C_RANDMT].Name[1]:='randommt';
FuncList[C_RANDMT].Name[2]:='randmt';
FuncList[C_RANDMT].Name[3]:='rndmt';
FuncList[C_RANDMT].arg:=2;
FuncList[C_RANDMT].FType:=-15;
FuncList[C_RANDMT].ArgType:=_Double;
FuncList[C_RANDMT].NST:=3;
FuncList[C_RANDMT].CalcConst:=_No;
FuncList[C_RANDMT].ReplFunc:=_No;
FuncList[C_RANDMT].SaveReg:=15;
FuncList[C_RANDMT].Addr:=@MT_RAND2;


SetLength(FuncList[C_INTRANDMT].Name,4);
FuncList[C_INTRANDMT].Name[1]:='intrandommt';
FuncList[C_INTRANDMT].Name[2]:='irandmt';
FuncList[C_INTRANDMT].Name[3]:='irndmt';
FuncList[C_INTRANDMT].ADDR:=@MT_intrand;
FuncList[C_INTRANDMT].CLST:=0;
FuncList[C_INTRANDMT].Arrow:=0;
FuncList[C_INTRANDMT].ArgType:=_Integer;
FuncList[C_INTRANDMT].arg:=1;
FuncList[C_INTRANDMT].NST:=3;   //need to ceck!!!
FuncList[C_INTRANDMT].FType:=-10;
FuncList[C_INTRANDMT].SaveReg:=15;            {!!!}
FuncList[C_INTRANDMT].ReplFunc:=_No;
FuncList[C_INTRANDMT].CalcConst:=_No;


SetLength(FuncList[C_RAND01MT].Name,4);
FuncList[C_RAND01MT].Name[1]:='random01mt';
FuncList[C_RAND01MT].Name[2]:='rand01mt';
FuncList[C_RAND01MT].Name[3]:='rnd01mt';
FuncList[C_RAND01MT].ADDR:=@MT_rand01;
FuncList[C_RAND01MT].CLST:=0;
FuncList[C_RAND01MT].Arrow:=0;
FuncList[C_RAND01MT].ArgType:=_Integer;
FuncList[C_RAND01MT].arg:=0;
FuncList[C_RAND01MT].NST:=2;   //need to ceck!!!
FuncList[C_RAND01MT].FType:=-11;  //eax
FuncList[C_RAND01MT].SaveReg:=15;
FuncList[C_RAND01MT].ReplFunc:=_No;
FuncList[C_RAND01MT].CalcConst:=_No;


SetLength(FuncList[C_RANDPLE].Name,5);     {!!!}
FuncList[C_RANDPLE].Name[1]:='randomple';
FuncList[C_RANDPLE].Name[2]:='randf';
FuncList[C_RANDPLE].Name[3]:='rndple';
FuncList[C_RANDPLE].Name[4]:='randple';
FuncList[C_RANDPLE].arg:=2;
FuncList[C_RANDPLE].FType:=4;
FuncList[C_RANDPLE].ArgType:=_Double;
FuncList[C_RANDPLE].NST:=2;
FuncList[C_RANDPLE].CalcConst:=_No;
FuncList[C_RANDPLE].ReplFunc:=_No;
FuncList[C_RANDPLE].SaveReg:=15;            {!!!}
FuncList[C_RANDPLE].Addr:=@PLE_RAND2_K;


SetLength(FuncList[C_INTRANDPLE].Name,4);
FuncList[C_INTRANDPLE].Name[1]:='intrandomple';
FuncList[C_INTRANDPLE].Name[2]:='irndple';
FuncList[C_INTRANDPLE].Name[3]:='irandple';
FuncList[C_INTRANDPLE].ADDR:=@PLE_intrand;
FuncList[C_INTRANDPLE].CLST:=0;
FuncList[C_INTRANDPLE].Arrow:=0;
FuncList[C_INTRANDPLE].ArgType:=_Integer;
FuncList[C_INTRANDPLE].arg:=1;
FuncList[C_INTRANDPLE].NST:=3;   //need to ceck!!!
FuncList[C_INTRANDPLE].FType:=-10;
FuncList[C_INTRANDPLE].SaveReg:=15;            {!!!}
FuncList[C_INTRANDPLE].ReplFunc:=_No;
FuncList[C_INTRANDPLE].CalcConst:=_No;


SetLength(FuncList[C_RAND01PLE].Name,4);
FuncList[C_RAND01PLE].Name[1]:='random01ple';
FuncList[C_RAND01PLE].Name[2]:='rand01ple';
FuncList[C_RAND01PLE].Name[3]:='rnd01ple';
FuncList[C_RAND01PLE].ADDR:=@PLE_rand01;
FuncList[C_RAND01PLE].CLST:=0;
FuncList[C_RAND01PLE].Arrow:=0;
FuncList[C_RAND01PLE].ArgType:=_Integer;
FuncList[C_RAND01PLE].arg:=0;
FuncList[C_RAND01PLE].NST:=2;   //need to ceck!!!
FuncList[C_RAND01PLE].FType:=-11;  //eax
FuncList[C_RAND01PLE].SaveReg:=15;
FuncList[C_RAND01PLE].ReplFunc:=_No;
FuncList[C_RAND01PLE].CalcConst:=_No;



SetLength(FuncList[C_RANDXOR].Name,4);     {!!!}
FuncList[C_RANDXOR].Name[1]:='randomxor';
FuncList[C_RANDXOR].Name[2]:='randxor';
FuncList[C_RANDXOR].Name[3]:='rndxor';
FuncList[C_RANDXOR].arg:=2;
FuncList[C_RANDXOR].FType:=4;
FuncList[C_RANDXOR].ArgType:=_Double;
FuncList[C_RANDXOR].NST:=2;
FuncList[C_RANDXOR].CalcConst:=_No;
FuncList[C_RANDXOR].ReplFunc:=_No;
FuncList[C_RANDXOR].SaveReg:=15;            {!!!}
FuncList[C_RANDXOR].Addr:=@XOR_RAND2_K;


SetLength(FuncList[C_INTRANDXOR].Name,4);
FuncList[C_INTRANDXOR].Name[1]:='intrandomxor';
FuncList[C_INTRANDXOR].Name[2]:='irndxor';
FuncList[C_INTRANDXOR].Name[3]:='irandxor';
FuncList[C_INTRANDXOR].ADDR:=@XOR_intrand;
FuncList[C_INTRANDXOR].CLST:=0;
FuncList[C_INTRANDXOR].Arrow:=0;
FuncList[C_INTRANDXOR].ArgType:=_Integer;
FuncList[C_INTRANDXOR].arg:=1;
FuncList[C_INTRANDXOR].NST:=3;   //need to ceck!!!
FuncList[C_INTRANDXOR].FType:=-10;
FuncList[C_INTRANDXOR].SaveReg:=15;            {!!!}
FuncList[C_INTRANDXOR].ReplFunc:=_No;
FuncList[C_INTRANDXOR].CalcConst:=_No;


SetLength(FuncList[C_RAND01XOR].Name,4);
FuncList[C_RAND01XOR].Name[1]:='random01xor';
FuncList[C_RAND01XOR].Name[2]:='rnd01xor';
FuncList[C_RAND01XOR].Name[3]:='rand01xor';
FuncList[C_RAND01XOR].ADDR:=@XOR_rand01;
FuncList[C_RAND01XOR].CLST:=0;
FuncList[C_RAND01XOR].Arrow:=0;
FuncList[C_RAND01XOR].ArgType:=_Integer;
FuncList[C_RAND01XOR].arg:=0;
FuncList[C_RAND01XOR].NST:=2;   //need to ceck!!!
FuncList[C_RAND01XOR].FType:=-11;  //eax
FuncList[C_RAND01XOR].SaveReg:=15;
FuncList[C_RAND01XOR].ReplFunc:=_No;
FuncList[C_RAND01XOR].CalcConst:=_No;

{.199}
(*
SetLength(FuncList[C_RANDF].Name,3);     {!!!}
FuncList[C_RANDF].Name[1]:='randf';
FuncList[C_RANDF].Name[2]:='randple';
FuncList[C_RANDF].arg:=2;
FuncList[C_RANDF].FType:=-15;
FuncList[C_RANDF].ArgType:=_Double;
FuncList[C_RANDF].NST:=2;
FuncList[C_RANDF].CalcConst:=_No;
FuncList[C_RANDF].SaveReg:=15;            {!!!}
FuncList[C_RANDF].Addr:=@PLE_RAND2;
*)




SetLength(FuncList[C_RANDG].Name,5);   {.209}
FuncList[C_RANDG].Name[1]:='randomgausse';
FuncList[C_RANDG].Name[2]:='randG';
FuncList[C_RANDG].Name[3]:='rndG';
FuncList[C_RANDG].Name[4]:='randgausse';
FuncList[C_RANDG].arg:=2;
FuncList[C_RANDG].FType:=-15;
FuncList[C_RANDG].ArgType:=_Extended;
FuncList[C_RANDG].NST:=3;
FuncList[C_RANDG].CalcConst:=_No;
FuncList[C_RANDG].ReplFunc:=_No;
FuncList[C_RANDG].SaveReg:=15;
FuncList[C_RANDG].Addr:=@_RANDG;

//randomize()
SetLength(FuncList[C_RANDOMIZE].Name,2);
FuncList[C_RANDOMIZE].Name[1]:='randomize';
FuncList[C_RANDOMIZE].ADDR:=@_Randomize;
FuncList[C_RANDOMIZE].CLST:=0;
FuncList[C_RANDOMIZE].Arrow:=0;
FuncList[C_RANDOMIZE].ArgType:=_Integer;
FuncList[C_RANDOMIZE].arg:=0;
FuncList[C_RANDOMIZE].NST:=2;   //need to ceck!!!
FuncList[C_RANDOMIZE].FType:=-11;  //eax
FuncList[C_RANDOMIZE].SaveReg:=15;
FuncList[C_RANDOMIZE].ReplFunc:=_No;
FuncList[C_RANDOMIZE].CalcConst:=_No;
FuncList[C_RANDOMIZE].RetNum:=_No;



SetLength(FuncList[C_RANDARRAYD].Name,2);   {.209}
FuncList[C_RANDARRAYD].Name[1]:='arand';
FuncList[C_RANDARRAYD].arg:=1;
FuncList[C_RANDARRAYD].FType:=-10;
FuncList[C_RANDARRAYD].ArgType:=_Integer;
FuncList[C_RANDARRAYD].NST:=3;
FuncList[C_RANDARRAYD].CalcConst:=_No;
FuncList[C_RANDARRAYD].ReplFunc:=_No;
FuncList[C_RANDARRAYD].SaveReg:=15;
FuncList[C_RANDARRAYD].Addr:=@RAND_ARRAYD;




FuncList[C_GAMMA].Name[1]:='gamma';
FuncList[C_GAMMA].Addr:=@gamma;
FuncList[C_GAMMA].arg:=1;
FuncList[C_GAMMA].FType:=2;
FuncList[C_GAMMA].SaveReg:=15;            {!!!}
FuncList[C_GAMMA].NST:=3;         //need to ceck!!!

FuncList[C_PSI].Name[1]:='psi';
FuncList[C_PSI].Addr:=@psi;
FuncList[C_PSI].arg:=1;
FuncList[C_PSI].FType:=2;
FuncList[C_PSI].SaveReg:=15;            {!!!}
FuncList[C_PSI].NST:=3;            //need to ceck!!!

FuncList[C_PSID1].Name[1]:='psid1';
FuncList[C_PSID1].Addr:=@psid1;
FuncList[C_PSID1].arg:=1;
FuncList[C_PSID1].FType:=2;
FuncList[C_PSID1].SaveReg:=15;            {!!!}
FuncList[C_PSID1].NST:=3;           //need to ceck!!!

FuncList[C_PSID2].Name[1]:='psid2';
FuncList[C_PSID2].Addr:=@psid2;
FuncList[C_PSID2].arg:=1;
FuncList[C_PSID2].FType:=2;
FuncList[C_PSID2].SaveReg:=15;            {!!!}
FuncList[C_PSID2].NST:=3;            //need to ceck!!!

FuncList[C_PSID3].Name[1]:='psid3';
FuncList[C_PSID3].Addr:=@psid3;
FuncList[C_PSID3].arg:=1;
FuncList[C_PSID3].FType:=2;
FuncList[C_PSID3].SaveReg:=15;            {!!!}
FuncList[C_PSID3].NST:=3;
                                  //need to ceck!!!

FuncList[C_ZETA].Name[1]:='zeta';
FuncList[C_ZETA].Addr:=@zeta;
FuncList[C_ZETA].arg:=1;
FuncList[C_ZETA].FType:=2;
FuncList[C_ZETA].SaveReg:=15;            {!!!}
FuncList[C_ZETA].NST:=3;

FuncList[C_ERF].Name[1]:='erf';
FuncList[C_ERF].Addr:=@erf;
FuncList[C_ERF].arg:=1;
FuncList[C_ERF].FType:=2;
FuncList[C_ERF].SaveReg:=15;            {!!!}
FuncList[C_ERF].NST:=3;         //need to ceck!!!

FuncList[C_ERFC].Name[1]:='erfc';
FuncList[C_ERFC].Addr:=@erfc;
FuncList[C_ERFC].arg:=1;
FuncList[C_ERFC].FType:=2;
FuncList[C_ERFC].SaveReg:=15;            {!!!}
FuncList[C_ERFC].NST:=3;         //need to ceck!!!



SetLength(FuncList[C_BESSELJ0].Name,3);
FuncList[C_BESSELJ0].Name[1]:='besselj0';
FuncList[C_BESSELJ0].Name[2]:='bj0';
FuncList[C_BESSELJ0].Addr:=@bessel_j0;
FuncList[C_BESSELJ0].arg:=1;
FuncList[C_BESSELJ0].FType:=2;
FuncList[C_BESSELJ0].SaveReg:=15;            {!!!}
FuncList[C_BESSELJ0].NST:=3;      //need to ceck!!!


SetLength(FuncList[C_BESSELJ1].Name,3);
FuncList[C_BESSELJ1].Name[1]:='besselj1';
FuncList[C_BESSELJ1].Name[2]:='bj1';
FuncList[C_BESSELJ1].Addr:=@bessel_j1;
FuncList[C_BESSELJ1].arg:=1;
FuncList[C_BESSELJ1].FType:=2;
FuncList[C_BESSELJ1].SaveReg:=15;            {!!!}
FuncList[C_BESSELJ1].NST:=3;       //need to ceck!!!

SetLength(FuncList[C_BESSELJN].Name,4);
FuncList[C_BESSELJN].Name[1]:='besseljn';
FuncList[C_BESSELJN].Name[2]:='besselj';
FuncList[C_BESSELJN].Name[3]:='bjn';
FuncList[C_BESSELJN].Addr:=@bessel_jn;
FuncList[C_BESSELJN].arg:=2;
FuncList[C_BESSELJN].FType:=4;
FuncList[C_BESSELJN].SaveReg:=15;            {!!!}
FuncList[C_BESSELJN].NST:=3;    //need to ceck!!!



SetLength(FuncList[C_BESSELY0].Name,3);
FuncList[C_BESSELY0].Name[1]:='bessely0';
FuncList[C_BESSELY0].Name[2]:='by0';
FuncList[C_BESSELY0].Addr:=@bessel_y0;
FuncList[C_BESSELy0].arg:=1;
FuncList[C_BESSELY0].FType:=2;
FuncList[C_BESSELY0].SaveReg:=15;            {!!!}
FuncList[C_BESSELY0].NST:=3;      //need to ceck!!!


SetLength(FuncList[C_BESSELY1].Name,3);
FuncList[C_BESSELY1].Name[1]:='bessely1';
FuncList[C_BESSELY1].Name[2]:='by1';
FuncList[C_BESSELY1].Addr:=@bessel_y1;
FuncList[C_BESSELY1].arg:=1;
FuncList[C_BESSELY1].FType:=2;
FuncList[C_BESSELY1].SaveReg:=15;            {!!!}
FuncList[C_BESSELY1].NST:=3;       //need to ceck!!!

SetLength(FuncList[C_BESSELYN].Name,4);
FuncList[C_BESSELYN].Name[1]:='besselyn';
FuncList[C_BESSELYN].Name[2]:='byn';
FuncList[C_BESSELYN].Name[3]:='bessely';
FuncList[C_BESSELYN].Addr:=@bessel_yn;
FuncList[C_BESSELYN].arg:=2;
FuncList[C_BESSELYN].FType:=4;
FuncList[C_BESSELYN].SaveReg:=15;            {!!!}
FuncList[C_BESSELYN].NST:=3;    //need to ceck!!!


SetLength(FuncList[C_SPHBESSELJN].Name,7);
FuncList[C_SPHBESSELJN].Name[1]:='sphbesseljn';
FuncList[C_SPHBESSELJN].Name[2]:='sphbjn';
FuncList[C_SPHBESSELJN].Name[3]:='sbjn';
FuncList[C_SPHBESSELJN].Name[4]:='sphericalbesseljn';
FuncList[C_SPHBESSELJN].Name[5]:='sphericalbesselj';
FuncList[C_SPHBESSELJN].Name[6]:='sphbesselj';
FuncList[C_SPHBESSELJN].Addr:=@sph_bessel_jn;
FuncList[C_SPHBESSELJN].arg:=2;
FuncList[C_SPHBESSELJN].FType:=4;
FuncList[C_SPHBESSELJN].SaveReg:=15;            {!!!}
FuncList[C_SPHBESSELJN].NST:=3;    //need to ceck!!!


SetLength(FuncList[C_LEGENDRE].Name,5);
FuncList[C_LEGENDRE].Name[1]:='legendrep';//    @@legpl,@@legpla
FuncList[C_LEGENDRE].Name[2]:='legp';
FuncList[C_LEGENDRE].Name[3]:='polyleg';
FuncList[C_LEGENDRE].Name[4]:='legpoly';
FuncList[C_LEGENDRE].FType:=-11;



SetLength(FuncList[C_LEGENDREP].Name,2);
FuncList[C_LEGENDREP].Name[1]:='@@legp';
FuncList[C_LEGENDREP].ADDR:=@sfc_legendre_p;
FuncList[C_LEGENDREP].CLST:=0;
FuncList[C_LEGENDREP].Arrow:=0;
FuncList[C_LEGENDREP].ArgType:=_Differ;
SetLength(FuncList[C_LEGENDREP].ArgTypeList,2);
FuncList[C_LEGENDREP].ArgTypeList[0]:=_Integer;
FuncList[C_LEGENDREP].ArgTypeList[1]:=_Extended;
FuncList[C_LEGENDREP].arg:=2;
FuncList[C_LEGENDREP].NST:=3;   //need to ceck!!!
FuncList[C_LEGENDREP].FType:=-10;
FuncList[C_LEGENDREP].SaveReg:=15;            {!!!}
FuncList[C_LEGENDREP].ReplFunc:=_Yes;


SetLength(FuncList[C_LEGENDREPA].Name,2);
FuncList[C_LEGENDREPA].Name[1]:='@@legpa';
FuncList[C_LEGENDREPA].ADDR:=@sfc_legendre_plm;
FuncList[C_LEGENDREPA].CLST:=0;
FuncList[C_LEGENDREPA].Arrow:=0;
FuncList[C_LEGENDREPA].ArgType:=_Differ;
SetLength(FuncList[C_LEGENDREPA].ArgTypeList,3);
FuncList[C_LEGENDREPA].ArgTypeList[0]:=_Integer;
FuncList[C_LEGENDREPA].ArgTypeList[1]:=_Integer;
FuncList[C_LEGENDREPA].ArgTypeList[2]:=_Extended;
FuncList[C_LEGENDREPA].arg:=3;
FuncList[C_LEGENDREPA].NST:=3;   //need to ceck!!!
FuncList[C_LEGENDREPA].FType:=-10;
FuncList[C_LEGENDREPA].SaveReg:=15;            {!!!}
FuncList[C_LEGENDREPA].ReplFunc:=_Yes;


{.211}
SetLength(FuncList[C_CHEBYSHEV_T].Name,8);
FuncList[C_CHEBYSHEV_T].Name[1]:='chebyshevt';
FuncList[C_CHEBYSHEV_T].Name[2]:='tcheb';
FuncList[C_CHEBYSHEV_T].Name[3]:='chebt';
FuncList[C_CHEBYSHEV_T].Name[4]:='tchebpoly';
FuncList[C_CHEBYSHEV_T].Name[5]:='chebtpoly';
FuncList[C_CHEBYSHEV_T].Name[6]:='polychebt';
FuncList[C_CHEBYSHEV_T].Name[7]:='polytcheb';
FuncList[C_CHEBYSHEV_T].ADDR:=@sfc_chebyshev_t;
FuncList[C_CHEBYSHEV_T].CLST:=0;
FuncList[C_CHEBYSHEV_T].Arrow:=0;
FuncList[C_CHEBYSHEV_T].ArgType:=_Differ;
SetLength(FuncList[C_CHEBYSHEV_T].ArgTypeList,2);
FuncList[C_CHEBYSHEV_T].ArgTypeList[0]:=_Integer;
FuncList[C_CHEBYSHEV_T].ArgTypeList[1]:=_Extended;
FuncList[C_CHEBYSHEV_T].arg:=2;
FuncList[C_CHEBYSHEV_T].NST:=3;   //need to ceck!!!
FuncList[C_CHEBYSHEV_T].FType:=-10;
FuncList[C_CHEBYSHEV_T].SaveReg:=15;            {!!!}
FuncList[C_CHEBYSHEV_T].ReplFunc:=_Yes;


{.211}
SetLength(FuncList[C_CHEBYSHEV_U].Name,8);
FuncList[C_CHEBYSHEV_U].Name[1]:='chebyshevu';
FuncList[C_CHEBYSHEV_U].Name[2]:='ucheb';
FuncList[C_CHEBYSHEV_U].Name[3]:='chebu';
FuncList[C_CHEBYSHEV_U].Name[4]:='polychebu';
FuncList[C_CHEBYSHEV_U].Name[5]:='uchebpoly';
FuncList[C_CHEBYSHEV_U].Name[6]:='polyucheb';
FuncList[C_CHEBYSHEV_U].Name[7]:='chebupoly';
FuncList[C_CHEBYSHEV_U].ADDR:=@sfc_chebyshev_u;
FuncList[C_CHEBYSHEV_U].CLST:=0;
FuncList[C_CHEBYSHEV_U].Arrow:=0;
FuncList[C_CHEBYSHEV_U].ArgType:=_Differ;
SetLength(FuncList[C_CHEBYSHEV_U].ArgTypeList,2);
FuncList[C_CHEBYSHEV_U].ArgTypeList[0]:=_Integer;
FuncList[C_CHEBYSHEV_U].ArgTypeList[1]:=_Extended;
FuncList[C_CHEBYSHEV_U].arg:=2;
FuncList[C_CHEBYSHEV_U].NST:=3;   //need to ceck!!!
FuncList[C_CHEBYSHEV_U].FType:=-10;
FuncList[C_CHEBYSHEV_U].SaveReg:=15;            {!!!}
FuncList[C_CHEBYSHEV_U].ReplFunc:=_Yes;



{.211}
SetLength(FuncList[C_HERMITE].Name,5);
FuncList[C_HERMITE].Name[1]:='hermitep';
FuncList[C_HERMITE].Name[2]:='polyher';
FuncList[C_HERMITE].Name[3]:='herpoly';
FuncList[C_HERMITE].Name[4]:='herp';
FuncList[C_HERMITE].ADDR:=@sfc_hermite_h;
FuncList[C_HERMITE].CLST:=0;
FuncList[C_HERMITE].Arrow:=0;
FuncList[C_HERMITE].ArgType:=_Differ;
SetLength(FuncList[C_HERMITE].ArgTypeList,2);
FuncList[C_HERMITE].ArgTypeList[0]:=_Integer;
FuncList[C_HERMITE].ArgTypeList[1]:=_Extended;
FuncList[C_HERMITE].arg:=2;
FuncList[C_HERMITE].NST:=3;   //need to ceck!!!
FuncList[C_HERMITE].FType:=-10;
FuncList[C_HERMITE].SaveReg:=15;            {!!!}
FuncList[C_HERMITE].ReplFunc:=_Yes;


(*
{.211}
SetLength(FuncList[C_LAGUERRE].Name,5);
FuncList[C_LAGUERRE].Name[1]:='laguerrep';
FuncList[C_LAGUERRE].Name[2]:='polylag';
FuncList[C_LAGUERRE].Name[3]:='lagpoly';
FuncList[C_LAGUERRE].Name[4]:='lagp';
FuncList[C_LAGUERRE].ADDR:=@sfc_LAGUERRE;
FuncList[C_LAGUERRE].CLST:=0;
FuncList[C_LAGUERRE].Arrow:=0;
FuncList[C_LAGUERRE].ArgType:=_Differ;
SetLength(FuncList[C_LAGUERRE].ArgTypeList,3);
FuncList[C_LAGUERRE].ArgTypeList[0]:=_Integer;
FuncList[C_LAGUERRE].ArgTypeList[1]:=_Extended;
FuncList[C_LAGUERRE].ArgTypeList[2]:=_Extended;
FuncList[C_LAGUERRE].arg:=3;
FuncList[C_LAGUERRE].NST:=3;   //need to ceck!!!
FuncList[C_LAGUERRE].FType:=-10;
FuncList[C_LAGUERRE].SaveReg:=15;            {!!!}
FuncList[C_LAGUERRE].ReplFunc:=_Yes;
*)

SetLength(FuncList[C_LAGUERRE].Name,5);
FuncList[C_LAGUERRE].Name[1]:='laguerrep';//    @@lagpl,@@lagpla
FuncList[C_LAGUERRE].Name[2]:='lagp';
FuncList[C_LAGUERRE].Name[3]:='polylag';
FuncList[C_LAGUERRE].Name[4]:='lagpoly';
FuncList[C_LAGUERRE].FType:=-11;



SetLength(FuncList[C_LAGUERREP].Name,2);
FuncList[C_LAGUERREP].Name[1]:='@@lagp';
FuncList[C_LAGUERREP].ADDR:=@sfc_LAGUERRE;
FuncList[C_LAGUERREP].CLST:=0;
FuncList[C_LAGUERREP].Arrow:=0;
FuncList[C_LAGUERREP].ArgType:=_Differ;
SetLength(FuncList[C_LAGUERREP].ArgTypeList,2);
FuncList[C_LAGUERREP].ArgTypeList[0]:=_Integer;
FuncList[C_LAGUERREP].ArgTypeList[1]:=_Extended;
FuncList[C_LAGUERREP].arg:=2;
FuncList[C_LAGUERREP].NST:=3;   //need to ceck!!!
FuncList[C_LAGUERREP].FType:=-10;
FuncList[C_LAGUERREP].SaveReg:=15;            {!!!}
FuncList[C_LAGUERREP].ReplFunc:=_Yes;


SetLength(FuncList[C_LAGUERREPA].Name,2);
FuncList[C_LAGUERREPA].Name[1]:='@@lagpa';
FuncList[C_LAGUERREPA].ADDR:=@sfc_LAGUERRE_A;
FuncList[C_LAGUERREPA].CLST:=0;
FuncList[C_LAGUERREPA].Arrow:=0;
FuncList[C_LAGUERREPA].ArgType:=_Differ;
SetLength(FuncList[C_LAGUERREPA].ArgTypeList,3);
FuncList[C_LAGUERREPA].ArgTypeList[0]:=_Integer;
FuncList[C_LAGUERREPA].ArgTypeList[1]:=_Extended;//Integer
FuncList[C_LAGUERREPA].ArgTypeList[2]:=_Extended;
FuncList[C_LAGUERREPA].arg:=3;
FuncList[C_LAGUERREPA].NST:=3;   //need to ceck!!!
FuncList[C_LAGUERREPA].FType:=-10;
FuncList[C_LAGUERREPA].SaveReg:=15;            {!!!}
FuncList[C_LAGUERREPA].ReplFunc:=_Yes;


SetLength(FuncList[C_POWER].Name,3);
FuncList[C_POWER].Name[1]:='power';//    FPWR,IPWR,IPWRC,IPWRSGN,I2PWR,sqrt,sqr
FuncList[C_POWER].Name[2]:='pow';
FuncList[C_POWER].arg:=2;
FuncList[C_POWER].FType:=4;


SetLength(FuncList[C_LOGN].Name,3);
FuncList[C_LOGN].Name[1]:='logn';
FuncList[C_LOGN].Name[2]:='log';
FuncList[C_LOGN].Addr:=@R_LOGN;
FuncList[C_LOGN].arg:=2;
FuncList[C_LOGN].FType:=4;
FuncList[C_LOGN].NST:=1;


FuncList[C_ROOT].Name[1]:='root';
FuncList[C_ROOT].arg:=2;
FuncList[C_ROOT].FType:=4;


FuncList[C_IROOT].Name[1]:='@@IROOT';
FuncList[C_IROOT].Addr:=@R_IROOT;
FuncList[C_IROOT].arg:=2;
FuncList[C_IROOT].FType:=3;
FuncList[C_IROOT].SaveReg:=3; //EAX+EBX
FuncList[C_IROOT].NST:=2;

FuncList[C_FROOT].Name[1]:='@@FROOT';
FuncList[C_FROOT].Addr:=@R_FROOT;
FuncList[C_FROOT].arg:=2;
FuncList[C_FROOT].FType:=4;
FuncList[C_FROOT].SaveReg:=3; //EAX+EBX
FuncList[C_FROOT].NST:=1;


FuncList[C_IPWR].Name[1]:='@@IPWR';
FuncList[C_IPWR].Addr:=@R_IPWR;
FuncList[C_IPWR].arg:=1;
FuncList[C_IPWR].FType:=3;
FuncList[C_IPWR].SaveReg:=13; //EAX+ECX+EDX
FuncList[C_IPWR].NST:=1;


FuncList[C_FPWR].Name[1]:='@@FPWR';
FuncList[C_FPWR].Addr:=@R_FPWR;  //@R_FPWR_I2;
FuncList[C_FPWR].arg:=2;
FuncList[C_FPWR].FType:=4;
FuncList[C_FPWR].SaveReg:=13;  //EAX+ECX+EDX
FuncList[C_FPWR].NST:=2;


FuncList[C_F2PWR].Name[1]:='@@F2PWR';     {.243}
FuncList[C_F2PWR].Addr:=@R_FPWR2F;
FuncList[C_F2PWR].arg:=1;
FuncList[C_F2PWR].FType:=2;
FuncList[C_F2PWR].NST:=2;
FuncList[C_F2PWR].SaveReg:=1;


FuncList[C_I2PWR].Name[1]:='@@I2PWR';     {.249}
FuncList[C_I2PWR].Addr:=@R_FPWR2I;
FuncList[C_I2PWR].arg:=1;
FuncList[C_I2PWR].FType:=1;
FuncList[C_I2PWR].NST:=1;
FuncList[C_I2PWR].SaveReg:=1;



SetLength(FuncList[C_SIGN].Name,3);
FuncList[C_SIGN].Name[1]:='sign';
FuncList[C_SIGN].Name[2]:='sgn';
FuncList[C_SIGN].Addr:=@R_SIGN;
FuncList[C_SIGN].arg:=1;
FuncList[C_SIGN].FType:=2;
FuncList[C_SIGN].NST:=0;

//C_IPWRSGNF:          {.219}
SetLength(FuncList[C_IPWRSGNF].Name,2);
FuncList[C_IPWRSGNF].Name[1]:='@@IPWRSGNF';
FuncList[C_IPWRSGNF].Addr:=@R_IPWRSGNF;
FuncList[C_IPWRSGNF].arg:=1;
FuncList[C_IPWRSGNF].FType:=2;
FuncList[C_IPWRSGNF].NST:=1;

FuncList[C_IF].Name[1]:='if';         {.199}
FuncList[C_IF].FType:=-17;
FuncList[C_IF].arg:=-1;

FuncList[C_IFNOP].Name[1]:='@@IFNOP';
FuncList[C_IFNOP].FType:=-17;
FuncList[C_IFNOP].arg:=6;
FuncList[C_IFNOP].NST:=0;

FuncList[C_JMP].Name[1]:='@@JMP';
FuncList[C_JMP].FType:=0;
FuncList[C_JMP].NST:=0;
FuncList[C_JMP].RetNum:=_NO;

FuncList[C_FLD].Name[1]:='@@FLD';
FuncList[C_FLD].FType:=0;
FuncList[C_FLD].NST:=0;

FuncList[C_LDA].Name[1]:='LDA';
FuncList[C_LDA].FType:=0;
FuncList[C_LDA].NST:=0;
FuncList[C_LDA].FType:=-17;

FuncList[C_COMPILED].Name[1]:='COMPILE';
FuncList[C_COMPILED].FType:=0;
FuncList[C_COMPILED].NST:=0;
FuncList[C_COMPILED].FType:=-17;
FuncList[C_COMPILED].arg:=1;

{.199}
{
FuncList[C_MAX2].Name[1]:='MAX2';
FuncList[C_MAX2].Addr:=@_MAX2;
FuncList[C_MAX2].arg:=2;
FuncList[C_MAX2].FType:=4;
FuncList[C_MAX2].SaveReg:=3; //EAX
FuncList[C_MAX2].NST:=1;

FuncList[C_MIN2].Name[1]:='MIN2';
FuncList[C_MIN2].Addr:=@_MIN2;
FuncList[C_MIN2].arg:=2;
FuncList[C_MIN2].FType:=4;
FuncList[C_MIN2].SaveReg:=3; //EAX
FuncList[C_MIN2].NST:=1;
}

FuncList[C_GOTO].Name[1]:='goto';        {.199}
FuncList[C_GOTO].FType:=-17;
FuncList[C_GOTO].NST:=0;
FuncList[C_GOTO].arg:=1;
FuncList[C_GOTO].RetNum:=_NO;


FuncList[C_SETLENX].Name[1]:='setlen';
FuncList[C_SETLENX].FType:=-17;
FuncList[C_SETLENX].arg:=2;
FuncList[C_SETLENX].RetNum:=_No;
FuncList[C_SETLENX].CalcConst:=_No;

FuncList[C_SETLEN].Name[1]:='@@setlen';
FuncList[C_SETLEN].FType:=-10;
FuncList[C_SETLEN].ArgType:=_INTEGER;
FuncList[C_SETLEN].SaveReg:=255;            {!!!}
FuncList[C_SETLEN].NST:=1;
FuncList[C_SETLEN].arg:=3;
FuncList[C_SETLEN].Addr:=@SetLengthArray;
FuncList[C_SETLEN].RetNum:=_No;
FuncList[C_SETLEN].CalcConst:=_No;
FuncList[C_SETLEN].CLST:=0;
FuncList[C_SETLEN].Arrow:=0;

(*
{.248}
SetLength(FuncList[C_TRY].Name,2);
FuncList[C_TRY].Name[1]:='try';
FuncList[C_TRY].ADDR:=@MaskFPUException;
FuncList[C_TRY].CLST:=0;
FuncList[C_TRY].Arrow:=0;
FuncList[C_TRY].ArgType:=_None;
FuncList[C_TRY].arg:=0;
FuncList[C_TRY].NST:=0;
FuncList[C_TRY].FType:=-10;
FuncList[C_TRY].SaveReg:=0;
FuncList[C_TRY].ReplFunc:=_NO;
FuncList[C_TRY].CalcConst:=_NO;
FuncList[C_TRY].RetNum:=_NO;

{.248}
SetLength(FuncList[C_EXCEPTION].Name,2);
FuncList[C_EXCEPTION].Name[1]:='exception';
FuncList[C_EXCEPTION].ADDR:=@IsFPUException;
FuncList[C_EXCEPTION].CLST:=0;
FuncList[C_EXCEPTION].Arrow:=0;
FuncList[C_EXCEPTION].ArgType:=_None;
FuncList[C_EXCEPTION].arg:=0;
FuncList[C_EXCEPTION].NST:=1;
FuncList[C_EXCEPTION].FType:=-10;
FuncList[C_EXCEPTION].SaveReg:=0;
FuncList[C_EXCEPTION].ReplFunc:=_NO;
FuncList[C_EXCEPTION].CalcConst:=_NO;
FuncList[C_EXCEPTION].RetNum:=_YES;
*)


{.249}
SetLength(FuncList[C_TRY].Name,2);
FuncList[C_TRY].Name[1]:='try';
FuncList[C_TRY].ADDR:=@MaskFPUExceptionA;
FuncList[C_TRY].CLST:=0;
FuncList[C_TRY].Arrow:=0;
FuncList[C_TRY].ArgType:=_None;
FuncList[C_TRY].arg:=0;
FuncList[C_TRY].NST:=0;
FuncList[C_TRY].FType:=-8;
FuncList[C_TRY].SaveReg:=5;  //EAX+ECX
FuncList[C_TRY].ReplFunc:=_NO;
FuncList[C_TRY].CalcConst:=_NO;
FuncList[C_TRY].RetNum:=_NO;

{.249}
SetLength(FuncList[C_EXCEPTION].Name,2);
FuncList[C_EXCEPTION].Name[1]:='exception';
FuncList[C_EXCEPTION].ADDR:=@IsFPUExceptionA;
FuncList[C_EXCEPTION].CLST:=0;
FuncList[C_EXCEPTION].Arrow:=0;
FuncList[C_EXCEPTION].ArgType:=_None;
FuncList[C_EXCEPTION].arg:=0;
FuncList[C_EXCEPTION].NST:=1;
FuncList[C_EXCEPTION].FType:=-8;
FuncList[C_EXCEPTION].SaveReg:=5;     //EAX+ECX
FuncList[C_EXCEPTION].ReplFunc:=_NO;
FuncList[C_EXCEPTION].CalcConst:=_NO;
FuncList[C_EXCEPTION].RetNum:=_YES;


for i:=1 to  High(FuncList) do
begin
 if FuncList[i].FType > 0   then FuncList[i].ReplFunc:=_Yes;
 if FuncList[i].FType >= 0  then FuncList[i].XchArg:=_Yes; {.205}
end;

FuncList[C_RAND].ReplFunc:=_No;
FuncList[C_RANDPLE].ReplFunc:=_No;
FuncList[C_RANDXOR].ReplFunc:=_No;
FuncList[C_INTRANDMT].ReplFunc:=_No;
FuncList[C_RAND01MT].ReplFunc:=_No;
FuncList[C_RAND01PLE].ReplFunc:=_No;
FuncList[C_INTRANDPLE].ReplFunc:=_No;
FuncList[C_RAND01XOR].ReplFunc:=_No;
FuncList[C_INTRANDXOR].ReplFunc:=_No;
FuncList[C_RANDG].ReplFunc:=_No;
FuncList[C_RANDMT].ReplFunc:=_No;


FuncList[C_MUL].ReplFunc:=_Yes;      {.199}
FuncList[C_ADD].ReplFunc:=_Yes;      {.199}
FuncList[C_MAF].ReplFunc:=_Yes;      {.201}
FuncList[C_AAF].ReplFunc:=_Yes;      {.201}
FuncList[C_FAVB].ReplFunc:=_Yes;      {.201}
end;



procedure CreateExchTable(FuncList: TArrayFuncList; var XchTable: TArray2I);
var
i,j,L: Integer;
begin
//XchTable:

{ Func1 -> Func2         }

{ ------> Func1 - }
{ |                      }
{ |                      }
{ |                      }
{ |                      }
{ V Func2 -    }

{
 :
|---------|     WCS -    (   (Func1) )
| `  WCS  |     RCS -   (   (Func2) )
|   `     |     NC  -     
| RCS `   |
|-------- |


RCS:
0 -  ;
1 -  ;
     RCS
}


//XchCode:
{XchCode[Number of Length of  XchTable] = Number Function in FuncList}




 L:=LFuncList+1;
 SetLength(XchTable,L,L);

 for i:=1 to L-1 do
 begin
  for j:=0 to L-1 do
  begin
   XchTable[i,j].RCS:=0;
   XchTable[i,j].WCS:=0;
   XchTable[i,j].NC:=0;
   if (i=j) and {(FuncList[i].FType > 0)} (FuncList[i].ReplFunc = _Yes)  then
   begin
    XchTable[i,j].RCS:=1;
    XchTable[i,j].WCS:=1;
    XchTable[i,j].NC:=1;
   end;
  end;
 end;


 //////   (S,1/S)  - :  sh(x)/(-exp(x))+ch(x)/(-sh(x))

 //. -: S=exp;
 // sh,ch ():
 //(RCS: 1 - (sh,ch); 2 - (exp,1/exp); 3 - (exp))
 //(WCS: -//-)
 // th,cth ():
 //(RCS: 1 - (th,cth); 2 - (cth,th); 3 - (exp))
 //(WCS: 1 - (th,cth); 2 - (exp))
(*
 //sh->ch (S; 1/S)
 //XchTable[C_SINH,C_COSH].RCS:=2;  XchTable[C_SINH,C_COSH].WCS:=2;  XchTable[C_SINH,C_COSH].NC:=2;
 //sh->ch (S)
 XchTable[C_SINH,C_COSH].RCS:=3;  XchTable[C_SINH,C_COSH].WCS:=3;  XchTable[C_SINH,C_COSH].NC:=1;
 //sh->th (S)
 XchTable[C_SINH,C_TANH].RCS:=3;  XchTable[C_SINH,C_TANH].WCS:=3;  XchTable[C_SINH,C_TANH].NC:=1;
 //sh->cth (S)
 XchTable[C_SINH,C_COTANH].RCS:=3;  XchTable[C_SINH,C_COTANH].WCS:=3;  XchTable[C_SINH,C_COTANH].NC:=1;
 //sh->exp (S)
 XchTable[C_SINH,C_EXP].RCS:=1;  XchTable[C_SINH,C_EXP].WCS:=3;  XchTable[C_SINH,C_EXP].NC:=1;
 //exp->sh (S)
 XchTable[C_EXP,C_SINH].RCS:=3;  XchTable[C_EXP,C_SINH].WCS:=1;  XchTable[C_EXP,C_SINH].NC:=1;
 //exp->ch (S)
 XchTable[C_EXP,C_COSH].RCS:=3;  XchTable[C_EXP,C_COSH].WCS:=1;  XchTable[C_EXP,C_COSH].NC:=1;
 //exp->th (S)
 XchTable[C_EXP,C_TANH].RCS:=3;  XchTable[C_EXP,C_TANH].WCS:=1;  XchTable[C_EXP,C_TANH].NC:=1;
 //exp->cth (S)
 XchTable[C_EXP,C_COTANH].RCS:=3;  XchTable[C_EXP,C_COTANH].WCS:=1;  XchTable[C_EXP,C_COTANH].NC:=1;
 //ch->sh (S; 1/S)
 //XchTable[C_COSH,C_SINH].RCS:=2;  XchTable[C_COSH,C_SINH].WCS:=2;  XchTable[C_COSH,C_SINH].NC:=2;
 //ch->sh (S)
 XchTable[C_COSH,C_SINH].RCS:=3;  XchTable[C_COSH,C_SINH].WCS:=3;  XchTable[C_COSH,C_SINH].NC:=1;
 //ch->th (S)
 XchTable[C_COSH,C_TANH].RCS:=3;  XchTable[C_COSH,C_TANH].WCS:=3;  XchTable[C_COSH,C_TANH].NC:=1;
 //ch->cth (S)
 XchTable[C_COSH,C_COTANH].RCS:=3;  XchTable[C_COSH,C_COTANH].WCS:=3;  XchTable[C_COSH,C_COTANH].NC:=1;
 //ch->exp (S)
 XchTable[C_COSH,C_EXP].RCS:=1;  XchTable[C_COSH,C_EXP].WCS:=3;  XchTable[C_COSH,C_EXP].NC:=1;
 //th->ch (S)
 XchTable[C_TANH,C_COSH].RCS:=3;  XchTable[C_TANH,C_COSH].WCS:=2;  XchTable[C_TANH,C_COSH].NC:=1;
 //th->sh (S)
 XchTable[C_TANH,C_SINH].RCS:=3;  XchTable[C_TANH,C_SINH].WCS:=2;  XchTable[C_TANH,C_SINH].NC:=1;
 //th->cth (th)
 XchTable[C_TANH,C_COTANH].RCS:=2;  XchTable[C_TANH,C_COTANH].WCS:=1;  XchTable[C_TANH,C_COTANH].NC:=1;
 //th->exp (S)
 XchTable[C_TANH,C_EXP].RCS:=1;  XchTable[C_TANH,C_EXP].WCS:=2;  XchTable[C_TANH,C_EXP].NC:=1;
 //cth->ch (S)
 XchTable[C_COTANH,C_COSH].RCS:=3;  XchTable[C_COTANH,C_COSH].WCS:=2;  XchTable[C_COTANH,C_COSH].NC:=1;
 //cth->sh (S)
 XchTable[C_COTANH,C_SINH].RCS:=3;  XchTable[C_COTANH,C_SINH].WCS:=2;  XchTable[C_COTANH,C_SINH].NC:=1;
 //cth->th  (cth)
 XchTable[C_COTANH,C_TANH].RCS:=2;  XchTable[C_COTANH,C_TANH].WCS:=1;  XchTable[C_COTANH,C_TANH].NC:=1;
 //cth->exp (S)
 XchTable[C_COTANH,C_EXP].RCS:=1;  XchTable[C_COTANH,C_EXP].WCS:=2;  XchTable[C_COTANH,C_EXP].NC:=1;


 //arcsin->arccos
 XchTable[C_ARCSIN,C_ARCCOS].RCS:=2;   XchTable[C_ARCSIN,C_ARCCOS].WCS:=1;   XchTable[C_ARCSIN,C_ARCCOS].NC:=1;
 //arccos->arcsin
 XchTable[C_ARCCOS,C_ARCSIN].RCS:=2;   XchTable[C_ARCCOS,C_ARCSIN].WCS:=1;   XchTable[C_ARCCOS,C_ARCSIN].NC:=1;


 //arctg->arcctg
 XchTable[C_ARCTAN,C_ARCCOTAN].RCS:=2;   XchTable[C_ARCTAN,C_ARCCOTAN].WCS:=1;   XchTable[C_ARCTAN,C_ARCCOTAN].NC:=1;
 //arcctg->arctg
 XchTable[C_ARCCOTAN,C_ARCTAN].RCS:=1;   XchTable[C_ARCCOTAN,C_ARCTAN].WCS:=2;   XchTable[C_ARCCOTAN,C_ARCTAN].NC:=1;

//tg->ctg
 XchTable[C_TAN,C_COTAN].RCS:=2;   XchTable[C_TAN,C_COTAN].WCS:=1;   XchTable[C_TAN,C_COTAN].NC:=1;
 //ctg->tg
 XchTable[C_COTAN,C_TAN].RCS:=1;   XchTable[C_COTAN,C_TAN].WCS:=2;   XchTable[C_COTAN,C_TAN].NC:=1;

 //sin->cos  ( fsincos)
 XchTable[C_SIN,C_COS].RCS:=1;  XchTable[C_SIN,C_COS].WCS:=2;  XchTable[C_SIN,C_COS].NC:=1;
 //cos->sin ( fsincos)
 XchTable[C_COS,C_SIN].RCS:=1;  XchTable[C_COS,C_SIN].WCS:=2;  XchTable[C_COS,C_SIN].NC:=1;
*)



end;





procedure  R_FPWR2I;   //2^int
var
MX: Extended;
M: Cardinal;
asm
//******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)

        {
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        }

        mov      ebx,Max2Power
        cmp      eax,ebx
        JL       @@LowMaxI
         //call Overflow Error:
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxI:  //ipower < Max2PowerN ??

        mov      ebx,Max2PowerN
        cmp      eax,ebx
        JG       @@pwr2i
        fldz
        JMP      @@end

@@pwr2i:

        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
@@end:

//**********************
end;





procedure R_FPWR2F;   //2^X
const F2: double =2.0;
var
M: Cardinal;
MX: Extended;
asm
//2^X
//ST(0) <- X




     FLD       ST
     //FRNDINT
     FISTP M
     FILD  M
     FCOMIP   ST(0),ST(1)
     JZ        @@int
     JMP       @@flow

@@int:


       //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxI
         //call Overflow Error:
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxI:  //ipower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2i
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2i:
        FISTP M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
    //**********************

     JMP     @@end


@@flow:

     // X=0 ??
     FLD F2

     FLDZ            {.}
     FCOMP           {.}
     //FTST
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@NoZero     //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??
     //FXCH
     //FLDZ
     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz
     JB    @@end              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)
     fld1
     fchs
     fsqrt
     JMP     @@end      {.}

@@NoZero:
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP  ST(1)
     }
      //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)

        //fpower >  Max2Power ??
        FXCH ST(1)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2f
        fstp st(0)
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)

    //**********************

@@end:

end;




{.245}
procedure R_FPWR;
const
MaxInt: double= 2147483647;
var
M: Cardinal;
MX: Extended;
asm
//X^Y
//ST(0) <- X
//ST(1) <- Y



     FXCH

     FLD       ST
     FABS
     FLD      MaxInt
     fcompp
     fstsw   ax
     sahf
     jb      @@flow


     FIST      M
     FICOM     M
     {fld      st
     frndint
     fcomp}


     FNSTSW    AX
     SAHF;
     JZ        @@int
     JMP       @@flow

@@int:

     FXCH
     FSTP     ST(1)
     //FISTP    M


     MOV      EAX, M



     PUSH     ECX
     PUSH     EDX

     fld1
     fxch
     mov     ecx, eax
     cdq
     xor     eax, edx
     sub     eax, edx
     jnz     @@5
     fstp    ST(0)

     jmp     @@endInt
@@4: fmul    ST, ST
@@5: shr     eax,1
     jnc     @@4
     fmul    ST(1),ST
     jnz     @@4
     fstp    st
     cmp     ecx, 0



     jge     @@endInt
     //fdivr   S_c1
     fld1
     fdivrp

@@endInt:
     POP     EDX
     POP     ECX
     JMP     @@end

@@flow:

     // X=0 ??
     FXCH

     FLDZ            {.}
     FCOMP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@NoZero           //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??

     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz
     JB    @@end              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)
     fld1
     fchs
     fsqrt
     JMP     @@end      {.}

@@NoZero:
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP  ST(1)
     }
      //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)

        //fpower >  Max2Power ??
        FXCH ST(1)
        FILD Max2Power
        FCOMIP   ST(0),ST(1)
        JNB   @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP     @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB   @@pwr2f
        fstp st(0)
        fstp st(0)
        fldz
        JMP     @@end

@@pwr2f:
        FISTP M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)

@@end:

end;






{.241}
procedure  R_IROOT; assembler;
//EAX - ; ST(0) - ;
var
M: Cardinal;
MX: Extended;
asm
     //push    bx

     xor     bx,bx    //  '+'
     mov     M,eax  //
     fldz
     fcomp
     fnstsw  ax
     sahf
     mov     eax,M //
     ja      @@1  //  < 0
     jb      @@2  //  > 0
                  //  = 0:    root(0,P)
     fstp    ST(0)
     fldz

     // P = ??
     fild    M
     fcomp
     fnstsw  ax
     sahf
     ja      @@END    // P > 0

     // P <=0: call error
     fld1
     fchs
     fsqrt
     jmp     @@END

@@1:
     bt      eax,0
     jc      @@3      //  

     //xor     eax,eax    //    
     //div     eax
     fsqrt     {.203}
@@3:
     or      bx,1b     // '-'
     fabs
@@2:

     fld1
     fild    dword ptr [M]
     fdivp
     fxch
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD

     {
     FSCALE
     FSTP    ST(1)
     }

     //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
     (*
        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
        //cal error
        fld1
        fchs
        fsqrt
        JMP       @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP       @@end

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        *)


        FXCH    st(1)
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)


    //**********************

     bt      bx,0
     jnc     @@END
     fchs

@@END:
     //pop     bx

end;


{.241}
procedure  R_FROOT; assembler;
//ST(1) - ; ST(0) - ;
var
M: Cardinal;
MX: Extended;
asm
     //push      eax

     FXCH
     FIST      M
     FICOM     M
     FNSTSW    AX
     SAHF;
     JNZ        @@F   // 

     FXCH
     FSTP     ST(1)
     MOV      EAX, M
     CALL     R_IROOT




     JMP      @@END



@@F:
       // X=0 ??
     FXCH

     FLDZ            {.}
     FCOMP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     JNZ    @@ROOT   //X<>0
                          //   0^P; P -  

     // X=0
     // P=0 ??
     //FXCH            {.215}
     //FLDZ            {.215}
     FCOMPP           {.}
     FSTSW   AX      {.}
     SAHF            {.}
     fldz            {.215}
     JB    @@END              //P>=0
                              //0^(-P)     P -  
                              //call error
     fstp st(0)               {.215}
     fld1
     fchs
     fsqrt
     JMP     @@END      {.}

@@ROOT:
     FXCH
     fld1
     fdivrp
     fxch
     FYL2X
     FLD     ST(0)
     //FRNDINT
     FISTP M
     FILD  M
     FSUB    ST(1), ST
     FXCH    ST(1)
     F2XM1
     FLD1
     FADD
     {
     FSCALE
     FSTP    ST(1)
     }
    //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        //PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        //POP   EAX
    //**********************

@@END:
     //pop      eax
end;







procedure R_SINH; assembler;
var
MX: Extended;
M:  Cardinal;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
  //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
      {  PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }

             //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD   ST(0)
  FLD1
  FDIVR
  FSUB
  FMUL   C05

@@end:
end;


procedure R_COSECH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
       { PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }

        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD   ST(0)
  FLD1
  FDIVR
  FSUB
  FMUL   C05
  FDIVR  C1
@@end:
end;



procedure R_EXP; assembler;
var
M: Cardinal;
MX: Extended;
asm
        FLDL2E
        FMUL
        FLD     ST(0)
        //FRNDINT
        FISTP M
        FILD  M
        FSUB    ST(1), ST
        FXCH    ST(1)
        F2XM1
        FLD1
        FADD
       {
        FSCALE
        FSTP ST(1)
       }
        //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX    }

        //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@end

@@pwr2f:
        //PUSH    EAX
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        //POP     EAX
    //**********************
@@end:
end;




procedure R_LogN; assembler;
//ST(0) <- Y
//ST(1) <- X
asm
        FXCH
        FLD1
        FXCH
        FYL2X
        FXCH
        FLD1
        FXCH
        FYL2X
        FDIV
end;





procedure R_ARCSIN; assembler;
asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FPATAN
end;



procedure R_ARCCOS; assembler;
asm      //arctan(sqrt(1-sqr(S))/S)
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FXCH
  FPATAN
end;


procedure R_ARCSEC; assembler;
asm      //arccos(1/x)
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FXCH
  FPATAN
end;


procedure R_ARCCOSEC; assembler;
asm            //arcsin(1/x)
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUBR
  FSQRT
  FPATAN
end;



procedure R_COSH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }
         //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD     Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        //PUSH    EAX
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
        //POP     EAX
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FDIVR
  FADD
  FMUL  C05
@@end:
end;



procedure R_SECH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FDIVR
  FADD
  FMUL  C05
  FDIVR C1
@@end:
end;




procedure R_TANH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FIMUL   C2
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP    ST(1)
  }
   //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX   }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FSUB
  FLD1
  FADD    ST,ST(2)
  FDIV
  FSTP    ST(1)
@@end:
end;



procedure R_COTANH; assembler;
var
M: Cardinal;
MX: Extended;
asm
  FIMUL   C2
  FLDL2E
  FMUL
  FLD     ST(0)
  //FRNDINT
  FISTP M
  FILD  M
  FSUB    ST(1), ST
  FXCH    ST(1)
  F2XM1
  FLD1
  FADD
  {
  FSCALE
  FSTP   ST(1)
   }
    //******ISCALE********* {.243}    (instead FSCALE; FSTP ST(1) ; Agner Fog)
        {PUSH  EAX
        FXCH  st(1)
        FISTP M
        MOV   EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP st(1),st(0)
        POP   EAX  }
          //fpower >  Max2Power ??
        FXCH     ST(1)
        FILD     Max2Power
        FCOMIP   ST(0),ST(1)
        JNB      @@LowMaxF
         //call Overflow Error:
        fstp st(0)
        fstp st(0)
        fild Power2Overflow
        fld1
        fscale  //Float Overflow
        JMP      @@end

@@LowMaxF:  //fpower < Max2PowerN ??
        FILD      Max2PowerN
        FCOMIP   ST(0),ST(1)
        JB       @@pwr2f
        fstp     st(0)
        fstp     st(0)
        fldz
        JMP      @@exp

@@pwr2f:
        FISTP   M
        MOV     EAX,M
        ADD     EAX, 00003FFFH
        MOV     DWORD PTR [MX],   0
        MOV     DWORD PTR [MX+4], 80000000H
        MOV     DWORD PTR [MX+8], EAX
        FLD     TBYTE PTR [MX]
        FMULP   st(1),st(0)
    //**********************
@@exp:
  FLD     ST(0)
  FLD1
  FSUB
  FLD1
  FADD    ST,ST(2)
  FDIVR
  FSTP    ST(1)
@@end:
end;



procedure R_ARCSINH; assembler;
asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FADD
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;



procedure R_ARCCOSECH; assembler;
asm
  push  ecx
  push  eax
  xor   ecx,ecx

  fldz
  fcomp
  fstsw   ax
  sahf
  jb      @@bz
  mov ecx,1

@@bz:
  fabs

  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FADD
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X

  cmp     ecx, 0
  je     @@end
  fchs
  @@end:

  pop  eax
  pop  ecx
end;





procedure R_ARCCOSH; assembler;
asm
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUB
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;


procedure R_ARCSECH; assembler;
asm
  FDIVR C1
  FLD   ST
  FMUL  ST,ST
  FLD1
  FSUB
  FSQRT
  FXCH
  FADD
  FLDLN2
  FXCH  ST(1)
  FYL2X
end;


procedure R_ARCTANH; assembler;
asm     //ln((1+S)/(1-S))*0.5
  FLD  ST
  FLD1
  FADD
  FXCH
  FLD1
  FSUBR
  FDIV
  FLDLN2
  FXCH  ST(1)
  FYL2X
  FMUL  C05
end;




procedure R_ARCCOTANH; assembler;
asm     //ln((1+S)/(S-1))*0.5
  FLD  ST
  FLD1
  FADD
  FXCH
  FLD1
  FSUB
  FDIV
  FLDLN2
  FXCH  ST(1)
  FYL2X
  FMUL  C05
end;





procedure R_IPWR;assembler;
asm
        //PUSH     ECX
        //PUSH     EDX

        FLD1
        FXCH


        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jnz     @@2
        fstp    ST(0)
        jmp     @@3
   @@1: fmul    ST, ST
   @@2: shr     eax,1
        jnc     @@1
        fmul    ST(1),ST
        jnz     @@1
        fstp    st
        cmp     ecx, 0
        jge     @@3
        fdivr   c1
        {fld1
        fdivrp}
        //  :  "", : sin((3*x+4)^(2*i-1)) [x=2,i=5]
        //   wait
   @@3:
         //POP     EDX
         //POP     ECX
        {wait
        wait
        wait
        wait
        wait}
end;



procedure R_IPWR_Correct;assembler;
asm
        //PUSH     EDX
        //PUSH     ECX
        FLD1
        FXCH


        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jnz     @@2
        fstp    ST(0)
        jmp     @@3
   @@1: fmul    ST, ST
   @@2: shr     eax,1
        jnc     @@1
        fmul    ST(1),ST
        jnz     @@1
        fstp    st
        cmp     ecx, 0
        jge     @@3
        fld1
        fdivrp
        //  :  "", : sin((3*x+4)^(2*i-1)) [x=2,i=5]
        //   wait
   @@3:
        //POP     EDX
        //POP     ECX

        wait
        wait
        wait
        wait
        wait
end;




procedure R_IPWRSGN; assembler;
asm
{    and  eax,$80000001
    jns  @1
    dec  eax
    or   eax,-$02
    inc  eax
@1: test eax,eax
    fld1
    jz  @2
    fchs
@2:}
       fld1
       and  eax,1
       jz   @end
       fchs 
@end:
end;


//   Lib (flGetAddress)
procedure R_FACT; assembler;
//N -> EAX
asm
 push  ecx
 mov   ecx, dword ptr [Factorial]
 lea   ecx, [ecx+eax*8]
 lea   eax, [ecx+eax*2]
 fld   tbyte ptr [eax]
 pop   ecx
 ret
end;




procedure R_FACT2; assembler;
asm
 push  ecx
 mov   ecx, dword ptr [Factorial2F]
 lea   ecx, [ecx+eax*8]
 lea   eax, [ecx+eax*2]
 fld   tbyte ptr [eax]
 pop   ecx

(*
      //push eax
      //push ecx
      mov  ecx,eax
      and  ecx,1
      test ecx,ecx
      jnz  @_fct1
      shr  eax,1
      mov  ecx,dword ptr [Factorial2]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
      jmp  @end
  @_fct1:
      add  eax,1
      shr  eax,1
      mov  ecx,dword ptr [Factorial1]
      lea  ecx,[ecx+eax*8]
      lea  eax,[ecx+eax*2]
      fld  tbyte ptr [eax]
  @end:
      //pop  ecx
      //pop  eax
*)
end;



procedure R_SIGN; assembler;
asm
     push    ax
     ftst
     fstp  st(0)
     fstsw   ax
     sahf
     jnz     @@nz
     fldz
     jmp     @@end
 @@nz:
     fld1
     jnb      @@end
     fchs
 @@end:
     pop     ax
end;




procedure R_IPWRSGNF;   {.219}
var
M: Cardinal;
asm
     push      eax

     FIST      M
     FICOM     M
     FNSTSW    AX
     SAHF;
     JZ        @@1
                          //call error : (-1)^float
     fabs
     fchs
     fsqrt
     JMP       @@end
 @@1:
     mov eax, M
     fld1
     and   eax,$01
     fstp st(1)
     jz    @@end
     fchs
 @@end:
     pop    eax
end;




procedure RD_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   fld  qword ptr [ebx+8*eax]
end;



procedure RS_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   fld  dword ptr [ebx+4*eax]
end;



procedure RI_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   fild  dword ptr [ebx+4*eax]
end;


procedure RE_VEC; assembler;
asm              //load vec[num]: EBX<-addr(vec);  EAX<-num
   mov  ebx, [ebx]
   //fld  tbyte ptr [ebx+10*ecx]
   lea   eax,[eax+4*eax]
   fld   tbyte ptr [ebx+2*eax]
end;





procedure R_IPWRB;assembler;
asm

        PUSH     EAX
        PUSH     EBX
        PUSH     ECX
        PUSH     EDX

        FLD1
        FXCH


        mov     ecx, eax
        cdq
        xor     eax, edx
        sub     eax, edx
        jnz     @@2
        fstp    ST(0)
        jmp     @@3
   @@1: fmul    ST, ST
   @@2: shr     eax,1
        jnc     @@1
        fmul    ST(1),ST
        jnz     @@1
        fstp    st
        cmp     ecx, 0
        jge     @@3
        fdivr   c1

   @@3:
        POP     EDX
        POP     ECX
        POP     EBX
        POP     EAX
end;



{************************************begin**********************************}

procedure RD_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       faddp
       loop @@L
       @@end:
       pop  ecx
end;





procedure RS_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RI_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@end:
       pop  ecx
end;


procedure RE_SUM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       faddp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;




procedure RD_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fld  qword ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ecx
end;




procedure RS_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fld  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       fld  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ecx
end;



procedure RI_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fild  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ecx
end;


procedure RE_NORM; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       jecxz @@sqrt
       dec  ecx
       fld  tbyte ptr [eax]
       fmul st,st
       faddp
       jecxz @@sqrt
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       faddp
       loop @@L
       @@sqrt:
       fsqrt
       @@end:
       pop  ebx
       pop  ecx
end;





procedure RD_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@res
       dec  ecx
       fld  qword ptr [eax]
       faddp
       jecxz @@res
       @@L:
       fld  qword ptr [eax+8*ecx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ecx
end;




procedure RS_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       jecxz @@res
       dec  ecx
       fld  dword ptr [eax]
       faddp
       jecxz @@res
       @@L:
       fld  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ecx
end;




procedure RI_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@res
       dec  ecx
       fild  dword ptr [eax]
       faddp
       jecxz @@res
       @@L:
       fild  dword ptr [eax+4*ecx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ecx
end;




procedure RE_AVR; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       push ecx
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@res
       dec  ecx
       fld  tbyte ptr [eax]
       faddp
       jecxz @@res
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       faddp
       loop @@L
       @@res:
       fidiv [esp]
       pop  ecx
       @@end:
       pop  ebx
       pop  ecx
end;






procedure RS_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ecx
end;




procedure RI_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ecx
end;




procedure RD_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RE_SUMQ; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fmul st,st
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmul st,st
       faddp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;




procedure R_LEN; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       mov  ecx,eax
       test ecx,ecx
       jz @@nil
       mov  ecx,[ecx-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       @@nil:
       push ecx
       fild dword ptr [esp]
       pop  ecx
       pop  ecx
end;





procedure RD_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   qword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,8
 @@cycl:
     fld   qword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fld   qword ptr [ebx]
 @@next:
     add   ebx,8
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RS_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fld   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fld   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;






procedure RI_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fild   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fild   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fild   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RE_MAX; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   tbyte ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,10
 @@cycl:
     fld   tbyte ptr [ebx]
     fcomp
     fstsw ax
     sahf
     jb    @@next
     fstp  st(0)
     fld   tbyte ptr [ebx]
 @@next:
     add   ebx,10
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RD_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   qword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,8
 @@cycl:
     fld   qword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fld   qword ptr [ebx]
 @@next:
     add   ebx,8
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;




procedure RS_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fld   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fld   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;




procedure RI_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fild   dword ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,4
 @@cycl:
     fild   dword ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fild   dword ptr [ebx]
 @@next:
     add   ebx,4
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;



procedure RE_MIN; assembler;
//EAX<-adr(vector)
//EBX Adr
//ECX Len
asm
     push  ebx
     push  ecx
     //mov   eax, [eax]
     fldz
     test eax,eax
     jz @@end
     mov   ebx,eax
     mov   ecx,[eax-4]
     {$IFDEF FPC}
     add ecx,1
     {$ENDIF}
     jecxz @@end
     fstp st
     fld   tbyte ptr [ebx]
     sub   ecx,1
     jle   @@end
     add   ebx,10
 @@cycl:
     fld   tbyte ptr [ebx]
     fcomp
     fstsw ax
     sahf
     ja    @@next
     fstp  st(0)
     fld   tbyte ptr [ebx]
 @@next:
     add   ebx,10
     loop  @@cycl
 @@end:
     pop   ecx
     pop   ebx
end;







procedure RD_PROD; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fmulp
       loop @@L
       @@end:
       pop  ecx
end;





procedure RS_PROD; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       fmulp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RI_PROD; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fmulp
       loop @@L
       @@end:
       pop  ecx
end;





procedure RE_PROD;  assembler;
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fmulp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*ecx]
       lea   ebx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*ebx]
       fmulp
       loop @@L
       @@end:
       pop  ebx
       pop  ecx
end;






procedure RI_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call RI_AVR
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;



procedure RI_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RI_AVR
       //mov  edi,ecx
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;







procedure RS_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call RS_AVR
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;



procedure RS_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RS_AVR
       //mov  edi,ecx
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;



procedure RD_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RD_AVR
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;





procedure RD_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RD_AVR
       //mov  edi,ecx
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  ecx
end;






procedure RE_DEV; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RE_AVR
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  edx
       pop  ecx
end;





procedure RE_DEVS; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //push edi
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@endp
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@endp
       fstp st
       push ecx
       dec  ecx
       call  RE_AVR
       //mov  edi,ecx
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(1)
       fmul st(0),st(0)
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(2)
       fmul st(0),st(0)
       faddp
       loop @@L
       @@end:
       //pop  ecx
       dec [esp]
       fidiv [esp]
       fsqrt
       fstp st(1)
       //pop  edi
       pop  ecx
       @@endp:
       pop  edx
       pop  ecx
end;








procedure RI_SUMP; assembler;     //sum(vec[i]^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RS_SUMP; assembler;     //sum(vec[i]^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       pop  ecx
end;



procedure RD_SUMP; assembler;     //sum(vec[i]^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       pop  ecx
end;







procedure RE_SUMP; assembler;     //sum(vec[i]^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower
       push ecx
       push edx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       pop  edx
       pop  ecx
end;





procedure RI_SUMPS; assembler;        //sum((vec[i]-shift)^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower  ST(0)<-shift
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fild  dword ptr [eax]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fild  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       fstp st(1)
       pop  ecx
end;



procedure RS_SUMPS; assembler;        //sum((vec[i]-shift)^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower  ST(0)<-shift
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(1)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  dword ptr [eax]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fld  dword ptr [eax+4*ecx]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       fstp st(1)
       pop  ecx
end;







procedure RE_SUMPS; assembler;        //sum((vec[i]-shift)^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower  ST(0)<-shift
       push ecx
       push edx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(1)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  tbyte ptr [eax]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       //fld  tbyte ptr [eax+10*edx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       fstp st(1)
       pop  edx
       pop  ecx
end;



procedure RD_SUMPS; assembler;        //sum((vec[i]-shift)^intpower)
asm     //EAX<-adr(vector)  EBX<-intpower  ST(0)<-shift
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(1)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       jecxz @@end
       dec  ecx
       fld  qword ptr [eax]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       jecxz @@end
       @@L:
       fld  qword ptr [eax+8*ecx]
       fsub st(0),st(2)
       xchg  eax,ebx
       call R_IPWRB
       xchg  eax,ebx
       faddp
       loop @@L
       @@end:
       fstp st(1)
       pop  ecx
end;





procedure RD_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDD
end;


procedure RE_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDE
end;


procedure RI_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDI
end;


procedure RS_GCD; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_GCDS
end;





procedure RD_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCMD
end;


procedure RE_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCME
end;


procedure RI_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCMI
end;


procedure RS_LCM; assembler;
asm     //EAX<-adr(vector)
       push eax
       call V_LCMS
end;




procedure RE_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       push edx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       //fld  qword ptr [eax+8*ecx]  //vd[n-1]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       //fld  qword ptr [eax+8*ecx]
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  edx
       pop  ecx
end;




procedure RD_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       fld  qword ptr [eax+8*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;




procedure RS_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       fld  dword ptr [eax+4*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;



procedure RI_POLY; assembler;      //polynom(vec,x)=vec[n]*x^(n)+...+vec[1]*x+vec[0] ;n=Length(vec)-1
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end
       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       dec  ecx
       @@L:
       fmul st(0),st(1)   //s*x
       fild  dword ptr [eax+4*ecx]
       faddp
       dec ecx
       cmp ecx,-$1
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;





procedure RD_POLYDIFF1; assembler;         //d/dx(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$1
       jz @@end


       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end



       push ecx
       fild [esp]
       pop  ecx
       fmulp

       dec  ecx

       cmp ecx,$0
       jz @@end

       @@L:
       fmul st(0),st(1)   //s*x
       fld  qword ptr [eax+8*ecx]

       push ecx
       fild [esp]
       pop  ecx
       fmulp

       faddp
       dec ecx
       cmp ecx,$0
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;




procedure RD_POLYDIFF2; assembler;        // d2/dx2(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$2
       jle @@end
       //jb @@end

       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end


       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       dec  ecx
       cmp ecx,$1
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fld  qword ptr [eax+8*ecx]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       faddp
       dec ecx
       cmp ecx,$1
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;



procedure RD_POLYDIFF3; assembler;         // d3/dx3(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$3
       jle @@end

       fstp st
       dec  ecx
       fld  qword ptr [eax+8*ecx]  //vd[n-1]
       jecxz @@end
       {cmp ecx,$1
       jz @@end }

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12



       dec  ecx
       cmp ecx,$2
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fld  qword ptr [eax+8*ecx]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12


       faddp
       dec ecx
       cmp ecx,$2
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;




procedure RE_POLYDIFF1; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx

       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$1
       jz @@end


       fstp st
       dec  ecx
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end



       push ecx
       fild [esp]
       pop  ecx
       fmulp

       dec  ecx

       cmp ecx,$0
       jz @@end

       @@L:
       fmul st(0),st(1)   //s*x
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]

       push ecx
       fild [esp]
       pop  ecx
       fmulp

       faddp
       dec ecx
       cmp ecx,$0
       jnz @@L
       @@end:
       fstp st(1)

       pop  edx
       pop  ecx
end;




procedure RE_POLYDIFF2; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$2
       jle @@end
       //jb @@end

       fstp st
       dec  ecx
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end


       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       dec  ecx
       cmp ecx,$1
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       faddp
       dec ecx
       cmp ecx,$1
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  edx
       pop  ecx
end;



procedure RE_POLYDIFF3; assembler;
asm     //EAX<-adr(vector)
       push ecx
       push edx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$3
       jle @@end

       fstp st
       dec  ecx
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]
       jecxz @@end
       {cmp ecx,$1
       jz @@end }

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12



       dec  ecx
       cmp ecx,$2
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       lea   edx,[ecx+4*ecx]
       fld   tbyte ptr [eax+2*edx]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12


       faddp
       dec ecx
       cmp ecx,$2
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  edx
       pop  ecx
end;





procedure RS_POLYDIFF1; assembler;         //d/dx(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$1
       jz @@end


       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end



       push ecx
       fild [esp]
       pop  ecx
       fmulp

       dec  ecx

       cmp ecx,$0
       jz @@end

       @@L:
       fmul st(0),st(1)   //s*x
       fld  dword ptr [eax+4*ecx]  //vd[n-1]

       push ecx
       fild [esp]
       pop  ecx
       fmulp

       faddp
       dec ecx
       cmp ecx,$0
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;




procedure RS_POLYDIFF2; assembler;        // d2/dx2(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$2
       jle @@end
       //jb @@end

       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end


       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       dec  ecx
       cmp ecx,$1
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fld  dword ptr [eax+4*ecx]  //vd[n-1]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       faddp
       dec ecx
       cmp ecx,$1
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;



procedure RS_POLYDIFF3; assembler;         // d3/dx3(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$3
       jle @@end

       fstp st
       dec  ecx
       fld  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       {cmp ecx,$1
       jz @@end }

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12



       dec  ecx
       cmp ecx,$2
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fld  dword ptr [eax+4*ecx]  //vd[n-1]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12


       faddp
       dec ecx
       cmp ecx,$2
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;







procedure RI_POLYDIFF1; assembler;         //d/dx(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$1
       jz @@end


       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end



       push ecx
       fild [esp]
       pop  ecx
       fmulp

       dec  ecx

       cmp ecx,$0
       jz @@end

       @@L:
       fmul st(0),st(1)   //s*x
       fild  dword ptr [eax+4*ecx]  //vd[n-1]

       push ecx
       fild [esp]
       pop  ecx
       fmulp

       faddp
       dec ecx
       cmp ecx,$0
       jnz @@L
       @@end:
       fstp st(1)
       pop  ecx
end;




procedure RI_POLYDIFF2; assembler;        // d2/dx2(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$2
       jle @@end
       //jb @@end

       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end


       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       dec  ecx
       cmp ecx,$1
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fild  dword ptr [eax+4*ecx]  //vd[n-1]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       //pop  ebx
       fmulp
       //pop  ebx
       add esp, 8


       faddp
       dec ecx
       cmp ecx,$1
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;



procedure RI_POLYDIFF3; assembler;         // d3/dx3(polynom(vec,x)
asm     //EAX<-adr(vector)
       push ecx
       push ebx
       //mov  eax, [eax]
       fldz
       test eax,eax
       jz @@end
       mov  ecx,[eax-4]
       {$IFDEF FPC}
       add ecx,1
       {$ENDIF}
       jecxz @@end

       cmp ecx,$3
       jle @@end

       fstp st
       dec  ecx
       fild  dword ptr [eax+4*ecx]  //vd[n-1]
       jecxz @@end
       {cmp ecx,$1
       jz @@end }

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12



       dec  ecx
       cmp ecx,$2
       jz @@end
       @@L:
       fmul st(0),st(1)   //s*x
       fild  dword ptr [eax+4*ecx]  //vd[n-1]

       mov  ebx,ecx
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       sub  ebx,1
       fmulp
       push ebx
       fild [esp]
       fmulp
       add esp,12


       faddp
       dec ecx
       cmp ecx,$2
       jnz @@L
       @@end:
       fstp st(1)

       pop  ebx
       pop  ecx
end;



{**************************************end**********************************}







procedure R_MAX(adr,len: Cardinal); stdcall;
var
i:Integer;
MAX: Extended;
begin
MAX:=PExtended(Adr)^;
for I := 1 to len - 1 do
begin
  if PExtended(Adr+16*i)^ > MAX then  MAX:=PExtended(Adr+16*i)^;
end;

 asm
  fld MAX
 end
end;




procedure R_MIN(adr,len: Cardinal); stdcall;
var
i:Integer;
MIN: Extended;
begin
MIN:=PExtended(Adr)^;
for I := 1 to len - 1 do
begin
  if PExtended(Adr+16*i)^ < MIN then  MIN:=PExtended(Adr+16*i)^;
end;

 asm
  fld MIN
 end
end;



procedure R_AVR(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+PExtended(Adr+16*i)^;
end;

S:=S/len;
 asm
  fld S
 end
end;




procedure R_SUM(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+PExtended(Adr+16*i)^;
end;

 asm
  fld S
 end
end;




procedure R_SUMQ(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^);
end;

 asm
  fld S
 end
end;




procedure R_NORM(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^);
end;

 asm
  fld S
  fsqrt
 end
end;




procedure R_PROD(adr,len: Cardinal); stdcall;
var
i:Integer;
S: Extended;
begin
S:=1;
for I := 0 to len - 1 do
begin
  S:=S*PExtended(Adr+16*i)^;
end;

 asm
  fld S
 end
end;



procedure R_DEV(adr,len: Cardinal); stdcall;
var
i:Integer;
S,avg: Extended;
begin

avg:=0;
for I := 0 to len - 1 do
begin
  avg:=avg+PExtended(Adr+16*i)^;
end;

avg:=avg/len;

S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^-avg);
end;

S:=sqrt(S/len);

 asm
  fld S
 end
end;



procedure R_DEVS(adr,len: Cardinal); stdcall;
var
i:Integer;
S,avg: Extended;
begin

avg:=0;
for I := 0 to len - 1 do
begin
  avg:=avg+PExtended(Adr+16*i)^;
end;

avg:=avg/len;

S:=0;
for I := 0 to len - 1 do
begin
  S:=S+sqr(PExtended(Adr+16*i)^-avg);
end;

S:=sqrt(S/(len-1));

 asm
  fld S
 end
end;


{.199}
{
procedure _MAX2; assembler;
asm
     //push eax
     fcom
     fstsw ax
     sahf
     jb    @@end
     fxch st(1)
@@end:
     fstp  st(0)
     //pop eax
end;



procedure _MIN2; assembler;
asm
     //push eax
     fcom
     fstsw ax
     sahf
     jnb    @@end
     fxch st(1)
@@end:
     fstp  st(0)
     //pop eax
end;
}




procedure R_GCD(adr,len: Cardinal); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r: longint;
begin

if len = 0  then S:=0 else
if len = 1  then S:=PExtended(Adr)^ else
begin
  n:=1;
  u:=Trunc(PExtended(Adr)^);
  v:=Trunc(PExtended(Adr+16)^);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(PExtended(Adr+16*n)^);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;





procedure V_GCDD(VD: TArrayD); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,len: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;



procedure V_GCDE(VD: TArrayE); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,len: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_GCDI(VD: TArrayI); stdcall;
label lp,endp;
var
i,n:Integer;
Res: Extended;
u,v,t,r,len,S: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=(VD[0]);
  v:=(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=(S);
    u:=(VD[n]);
    goto lp;
  end;

end;

endp:
Res:=S;
 asm
  fld Res
 end
end;




procedure V_GCDS(VD: TArrayS); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,len: longint;
begin

len:=Length(VD);
if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);

lp:
  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  S := abs(u);
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;



procedure R_LCM(adr,len: Cardinal); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1: longint;
begin


if len = 0  then S:=0 else
if len = 1  then S:=PExtended(Adr)^ else
begin
  n:=1;
  u:=Trunc(PExtended(Adr)^);
  v:=Trunc(PExtended(Adr+16)^);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(PExtended(Adr+16*n)^);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_LCMD(VD: TArrayD); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1,len: longint;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_LCME(VD: TArrayE); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1,len: longint;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;



procedure V_LCMS(VD: TArrayS); stdcall;
label lp,endp;
var
i,n:Integer;
S: Extended;
u,v,t,r,u1,v1,len: longint;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=Trunc(VD[0]);
  v:=Trunc(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=Trunc(S);
    u:=Trunc(VD[n]);
    goto lp;
  end;

end;

endp:
 asm
  fld S
 end
end;




procedure V_LCMI(VD: TArrayI); stdcall;
label lp,endp;
var
i,n:Integer;
S: longint;
u,v,t,r,u1,v1,len: longint;
Res: Extended ;
begin

len:=Length(VD);

if len = 0  then S:=0 else
if len = 1  then S:=VD[0] else
begin
  n:=1;
  u:=(VD[0]);
  v:=(VD[1]);


lp:
  u1:=u; v1:=v;

  while v <> 0 do
  begin
    t := u;
    u := v;
    v := t mod v;
  end;

  u:=(u1*v1) div abs(u);

  S := u;
  inc(n);
  if n > len-1 then goto endp
  else
  begin
    v:=(S);
    u:=(VD[n]);
    goto lp;
  end;

end;

endp:
Res:=S;
 asm
  fld Res
 end
end;




function PolyDiffD(va: PArrayD; x: PExtended;  p: Pinteger): extended;  stdcall;

label endp;
var
  i,n: integer;
  s,c,res: extended;
begin
 n:=length(va^);

  if (n-p^)<=0 then begin
    s := 0.0;
    goto endp;
  end;


  c:=factorial[n-1]/factorial[n-1-p^];
  s := va^[n-1]*c;
  for i:=n-2 downto p^ do
  begin
    c:=factorial[i]/factorial[i-p^];
    s := s*x^ + va^[i]*c;
  end;


endp:
PolyDiffD:=s;
end;



function PolyDiffE(va: PArrayE; x: PExtended;  p: Pinteger): extended;  stdcall;

label endp;
var
  i,n: integer;
  s,c,res: extended;
begin
 n:=length(va^);

  if (n-p^)<=0 then begin
    s := 0.0;
    goto endp;
  end;


  c:=factorial[n-1]/factorial[n-1-p^];
  s := va^[n-1]*c;
  for i:=n-2 downto p^ do
  begin
    c:=factorial[i]/factorial[i-p^];
    s := s*x^ + va^[i]*c;
  end;


endp:
PolyDiffE:=s;
end;



function PolyDiffS(va: PArrayS; x: PExtended;  p: Pinteger): extended;  stdcall;

label endp;
var
  i,n: integer;
  s,c,res: extended;
begin
 n:=length(va^);

  if (n-p^)<=0 then begin
    s := 0.0;
    goto endp;
  end;


  c:=factorial[n-1]/factorial[n-1-p^];
  s := va^[n-1]*c;
  for i:=n-2 downto p^ do
  begin
    c:=factorial[i]/factorial[i-p^];
    s := s*x^ + va^[i]*c;
  end;


endp:
PolyDiffS:=s;
end;



function PolyDiffI(va: PArrayI; x: PExtended;  p: Pinteger): extended;  stdcall;

label endp;
var
  i,n: integer;
  s,c,res: extended;
begin
 n:=length(va^);

  if (n-p^)<=0 then begin
    s := 0.0;
    goto endp;
  end;


  c:=factorial[n-1]/factorial[n-1-p^];
  s := va^[n-1]*c;
  for i:=n-2 downto p^ do
  begin
    c:=factorial[i]/factorial[i-p^];
    s := s*x^ + va^[i]*c;
  end;


endp:
PolyDiffI:=s;
end;




function PolyDiffN(pvec: Pointer; px: PExtended;  p: Pinteger; VType: PInteger): extended;  stdcall;

begin

if VType^ = _ARRAYD then
begin
  PolyDiffN:=PolyDiffD(PArrayD(pvec),px,p)
end
else
if VType^ = _ARRAYE then
begin
  PolyDiffN:=PolyDiffE(PArrayE(pvec),px,p)
end
else
if VType^ = _ARRAYS then
begin
  PolyDiffN:=PolyDiffS(PArrayS(pvec),px,p)
end
else
if VType^ = _ARRAYI then
begin
  PolyDiffN:=PolyDiffI(PArrayI(pvec),px,p)
end
;

end;



procedure SetLengthArray(AdrV: Pointer; TypeV: Cardinal; Len: Cardinal);  stdcall;
var
L1: integer;
begin

if TypeV = _ARRAYD then
begin
  SetLength(PArrayD(AdrV)^,Len);
end
else
if TypeV = _ARRAYE then
begin
  SetLength(PArrayE(AdrV)^,Len);
end
else
if TypeV = _ARRAYI then
begin
  SetLength(PArrayI(AdrV)^,Len);
end
else
if TypeV = _ARRAYS then
begin
  SetLength(PArrayS(AdrV)^,Len);
end;

end;





function std_rand2(x1,x2:PExtended): extended;  stdcall;
begin
 std_rand2:=(x1^-x2^)*random+x2^;
end;



procedure _GetTickProc(var ht,lt:Cardinal) ;
begin
asm
   push   eax
   push   edx
   RDTSC  //eax:edx
   mov    lt,eax
   mov    ht,edx
   pop    edx
   pop    eax
end;
end;


//randomize()
procedure _Randomize;
var
ht,lt,ISeed: Cardinal;
b1,b2: Integer;
begin
InitRandomGenerator;
{  asm
   push   eax
   push   edx
   RDTSC  //eax:edx
   shr    eax,10
   mov    lt,eax
   mov    ht,edx
   pop    edx
   pop    eax
 end;

  b1:=1; b2:=1;
  if lt  mod 2 <> 0 then b1:=-1;
  if ht  mod 2 <> 0 then b2:=-1;

  ISeed:=b1*lt+b2*ht;

  mt19937_init(ctxMT,ISeed);
  taus113_init(ctxPLE,ISeed);
  xor4096_init(ctxXOR,ISeed);
  mt19937_init(ctxMTG,ISeed); //only for RandG
  }
end;

procedure InitRandomGenerator;
var

T1,i,nrnd,x1,x2,ht,lt,ISeed: Cardinal;

DT,DT2,Seed,b1,b2,b3,b4,b5,b6,b7,r: double;
DY,DM,DD,DHr,DMn,DSc,DMs: WORD;
begin
  {.209}
 DT:=Now;
 //DecodeDate( DT, DY, DM, DD);
 DecodeTime( DT, DHr, DMn, DSc, DMs);

 T1:=GetTickCount;

 asm
   push   eax
   push   edx
   RDTSC  //eax:edx
   mov    lt,eax
   mov    ht,edx
   pop    edx
   pop    eax
end;

 


 b1:=1; b2:=1; b3:=1; b4:=1; b5:=1; b6:=1;b7:=1;
 if DHr mod 2 <> 0 then b1:=-1;
 if DMn mod 2 <> 0 then b2:=-1;
 if DSc mod 2 <> 0 then b3:=-1;
 if DMs mod 2 <> 0 then b4:=-1;
 if T1  mod 2 <> 0 then b5:=-1;
 if lt  mod 2 <> 0 then b6:=-1;
 if ht  mod 2 <> 0 then b7:=-1;

 //ISeed:=Trunc(DHr*b1+DMs*b2+DSc*b3+DMs*b4+T1*b5+lt*b6+ht*b7);
 //ISeed:=DHr*DMs*DSc*DMs*T1*lt*ht;
 ISeed:=(DHr+1)*(DMs+1)*(DSc+1)*(DMs+1)*(T1+1)*(lt+1)*(ht+1);

 mt19937_init(ctxMT,ISeed);
 taus113_init(ctxPLE,ISeed);
 xor4096_init(ctxXOR,ISeed);
 mt19937_init(ctxMTG,ISeed); //only for RandG

 mt19937_next(ctxMT);
 taus113_next(ctxPLE);
 xor4096_next(ctxXOR);
 mt19937_next(ctxMTG);  //only for RandG



end;




function RandomMT: double;
  {-Next random double in [0..1) with full double 53 bit precision}
var
  hb,lb: longint;
begin
  mt19937_next(ctxMTG);
  hb := ctxMTG.nr shr 5;
  mt19937_next(ctxMTG);
  lb := ctxMTG.nr shr 6;
  RandomMT := (hb*67108864.0+lb)/9007199254740992.0;
end;


function _RandG(Mean, StdDev: PExtended): Extended;  stdcall;
{ Marsaglia-Bray algorithm }
var
  U1, S2: Extended;
begin
  repeat
    U1 := 2*RandomMT - 1;
    S2 := Sqr(U1) + Sqr(2*RandomMT-1);
  until S2 < 1;
  Result := Sqrt(-2*Ln(S2)/S2) * U1 * StdDev^ + Mean^;
end;
          //***************
//randomize()
//randmt(), randple(), randxor()  in (0,1)
//intrandmt(Num), intrandple(Num), intrandxor(Num)  in (0,Num) Num - int
//rand(array)



{********************SPLAIN3*************************************************}



procedure _LSE(A: TArray2E;  B: TArrayE; var B1: TArrayE; var D: TFloatType);
begin
 if _LSEType = f_Jordan then _Jordan(A,B,B1,D) else
 if _LSEType = f_Gauss then _Gausse(A,B,B1,D);
end;





procedure _Gausse(A: TArray2E; B: TArrayE; var B1: TArrayE; var D: TFloatType);
label endp;
var
i,j,k,N: Cardinal;
S,T,D1,x: TFloatType;
E,Z: Integer;
adrA,adrB,adrB1: TAddress;
P: Pointer;
begin
 N:=Length(A);
 SetLength(B1,N);
 dec(N);
 S_Error:=0;
 adrA:=TAddress(A);  adrB:=TAddress(B); adrB1:=TAddress(B1);
 F_Clm:=0;

 for I := 0 to N - 1 do
   begin
     B1[i]:=0;
   end;

 D1:=1;
 Z:=1;


//EXTENDED:

 for k:=0 to N do
 begin
  if A[k,k] = 0 then
  begin
   _XCHSC(k,k,A,B,E);
   if S_Error = 1 then goto endp;        //ERROR!!!!!!!!!!!!!!!
   Z:=-Z;
   end;
  T:=1/A[k,k];
  D1:=D1*A[k,k];
  for i:=k to N do
  begin
   for j:=k to N do
   begin
    if i=k then A[i,j]:=A[i,j]*T else
    if j>k then A[i,j]:=A[i,j]-A[i,k]*A[k,j];
   end;

   if i=k then B[i]:=B[i]*T else
   if i>k then B[i]:=B[i]-A[i,k]*B[k];
  end;
 end;

 B1[N]:=B[N];

 for k:=N-1 downto 0 do
 begin
  S:=0;
  for i:=N downto k do
  begin
   S:=S+A[k,i]*B1[i];
  end;
   B1[k]:=B[k]-S;
 end;



D:=D1*Z;

if F_Clm = 1 then
begin
 for i:=0 to High(G_TC) do
 begin
  x:=B1[G_TC[i].I1];
  B1[G_TC[i].I1]:=B1[G_TC[i].I2];
  B1[G_TC[i].I2]:=x;
 end;
end;

endp:
end;



procedure _Jordan(A: TArray2E; B: TArrayE; var B1: TArrayE; var D: TFloatType);
label endp;
var
N,i,j,k,N1,k1: TIntegType;
T,T1,T2,D1,x: TFloatType;
E,Z: Integer;
adrA,adrB: TAddress;
begin
S_Error:=0;
N:=High(A);
adrA:=TAddress(A);
adrB:=TAddress(B);
F_Clm:=0;
Z:=1;
D1:=1;



//EXTENDED:

for k:=0 to N do
begin

 if A[k,k] = 0 then
 begin
  _XCHSC(k,k,A,B,E);
  if S_Error = 1 then goto endp;   //ERROR!!!!!!!!!!!!!!!
  Z:=-Z;
 end;

 T:=1/A[k,k];

 D1:=D1*A[k,k];

 for j:=k to N do
  begin
    A[k,j]:=A[k,j]*T;
  end;

  B[k]:=B[k]*T;



 for i:=0 to N do
 begin

  if i <> k then
  begin

   T1:=A[i,k];

   for j:=k to N do
   begin
    A[i,j]:=A[i,j]-A[k,j]*T1;
   end;

    B[i]:=B[i]-B[k]*T1;

  end;
 end;
end;


D:=D1*Z;

if F_Clm = 1 then
begin
 for i:=0 to High(G_TC) do
 begin
  x:=B[G_TC[i].I1];
  B[G_TC[i].I1]:=B[G_TC[i].I2];
  B[G_TC[i].I2]:=x;
 end;
end;

B1:=B;
endp:
end;




procedure _CreateSplain3(MX: TArrayE; MF: TArrayE; var idf: cardinal; var E: integer);
label endp;
var
T,Max,D: TFloatType;
i,j,k,l,N: LongInt;
B,B1,h,M,F: TArrayE;
A: TArray2E;
MX1,MF1: TArrayE;
A3,A2,A1,A0,PS,Pk0,Pk: TArrayE;
SL: Integer;
SP: TSplain;
begin
N:=High(MX);
S_Error:=0;

SL:=N;
SetLength(F,N+1); SetLength(A,N,N);
SetLength(PS,N+1);  SetLength(B,N);    SetLength(h,N+1);  SetLength(M,N+1);
SetLength(A3,N+1); SetLength(A2,N+1); SetLength(A1,N+1); SetLength(A0,N+1);

SetLength(MX1,N+2);
SetLength(MF1,N+2);
for i:=1 to N+1 do
begin
 MX1[i]:=MX[i-1];
 MF1[i]:=MF[i-1];
end;


k:=N+1;
for i:=1 to N+1 do
begin
 Max:=MX1[1]; l:=1;
 for j:=1 to k do
  begin
   if MX1[j] > Max then
   begin
    Max:=MX1[j]; l:=j;
   end;
 end;

 if l <> k then
 begin
  T:=MX1[l];
  MX1[l]:=MX1[k];
  MX1[k]:=T;
  T:=MF1[l];
  MF1[l]:=MF1[k];
  MF1[k]:=T;
 end;
 dec(k);
end;


for i:=0 to N do
begin
F[i]:=MF1[i+1];
PS[i]:=MX1[i+1];
end;


for i:=1 to N do
begin
h[i]:=PS[i]-PS[i-1];
end;

for i:=1 to N-1 do
begin
B[i]:=(F[i+1]-F[i])/h[i+1]-(F[i]-F[i-1])/h[i];
end;


for i:=1 to N-1 do
begin
for j:=1 to N-1 do
begin
 if j = i-1 then A[i,j]:=h[i]/6
 else
 if j = i then A[i,j]:=(h[i]+h[i+1])/3
 else
 if j = i+1 then A[i,j]:=h[i+1]/6
 else A[i,j]:=0;
end;
end;


////////////////////
for i:=1 to N-1 do
begin
 for j:=1 to N-1 do
 begin
  A[i-1,j-1]:=A[i,j];
 end;
 B[i-1]:=B[i];
end;

SetLength(A,N-1,N-1);
SetLength(B,N-1);
///////////////////


_LSE(A,B,B1,D);
if S_Error = 1  then goto endp;




M[0]:=0; M[N]:=0;
for i:=1 to N-1 do
begin
 M[i]:=B1[i-1];
end;

for i:=1 to N do
begin
 A3[i]:=(M[i]-M[i-1])/(6*h[i]);
 A2[i]:=(M[i-1]*PS[i]-M[i]*PS[i-1])/(2*h[i]);
 A1[i]:=(3*M[i]*sqr(PS[i-1])-3*M[i-1]*sqr(PS[i])-6*F[i-1]+M[i-1]*sqr(h[i])+
        6*F[i]-M[i]*sqr(h[i]))/(6*h[i]);
 A0[i]:=(M[i-1]*intpower(PS[i],3)-M[i]*intpower(PS[i-1],3)+6*PS[i]*F[i-1]-
        M[i-1]*sqr(h[i])*PS[i]-6*PS[i-1]*F[i]+M[i]*sqr(h[i])*PS[i-1])/(6*h[i]);
end;


SetLength(SP.A0,N);
SetLength(SP.A1,N);
SetLength(SP.A2,N);
SetLength(SP.A3,N);
SetLength(SP.PS,N+1);
SP.SL:=SL-1;
for i:=0 to N-1 do
begin
 SP.A0[i]:=A0[i+1];
 SP.A1[i]:=A1[i+1];
 SP.A2[i]:=A2[i+1];
 SP.A3[i]:=A3[i+1];
 SP.PS[i]:=PS[i];
end;
 SP.PS[N]:=PS[N];
 SP.DS:=abs(SP.PS[0]-SP.PS[SP.SL])/(N-1);


 SetLength(GSPL3,Length(GSPL3)+1);
 GSPL3[High(GSPL3)]:=SP;
 idf:=High(GSPL3);


A0:=nil;  A1:=nil; A2:=nil; A3:=nil; PS:=nil;
B:=nil; B1:=nil; H:=nil; M:=nil; F:=nil;
A:=nil; MX1:=nil; MF1:=nil;
endp:
E:=S_Error;
end;





procedure _XCHSC(i,j: TIntegType; M: TArray2E; B: TArrayE; var E: Integer);
label endp;
var
x: TFloatType;
l,k,N: TIntegType;
begin
E:=0;
N:=High(M);
for k:=i to N do
begin
 if M[k,j] <> 0 then
 begin
  for l:=0 to N do
  begin
   x:=M[i,l]; M[i,l]:=M[k,l]; M[k,l]:=x;
  end;
   x:=B[i]; B[i]:=B[k]; B[k]:=x;
  goto endp;
 end;
end;



for k:=j to N do
begin
 if M[i,k] <> 0 then
 begin
  for l:=0 to N do
  begin
   x:=M[l,j]; M[l,j]:=M[l,k]; M[l,k]:=x;
   SetLength(G_TC,Length(G_TC)+1);
   G_TC[High(G_TC)].I1:=j;
   G_TC[High(G_TC)].I2:=k;
   F_Clm:=1;
  end;
  goto endp;
 end;
end;


S_Error:=1;

endp:
end;














function _FSPL3(x: PExtended; id: integer): TFloatType; stdcall;
label 1,2,3,4,endp;
var
SP: ^TSplain;
i,j,SL,Nd,Nu,M: integer;
begin

SP:=@GSPL3[id];
SL:=GSPL3[id].SL+1;


if (x^ < SP^.PS[0]) or (x^ > SP^.PS[SL]) then
begin
 _FSPL3:=0;
 goto endp;
end;

(*
//if SP.EqDs = True then
begin
  m:=trunc(abs(SP^.DS*x^));
  if m = SL then  dec(m);

  if (x^ >= SP^.PS[m]) and (x^ <= SP^.PS[m+1]) then goto 1;
end;
*)


Nd:=-1;   //min-1
Nu:=SL+1; //max+1
2:
//while Nd <> Nu-1 do
begin
  M:=(Nd+Nu) div 2;



  if (x^ >= SP^.PS[M]) and (x^ <= SP^.PS[M+1]) then goto 4
  else
  if (x^ >= SP^.PS[M])
  then Nd:=M
  else Nu:=M;

end;
goto 2;
4:
//j:=M;




{
for i:=0 to SP.SL do
begin
 if (x >= SP.PS[i]) and (x <= SP.PS[i+1]) then begin j:=i; goto 1; end
 else
 if (x <= SP.PS[SL-i]) and (x >= SP.PS[SL-i-1]) then
  begin
    j:=SL-i-1; goto 1;
  end;
end;
 }

1:
//_FSPL3:=SP.A3[j]*x*sqr(x)+SP.A2[j]*sqr(x)+SP.A1[j]*x+SP.A0[j];
_FSPL3:=x^*(x^*(SP^.A3[m]*x^+SP^.A2[m])+SP^.A1[m])+SP^.A0[m];
endp:

  //_FSPL3:=fSplain3(x,GSPL3[id]);
end;




function _FSPL3D1(x: PExtended; id: integer): TFloatType; stdcall;
label 1,2,endp;
var
SP: ^TSplain;
i,j,SL,Nd,Nu,M: integer;
begin

SP:=@GSPL3[id];
SL:=GSPL3[id].SL+1;


if (x^ < SP^.PS[0]) or (x^ > SP^.PS[SL]) then
begin
 _FSPL3D1:=0;
 goto endp;
end;


Nd:=-1;   //min-1
Nu:=SL+1; //max+1

2:
//while Nd <> Nu-1 do
begin
  M:=(Nd+Nu) div 2;
  if (x^ >= SP^.PS[M]) and (x^ <= SP^.PS[M+1]) then goto 1
  else
  if (x^ >= SP^.PS[M])
  then Nd:=M
  else Nu:=M;
end;
goto 2;

1:
_FSPL3D1:=(3*SP.A3[m]*x^+2*SP.A2[m])*x^+SP.A1[m];{3*SP^.A3[m]*sqr(x^)+2*SP^.A2[m]*x^+SP^.A1[m];}
endp:
end;




function _FSPL3D2(x: PExtended; id: integer): TFloatType; stdcall;
label 1,2,endp;
var
SP: ^TSplain;
i,j,SL,Nd,Nu,M: integer;
begin

SP:=@GSPL3[id];
SL:=GSPL3[id].SL+1;


if (x^ < SP^.PS[0]) or (x^ > SP^.PS[SL]) then
begin
 _FSPL3D2:=0;
 goto endp;
end;


Nd:=-1;   //min-1
Nu:=SL+1; //max+1

2:
//while Nd <> Nu-1 do
begin
  M:=(Nd+Nu) div 2;
  if (x^ >= SP^.PS[M]) and (x^ <= SP^.PS[M+1]) then goto 1
  else
  if (x^ >= SP^.PS[M])
  then Nd:=M
  else Nu:=M;
end;
goto 2;


1:
_FSPL3D2:=6*SP^.A3[m]*x^+2*SP^.A2[m];
endp:
end;




function _FSPL3D3(x: PExtended; id: integer): TFloatType; stdcall;
label 1,2,endp;
var
SP: ^TSplain;
i,j,SL,Nd,Nu,M: integer;
begin

SP:=@GSPL3[id];
SL:=GSPL3[id].SL+1;


if (x^ < SP^.PS[0]) or (x^ > SP^.PS[SL]) then
begin
 _FSPL3D3:=0;
 goto endp;
end;


Nd:=-1;   //min-1
Nu:=SL+1; //max+1

2:
//while Nd <> Nu-1 do
begin
  M:=(Nd+Nu) div 2;
  if (x^ >= SP^.PS[M]) and (x^ <= SP^.PS[M+1]) then goto 1
  else
  if (x^ >= SP^.PS[M])
  then Nd:=M
  else Nu:=M;
end;
goto 2;

1:
_FSPL3D3:=6*SP^.A3[m];
endp:
end;




function _FSPL3D4(x: PExtended; id: integer): TFloatType; stdcall;
begin
  _FSPL3D4:=0;
end;




 //vd[irand(len(vd))]
 //arand(vd)
procedure RAND_ARRAYD(vd: TArrayD); stdcall;  //assembler;
const
  maxint: cardinal = 2147483647;{4294967295}
  divmaxint: extended = 1/2147483647;
  //divmaxlong: extended = 1/4294967295;
var
  Len: Cardinal;
  rn,mem: Cardinal;
begin
   Len:=Length(vd);

   mt19937_next(ctxMT);
   rn:=ctxMT.nr shr 1;

  asm
    push eax
    push ebx

    fild  rn
    fimul Len
    fld tbyte ptr divmaxint
    fmulp
    //FRNDINT

    fistp mem
    //pop eax
    mov  eax,mem
    mov  ebx, vd
    //mov  ebx, [ebx]
    fld  qword ptr [ebx+8*eax]

    pop  ebx
    pop  eax
  end;

end;
(*
asm     //EAX<-adr(vector)
       push ecx
       //mov  eax, [eax]
       mov  ecx,eax
       test ecx,ecx
       jz @@nil
       mov  ecx,[ecx-4]
       {$IFDEF FPC}
        add ecx,1
       {$ENDIF}
       @@nil:


       {push ecx
       fild dword ptr [esp]}

       pop  ecx
       pop  ecx
end;
*)




 (*
procedure MaskFPUException;    {.248}
asm
    fnstcw word ptr [CWMEM];       //->store
    sub esp,4
    fnstcw [esp];
    or word ptr [esp], 003Fh;
    fldcw [esp];
    add esp, 4
end;



  {.248}
procedure IsFPUException; //function: st(0)
asm
    push eax
    xor  eax,eax
    fnstsw ax
    sahf
    test ax, 001Fh
    fldz
    jz @@Z
    fstp st(0)
    fld1
    @@Z:
    pop eax
    fnclex
    fldcw word ptr [CWMEM]   //<-load
end;

 *)

procedure MaskFPUExceptionA;    {.249}
asm
    fnstcw word ptr [ECX];       //->store
    sub esp,4
    fnstcw [esp];
    or word ptr [esp], 003Fh;
    fldcw [esp];
    add esp, 4
 end;



 {.249}
procedure IsFPUExceptionA;         {.249}
asm
    {xor  eax,eax
    fnstsw ax
    sahf
    test ax, 001Fh
    fldz
    jz @@Z
    fstp st(0)
    fld1
    @@Z:
    fnclex
    fldcw word ptr [ECX]   //<-load }

    xor  eax,eax
    fnstsw ax
    sahf
    and ax, 001Fh
    push eax
    fild [esp]
    pop eax
    fnclex
    fldcw word ptr [ECX]   //<-load

end;






initialization
begin
 //init random numbers generator
 _LSEType:=f_Jordan;
 InitRandomGenerator;
 F_FastSpec:=True;
  CWMEM:=$1372;   {.248}
end;


end.
