unit Foreval_SpecFunc;


 {
              Random numbers generator  algorithm( taus113 (Pierre L'Ecuyer),mt19937 (Mersenne Twister), xor4096)
              Gamma function,
              Bessel   functions,
              Legendre, Chebyshev polynoms,
              Zeta functions, and other spec. functions are  from
                  (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/
 }


                         { ver. 8.4.17.248 }



 {$IFDEF FPC}
  {$MODE DELPHI}
{$ENDIF}

//{$OverFlowChecks Off}
//{$RangeChecks Off}
//{$ZEROBASEDSTRINGS Off}


{$R-}
{$Q-}


interface
uses math;



//RAND  PLE

  const
   TC1 = longint($FFFFFFFE);  {4294967294}
   TC2 = longint($FFFFFFF8);  {4294967288}
   TC3 = longint($FFFFFFF0);  {4294967280}
   TC4 = longint($FFFFFF80);  {4294967168}
   ICNT = 10;                  {Init count to reach valid state}

    type
  taus113_ctx = record
               nr         : longint;  {next rand = s1 xor s2 xor s3 xor s4}
               s1,s2,s3,s4: longint;  {state variables }
             end;


//RAND  MT

const
  N19937 = 624;
  M19937 = 397;


  const
  mag01: array[0..1] of longint = (0, longint($9908b0df));

const
  Upper_Mask = longint($80000000);
  Lower_mask = longint($7fffffff);




type TDA02 = array[0..2] of double;
type THexDblW = packed array[0..3] of word;  {Double   as array of word}
type THexDblA = packed array[0..7] of byte;  {Double   as array of bytes}

type

  mt19937_state = array[0..N19937-1] of longint;

  mt19937_ctx   = record
                    mt : mt19937_state; {state vector}
                    mti: integer;       {state index }
                    nr : longint;       {next random }
                  end;



type
  TExtVector = array[0..6528] of extended;
  THexExtW = packed array[0..4] of word;
  type
  TExtRec  = packed record     {Extended as sign, exponent, significand}
               lm: longint;    {low  32 bit of significand}
               hm: longint;    {high 32 bit of significand}
               xp: word;       {biased exponent and sign  }
             end;



  //XOR 4096
 type
  xor4096_ctx   = record
                    x : array[0..127] of longint; {state vector}
                    i : integer;                  {state index }
                    w : longint;                  {Weyl state  }
                    nr: longint;                  {next random }
                  end;
  const
  Weyl = $61c88647; {Weyl generator increment}

  {Bernoulli numbers. The B(2n), n = 0..MaxB2nSmall are in AMath}
  const
  MaxBernoulli = 2312;
  const
  MaxB2nSmall  = 60;


  const ph_cutoff   : extended = 549755813888.0;     {2^39, threshold for Payne/Hanek}
  const Pi_4hex     : THexExtW = ($c235,$2168,$daa2,$c90f,$3ffe); {Pi/4}
  const eps_x: extended = 1.084202172485504434E-19; {Hex: 0000000000000080C03F}
  const Sqrt_MinXH  : THexExtW = ($0000,$0000,$0000,$8000,$2000); {sqrt(MinExtended) as Hex}
  const ln_succx0   = -11398.80538430830061;
  const c2dPi:extended = 2/pi;
  const MachineEpsilon: extended = 5E-16;
  const PosInfXHex  : THexExtW = ($0000,$0000,$0000,$8000,$7fff);
  const RTE_ArgumentRange: integer = 235;
  const MAXGAMX = 1755.455;
  const NaNXHex     : THexExtW = ($ffff,$ffff,$ffff,$ffff,$7fff); {an extended NaN as hex}
  const Sqrt_2Pihex : THexExtW = ($2cb3,$b138,$98ff,$a06c,$4000); {sqrt(2*Pi)}
  const XLGAE   = 8.0;
  const MaxExtHex   : THexExtW = ($ffff,$ffff,$ffff,$ffff,$7ffe); {MaxExtended as Hex}
  const LnSqrt2Pihex: THexExtW = ($a535,$25f5,$8e43,$eb3f,$3ffe); {ln(sqrt(2*Pi)=}
  const lgm05_hex : THexExtW = ($0000,$B257,$2363,$CA50,$4002);
  const Pi_2hex     : THexExtW = ($c235,$2168,$daa2,$c90f,$3fff); {Pi/2}
  const NegInfXHex  : THexExtW = ($0000,$0000,$0000,$8000,$ffff); {extended -INF   as hex}
  const ln_MaxXH    : THexExtW = ($79ab,$d1cf,$17f7,$b172,$400c); {predx(ln(MaxExtended))}
  const EulerGamHex : THexExtW = ($c7a5,$7db0,$67e3,$93c4,$3ffe); {Euler's constant}
  const RTE_NoConvergence: integer = 234;  {shb}
  const BT_J = 1; BT_Y = 2;
  const  twon24: double = 5.9604644775390625e-08; {1/2^24}
  const two24:  double = 16777216.0;             {2^24}
  const ln2_hi: THexDblA = ($00,$00,$E0,$FE,$42,$2E,$E6,$3F);
  const ln2_lo: THexDblA = ($76,$3C,$79,$35,$EF,$39,$EA,$3D);
  const log2ehex    : THexExtW = ($f0bc,$5c17,$3b29,$b8aa,$3fff); {log2(e)}
  const ln_MinXH    : THexExtW = ($eb2f,$1210,$8c67,$b16c,$c00c); {succx(ln(MinExtended))}
  const ln2hex      : THexExtW = ($79ac,$d1cf,$17f7,$b172,$3ffe); {ln(2)}
  const ln3hex      : THexDblW = ($030A,$7AAD,$93EA,$3FF1); {ln(3)}
  const PiSqrHex    : THexExtW = ($F2D2,$F22E,$E64D,$9DE9,$4002); {9.8696044010893586185}
  const PosInfDHex  : THexDblW = ($0000,$0000,$0000,$7ff0); {double +INF  as hex}
  const TwoPihex    : THexExtW = ($c235,$2168,$daa2,$c90f,$4001); {2*Pi}
  const ETAEPS = 1e-9;
  const x0e :double = 6.5;         {erf(x)  = 1 for |x| > x0e}
  const x1e:double = 106.75;      {erfc(x) = 0 for |x| > x1e}

  {$ifndef use_fast_exp}
  const
  half: single = 0.5;     {used for roundmode-safe exp routines}
  two : single = 2.0;     {used for roundmode-safe exp routines}
  ebig: single = 24576.0; {used for roundmode-safe exp routines}
 {$endif}

  const
  B2nHex: array[0..MaxB2nSmall] of THexExtw = ( {Bernoulli(2n), n=0..}
            ($0000,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000000}
            ($AAAB,$AAAA,$AAAA,$AAAA,$3FFC),  {+1.6666666666666666667E-1}
            ($8889,$8888,$8888,$8888,$BFFA),  {-3.3333333333333333335E-2}
            ($C30C,$0C30,$30C3,$C30C,$3FF9),  {+2.3809523809523809523E-2}
            ($8889,$8888,$8888,$8888,$BFFA),  {-3.3333333333333333335E-2}
            ($26CA,$6C9B,$C9B2,$9B26,$3FFB),  {+7.5757575757575757578E-2}
            ($8198,$9819,$1981,$8198,$BFFD),  {-2.5311355311355311355E-1}
            ($5555,$5555,$5555,$9555,$3FFF),  {+1.1666666666666666666}
            ($F2F3,$F2F2,$F2F2,$E2F2,$C001),  {-7.0921568627450980392}
            ($27C8,$9F1E,$7C78,$DBE2,$4004),  {+5.4971177944862155390E+1}
            ($67F4,$7F39,$F396,$8447,$C008),  {-5.2912424242424242426E+2}
            ($28D0,$33F1,$FC4A,$C180,$400B),  {+6.1921231884057971016E+3}
            ($6606,$0660,$2066,$A91A,$C00F),  {-8.6580253113553113550E+4}
            ($5555,$5555,$6955,$AE03,$4013),  {+1.4255171666666666666E+6}
            ($9C8B,$AE32,$DB88,$D044,$C017),  {-2.7298231067816091954E+7}
            ($FE36,$9A41,$9527,$8F6D,$401C),  {+6.0158087390064236836E+8}
            ($E5E6,$C5E5,$2B1D,$E140,$C020),  {-1.5116315767092156863E+10}
            ($5555,$EA55,$0E6E,$C80E,$4025),  {+4.2961464306116666666E+11}
            ($5309,$8E05,$E567,$C787,$C02A),  {-1.3711655205088332772E+13}
            ($9555,$CF4C,$5D33,$DE11,$402F),  {+4.8833231897359316666E+14}
            ($C84C,$1CEA,$45FA,$891C,$C035),  {-1.9296579341940068148E+16}
            ($9B70,$68B9,$B5E0,$BAE4,$403A),  {+8.4169304757368261500E+17}
            ($60ED,$6259,$67A8,$8BF3,$C040),  {-4.0338071854059455412E+19}
            ($D4E6,$F92A,$1ECC,$E551,$4045),  {+2.1150748638081991606E+21}
            ($8EB9,$1AF2,$630E,$CCC1,$C04B),  {-1.2086626522296525934E+23}
            ($C97F,$74B7,$D9A2,$C68B,$4051),  {+7.5008667460769643668E+24}
            ($50C0,$269C,$2369,$D066,$C057),  {-5.0387781014810689143E+26}
            ($D9ED,$4EC6,$CA9A,$EC0F,$405D),  {+3.6528776484818123334E+28}
            ($E7CF,$899B,$CBC7,$8FE1,$C064),  {-2.8498769302450882226E+30}
            ($5070,$15A2,$D8D6,$BC43,$406A),  {+2.3865427499683627645E+32}
            ($65EE,$E640,$2AAB,$83E3,$C071),  {-2.1399949257225333666E+34}
            ($707D,$1C11,$CE9D,$C56A,$4077),  {+2.0500975723478097570E+36}
            ($1E1C,$7AC8,$21EF,$9D85,$C07E),  {-2.0938005911346378408E+38}
            ($E927,$8376,$73E5,$85BA,$4085),  {+2.2752696488463515559E+40}
            ($8500,$E7B3,$9488,$F123,$C08B),  {-2.6257710286239576047E+42}
            ($B991,$385E,$7503,$E67C,$4092),  {+3.2125082102718032518E+44}
            ($0E52,$1EBF,$9A0F,$E92A,$C099),  {-4.1598278166794710914E+46}
            ($7C1E,$0021,$4E79,$F942,$40A0),  {+5.6920695482035280023E+48}
            ($970D,$F26E,$AF7E,$8C94,$C0A8),  {-8.2183629419784575694E+50}
            ($2C49,$524C,$2BE5,$A716,$40AF),  {+1.2502904327166993017E+53}
            ($C0A8,$EA20,$F1CB,$D0F8,$C0B6),  {-2.0015583233248370275E+55}
            ($8E7B,$4C4E,$5298,$8956,$40BE),  {+3.3674982915364374232E+57}
            ($7DC5,$B6DB,$4610,$BD7C,$C0C5),  {-5.9470970503135447718E+59}
            ($12CC,$59DB,$F874,$890D,$40CD),  {+1.1011910323627977560E+62}
            ($4590,$25D4,$A47F,$CFA5,$C0D4),  {-2.1355259545253501188E+64}
            ($A679,$4EF1,$AF84,$A492,$40DC),  {+4.3328896986641192418E+66}
            ($5E81,$3FBE,$35D8,$8854,$C0E4),  {-9.1885528241669328228E+68}
            ($8A90,$C146,$AAF0,$EBD8,$40EB),  {+2.0346896776329074493E+71}
            ($C12C,$4593,$68CD,$D4D3,$C0F3),  {-4.7003833958035731077E+73}
            ($FF4E,$F062,$48DC,$C82E,$40FB),  {+1.1318043445484249271E+76}
            ($3F3B,$AFB3,$532E,$C417,$C103),  {-2.8382249570693706958E+78}
            ($7E40,$DF17,$81CD,$C7E2,$410B),  {+7.4064248979678850632E+80}
            ($45B1,$D38C,$5DAE,$D3DC,$C113),  {-2.0096454802756604484E+83}
            ($5F41,$4EA9,$1C33,$E951,$411B),  {+5.6657170050805941445E+85}
            ($6A4B,$0BDB,$E3F8,$8563,$C124),  {-1.6584511154136216916E+88}
            ($4724,$DD5F,$F84A,$9E3F,$412C),  {+5.0368859950492377418E+90}
            ($328E,$4648,$E010,$C2A9,$C134),  {-1.5861468237658186369E+93}
            ($826C,$BA97,$AC47,$F81F,$413C),  {+5.1756743617545626984E+95}
            ($93F0,$781D,$43FD,$A3C1,$C145),  {-1.7488921840217117340E+98}
            ($0AFB,$0494,$C087,$DFB2,$414D),  {+6.1160519994952185254E+100}
            ($8384,$4095,$B028,$9E09,$C156)); {-2.2122776912707834942E+103}

  const
  ipio2: array[0..689] of longint = (
           $A2F983, $6E4E44, $1529FC, $2757D1, $F534DD, $C0DB62,
           $95993C, $439041, $FE5163, $ABDEBB, $C561B7, $246E3A,
           $424DD2, $E00649, $2EEA09, $D1921C, $FE1DEB, $1CB129,
           $A73EE8, $8235F5, $2EBB44, $84E99C, $7026B4, $5F7E41,
           $3991D6, $398353, $39F49C, $845F8B, $BDF928, $3B1FF8,
           $97FFDE, $05980F, $EF2F11, $8B5A0A, $6D1F6D, $367ECF,
           $27CB09, $B74F46, $3F669E, $5FEA2D, $7527BA, $C7EBE5,
           $F17B3D, $0739F7, $8A5292, $EA6BFB, $5FB11F, $8D5D08,
           $560330, $46FC7B, $6BABF0, $CFBC20, $9AF436, $1DA9E3,
           $91615E, $E61B08, $659985, $5F14A0, $68408D, $FFD880,
           $4D7327, $310606, $1556CA, $73A8C9, $60E27B, $C08C6B,
           {the following bits are needed for extended exponents only}
           $47C419, $C367CD, $DCE809, $2A8359, $C4768B, $961CA6,
           $DDAF44, $D15719, $053EA5, $FF0705, $3F7E33, $E832C2,
           $DE4F98, $327DBB, $C33D26, $EF6B1E, $5EF89F, $3A1F35,
           $CAF27F, $1D87F1, $21907C, $7C246A, $FA6ED5, $772D30,
           $433B15, $C614B5, $9D19C3, $C2C4AD, $414D2C, $5D000C,
           $467D86, $2D71E3, $9AC69B, $006233, $7CD2B4, $97A7B4,
           $D55537, $F63ED7, $1810A3, $FC764D, $2A9D64, $ABD770,
           $F87C63, $57B07A, $E71517, $5649C0, $D9D63B, $3884A7,
           $CB2324, $778AD6, $23545A, $B91F00, $1B0AF1, $DFCE19,
           $FF319F, $6A1E66, $615799, $47FBAC, $D87F7E, $B76522,
           $89E832, $60BFE6, $CDC4EF, $09366C, $D43F5D, $D7DE16,
           $DE3B58, $929BDE, $2822D2, $E88628, $4D58E2, $32CAC6,
           $16E308, $CB7DE0, $50C017, $A71DF3, $5BE018, $34132E,
           $621283, $014883, $5B8EF5, $7FB0AD, $F2E91E, $434A48,
           $D36710, $D8DDAA, $425FAE, $CE616A, $A4280A, $B499D3,
           $F2A606, $7F775C, $83C2A3, $883C61, $78738A, $5A8CAF,
           $BDD76F, $63A62D, $CBBFF4, $EF818D, $67C126, $45CA55,
           $36D9CA, $D2A828, $8D61C2, $77C912, $142604, $9B4612,
           $C459C4, $44C5C8, $91B24D, $F31700, $AD43D4, $E54929,
           $10D5FD, $FCBE00, $CC941E, $EECE70, $F53E13, $80F1EC,
           $C3E7B3, $28F8C7, $940593, $3E71C1, $B3092E, $F3450B,
           $9C1288, $7B20AB, $9FB52E, $C29247, $2F327B, $6D550C,
           $90A772, $1FE76B, $96CB31, $4A1679, $E27941, $89DFF4,
           $9794E8, $84E6E2, $973199, $6BED88, $365F5F, $0EFDBB,
           $B49A48, $6CA467, $427271, $325D8D, $B8159F, $09E5BC,
           $25318D, $3974F7, $1C0530, $010C0D, $68084B, $58EE2C,
           $90AA47, $02E774, $24D6BD, $A67DF7, $72486E, $EF169F,
           $A6948E, $F691B4, $5153D1, $F20ACF, $339820, $7E4BF5,
           $6863B2, $5F3EDD, $035D40, $7F8985, $295255, $C06437,
           $10D86D, $324832, $754C5B, $D4714E, $6E5445, $C1090B,
           $69F52A, $D56614, $9D0727, $50045D, $DB3BB4, $C576EA,
           $17F987, $7D6B49, $BA271D, $296996, $ACCCC6, $5414AD,
           $6AE290, $89D988, $50722C, $BEA404, $940777, $7030F3,
           $27FC00, $A871EA, $49C266, $3DE064, $83DD97, $973FA3,
           $FD9443, $8C860D, $DE4131, $9D3992, $8C70DD, $E7B717,
           $3BDF08, $2B3715, $A0805C, $93805A, $921110, $D8E80F,
           $AF806C, $4BFFDB, $0F9038, $761859, $15A562, $BBCB61,
           $B989C7, $BD4010, $04F2D2, $277549, $F6B6EB, $BB22DB,
           $AA140A, $2F2689, $768364, $333B09, $1A940E, $AA3A51,
           $C2A31D, $AEEDAF, $12265C, $4DC26D, $9C7A2D, $9756C0,
           $833F03, $F6F009, $8C402B, $99316D, $07B439, $15200C,
           $5BC3D8, $C492F5, $4BADC6, $A5CA4E, $CD37A7, $36A9E6,
           $9492AB, $6842DD, $DE6319, $EF8C76, $528B68, $37DBFC,
           $ABA1AE, $3115DF, $A1AE00, $DAFB0C, $664D64, $B705ED,
           $306529, $BF5657, $3AFF47, $B9F96A, $F3BE75, $DF9328,
           $3080AB, $F68C66, $15CB04, $0622FA, $1DE4D9, $A4B33D,
           $8F1B57, $09CD36, $E9424E, $A4BE13, $B52333, $1AAAF0,
           $A8654F, $A5C1D2, $0F3F0B, $CD785B, $76F923, $048B7B,
           $721789, $53A6C6, $E26E6F, $00EBEF, $584A9B, $B7DAC4,
           $BA66AA, $CFCF76, $1D02D1, $2DF1B1, $C1998C, $77ADC3,
           $DA4886, $A05DF7, $F480C6, $2FF0AC, $9AECDD, $BC5C3F,
           $6DDED0, $1FC790, $B6DB2A, $3A25A3, $9AAF00, $9353AD,
           $0457B6, $B42D29, $7E804B, $A707DA, $0EAA76, $A1597B,
           $2A1216, $2DB7DC, $FDE5FA, $FEDB89, $FDBE89, $6C76E4,
           $FCA906, $70803E, $156E85, $FF87FD, $073E28, $336761,
           $86182A, $EABD4D, $AFE7B3, $6E6D8F, $396795, $5BBF31,
           $48D784, $16DF30, $432DC7, $356125, $CE70C9, $B8CB30,
           $FD6CBF, $A200A4, $E46C05, $A0DD5A, $476F21, $D21262,
           $845CB9, $496170, $E0566B, $015299, $375550, $B7D51E,
           $C4F133, $5F6E13, $E4305D, $A92E85, $C3B21D, $3632A1,
           $A4B708, $D4B1EA, $21F716, $E4698F, $77FF27, $80030C,
           $2D408D, $A0CD4F, $99A520, $D3A2B3, $0A5D2F, $42F9B4,
           $CBDA11, $D0BE7D, $C1DB9B, $BD17AB, $81A2CA, $5C6A08,
           $17552E, $550027, $F0147F, $8607E1, $640B14, $8D4196,
           $DEBE87, $2AFDDA, $B6256B, $34897B, $FEF305, $9EBFB9,
           $4F6A68, $A82A4A, $5AC44F, $BCF82D, $985AD7, $95C7F4,
           $8D4D0D, $A63A20, $5F57A4, $B13F14, $953880, $0120CC,
           $86DD71, $B6DEC9, $F560BF, $11654D, $6B0701, $ACB08C,
           $D0C0B2, $485551, $0EFB1E, $C37295, $3B06A3, $3540C0,
           $7BDC06, $CC45E0, $FA294E, $C8CAD6, $41F3E8, $DE647C,
           $D8649B, $31BED9, $C397A4, $D45877, $C5E369, $13DAF0,
           $3C3ABA, $461846, $5F7555, $F5BDD2, $C6926E, $5D2EAC,
           $ED440E, $423E1C, $87C461, $E9FD29, $F3D6E7, $CA7C22,
           $35916F, $C5E008, $8DD7FF, $E26A6E, $C6FDB0, $C10893,
           $745D7C, $B2AD6B, $9D6ECD, $7B723E, $6A11C6, $A9CFF7,
           $DF7329, $BAC9B5, $5100B7, $0DB2E2, $24BA74, $607DE5,
           $8AD874, $2C150D, $0C1881, $94667E, $162901, $767A9F,
           $BEFDFD, $EF4556, $367ED9, $13D9EC, $B9BA8B, $FC97C4,
           $27A831, $C36EF1, $36C594, $56A8D8, $B5A8B4, $0ECCCF,
           $2D8912, $34576F, $89562C, $E3CE99, $B920D6, $AA5E6B,
           $9C2A3E, $CC5F11, $4A0BFD, $FBF4E1, $6D3B8E, $2C86E2,
           $84D4E9, $A9B4FC, $D1EEEF, $C9352E, $61392F, $442138,
           $C8D91B, $0AFC81, $6A4AFB, $D81C2F, $84B453, $8C994E,
           $CC2254, $DC552A, $D6C6C0, $96190B, $B8701A, $649569,
           $605A26, $EE523F, $0F117F, $11B5F4, $F5CBFC, $2DBC34,
           $EEBC34, $CC5DE8, $605EDD, $9B8E67, $EF3392, $B817C9,
           $9B5861, $BC57E1, $C68351, $103ED8, $4871DD, $DD1C2D,
           $A118AF, $462C21, $D7F359, $987AD9, $C0549E, $FA864F,
           $FC0656, $AE79E5, $362289, $22AD38, $DC9367, $AAE855,
           $382682, $9BE7CA, $A40D51, $B13399, $0ED7A9, $480569,
           $F0B265, $A7887F, $974C88, $36D1F9, $B39221, $4A827B,
           $21CF98, $DC9F40, $5547DC, $3A74E1, $42EB67, $DF9DFE,
           $5FD45E, $A4677B, $7AACBA, $A2F655, $23882B, $55BA41,
           $086E59, $862A21, $834739, $E6E389, $D49EE5, $40FB49,
           $E956FF, $CA0F1C, $8A59C5, $2BFA94, $C5C1D3, $CFC50F,
           $AE5ADB, $86C547, $624385, $3B8621, $94792C, $876110,
           $7B4C2A, $1A2C80, $12BF43, $902688, $893C78, $E4C4A8,
           $7BDBE5, $C23AC4, $EAF426, $8A67F7, $BF920D, $2BA365,
           $B1933D, $0B7CBD, $DC51A4, $63DD27, $DDE169, $19949A,
           $9529A8, $28CE68, $B4ED09, $209F44, $CA984E, $638270,
           $237C7E, $32B90F, $8EF5A7, $E75614, $08F121, $2A9DB5,
           $4D7E6F, $5119A5, $ABF9B5, $D6DF82, $61DD96, $023616,
           $9F3AC4, $A1A283, $6DED72, $7A8D39, $A9B882, $5C326B,
           $5B2746, $ED3400, $7700D2, $55F4FC, $4D5901, $8071E0);

const
  PIo2: array[0..7] of THexDblW = (
          ($0000, $4000, $21FB, $3FF9),  {1.5707962512969971    }
          ($0000, $0000, $442D, $3E74),  {7.5497894158615964e-08}
          ($0000, $8000, $4698, $3CF8),  {5.3903025299577648e-15}
          ($0000, $6000, $CC51, $3B78),  {3.2820034158079130e-22}
          ($0000, $8000, $1B83, $39F0),  {1.2706557530806761e-29}
          ($0000, $4000, $2520, $387A),  {1.2293330898111133e-36}
          ($0000, $8000, $8222, $36E3),  {2.7337005381646456e-44}
          ($0000, $0000, $F31D, $3569)); {2.1674168387780482e-51}
  const
  init_jk: array[0..3] of integer = (2,3,4,6); {initial value for jk}
  const
  NBoF  = 20;
  BoFHex: array[0..NBoF] of THexExtW = (      {Bernoulli(2k+2)/(2k+2)! }
            ($AAAB,$AAAA,$AAAA,$AAAA,$3FFB),  {+8.3333333333333333336E-2}
            ($B60B,$0B60,$60B6,$B60B,$BFF5),  {-1.3888888888888888888E-3}
            ($355E,$08AB,$55E0,$8AB3,$3FF0),  {+3.3068783068783068783E-5}
            ($5563,$A778,$BC99,$DDEB,$BFEA),  {-8.2671957671957671957E-7}
            ($ED15,$B875,$795F,$B354,$3FE5),  {+2.0876756987868098980E-8}
            ($F7BA,$2133,$2EB2,$9140,$BFE0),  {-5.2841901386874931847E-10}
            ($0FDC,$048B,$9627,$EB6D,$3FDA),  {+1.3382536530684678833E-11}
            ($E260,$CCAA,$70FB,$BED2,$BFD5),  {-3.3896802963225828668E-13}
            ($ECEE,$2974,$38EB,$9AAC,$3FD0),  {+8.5860620562778445639E-15}
            ($8664,$5567,$CB46,$FABE,$BFCA),  {-2.1748686985580618731E-16}
            ($77D8,$E1BB,$0F84,$CB3F,$3FC5),  {+5.5090028283602295151E-18}
            ($E371,$D15A,$C819,$A4BE,$BFC0),  {-1.3954464685812523341E-19}
            ($3204,$3568,$96BE,$8589,$3FBB),  {+3.5347070396294674718E-21}
            ($B3E6,$F8E0,$96AC,$D87B,$BFB5),  {-8.9535174270375468504E-23}
            ($6AD7,$6513,$6D26,$AF79,$3FB0),  {+2.2679524523376830603E-24}
            ($57E2,$6F1C,$EFCB,$8E3B,$BFAB),  {-5.7447906688722024451E-26}
            ($0D7F,$90DB,$CEF2,$E694,$3FA5),  {+1.4551724756148649018E-27}
            ($3676,$2BA5,$F32B,$BAE6,$BFA0),  {-3.6859949406653101781E-29}
            ($F634,$DA56,$46D9,$977F,$3F9B),  {+9.3367342570950446721E-31}
            ($F768,$3B0F,$1482,$F599,$BF95),  {-2.3650224157006299346E-32}
            ($2185,$9423,$FFC9,$C712,$3F90)); {+5.9906717624821343044E-34}




  var
  Pi_4        : extended absolute Pi_4hex;
  Sqrt_MinExt : extended absolute Sqrt_MinXH;  {= 1.833603867554847166E-2466}  {= 0.5^8191}
  PosInf_x    : extended absolute PosInfXHex;  {extended +INF  }
  MaxExtended : extended absolute MaxExtHex;   {= 1.189731495357231764E+4932}  {2^16384-2^16320 = (2^64-1)*2^16320}
  NaN_x       : extended absolute NaNXHex;
  Sqrt_TwoPi  : extended absolute Sqrt_2Pihex;
  LnSqrt2Pi   : extended absolute LnSqrt2Pihex;
  ctxPLE      :  taus113_ctx;
  ctxMT,ctxMTG:  mt19937_ctx;
  ctxXOR      : xor4096_ctx;
  lanczos_gm05: extended absolute lgm05_hex;
  Pi_2        : extended absolute Pi_2hex;
  NegInf_x    : extended absolute NegInfXHex;
  ln_MaxExt   : extended absolute ln_MaxXH;    {= 11356.52340629414394}
  EulerGamma  : extended absolute EulerGamHex; {= 0.57721566490153286061}
  ln_MinExt   : extended absolute ln_MinXH;    {=-11355.13711193302405}
  log2e       : extended absolute log2ehex;    {= 1.4426950408889634079 }
  ln2         : extended absolute ln2hex;      {= 0.69314718055994530942}
  PiSqr       : extended absolute PiSqrHex;    {= 9.8696044010893586185}
  PosInf_d    : double absolute PosInfDHex;    {double +INF }
  TwoPi       : extended absolute TwoPihex;    {= 6.2831853071795864769 }




  function bessel_j0: extended;
  function bessel_j1: extended;
  function bessel_jn: extended;
  function bessel_y0: extended;
  function bessel_y1: extended;
  function bessel_yn: extended;
  function gamma: extended;
  function sph_bessel_jn: extended;
  function psi: extended;
  function psid1: extended;
  function psid2: extended;
  function psid3: extended;
  function zeta: extended;



  function sfc_j0(x: extended): extended;
  function sfc_j1(x: extended): extended;
  function sfc_jn(n: integer; x: extended): extended;
  function sfc_y0(x: extended): extended;
  function sfc_y1(x: extended): extended;
  function sfc_yn(n: integer; x: extended): extended;

  function sfc_sph_jn(n: integer; x: extended): extended;

  function sfc_gamma(x: extended):extended; stdcall;
  function sfc_psi(x: extended): extended;  stdcall;
  function sfc_trigamma(x: extended): extended;   stdcall;
  function sfc_tetragamma(x: extended): extended;   stdcall;
  function sfc_pentagamma(x: extended): extended;   stdcall;
  function sfc_zeta(s: extended): extended; stdcall;

  function PLE_rand2(x1,x2:PDouble): double;  stdcall;
  function MT_rand2(x1,x2:PDouble): double;  stdcall;
  function XOR_rand2(x1,x2:PDouble): double;  stdcall;
  procedure PLE_rand2_K;
  procedure XOR_rand2_K;  stdcall;
  procedure MT_intrand(Num: Cardinal);  stdcall;
  procedure MT_rand01;  stdcall;
  procedure PLE_intrand(Num: Cardinal);  stdcall;
  procedure PLE_rand01;  stdcall;
  procedure XOR_intrand(Num: Cardinal);  stdcall;
  procedure XOR_rand01;  stdcall;
  //procedure _Randomize;  stdcall;


    procedure taus113_init4(var ctx: taus113_ctx; seed1,seed2,seed3,seed4: longint);
    function  taus113_double(var ctx: taus113_ctx): double;
    function taus113_double53(var ctx: taus113_ctx): double;
    procedure taus113_init(var ctx: taus113_ctx; seed: longint);
    procedure mt19937_init(var ctxMT: mt19937_ctx; seed: longint);
    procedure mt19937_next(var ctx: mt19937_ctx);
    function mt19937_double(var ctxMT: mt19937_ctx): double;
    function mt19937_double53(var ctxMT: mt19937_ctx): double;
    procedure xor4096_init(var ctx: xor4096_ctx; seed: longint);
    procedure xor4096_next(var ctx: xor4096_ctx);
    function xor4096_double53(var ctx: xor4096_ctx): double;
    function _exp_x(x: extended): extended;
    function _power_x(x,y: extended): extended;
    function rem_pio2_x(x: extended; var z: extended): integer;
    procedure sincosPi(x: extended; var s,c: extended);
    procedure _sincos(x: extended; var s,c: extended);
    function exp2m1(x: extended): extended;
    function coshm1(x: extended): extended;
    function sinh_small(x: extended): extended;
    function sfc_zetaint(n: integer): extended;
    function powm1(x,y: extended): extended;
    function expm1(x: extended): extended;
    function exp2(x: extended): extended;
    function sfc_bernoulli(n: integer): extended;
    function sfc_lngammas(x: extended; var s: integer): extended;
    function zetap(s,sc: extended): extended;
    function maxx(x, y: extended): extended;  {$ifdef HAS_INLINE} inline;{$endif}
    function CSEvalX(x: extended; const a: array of extended; n: integer): extended;
    function sfc_zetam1(s: extended): extended;
    function etam1pos(s: extended): extended;
    function exp3(x: extended): extended;
    function exp10(x: extended): extended;
    function exp7(x: extended): extended;
    function bernpoly_intern(n: integer; x: extended): extended;
    function hz_a1(s,a: extended; var ok: boolean): extended;
    function hurwitz_formula(s,a: extended): extended;
    function sfc_chebyshev_t(n: integer; x: extended): extended;  stdcall;
    function sfc_chebyshev_u(n: integer; x: extended): extended;  stdcall;
    function sfc_laguerre(n: integer; x: extended): extended; stdcall;
    function sfc_laguerre_a(n: integer; a,x: extended): extended; stdcall;
    function sfc_hermite_h(n: integer; x: extended): extended; stdcall;
    function erf_small(x: extended): extended;
    function sfc_erf(x: extended): extended;
    function sfc_erfc(x: extended): extended;
    procedure sfc_erfc_iusc(x: extended; var p,q: extended);
    function IsInf(x: extended): boolean;
    function expx2(x: extended): extended;
    function erf: extended;
    function erfc: extended;


   function sfc_lngamma(x: extended): extended;
   function sfc_lngcorr(x: extended): extended;
   function sfc_legendre_p(l: integer; x: extended): extended;      stdcall;
   function sfc_legendre_plm(l,m: integer; x: extended): extended;  stdcall;
   function lngamma_small(x,xm1,xm2: extended; useln1p: boolean): extended;
   procedure taus113_next(var ctx: taus113_ctx);



implementation
uses Foreval_Command, Foreval_Main;




{
              Random numbers generator  algorithm
                  (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/

                    algorithm taus113 (Pierre L'Ecuyer)
                    algorithm mt19937 (Mersenne Twister)
                    algorithm xor4096 ()
}




procedure taus113_init(var ctx: taus113_ctx; seed: longint);
  {-Init context from seed}
const
  M=69069;
  A=1;
var
  seed2,seed3,seed4: longint;
begin
  seed2 := M*seed+A;
  seed3 := M*seed2+A;
  seed4 := M*seed3+A;
  taus113_init4(ctx,seed,seed2,seed3,seed4);
end;


procedure taus113_next(var ctx: taus113_ctx);
  {-Next step of PRNG}
begin
  with ctx do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;
end;





function taus113_double(var ctx: taus113_ctx): double;
  {-Next random double [0..1) with 32 bit precision}
begin
  taus113_next(ctx);
  taus113_double := (ctx.nr + 2147483648.0) / 4294967296.0;
end;


function taus113_double53(var ctx: taus113_ctx): double;
  {-Next random double in [0..1) with 53 bit precision}
var
  hb,lb: longint;
begin
  taus113_next(ctx);
  hb := ctx.nr shr 5;
  taus113_next(ctx);
  lb := ctx.nr shr 6;
  taus113_double53 := (hb*67108864.0+lb)/9007199254740992.0;
end;


function PLE_rand2(x1,x2:PDouble): double;  stdcall;
var
  hb,lb: longint;
begin
{
 taus113_next(ctx);
 PLE_rand2:=(x1^-x2^)*taus113_double(ctx)+x2^;
}

//32bit  float:
{ with ctx do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

 PLE_rand2:=(x1^-x2^)*((ctx.nr + 2147483648.0) / 4294967296.0)+x2^;
 }

//53bit floar  :
with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

hb := ctxPLE.nr shr 5;

with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

lb := ctxPLE.nr shr 6;

PLE_rand2:=(x1^-x2^)*((hb*67108864.0+lb)/9007199254740992.0)+x2^;

end;



procedure PLE_rand2_K;
const
c1: double=67108864.0;
c2: double= 9007199254740992.0;
var
  hb,lb: longint;
  r:double;
begin


//53bit floar  :
with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

hb := ctxPLE.nr shr 5;

with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

lb := ctxPLE.nr shr 6;

asm
  fld st(0)
  fsubp st(2),st(0)
  fild  dword ptr [hb]
  fmul  qword ptr [c1]
  fadd  qword ptr [lb]        {.209}
  fdiv  qword ptr [c2]
  fmulp st(2),st(0)
  faddp
end;
//r:=(x1^-x2^)*((hb*c1+lb)/c2)+x2^;

end;



procedure PLE_intrand(Num: Cardinal){extended};  stdcall;
{const
c3: extended}
const
maxint: cardinal = 2147483647;{4294967295}
divmaxint: extended = 1/2147483647;
divmaxlong: extended = 1/4294967295;
var
rn: Cardinal;
res,res1: double;
begin

{in range max integer: 0..2^31-1=2147483647}
// Num*(ctxMT.nr sh1 1) / 4294967296.0;

  //taus113_next(ctxPLE);
   with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;
  rn:=ctxPLE.nr shr 1;

  asm
    fild  rn
    fimul Num
    fld tbyte ptr divmaxint
    fmulp
    FRNDINT
  end;


{in range max cardinal: 0..2^32-1=4294967295}
// Num*(ctxPLE.nr + 2147483648.0) / 4294967296.0;

{ PLE19937_next(ctxPLE);


  asm
    fild  ctxPLE.nr
    fiadd maxint
    fimul Num
    fld tbyte ptr divmaxlong
    fmulp
    FRNDINT
  end;
 }
end;


procedure PLE_rand01;  stdcall;
const
c1: extended = 1/4294967296.0;
c2: double =  2147483648.0;
c3: double =  4294967296.0;
var
rn: Cardinal;
res: extended;
begin
  //taus113_next(ctxPLE);
  with ctxPLE do begin
    s1 := ((s1 and TC1) shl 18) xor (((s1 shl  6) xor s1) shr 13);
    s2 := ((s2 and TC2) shl  2) xor (((s2 shl  2) xor s2) shr 27);
    s3 := ((s3 and TC3) shl  7) xor (((s3 shl 13) xor s3) shr 21);
    s4 := ((s4 and TC4) shl 13) xor (((s4 shl  3) xor s4) shr 12);
    nr := s1 xor s2 xor s3 xor s4;
  end;

  asm
    fild   ctxPLE.nr
    fadd   c2
    fld c1
    fmulp
    //fdiv   c3
  end;

end;




procedure taus113_init4(var ctx: taus113_ctx; seed1,seed2,seed3,seed4: longint);
  {-Init all context with separate seeds, the initial seeds}
  { seed1, seed2, seed3, seed4 should be >= 2, 8, 16, 128 resp. or < 0  }
var
  i: integer;
begin
  with ctx do begin
    s1 := seed1;
    s2 := seed2;
    s3 := seed3;
    s4 := seed4;
    {Make sure most significant bits are set}
    if s1 and TC1=0 then inc(s1,2);
    if s2 and TC2=0 then inc(s2,8);
    if s3 and TC3=0 then inc(s3,16);
    if s4 and TC4=0 then inc(s4,128);
    {Make sure ctx has valid state}
    for i:=1 to ICNT do taus113_next(ctx);
  end;
end;




procedure mt19937_init(var ctxMT: mt19937_ctx; seed: longint);
  {-Init context from seed}

begin
  with ctxMT do begin
    mt[0]:= seed;
    mti  := 1;
    while mti<N19937 do begin
      {See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier.}
      {In the previous versions, MSBs of the seed affect  }
      {only MSBs of the array mt[].                       }
      {2002/01/09 modified by Makoto Matsumoto}
      mt[mti] := (1812433253*(mt[mti-1] xor (mt[mti-1] shr 30))+mti);
      inc(mti);
    end;
  end;
end;


procedure mt19937_next(var ctx: mt19937_ctx);
  {-Next step of PRNG}
var
  kk: integer;
  y : longint;
begin
  with ctx do begin
    if mti >= N19937 then begin
      {generate N19937 numbers at one time}
      kk := 0;
      while kk<N19937-M19937 do begin
        y := (mt[kk] and Upper_Mask) or (mt[kk+1] and Lower_mask);
        mt[kk] := mt[kk+M19937] xor (y shr 1) xor mag01[y and 1];
        inc(kk);
      end;
      while kk<N19937-1 do begin
        y := (mt[kk] and Upper_Mask) or (mt[kk+1] and Lower_mask);
        mt[kk] := mt[kk+(M19937-N19937)] xor (y shr 1) xor mag01[y and 1];
        inc(kk);
      end;
      y := (mt[N19937-1] and Upper_Mask) or (mt[0] and Lower_mask);
      mt[N19937-1] := mt[M19937-1] xor (y shr 1) xor mag01[y and 1];
      mti := 0;
    end;
    y := mt[mti];
    inc(mti);

      y := y xor (y shr 11);
      y := y xor ((y shl  7) and longint($9d2c5680));
      y := y xor ((y shl 15) and longint($efc60000));
      nr:= y xor (y shr 18);

  end;
end;



{---------------------------------------------------------------------------}


function mt19937_long(var ctx: mt19937_ctx): longint;
  {-Next random positive longint}
begin
  mt19937_next(ctx);
  mt19937_long := ctx.nr shr 1;
end;


procedure MT_intrand(Num: Cardinal){extended};  stdcall;
{const
c3: extended}
const
maxint: cardinal = 2147483647;{4294967295}
divmaxint: extended = 1/2147483647;
divmaxlong: extended = 1/4294967295;
var
rn: Cardinal;
res,res1: double;
begin

{in range max integer: 0..2^31-1=2147483647}
// Num*(ctxMT.nr sh1 1) / 4294967296.0;

  mt19937_next(ctxMT);
  rn:=ctxMT.nr shr 1;

  asm
    fild  rn
    fimul Num
    fld tbyte ptr divmaxint
    fmulp
    FRNDINT
  end;


{in range max cardinal: 0..2^32-1=4294967295}
// Num*(ctxMT.nr + 2147483648.0) / 4294967296.0;

{ mt19937_next(ctxMT);


  asm
    fild  ctxMT.nr
    fiadd maxint
    fimul Num
    fld tbyte ptr divmaxlong
    fmulp
    FRNDINT
  end;
 }
end;


procedure MT_rand01;  stdcall;
const
c1: extended = 1/4294967296.0;
c2: double =  2147483648.0;
c3: double =  4294967296.0;
var
rn: Cardinal;
res: extended;
begin
  mt19937_next(ctxMT);

  asm
    fild   ctxMT.nr
    fadd   c2
    fld c1
    fmulp
    //fdiv   c3
  end;

end;

function mt19937_double(var ctxMT: mt19937_ctx): double;
  {-Next random double [0..1) with 32 bit precision}
begin
  mt19937_next(ctxMT);
  mt19937_double := (ctxMT.nr + 2147483648.0) / 4294967296.0;
end;


function mt19937_double53(var ctxMT: mt19937_ctx): double;
  {-Next random double in [0..1) with full double 53 bit precision}
var
  hb,lb: longint;
begin
  mt19937_next(ctxMT);
  hb := ctxMT.nr shr 5;
  mt19937_next(ctxMT);
  lb := ctxMT.nr shr 6;
  mt19937_double53 := (hb*67108864.0+lb)/9007199254740992.0;
end;



function MT_rand2(x1,x2:PDouble): double;  stdcall;
var
  hb,lb: longint;
begin
  mt19937_next(ctxMT);
  hb := ctxMT.nr shr 5;
  mt19937_next(ctxMT);
  lb := ctxMT.nr shr 6;

  MT_rand2:=(x1^-x2^)*((hb*67108864.0+lb)/9007199254740992.0)+x2^;
end;



{*******************************xor4096**************************************}

procedure xor4096_init(var ctx: xor4096_ctx; seed: longint);
  {-Init context from seed}
var
  k: integer;
  v: longint;
begin
  v := seed;
  {starting value must be non-zero}
  if v=0 then v := not v;
  with ctx do begin
    {Avoid correlations for close seeds, recurrence has period 2^32-1}
    for k:=0 to 31 do begin
      v := v xor (v shl 10);  v := v xor (v shr 15);
      v := v xor (v shl  4);  v := v xor (v shr 13);
    end;
    {Initialize circular array}
    w := v;
    for k:=0 to 127 do begin
      v := v xor (v shl 10);  v := v xor (v shr 15);
      v := v xor (v shl  4);  v := v xor (v shr 13);
      w := w + Weyl;
      x[k] := v + w;
    end;
    {Discard first 512 results}
    i := 127;
    {Because xor4096_next changes w, the current Weyl state is saved and}
    {restored. The original c code doubles the source except that w is  }
    {unchanged. As the init code is normally called only once, the small}
    {speed penalty is no problem here and optimization is done in next. }
    v := w;
    for k:=0 to 511 do xor4096_next(ctx);
    {Restore Weyl state}
    w := v;
  end;
end;


procedure xor4096_next(var ctx: xor4096_ctx);
  {-Next step of PRNG}
var
  v,t: longint;
const
  a=17; b=12; c=13; d=15;
begin
  with ctx do begin
    {(I + L^a)(I + R^b) */}
    i := succ(i) and 127;
    t := x[i];
    t := t xor (t shl a);
    t := t xor (t shr b);
    {(I + L^c)(I + R^d) and xor with t}
    v := x[(i+33) and 127];
    v := v xor (v shl c);
    v := (v xor (v shr d)) xor t;
    {Update circular array}
    x[i] := v;
    {Update Weyl generator}
    w := w + Weyl;
    {Store next random result}
      nr := v + (w xor (w shr 16));
  end;
end;



function xor4096_double53(var ctx: xor4096_ctx): double;
  {-Next random double in [0..1) with full double 53 bit precision}
var
  hb,lb: longint;
begin
  xor4096_next(ctx);
  hb := ctx.nr shr 5;
  xor4096_next(ctx);
  lb := ctx.nr shr 6;
  xor4096_double53 := (hb*67108864.0+lb)/9007199254740992.0;
end;



function XOR_rand2(x1,x2:PDouble): double;  stdcall;
var
  hb,lb: longint;
begin
  xor4096_next(ctxXOR);
  hb := ctxXOR.nr shr 5;
  xor4096_next(ctxXOR);
  lb := ctxXOR.nr shr 6;


  XOR_rand2:=(x1^-x2^)*((hb*67108864.0+lb)/9007199254740992.0)+x2^;
end;


procedure XOR_rand2_K;  stdcall;
const
c1: double=67108864.0;
c2: double= 9007199254740992.0;
c3:extended = 1/9007199254740992.0;
var
  hb,lb: longint;
begin
  xor4096_next(ctxXOR);
  hb := ctxXOR.nr shr 5;
  xor4096_next(ctxXOR);
  lb := ctxXOR.nr shr 6;


  //XOR_rand2:=(x1^-x2^)*((hb*67108864.0+lb)/9007199254740992.0)+x2^;

asm
  fld st(0)
  fsubp st(2),st(0)
  fild  dword ptr [hb]
  fmul  qword ptr [c1]
  fadd  qword ptr [lb]
  fdiv  qword ptr [c2]
  {fld tbyte ptr [c3]
  fmulp }
  fmulp st(2),st(0)
  faddp
end;
end;



procedure XOR_intrand(Num: Cardinal){extended};  stdcall;
{const
c3: extended}
const
maxint: cardinal = 2147483647;{4294967295}
divmaxint: extended = 1/2147483647;
divmaxlong: extended = 1/4294967295;
var
rn: Cardinal;
res,res1: double;
begin

{in range max integer: 0..2^31-1=2147483647}
// Num*(ctxXOR.nr sh1 1) / 4294967296.0;

  xor4096_next(ctxXOR);

  rn:=ctxXOR.nr shr 1;

  asm
    fild  rn
    fimul Num
    fld tbyte ptr divmaxint
    fmulp
    FRNDINT
  end;


{in range max cardinal: 0..2^32-1=4294967295}
// Num*(ctxXOR.nr + 2147483648.0) / 4294967296.0;

{ XOR4096_next(ctxXOR);


  asm
    fild  ctxXOR.nr
    fiadd maxint
    fimul Num
    fld tbyte ptr divmaxlong
    fmulp
    FRNDINT
  end;
 }
end;


procedure XOR_rand01;  stdcall;
const
c1: extended = 1/4294967296.0;
c2: double =  2147483648.0;
c3: double =  4294967296.0;
var
rn: Cardinal;
res: extended;
begin
  xor4096_next(ctxXOR);


  asm
    fild   ctxXOR.nr
    fadd   c2
    fld c1
    fmulp
    //fdiv   c3
  end;

end;




{*****************************end xor4096************************************}



{*******************************GAMMA***********************************************}
        { (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/ }


function copysign(x,y: extended): extended; {$ifdef HAS_INLINE} inline;{$endif}
  {-Return abs(x)*sign(y)}
begin
  THexExtW(x)[4] := (THexExtW(x)[4] and $7FFF) or (THexExtW(y)[4] and $8000);
  copysign := x;
end;


function IsNaNorInf(x: extended): boolean;  {$ifdef HAS_INLINE} inline;{$endif}
  {-Return true if x is a NaN or infinite}
begin
  IsNaNorInf := TExtRec(x).xp and $7FFF=$7FFF;
end;

function ilogb(x: extended): longint;
  {-Return base 2 exponent of x. For finite x ilogb = floor(log2(|x|))}
  { otherwise -MaxLongint for x=0 and MaxLongint if x = +-INF or Nan. }
var
  e: integer;
  f: longint;
  m: extended;
begin
  e := THexExtW(x)[4] and $7FFF;
  if e=$7FFF then ilogb := MaxLongint
  else if x=0.0 then ilogb := -MaxLongint
  else if e<>0 then  ilogb := e-$3FFF
  else begin
    {e=0, x<>0: denormal use frexp exponent}
    frexp(x,m,f);
    ilogb := f-1;
  end;
end;


function floorx(x: extended): extended;
  {-Return the largest integer <= x}
var
  t: extended;
begin
  t := int(x);
  if (x>=0.0) or (x=t) then floorx := t
  else floorx := t - 1.0;
end;



function power_x(x, y : extended): extended;
  {-Return x^y; if frac(y)<>0 then x must be > 0}

  {This is my Pascal translation of the ldouble/powl.c file from the   }
  {freely distributable Cephes Mathematical Library V2.8, (June 2000)  }
  {Copyright 1984, 1991, 1998 by Stephen L. Moshier                    }

  {Following Cody and Waite, the Cephes function uses a lookup table of}
  {2**-i/NXT and pseudo extended precision arithmetic to obtain several}
  {extra bits of accuracy in both the logarithm and the exponential.   }

  {My routine uses a table size of 512 entries resulting in about four }
  {additional bits of accuracy; the tables are calculated with MPArith.}

const
  NXT = 512;  {table size, must be power of 2}

const
  MEXP   =  NXT*16384.0;
  MNEXP  = -NXT*(16384.0+64.0);    {+64 for denormal support}

const
  LOG2EA : THexExtW = ($c2ef,$705f,$eca5,$e2a8,$3ffd); { = 0.44269504088896340736 = log2(e)-1}

const
  {Coefficients for log(1+x) =  x - .5x^2 + x^3 *  P(z)/Q(z) }
  {on the domain  2^(-1/32) - 1  <=  x  <=  2^(1/32) - 1     }
  PH: array[0..3] of THexExtW = (
        ($b804,$a8b7,$c6f4,$da6a,$3ff4),
        ($7de9,$cf02,$58c0,$fae1,$3ffd),
        ($405a,$3722,$67c9,$e000,$3fff),
        ($cd99,$6b43,$87ca,$b333,$3fff));
  QH: array[0..2] of THexExtW = (
        ($6307,$a469,$3b33,$a800,$4001),
        ($fec2,$62d7,$a51c,$8666,$4002),
        ($da32,$d072,$a5d7,$8666,$4001));
  {Coefficients for 2^x = 1 + x P(x),  on the interval -1/32 <= x <= 0}
  RH: array[0..6] of THexExtW = (
        ($a69b,$530e,$ee1d,$fd2a,$3fee),
        ($c746,$8e7e,$5960,$a182,$3ff2),
        ($63b6,$adda,$fd6a,$aec3,$3ff5),
        ($c104,$fd99,$5b7c,$9d95,$3ff8),
        ($e05e,$249d,$46b8,$e358,$3ffa),
        ($5d1d,$162c,$effc,$f5fd,$3ffc),
        ($79aa,$d1cf,$17f7,$b172,$3ffe));
var
  P: array[0..3] of extended absolute PH;
  Q: array[0..2] of extended absolute QH;
  R: array[0..6] of extended absolute RH;

const
  {A[i] = 2^(-i/NXT), calculated with MPArith/t_powtab.pas}
  {If i is even, A[i] + B[i/2] gives additional accuracy. }
  A: array[0..NXT] of THexExtW = (
       ($0000,$0000,$0000,$8000,$3fff), ($aed2,$1c8d,$5652,$ffa7,$3ffe), ($c8a5,$511e,$cb59,$ff4e,$3ffe),
       ($a3ca,$fb18,$5f0a,$fef6,$3ffe), ($884c,$7b8f,$115c,$fe9e,$3ffe), ($695d,$3745,$e243,$fe45,$3ffe),
       ($9f35,$96a8,$d1b4,$fded,$3ffe), ($a15e,$05d2,$dfa6,$fd95,$3ffe), ($c175,$f486,$0c0c,$fd3e,$3ffe),
       ($e654,$d630,$56de,$fce6,$3ffe), ($47bb,$21e4,$c011,$fc8e,$3ffe), ($2a57,$525a,$4799,$fc37,$3ffe),
       ($9c49,$e5f0,$ed6c,$fbdf,$3ffe), ($3211,$5ea9,$b181,$fb88,$3ffe), ($c3f4,$4227,$93cc,$fb31,$3ffe),
       ($2bca,$19b1,$9443,$fada,$3ffe), ($033a,$722a,$b2db,$fa83,$3ffe), ($6270,$dc15,$ef8a,$fa2c,$3ffe),
       ($9f35,$eb93,$4a46,$f9d6,$3ffe), ($0c81,$3861,$c305,$f97f,$3ffe), ($ba74,$5dd4,$59bb,$f929,$3ffe),
       ($36c6,$fadf,$0e5e,$f8d3,$3ffe), ($4d9c,$b209,$e0e5,$f87c,$3ffe), ($cad3,$2972,$d145,$f826,$3ffe),
       ($3bb9,$0ad1,$df73,$f7d0,$3ffe), ($b12e,$036e,$0b65,$f77b,$3ffe), ($8239,$c428,$5510,$f725,$3ffe),
       ($0f0b,$016e,$bc6c,$f6cf,$3ffe), ($846e,$733f,$416c,$f67a,$3ffe), ($9f9c,$d52c,$e407,$f624,$3ffe),
       ($7290,$e653,$a433,$f5cf,$3ffe), ($28bd,$695f,$81e6,$f57a,$3ffe), ($cc2c,$2486,$7d15,$f525,$3ffe),
       ($0b17,$e18c,$95b5,$f4d0,$3ffe), ($fddf,$6db9,$cbbe,$f47b,$3ffe), ($ed7e,$99e3,$1f24,$f427,$3ffe),
       ($1a5b,$3a64,$8fde,$f3d2,$3ffe), ($8390,$271a,$1de1,$f37e,$3ffe), ($ae9c,$3b6b,$c923,$f329,$3ffe),
       ($6f7d,$563f,$919a,$f2d5,$3ffe), ($b13a,$59ff,$773c,$f281,$3ffe), ($3eda,$2c97,$79ff,$f22d,$3ffe),
       ($8cc1,$b770,$99d8,$f1d9,$3ffe), ($8280,$e774,$d6be,$f185,$3ffe), ($4509,$ad09,$30a7,$f132,$3ffe),
       ($0154,$fc11,$a788,$f0de,$3ffe), ($b76a,$cbe8,$3b58,$f08b,$3ffe), ($05e5,$1767,$ec0d,$f037,$3ffe),
       ($f5cb,$dcda,$b99b,$efe4,$3ffe), ($c6e3,$1e0a,$a3fb,$ef91,$3ffe), ($bc6b,$e032,$ab20,$ef3e,$3ffe),
       ($ea3d,$2c03,$cf03,$eeeb,$3ffe), ($025b,$0da3,$0f98,$ee99,$3ffe), ($22ea,$94a7,$6cd5,$ee46,$3ffe),
       ($a491,$d418,$e6b1,$edf3,$3ffe), ($e946,$e26f,$7d22,$eda1,$3ffe), ($2b84,$d994,$301e,$ed4f,$3ffe),
       ($4dea,$d6da,$ff9b,$ecfc,$3ffe), ($ab41,$fb03,$eb8f,$ecaa,$3ffe), ($e6f1,$6a3c,$f3f1,$ec58,$3ffe),
       ($bddc,$4c1c,$18b6,$ec07,$3ffe), ($d79f,$cba2,$59d4,$ebb5,$3ffe), ($9840,$1736,$b743,$eb63,$3ffe),
       ($f244,$60a5,$30f7,$eb12,$3ffe), ($392f,$dd24,$c6e7,$eac0,$3ffe), ($f464,$c548,$790a,$ea6f,$3ffe),
       ($b27b,$550e,$4756,$ea1e,$3ffe), ($dcf5,$cbd1,$31c0,$e9cd,$3ffe), ($8c57,$6c4f,$3840,$e97c,$3ffe),
       ($5cb8,$7ca4,$5acb,$e92b,$3ffe), ($42ab,$464b,$9958,$e8da,$3ffe), ($6094,$161c,$f3dd,$e889,$3ffe),
       ($dc68,$3c4b,$6a50,$e839,$3ffe), ($b5d3,$0c68,$fca8,$e7e8,$3ffe), ($9cbf,$dd5b,$aada,$e798,$3ffe),
       ($c84b,$0965,$74df,$e748,$3ffe), ($ce22,$ee1f,$5aaa,$e6f8,$3ffe), ($7a41,$ec78,$5c34,$e6a8,$3ffe),
       ($a717,$68b3,$7973,$e658,$3ffe), ($1616,$ca69,$b25c,$e608,$3ffe), ($48a8,$7c83,$06e7,$e5b9,$3ffe),
       ($5988,$ed3e,$7709,$e569,$3ffe), ($d681,$8e26,$02ba,$e51a,$3ffe), ($9a96,$d418,$a9ef,$e4ca,$3ffe),
       ($a88d,$373d,$6ca0,$e47b,$3ffe), ($05e2,$330d,$4ac2,$e42c,$3ffe), ($9619,$4649,$444c,$e3dd,$3ffe),
       ($f681,$f300,$5934,$e38e,$3ffe), ($5a51,$be8a,$8972,$e33f,$3ffe), ($6732,$3185,$d4fc,$e2f0,$3ffe),
       ($1226,$d7d9,$3bc7,$e2a2,$3ffe), ($7cdd,$40b2,$bdcc,$e253,$3ffe), ($d369,$fe83,$5aff,$e205,$3ffe),
       ($2a54,$a703,$1359,$e1b7,$3ffe), ($5d23,$d329,$e6cf,$e168,$3ffe), ($ed37,$1f30,$d559,$e11a,$3ffe),
       ($e111,$2a94,$deec,$e0cc,$3ffe), ($a3fe,$980f,$037f,$e07f,$3ffe), ($e627,$0d99,$430a,$e031,$3ffe),
       ($7d00,$3469,$9d82,$dfe3,$3ffe), ($4420,$b8f0,$12de,$df96,$3ffe), ($fe78,$4ada,$a316,$df48,$3ffe),
       ($37f2,$9d10,$4e1f,$defb,$3ffe), ($276d,$65af,$13f1,$deae,$3ffe), ($9124,$5e0e,$f482,$de60,$3ffe),
       ($a96f,$42bb,$efc9,$de13,$3ffe), ($f7f0,$d378,$05bc,$ddc7,$3ffe), ($3b1a,$d33d,$3653,$dd7a,$3ffe),
       ($4c20,$0832,$8185,$dd2d,$3ffe), ($0346,$3bb4,$e747,$dce0,$3ffe), ($1c92,$3a4f,$6791,$dc94,$3ffe),
       ($1cde,$d3c0,$0259,$dc48,$3ffe), ($3755,$daf2,$b797,$dbfb,$3ffe), ($3343,$25fe,$8742,$dbaf,$3ffe),
       ($5255,$8e29,$714f,$db63,$3ffe), ($3730,$efe4,$75b6,$db17,$3ffe), ($cc72,$2ac9,$946f,$dacb,$3ffe),
       ($2c0c,$219e,$cd6f,$da7f,$3ffe), ($8704,$ba4d,$20ad,$da34,$3ffe), ($0d97,$ddeb,$8e21,$d9e8,$3ffe),
       ($d7b6,$78af,$15c2,$d99d,$3ffe), ($cdeb,$79f9,$b786,$d951,$3ffe), ($929c,$d44a,$7364,$d906,$3ffe),
       ($6bac,$7d46,$4954,$d8bb,$3ffe), ($2c84,$6db3,$394c,$d870,$3ffe), ($2070,$a177,$4343,$d825,$3ffe),
       ($f56a,$1797,$6731,$d7da,$3ffe), ($a737,$d239,$a50b,$d78f,$3ffe), ($6af4,$d69d,$fcca,$d744,$3ffe),
       ($9af2,$2d20,$6e65,$d6fa,$3ffe), ($a2fe,$e13b,$f9d1,$d6af,$3ffe), ($ed01,$0180,$9f08,$d665,$3ffe),
       ($ce07,$9f9b,$5dfe,$d61b,$3ffe), ($739a,$d04f,$36ac,$d5d1,$3ffe), ($d18a,$ab75,$2909,$d587,$3ffe),
       ($9006,$4bfe,$350c,$d53d,$3ffe), ($fa1f,$cfed,$5aab,$d4f3,$3ffe), ($eca5,$585b,$99df,$d4a9,$3ffe),
       ($c561,$0972,$f29e,$d45f,$3ffe), ($52af,$0a6e,$64df,$d416,$3ffe), ($c379,$859a,$f099,$d3cc,$3ffe),
       ($978e,$a853,$95c4,$d383,$3ffe), ($9054,$a302,$5457,$d33a,$3ffe), ($a1de,$a91e,$2c49,$d2f1,$3ffe),
       ($e45a,$f12a,$1d91,$d2a8,$3ffe), ($85e3,$b4b5,$2827,$d25f,$3ffe), ($bcab,$3056,$4c02,$d216,$3ffe),
       ($b983,$a3af,$8918,$d1cd,$3ffe), ($9ac6,$5169,$df62,$d184,$3ffe), ($5f98,$7f34,$4ed6,$d13c,$3ffe),
       ($db8d,$75c5,$d76c,$d0f3,$3ffe), ($aaa0,$80d8,$791b,$d0ab,$3ffe), ($2595,$ef2b,$33da,$d063,$3ffe),
       ($56ab,$127e,$07a2,$d01b,$3ffe), ($eeb5,$3f94,$f468,$cfd2,$3ffe), ($3a86,$ce32,$fa24,$cf8a,$3ffe),
       ($18c1,$1919,$18cf,$cf43,$3ffe), ($f001,$7e0a,$505e,$cefb,$3ffe), ($a55d,$5dc6,$a0ca,$ceb3,$3ffe),
       ($934d,$1c07,$0a0a,$ce6c,$3ffe), ($80e4,$1f84,$8c15,$ce24,$3ffe), ($9967,$d1ee,$26e2,$cddd,$3ffe),
       ($6445,$9ff0,$da6a,$cd95,$3ffe), ($bd65,$f92c,$a6a3,$cd4e,$3ffe), ($cdd2,$503d,$8b86,$cd07,$3ffe),
       ($04bd,$1ab4,$8909,$ccc0,$3ffe), ($10e5,$d115,$9f23,$cc79,$3ffe), ($da4f,$eeda,$cdcd,$cc32,$3ffe),
       ($7c5d,$f272,$14fe,$cbec,$3ffe), ($4041,$5d3b,$74ae,$cba5,$3ffe), ($97c9,$b385,$ecd3,$cb5e,$3ffe),
       ($1886,$7c92,$7d66,$cb18,$3ffe), ($774e,$4290,$265e,$cad2,$3ffe), ($8414,$929e,$e7b2,$ca8b,$3ffe),
       ($2624,$fcc7,$c15a,$ca45,$3ffe), ($58aa,$1401,$b34f,$c9ff,$3ffe), ($27a3,$6e2f,$bd86,$c9b9,$3ffe),
       ($ad1a,$a41c,$dff8,$c973,$3ffe), ($0ecc,$517f,$1a9d,$c92e,$3ffe), ($7c14,$14f3,$6d6c,$c8e8,$3ffe),
       ($2c45,$8ffe,$d85c,$c8a2,$3ffe), ($5d4c,$6709,$5b66,$c85d,$3ffe), ($52b2,$4164,$f681,$c817,$3ffe),
       ($54fd,$c942,$a9a4,$c7d2,$3ffe), ($b15d,$abb9,$74c8,$c78d,$3ffe), ($b9ba,$98c2,$57e4,$c748,$3ffe),
       ($c51e,$4336,$52f0,$c703,$3ffe), ($306b,$60cf,$65e3,$c6be,$3ffe), ($5f79,$aa24,$90b5,$c679,$3ffe),
       ($be7f,$daac,$d35e,$c634,$3ffe), ($c3d9,$b0bb,$2dd6,$c5f0,$3ffe), ($f22e,$ed80,$a014,$c5ab,$3ffe),
       ($dadd,$5506,$2a11,$c567,$3ffe), ($20d3,$ae32,$cbc3,$c522,$3ffe), ($7baa,$c2c0,$8523,$c4de,$3ffe),
       ($bb30,$5f47,$5629,$c49a,$3ffe), ($cb33,$5334,$3ecc,$c456,$3ffe), ($b7b1,$70ca,$3f04,$c412,$3ffe),
       ($b15d,$8d21,$56c9,$c3ce,$3ffe), ($1279,$8026,$8613,$c38a,$3ffe), ($6407,$2497,$ccda,$c346,$3ffe),
       ($6351,$5807,$2b15,$c303,$3ffe), ($07c9,$fad9,$a0bc,$c2bf,$3ffe), ($8940,$f03f,$2dc8,$c27c,$3ffe),
       ($6673,$1e3d,$d231,$c238,$3ffe), ($6be9,$6da3,$8ded,$c1f5,$3ffe), ($bb33,$ca0f,$60f5,$c1b2,$3ffe),
       ($d278,$21ec,$4b42,$c16f,$3ffe), ($9456,$6670,$4cca,$c12c,$3ffe), ($5028,$8b9b,$6586,$c0e9,$3ffe),
       ($ca8d,$8836,$956e,$c0a6,$3ffe), ($4652,$55d5,$dc7a,$c063,$3ffe), ($8db0,$f0d0,$3aa1,$c021,$3ffe),
       ($fbdf,$5848,$afdd,$bfde,$3ffe), ($86f8,$8e24,$3c24,$bf9c,$3ffe), ($ca35,$970d,$df6f,$bf59,$3ffe),
       ($1083,$7a73,$99b6,$bf17,$3ffe), ($5f66,$4285,$6af1,$bed5,$3ffe), ($8238,$fc37,$5317,$be93,$3ffe),
       ($15b4,$b73d,$5222,$be51,$3ffe), ($93e2,$8609,$6809,$be0f,$3ffe), ($6049,$7dcf,$94c4,$bdcd,$3ffe),
       ($d483,$b67e,$d84b,$bd8b,$3ffe), ($4d18,$4ac5,$3297,$bd4a,$3ffe), ($36bf,$580c,$a39f,$bd08,$3ffe),
       ($1bdc,$fe78,$2b5b,$bcc7,$3ffe), ($b269,$60e7,$c9c5,$bc85,$3ffe), ($ea22,$a4f2,$7ed3,$bc44,$3ffe),
       ($fb0d,$f2e9,$4a7e,$bc03,$3ffe), ($7453,$75d4,$2cbf,$bbc2,$3ffe), ($4b6f,$5b70,$258d,$bb81,$3ffe),
       ($ebac,$d430,$34e0,$bb40,$3ffe), ($45fb,$133e,$5ab2,$baff,$3ffe), ($e119,$4e73,$96f9,$babe,$3ffe),
       ($ea09,$be5f,$e9ae,$ba7d,$3ffe), ($44e1,$9e42,$52ca,$ba3d,$3ffe), ($9deb,$2c0b,$d245,$b9fc,$3ffe),
       ($7b17,$a85c,$6816,$b9bc,$3ffe), ($4dbf,$5684,$1437,$b97c,$3ffe), ($84c1,$7c80,$d69f,$b93b,$3ffe),
       ($9ee9,$62fb,$af47,$b8fb,$3ffe), ($3dab,$554c,$9e27,$b8bb,$3ffe), ($3834,$a174,$a337,$b87b,$3ffe),
       ($aec9,$981f,$be70,$b83b,$3ffe), ($1e7c,$8ca4,$efca,$b7fb,$3ffe), ($752f,$d4ff,$373d,$b7bc,$3ffe),
       ($25e9,$c9d7,$94c2,$b77c,$3ffe), ($3d80,$c677,$0851,$b73d,$3ffe), ($7791,$28d1,$91e3,$b6fd,$3ffe),
       ($53cc,$517c,$316f,$b6be,$3ffe), ($2b8f,$a3b2,$e6ee,$b67e,$3ffe), ($47d3,$8550,$b259,$b63f,$3ffe),
       ($f76c,$5ed5,$93a8,$b600,$3ffe), ($a594,$9b63,$8ad3,$b5c1,$3ffe), ($f0d2,$a8b9,$97d3,$b582,$3ffe),
       ($c224,$f738,$baa0,$b543,$3ffe), ($6484,$f9de,$f333,$b504,$3ffe), ($9cbb,$2646,$4185,$b4c6,$3ffe),
       ($c180,$f4a9,$a58c,$b487,$3ffe), ($d3ed,$dfdb,$1f43,$b449,$3ffe), ($9841,$654b,$aea2,$b40a,$3ffe),
       ($aef3,$0501,$53a1,$b3cc,$3ffe), ($ae18,$419f,$0e38,$b38e,$3ffe), ($3b0e,$a05f,$de60,$b34f,$3ffe),
       ($2489,$a911,$c412,$b311,$3ffe), ($7cdd,$e61c,$bf46,$b2d3,$3ffe), ($b4a4,$e47d,$cff5,$b295,$3ffe),
       ($b5ac,$33c5,$f618,$b257,$3ffe), ($fe3b,$6618,$31a6,$b21a,$3ffe), ($bc9f,$102e,$8299,$b1dc,$3ffe),
       ($eb09,$c94f,$e8e8,$b19e,$3ffe), ($6bbd,$2b56,$648e,$b161,$3ffe), ($2590,$d2ac,$f581,$b123,$3ffe),
       ($20b2,$5e4a,$9bbc,$b0e6,$3ffe), ($a3c9,$6fb7,$5736,$b0a9,$3ffe), ($515c,$ab09,$27e8,$b06c,$3ffe),
       ($4584,$b6e0,$0dcb,$b02f,$3ffe), ($33f9,$3c69,$08d8,$aff2,$3ffe), ($8661,$e75b,$1906,$afb5,$3ffe),
       ($7af7,$65f8,$3e50,$af78,$3ffe), ($4375,$690a,$78ad,$af3b,$3ffe), ($2457,$a3e3,$c816,$aefe,$3ffe),
       ($9465,$cc5c,$2c84,$aec2,$3ffe), ($5c8e,$9ad6,$a5f0,$ae85,$3ffe), ($b80e,$ca35,$3452,$ae49,$3ffe),
       ($74e3,$17e4,$d7a4,$ae0c,$3ffe), ($1491,$43d0,$8fdd,$add0,$3ffe), ($ed2f,$1068,$5cf7,$ad94,$3ffe),
       ($4ac6,$42a1,$3eea,$ad58,$3ffe), ($90fb,$a1ec,$35af,$ad1c,$3ffe), ($5d04,$f83e,$413f,$ace0,$3ffe),
       ($a7ed,$1209,$6194,$aca4,$3ffe), ($e929,$be3f,$96a4,$ac68,$3ffe), ($3971,$ce50,$e06a,$ac2c,$3ffe),
       ($75e9,$1626,$3edf,$abf1,$3ffe), ($639b,$6c2a,$b1fa,$abb5,$3ffe), ($d337,$a93e,$39b5,$ab7a,$3ffe),
       ($c526,$a8c0,$d609,$ab3e,$3ffe), ($8de1,$4886,$86ef,$ab03,$3ffe), ($fa98,$68de,$4c5f,$aac8,$3ffe),
       ($7629,$ec90,$2652,$aa8d,$3ffe), ($2e5f,$b8d8,$14c2,$aa52,$3ffe), ($3979,$b569,$17a7,$aa17,$3ffe),
       ($bc03,$cc6b,$2efa,$a9dc,$3ffe), ($0ef8,$ea7c,$5ab4,$a9a1,$3ffe), ($e631,$fea9,$9ace,$a966,$3ffe),
       ($771b,$fa77,$ef41,$a92b,$3ffe), ($9fbe,$d1d8,$5806,$a8f1,$3ffe), ($0e09,$7b32,$d516,$a8b6,$3ffe),
       ($6770,$ef58,$6669,$a87c,$3ffe), ($70d1,$298f,$0bfa,$a842,$3ffe), ($36a1,$2789,$c5c0,$a807,$3ffe),
       ($356a,$e965,$93b4,$a7cd,$3ffe), ($828e,$71af,$75d1,$a793,$3ffe), ($f55b,$c55f,$6c0e,$a759,$3ffe),
       ($5062,$ebd9,$7665,$a71f,$3ffe), ($6b1e,$eee8,$94cf,$a6e5,$3ffe), ($5bdf,$dac3,$c745,$a6ab,$3ffe),
       ($a20c,$be08,$0dc0,$a672,$3ffe), ($509c,$a9be,$6839,$a638,$3ffe), ($38ea,$b151,$d6a9,$a5fe,$3ffe),
       ($15cb,$ea94,$5909,$a5c5,$3ffe), ($b6ee,$6dbe,$ef53,$a58b,$3ffe), ($2c84,$556d,$997f,$a552,$3ffe),
       ($f339,$be9e,$5786,$a519,$3ffe), ($2070,$c8b6,$2962,$a4e0,$3ffe), ($8ec5,$9576,$0f0c,$a4a7,$3ffe),
       ($0ae4,$4905,$087d,$a46e,$3ffe), ($809e,$09e6,$15ae,$a435,$3ffe), ($284c,$00ff,$3698,$a3fc,$3ffe),
       ($b47c,$5991,$6b34,$a3c3,$3ffe), ($7fe0,$413e,$b37c,$a38a,$3ffe), ($bb93,$e802,$0f68,$a352,$3ffe),
       ($9d94,$8037,$7ef3,$a319,$3ffe), ($8f9e,$3e91,$0215,$a2e1,$3ffe), ($5e39,$5a1f,$98c7,$a2a8,$3ffe),
       ($6819,$0c49,$4303,$a270,$3ffe), ($cdc5,$90d0,$00c1,$a238,$3ffe), ($a188,$25ce,$d1fc,$a1ff,$3ffe),
       ($17a8,$0bb3,$b6ac,$a1c7,$3ffe), ($b6e4,$8544,$aeca,$a18f,$3ffe), ($8939,$d79f,$ba50,$a157,$3ffe),
       ($4cf7,$4a34,$d938,$a11f,$3ffe), ($a612,$26c7,$0b7a,$a0e8,$3ffe), ($4fc2,$b971,$510f,$a0b0,$3ffe),
       ($4e69,$509b,$a9f2,$a078,$3ffe), ($21be,$3d01,$161b,$a041,$3ffe), ($f745,$d1ae,$9583,$a009,$3ffe),
       ($dd06,$6400,$2825,$9fd2,$3ffe), ($f493,$4ba1,$cdf9,$9f9a,$3ffe), ($a651,$e28b,$86f8,$9f63,$3ffe),
       ($d504,$8504,$531d,$9f2c,$3ffe), ($11ae,$91a1,$3260,$9ef5,$3ffe), ($cfa5,$693f,$24bb,$9ebe,$3ffe),
       ($98ff,$6f0b,$2a27,$9e87,$3ffe), ($4337,$0879,$429e,$9e50,$3ffe), ($2420,$9d47,$6e18,$9e19,$3ffe),
       ($4720,$977c,$ac90,$9de2,$3ffe), ($a2aa,$6367,$fdff,$9dab,$3ffe), ($4e03,$6f9f,$625e,$9d75,$3ffe),
       ($b751,$2cff,$d9a7,$9d3e,$3ffe), ($d9e2,$0eaa,$63d3,$9d08,$3ffe), ($74cb,$8a07,$00db,$9cd2,$3ffe),
       ($41be,$16c0,$b0ba,$9c9b,$3ffe), ($2c2d,$2ec3,$7368,$9c65,$3ffe), ($88b4,$4e40,$48df,$9c2f,$3ffe),
       ($4cc1,$f3aa,$3118,$9bf9,$3ffe), ($4688,$9fb3,$2c0e,$9bc3,$3ffe), ($5539,$d54e,$39b9,$9b8d,$3ffe),
       ($a17e,$19ad,$5a14,$9b57,$3ffe), ($d63d,$f441,$8d16,$9b21,$3ffe), ($599b,$eeb9,$d2bb,$9aeb,$3ffe),
       ($864a,$94ff,$2afc,$9ab6,$3ffe), ($e51a,$753b,$95d2,$9a80,$3ffe), ($66ca,$1fd1,$1337,$9a4b,$3ffe),
       ($9e27,$275d,$a324,$9a15,$3ffe), ($fa65,$20b7,$4593,$99e0,$3ffe), ($01c4,$a2f1,$fa7d,$99aa,$3ffe),
       ($8c77,$4751,$c1dd,$9975,$3ffe), ($ffcf,$a959,$9bab,$9940,$3ffe), ($89aa,$66c1,$87e2,$990b,$3ffe),
       ($5c25,$1f75,$867b,$98d6,$3ffe), ($e996,$7597,$976f,$98a1,$3ffe), ($20c6,$0d80,$bab9,$986c,$3ffe),
       ($a96f,$8db8,$f051,$9837,$3ffe), ($2103,$9eff,$3832,$9803,$3ffe), ($57ab,$ec43,$9255,$97ce,$3ffe),
       ($8d99,$22a6,$feb5,$9799,$3ffe), ($b08e,$f17a,$7d49,$9765,$3ffe), ($99b3,$0a41,$0e0e,$9731,$3ffe),
       ($4ba3,$20ac,$b0fb,$96fc,$3ffe), ($30cb,$ea9a,$660a,$96c8,$3ffe), ($5a00,$2018,$2d37,$9694,$3ffe),
       ($bd5c,$7b60,$0679,$9660,$3ffe), ($7560,$b8d9,$f1cb,$962b,$3ffe), ($0054,$9714,$ef27,$95f7,$3ffe),
       ($7fef,$d6cc,$fe86,$95c3,$3ffe), ($f940,$3ae8,$1fe3,$9590,$3ffe), ($94d5,$8878,$5336,$955c,$3ffe),
       ($df2c,$86b2,$987a,$9528,$3ffe), ($0961,$fef7,$efa8,$94f4,$3ffe), ($2a1f,$bccb,$58bb,$94c1,$3ffe),
       ($7ed3,$8ddb,$d3ac,$948d,$3ffe), ($ad28,$41f9,$6075,$945a,$3ffe), ($04b6,$ab1c,$ff0f,$9426,$3ffe),
       ($c105,$9d5c,$af75,$93f3,$3ffe), ($4bc1,$eef9,$71a0,$93c0,$3ffe), ($7f3c,$7851,$458b,$938d,$3ffe),
       ($e92c,$13e6,$2b2f,$935a,$3ffe), ($0dac,$9e5c,$2285,$9327,$3ffe), ($aa7c,$f673,$2b88,$92f4,$3ffe),
       ($fa89,$fd0f,$4632,$92c1,$3ffe), ($f9ac,$9531,$727d,$928e,$3ffe), ($a8b4,$a3f8,$b062,$925b,$3ffe),
       ($51ad,$10a0,$ffdc,$9228,$3ffe), ($cc64,$c481,$60e3,$91f6,$3ffe), ($c336,$ab11,$d373,$91c3,$3ffe),
       ($f815,$b1df,$5785,$9191,$3ffe), ($89d3,$c896,$ed13,$915e,$3ffe), ($39b3,$e0f9,$9417,$912c,$3ffe),
       ($b12b,$eee4,$4c8b,$90fa,$3ffe), ($c7f9,$e84d,$1669,$90c8,$3ffe), ($ca6b,$c540,$f1ab,$9095,$3ffe),
       ($bfef,$7fe0,$de4b,$9063,$3ffe), ($b1dc,$1466,$dc43,$9031,$3ffe), ($f285,$8120,$eb8c,$8fff,$3ffe),
       ($6481,$c672,$0c21,$8fce,$3ffe), ($c23d,$e6d1,$3dfc,$8f9c,$3ffe), ($e5c4,$e6c8,$8117,$8f6a,$3ffe),
       ($10d3,$ccf4,$d56c,$8f38,$3ffe), ($3520,$a201,$3af5,$8f07,$3ffe), ($3ce9,$70af,$b1ac,$8ed5,$3ffe),
       ($53c0,$45cd,$398b,$8ea4,$3ffe), ($2f95,$303a,$d28c,$8e72,$3ffe), ($5a01,$40e3,$7ca9,$8e41,$3ffe),
       ($79d3,$8ac4,$37dc,$8e10,$3ffe), ($9cd6,$22e6,$0420,$8ddf,$3ffe), ($81d8,$205f,$e16e,$8dad,$3ffe),
       ($e2f8,$9c50,$cfc0,$8d7c,$3ffe), ($c024,$b1e7,$cf11,$8d4b,$3ffe), ($a9e6,$7e5b,$df5b,$8d1a,$3ffe),
       ($0c60,$20ee,$0098,$8cea,$3ffe), ($7a95,$bae9,$32c1,$8cb9,$3ffe), ($f9e9,$6fa0,$75d2,$8c88,$3ffe),
       ($4dde,$646f,$c9c4,$8c57,$3ffe), ($4414,$c0b6,$2e91,$8c27,$3ffe), ($0085,$adde,$a434,$8bf6,$3ffe),
       ($4a01,$5754,$2aa7,$8bc6,$3ffe), ($d6e7,$ea8b,$c1e3,$8b95,$3ffe), ($9a18,$96fb,$69e4,$8b65,$3ffe),
       ($1032,$8e1e,$22a3,$8b35,$3ffe), ($8cfe,$0370,$ec1b,$8b04,$3ffe), ($8924,$2c72,$c645,$8ad4,$3ffe),
       ($f018,$40a4,$b11c,$8aa4,$3ffe), ($6e47,$7989,$ac9a,$8a74,$3ffe), ($bf7f,$12a1,$b8ba,$8a44,$3ffe),
       ($fd9a,$496e,$d575,$8a14,$3ffe), ($ef5e,$5d70,$02c6,$89e5,$3ffe), ($57a4,$9025,$40a7,$89b5,$3ffe),
       ($44b2,$2507,$8f13,$8985,$3ffe), ($5fdd,$618e,$ee03,$8955,$3ffe), ($3d5e,$8d2e,$5d72,$8926,$3ffe),
       ($ac6b,$f155,$dd5a,$88f6,$3ffe), ($078c,$d96e,$6db6,$88c7,$3ffe), ($8527,$92da,$0e80,$8898,$3ffe),
       ($8851,$6cf7,$bfb2,$8868,$3ffe), ($f1d4,$b919,$8146,$8839,$3ffe), ($717c,$ca8e,$5337,$880a,$3ffe),
       ($d792,$f698,$357f,$87db,$3ffe), ($66a0,$9473,$2819,$87ac,$3ffe), ($256c,$fd4e,$2afe,$877d,$3ffe),
       ($3131,$8c4e,$3e2a,$874e,$3ffe), ($1010,$9e8d,$6196,$871f,$3ffe), ($03c3,$9318,$953d,$86f0,$3ffe),
       ($5c88,$caef,$d919,$86c1,$3ffe), ($cc4a,$a905,$2d25,$8693,$3ffe), ($ba04,$923f,$915b,$8664,$3ffe),
       ($9563,$ed72,$05b5,$8636,$3ffe), ($2a9f,$2364,$8a2f,$8607,$3ffe), ($f694,$9ec9,$1ec1,$85d9,$3ffe),
       ($7b15,$cc48,$c367,$85aa,$3ffe), ($9376,$1a72,$781c,$857c,$3ffe), ($c95d,$f9c8,$3cd8,$854e,$3ffe),
       ($a9c0,$dcb8,$1198,$8520,$3ffe), ($1a29,$379c,$f656,$84f1,$3ffe), ($ae30,$80b8,$eb0b,$84c3,$3ffe),
       ($fd30,$303e,$efb3,$8495,$3ffe), ($f83c,$c049,$0447,$8468,$3ffe), ($4046,$acde,$28c3,$843a,$3ffe),
       ($7c8b,$73e9,$5d21,$840c,$3ffe), ($b132,$9541,$a15b,$83de,$3ffe), ($9628,$92a4,$f56c,$83b0,$3ffe),
       ($ee37,$efb6,$594e,$8383,$3ffe), ($de5a,$3203,$ccfd,$8355,$3ffe), ($4547,$e0fc,$5071,$8328,$3ffe),
       ($1333,$85f6,$e3a7,$82fa,$3ffe), ($a1d7,$ac2b,$8698,$82cd,$3ffe), ($0ca8,$e0bb,$393f,$82a0,$3ffe),
       ($894c,$b2a5,$fb97,$8272,$3ffe), ($c048,$b2ce,$cd9a,$8245,$3ffe), ($25ec,$73fc,$af43,$8218,$3ffe),
       ($5370,$8ad4,$a08c,$81eb,$3ffe), ($6056,$8dde,$a170,$81be,$3ffe), ($3bfd,$1581,$b1ea,$8191,$3ffe),
       ($0773,$bc03,$d1f3,$8164,$3ffe), ($6f7c,$1d88,$0188,$8138,$3ffe), ($06d4,$d814,$40a1,$810b,$3ffe),
       ($a0af,$8b85,$8f3b,$80de,$3ffe), ($ab6c,$d999,$ed4f,$80b1,$3ffe), ($8b84,$65e8,$5ad9,$8085,$3ffe),
       ($f6b1,$d5e5,$d7d2,$8058,$3ffe), ($4f51,$d0e0,$6436,$802c,$3ffe), ($0000,$0000,$0000,$8000,$3ffe));

  B: array[0..NXT div 2] of THexExtW = (
       ($0000,$0000,$0000,$0000,$0000), ($75f1,$bc63,$885f,$c06e,$3fbc), ($a1ed,$30c5,$4cd4,$a45b,$bfbd),
       ($7996,$d6b2,$a131,$ee62,$bfbc), ($23fa,$9c3e,$8b4d,$f581,$bfbd), ($ed9b,$4bb4,$c430,$8aba,$3fbd),
       ($3a32,$426a,$018d,$c4b4,$bfbd), ($9352,$0f8b,$4e4d,$decd,$3fbd), ($ff99,$62ba,$7628,$f84b,$3fbd),
       ($8143,$0442,$24f9,$b4b8,$3fbc), ($8d5b,$21a9,$86c7,$d2df,$3fbc), ($f811,$5153,$4607,$8019,$bfbd),
       ($8a8b,$f824,$b494,$b795,$bfb7), ($71b9,$b976,$bc54,$b92d,$bfbc), ($b9e0,$6310,$046b,$fced,$bfbd),
       ($f03c,$3c66,$a4f9,$cbc8,$3fbd), ($1f87,$db30,$18f5,$f73a,$3fbd), ($f2a3,$b0d9,$b054,$9480,$bfbb),
       ($a758,$8798,$7cdc,$b74d,$bfbd), ($7f23,$646f,$a919,$f15b,$3fb5), ($2da7,$b85c,$ab19,$bb3f,$bfbb),
       ($b86e,$5d58,$808e,$d6f3,$3fbd), ($cfcf,$b4c8,$42f5,$ec3f,$3fbc), ($e18c,$0bc4,$2a38,$ad64,$3fbd),
       ($7226,$291b,$39ed,$8cac,$3fbd), ($7cff,$5c4a,$6191,$ab5c,$3fbd), ($9321,$30a3,$3c06,$95de,$3fbd),
       ($f8e2,$0dde,$ca3a,$8736,$3fbc), ($f624,$4c97,$5b6d,$c01a,$3fbd), ($f23b,$506c,$1942,$b491,$bfbd),
       ($97d9,$0bf0,$0910,$9f3a,$3fbc), ($2748,$d2ff,$abf4,$bc2d,$bfbc), ($ac15,$3e46,$2932,$bf4a,$bfbc),
       ($20bd,$8514,$d685,$d4ef,$3fbd), ($259d,$efed,$3b0a,$f6e3,$bfba), ($d9ce,$8932,$32e4,$e4ef,$bfbc),
       ($8fbc,$58e1,$21a1,$f22f,$3fbd), ($43fe,$5717,$6dff,$e9b8,$bfbb), ($5160,$51be,$8fac,$f895,$3fbd),
       ($fbdb,$24a1,$8024,$83e1,$bfbb), ($8765,$76dd,$7a52,$f2f4,$3fbb), ($5d6e,$3d3f,$4342,$afb4,$bfbc),
       ($ddb5,$c442,$8805,$cbc4,$3fbd), ($8e61,$ab24,$c4aa,$d777,$bfbd), ($eecf,$5980,$9079,$9bfe,$3fba),
       ($3963,$50d1,$6c4a,$f8db,$bfbb), ($ac99,$7d75,$df0e,$b207,$bfbd), ($943c,$8a3a,$0dd6,$ba7a,$3fbd),
       ($7944,$ba66,$a091,$cb12,$3fb9), ($40a9,$b97a,$c086,$b4ab,$3fbd), ($9fa0,$e344,$2518,$8d70,$3fbd),
       ($b2ac,$6d49,$83a9,$988d,$bfbb), ($5e9c,$5ee6,$7498,$8be1,$bfbc), ($dc97,$9dfa,$3696,$ad2e,$3fbd),
       ($2972,$b47f,$6af5,$cb3c,$3fbd), ($7771,$7175,$32d9,$8595,$bfbd), ($a991,$78a6,$356a,$f610,$3fbc),
       ($43dd,$6e2f,$883f,$f352,$3fbb), ($1d1c,$f187,$dd36,$efdd,$bfbc), ($8358,$36fd,$6208,$9c21,$3fbd),
       ($2a20,$0f6a,$09ae,$bc61,$bfb7), ($83db,$4d18,$7e2a,$bf6d,$bfbd), ($1cbc,$ed94,$bf8d,$8559,$3fbc),
       ($ad2d,$b32d,$cc57,$bf03,$bfbd), ($ff78,$40b4,$2ee6,$e69a,$3fbc), ($6957,$1b4c,$4e47,$c449,$bfbb),
       ($1837,$3407,$b9db,$8d32,$bfbc), ($7a1e,$aecd,$45ae,$f389,$bfbc), ($58b5,$4c4c,$bdff,$b243,$3fbd),
       ($0795,$517f,$742b,$cf65,$bfba), ($8797,$f1a9,$b158,$dfb2,$3fbd), ($70d3,$3663,$0670,$f563,$3fbc),
       ($0143,$1ef2,$72be,$9124,$3fbb), ($8b0e,$bf14,$c09d,$ff4e,$3fba), ($3fd1,$56aa,$b86d,$b8fb,$3fba),
       ($01a1,$78c9,$ba0c,$f388,$bfbc), ($abf8,$996c,$8e6a,$a4ae,$bfbc), ($7873,$a64a,$a1c0,$c62a,$3fbd),
       ($21f2,$8dc0,$1cc9,$994f,$3fbc), ($7323,$e675,$113e,$a0ae,$3fbc), ($c895,$5069,$e383,$ee53,$bfbb),
       ($3036,$2fc1,$edaa,$9c9b,$bfba), ($14e6,$8c84,$73b9,$ef64,$bfbd), ($344f,$8ec8,$24c7,$bdaa,$3fbd),
       ($9231,$a140,$370b,$b6f8,$bfba), ($3b7d,$2256,$fc67,$9659,$bfbd), ($8383,$0390,$6a5f,$b7c9,$bfbd),
       ($a683,$6654,$5f4b,$f59e,$bfbc), ($2d04,$f5dd,$0dab,$fe3c,$bfbd), ($024d,$6fed,$c73b,$abf4,$bfbd),
       ($0385,$6e1c,$d3ed,$c368,$3fbc), ($aaed,$f499,$7f8f,$b2a1,$3fbd), ($6207,$24ff,$471a,$fb17,$bfbc),
       ($5d3b,$72dc,$e4d9,$a537,$bfbc), ($84aa,$c55d,$d161,$aa1c,$3fbd), ($4ef2,$1b9e,$10d1,$d7bf,$3fbd),
       ($7cde,$9376,$4325,$f8ab,$3fbc), ($0271,$0aa0,$1d48,$e551,$3fbd), ($9ed8,$b162,$20d2,$cf43,$bfbd),
       ($34c3,$ef34,$b6d1,$b5f4,$3fbc), ($e90a,$a2e0,$1584,$83b2,$3fbc), ($5a72,$5b3e,$e2c8,$9d23,$bfbd),
       ($c8f1,$bf6a,$6839,$d094,$bfbd), ($2880,$f54f,$77ca,$e68c,$3fbd), ($0f6b,$4a01,$fab3,$f88a,$3fbd),
       ($96b3,$aca3,$bab8,$f23c,$bfbd), ($a707,$354e,$649a,$de67,$3fbd), ($7c2d,$5246,$6cee,$ee30,$3fba),
       ($3d7b,$a07a,$7aa1,$bf51,$bfbb), ($da5f,$d319,$2351,$8901,$bfbd), ($b2dd,$7562,$4593,$9334,$3fbd),
       ($81bb,$2209,$1a37,$ed61,$bfbd), ($a10c,$25e0,$c093,$aefd,$bfbd), ($a77b,$9ef7,$556e,$d431,$3fbc),
       ($9a84,$e9a8,$fef4,$a3fa,$bfbc), ($a31d,$eff0,$228f,$ee2d,$3fba), ($0718,$8b27,$33a4,$e9aa,$3fbd),
       ($fcc0,$e4bd,$6478,$a79d,$bfb5), ($5478,$3b9f,$65d5,$d96c,$bfbb), ($46fe,$1c8c,$358b,$a80b,$bfbd),
       ($2d0d,$b35b,$bbc2,$dc3c,$3fbb), ($ece5,$6684,$ae95,$a6ec,$bfbd), ($ff37,$4277,$9e7c,$fc36,$3fbc),
       ($46da,$590d,$0d64,$ba4f,$bfbc), ($4b3f,$aa83,$e1bb,$e2cb,$3fb9), ($efe7,$c23c,$bc5c,$aa6d,$3fbd),
       ($0214,$b0cc,$7ff0,$9566,$bfbd), ($a511,$5e39,$6cac,$e15a,$bfbc), ($7d3e,$ea95,$1366,$b2fb,$3fbd),
       ($838b,$8b93,$5d68,$97b3,$3fbd), ($e500,$3363,$6118,$ea37,$bfbb), ($530c,$2670,$0d8c,$e64b,$bfbd),
       ($4670,$e629,$5371,$fb3c,$3fbc), ($0259,$822c,$69cf,$f572,$bfbd), ($e410,$d9a4,$4c4e,$f871,$3fbd),
       ($477b,$1aa3,$469b,$ef41,$bfbb), ($44e3,$25bd,$902d,$f05f,$bfbd), ($4b87,$09bd,$ae13,$cf92,$3fbd),
       ($cb15,$91c4,$d5b7,$90a6,$bfbd), ($70d3,$0dc8,$8646,$a443,$3fbd), ($bf35,$d132,$bf8c,$8367,$bfbc),
       ($dba0,$dea4,$a5d4,$ac2b,$3fbc), ($07b1,$0d44,$02d3,$9637,$3fbc), ($b6fe,$cca3,$b983,$bd67,$3fba),
       ($5d89,$eb34,$5191,$9301,$3fbd), ($79c4,$83b1,$19f9,$b3b9,$bfbd), ($ec93,$bcf2,$7343,$bc2b,$3fbd),
       ($7312,$e23a,$2bdc,$c645,$bfbc), ($00f3,$eb3c,$4764,$cb00,$3fbd), ($eed6,$fbe6,$a3ba,$db85,$bfbd),
       ($7264,$4ca6,$0243,$ec62,$3fbd), ($6cc4,$df77,$4b57,$9b46,$3fbc), ($f4e6,$6a63,$49d8,$a83c,$3fbd),
       ($b45b,$11d7,$8202,$ce57,$3fbc), ($6f19,$7dde,$37b2,$d0ad,$bfbd), ($2f65,$cbd8,$36e8,$9369,$3fbc),
       ($1aa2,$f8c1,$9655,$c274,$bfbd), ($c479,$6b70,$68a1,$a0b4,$3fbd), ($7b4e,$f33b,$e561,$c910,$bfbd),
       ($9774,$519b,$7aa1,$eb4b,$bfbc), ($80d9,$b883,$fb10,$e5eb,$3fbb), ($f6f0,$fdf9,$e512,$b70f,$bfbd),
       ($461e,$9586,$f46f,$d8bc,$3fbd), ($a3fa,$a7e0,$9267,$a257,$bfbc), ($1eec,$781e,$4831,$d1db,$3fba),
       ($b6b0,$331c,$4457,$f0ed,$bfbb), ($20c0,$6268,$a6dd,$ed0b,$bfbd), ($7a5c,$038e,$e827,$df33,$3fbd),
       ($42b1,$fe60,$f620,$c90b,$bfbd), ($da65,$029a,$04c0,$be97,$3fba), ($7e1e,$3313,$6bef,$fbbc,$bfbd),
       ($6a4d,$915c,$a675,$e91f,$3fbc), ($cd4c,$d87e,$3cf6,$c96e,$3fbb), ($74e4,$eded,$42cb,$81d2,$bfbc),
       ($73e3,$cd04,$5be7,$8fe5,$bfb9), ($379c,$7943,$7193,$cc6f,$bfbd), ($045d,$288c,$c1ec,$bedd,$bfbd),
       ($2374,$c710,$c284,$8d08,$3fbd), ($f035,$9bba,$5ac7,$f914,$3fb6), ($9707,$39e2,$6400,$e647,$bfbd),
       ($93d4,$bc59,$cc3e,$86da,$bfbc), ($1ff1,$9a23,$4714,$d2fe,$bfbc), ($53bd,$d0c8,$d9be,$9cb0,$3fbd),
       ($ae8d,$7102,$4c53,$8d58,$3fbd), ($81c2,$b867,$d0ba,$baaf,$bfbd), ($e89f,$2ec0,$365c,$a250,$bfbb),
       ($1adb,$75e7,$ec6e,$c468,$3fbc), ($714a,$cb7d,$9eef,$b14d,$3fbd), ($1a8c,$5a50,$c9a6,$de7b,$bfbb),
       ($8739,$39e4,$d2c3,$85f0,$3fbd), ($443a,$e0f2,$79fe,$c61c,$bfbc), ($8e76,$65ec,$8180,$cb81,$bfbd),
       ($eded,$5c85,$4630,$8d5a,$3fbd), ($8609,$5b52,$3509,$eaab,$3fba), ($af61,$3833,$5d52,$a0f4,$3fbd),
       ($523a,$33e5,$218f,$9802,$bfbc), ($458a,$7538,$36d0,$91d5,$3fbd), ($48b9,$923c,$75a8,$a2d3,$bfbc),
       ($dde8,$aa3b,$6381,$adcd,$bfba), ($4325,$58c4,$e96f,$bce6,$3fbb), ($7951,$9547,$eb44,$ba2b,$3fbc),
       ($bd1d,$d0c0,$eff8,$e375,$3fbc), ($7fdb,$a096,$f03c,$f15c,$3fbd), ($a970,$4a37,$030c,$c1fd,$bfbc),
       ($c463,$89b6,$afc5,$b319,$bfbc), ($74dd,$962b,$618d,$d125,$3fbc), ($6a72,$5204,$7a7a,$aadf,$3fbb),
       ($ce25,$a73d,$8d59,$cdbb,$bfb8), ($9d82,$e5ac,$8e0a,$fd6d,$3fba), ($1e38,$d3d9,$aefa,$9f6a,$3fbd),
       ($49e1,$424b,$6d6b,$d02d,$bfbd), ($cda6,$6474,$7e55,$9696,$3fbc), ($0fc5,$929d,$2950,$eeb0,$3fbd),
       ($3c94,$28f4,$da14,$b6d7,$3fbd), ($826a,$e8f2,$0890,$cffb,$3fba), ($ba60,$486f,$dcd2,$89fd,$3fbd),
       ($5317,$1e0f,$5132,$b700,$3fbc), ($c524,$6473,$087c,$bbc0,$3fbc), ($0653,$81ca,$4bbd,$e18d,$bfbd),
       ($8588,$002b,$92db,$f4b6,$bfba), ($a594,$e37c,$96d2,$9670,$bfbd), ($14b4,$820b,$7c5f,$b761,$3fbd),
       ($6840,$afa3,$5c3d,$8f46,$bfb9), ($06b3,$cdef,$a9d2,$e00e,$bfba), ($6dfd,$eb58,$af14,$8373,$bfb9),
       ($c4e4,$5cd8,$d15e,$dc47,$bfbb), ($92a3,$13bf,$dd56,$d573,$3fb9), ($bbfe,$55af,$dfa2,$f41c,$bfbb),
       ($0c22,$c368,$1d92,$80ca,$3fba), ($6cb6,$e77c,$d246,$847d,$bfbd), ($08d9,$971d,$cdb2,$d452,$bfbd),
       ($3455,$2381,$3f01,$eac2,$3fbd), ($eac6,$318c,$aed9,$bbf1,$3fbd), ($d5c0,$baa7,$d25f,$dd62,$3fbd),
       ($aeb2,$ebd9,$2275,$df6e,$bfbd), ($b52b,$7922,$eaff,$9190,$3fbd), ($a559,$ed4a,$9f15,$e85c,$3fbc),
       ($4b5f,$b50b,$1f98,$a6f2,$bfbc), ($eaa7,$4742,$8ed8,$91f4,$bfbc), ($e75c,$7784,$690e,$c483,$3fbc),
       ($f938,$7aac,$91cf,$e8da,$bfbc), ($0e4d,$9bc6,$9e8e,$b643,$3fbb), ($a5d1,$6b5b,$62c2,$f030,$3fba),
       ($e760,$a1eb,$147f,$c700,$bfba), ($ac10,$9fe8,$7650,$d7c9,$3fbb), ($b51d,$bff9,$1161,$cd15,$3fbc),
       ($4be1,$1360,$589e,$eff1,$bfbb), ($1f52,$e5ac,$e670,$ded5,$bfbd), ($2ffd,$1948,$1d6d,$f8a9,$3fbc),
       ($63ed,$2320,$a834,$de4e,$3fbc), ($9130,$5b81,$5df2,$c706,$bfbd), ($d9d8,$f72d,$8a90,$bf70,$bfbd),
       ($bde9,$7a29,$ca4f,$f7ca,$3fbd), ($54fb,$4a66,$3ab1,$cef0,$3fba), ($f480,$db9b,$5c66,$94cd,$3fbc),
       ($6abc,$ee23,$ec13,$d654,$bfbd), ($0000,$0000,$0000,$0000,$0000));
var
  w, z, ya, yb, u: extended;
  F, Fa, Fb, G, Ga, Gb, H, Ha, Hb: extended;
  e,i : longint;
  k: integer;
  nflg : boolean;
begin

  if IsNanOrInf(x) or IsNanOrInf(y) then begin
    power_x := NaN_x;
    exit;
  end;

  {Handle easy cases}
  if y=0.0 then begin
    power_x := 1.0;
    exit;
  end
  else if (x=0.0) then begin
    if y>0.0 then power_x := 0.0
    else begin
      {here y<0: if y is odd and x=-0 return -INF}
      if frac(0.5*y)=0.0 then power_x := PosInf_x
      else power_x := copysign(PosInf_x,x);
    end;
    exit;
  end
  else if y=1.0 then begin
    power_x := x;
    exit;
  end
  else if x=1.0 then begin
    power_x := 1.0;
    exit;
  end
  else if abs(y)=0.5 then begin
    if y>0.0 then power_x := sqrt(x)
    else power_x := 1.0/sqrt(x);
    exit;
  end;

  nflg := false;         {true if x<0 raised to integer power }
  w := floorx(y);
  if w=y then begin
    z := abs(w);
    u := ilogb(x);
    {intpower does not catch overflows, therefore call it only if safe.}
    if (z<=2048) and (z*(abs(u) + 1.0) < 16382.0) then begin
      power_x := intpower(x,trunc(w));
      exit;
    end;
  end;
  if x<0.0 then begin
    if w<>y then begin
      {noninteger power of negative number: exception or RTE}
      power_x := _exp_x(y*ln(x));
      exit;
    end;
    {Find out if the integer exponent is odd or even.}
    w := 2.0*floorx(0.5*y);
    nflg := w<>y;
    x := abs(x);
  end;

  {separate significand from exponent}
  frexp(x, x, e);

  {find significand in antilog table A[]}
  i := 1;
  if x <= extended(A[257])  then i := 257;
  if x <= extended(A[i+128])then inc(i,128);
  if x <= extended(A[i+64]) then inc(i,64);
  if x <= extended(A[i+32]) then inc(i,32);
  if x <= extended(A[i+16]) then inc(i,16);
  if x <= extended(A[i+8])  then inc(i, 8);
  if x <= extended(A[i+4])  then inc(i, 4);
  if x <= extended(A[i+2])  then inc(i, 2);
  if x >= extended(A[1])    then i := -1;
  inc(i);

  {Find (x - A[i])/A[i] in order to compute log(x/A[i]):}
  {log(x) = log( a x/a ) = log(a) + log(x/a)            }
  {log(x/a) = log(1+v),  v = x/a - 1 = (x-a)/a          }
  x := x - extended(A[i]);
  x := x - extended(B[i div 2]);
  x := x / extended(A[i]);

  {rational approximation for log(1+v) = v - v**2/2 + v**3 P(v) / Q(v)}
  z := x*x;

  {w:= polevl(x,P,3)}
  w := ((P[0]*x + P[1])*x + P[2])*x + P[3];
  {u = p1evl(x,Q,3)}
  u := ((x + Q[0])*x + Q[1])*x + Q[2];

  w := x*(z*w/u) - 0.5*z;

  {Convert to base 2 logarithm: multiply by log2(e) = 1+LOG2EA}
  z := extended(LOG2EA)*w;
  z := z + w;
  z := z + extended(LOG2EA)*x;
  z := z + x;

  {Compute exponent term of the base 2 logarithm.}
  w := -i;
  w := w/NXT + e;

  {Now base 2 log of x is w + z.  Multiply base 2 log by y, in pseudo multi}
  {precision. Separate y into large part ya and small part yb less than 1/NXT}

  {WE: mul/div with NXT is faster than original ldexp(,LNXT)}

  ya := floorx(y*NXT)/NXT;    {reduc(y)}
  yb := y - ya;

  F  := z*y + w*yb;
  Fa := floorx(F*NXT)/NXT;    {reduc(F)}
  Fb := F - Fa;

  G  := Fa + w*ya;
  Ga := floorx(G*NXT)/NXT;    {reduc(G)}
  Gb := G - Ga;

  H  := Fb + Gb;
  Ha := floorx(H*NXT)/NXT;    {reduc(H)}
  w  := NXT*(Ga+Ha);

  {Test the power of 2 for over/underflow}
  if w>MEXP then begin
    if nflg then power_x := NegInf_x
    else power_x := PosInf_x;
    exit;
  end
  else if w<MNEXP then begin
    power_x := 0.0;
    exit;
  end;

  e := trunc(w);
  Hb := H - Ha;
  if Hb>0.0 then begin
    inc(e);
    Hb := Hb - 1.0/NXT;
  end;

  {Now the product y * log2(x) = Hb + e/NXT. Compute base 2 exponential}
  {of Hb, where -1/NXT <= Hb <= 0.}

  {z := Hb*polevl(Hb,R,6);    z=2**Hb-1}
  z := R[0];
  for k:=1 to 6 do z := z*Hb + R[k];
  z := Hb*z;

  {Express e/NXT as an integer plus a negative number of (1/NXT)ths.}
  {Find lookup table entry for the fractional power of 2.}
  if e<0 then i := 0 else i := 1;
  inc(i, e div NXT);
  e := NXT*i - e;

  w := extended(A[e]);
  z := w + w*z;     {2**-e * ( 1 + (2**Hb-1) )     }
  z := ldexp(z,i);  {multiply by integer power of 2}

  if nflg then power_x := -z {odd integer exponent and x<0}
  else power_x := z;

end;


function exp_x(x: extended): extended;
  {-slow but accurate exp, result good to extended precision}
const
  xmin = -11398.80538430831;  { < -11398.8053843083006133663822374 = ln(succx(0))}
var
  z,t: extended;
  k: longint;
  ln2hi: double absolute ln2_hi;
  ln2lo: double absolute ln2_lo;
  zwa: array[0..4] of word absolute z;
begin
  if x > ln_MaxExt then exp_x := system.exp(x)
  else if x < xmin then exp_x := 0.0
  else if x < ln_MinExt then begin
    {-11398.80538430831 <= x < ln_MinExt: denormal try exp(x/2)^2}
    t := exp_x(0.5*x);
    exp_x := t*t;
  end
  else begin
    {rewrite exp(x) = exp(x - k*ln2 + k*ln2) = exp(k*ln2)*exp(x - k*ln2)}
    { = 2^k*exp(x - k*ln2) with small x-k*ln2, ie k ~= x/ln2 = x*log2(e)}
    k := trunc(log2e*x);
    t := k;
    z := x - t*ln2hi;
    t := t*ln2lo;
    x := z - t;
    {z = exp(x - k*ln2}
    z := system.exp(x);
    {multiply with 2^k, handle over/underflow}
    {z should be >=0, ie the sign bit should be clear, but..}
    k := k + (zwa[4] and $7FFF);
    if k<0 then begin
      {underflow}
      exp_x := 0.0;
    end
    else begin
      if k>=32767 then begin
        {generate floating point overflow}
        RunError(205);
      end
      else begin
        {store exponent of 2^k*exp(x - k*ln2)}
        zwa[4] := k;
      end;
      exp_x := z;
    end;
  end;
end;


function sin_x(x: extended): extended;
  {-Accurate version of circular sine, uses system.sin for |x| <= Pi/4}
var
  t: extended;
begin
  {reduction mod Pi/2, |t| <= Pi/4}
  case rem_pio2_x(x,t) and 3 of
      0:  sin_x :=  system.sin(t);
      1:  sin_x :=  system.cos(t);
      2:  sin_x := -system.sin(t);
    else  sin_x := -system.cos(t);
  end;
end;

function cos_x(x: extended): extended;
  {-Accurate version of circular cosine, uses system.cos for |x| <= Pi/4}
var
  t: extended;
begin
  {reduction mod Pi/2, |t| <= Pi/4}
  case rem_pio2_x(x,t) and 3 of
      0:  cos_x :=  system.cos(t);
      1:  cos_x := -system.sin(t);
      2:  cos_x := -system.cos(t);
    else  cos_x :=  system.sin(t);
  end;
end;


procedure sincos_x(x: extended; var s,c: extended);
  {-Return accurate values s=sin(x), c=cos(x)}
var
  t,ss,cc: extended;
  n: integer;
begin
  {reduction mod Pi/2, |t| <= Pi/4}
  n := rem_pio2_x(x,t) and 3;
  sincos(t,ss,cc);
  case n of
      0: begin s:= ss; c:= cc; end;
      1: begin s:= cc; c:=-ss; end;
      2: begin s:=-ss; c:=-cc; end;
    else begin s:=-cc; c:= ss; end;
  end;
end;



function cosh_x(x: extended): extended;
  {-Return the hyperbolic cosine of x}
const
  x0 = 23.0;     {ceil(-0.5*ln(2^-64)) = ceil(32*ln(2))}
begin
  x := abs(x);
  if x<=1.0 then cosh_x := 1.0 + coshm1(x)
  else begin
    if x>x0 then begin
      {x>x0: exp(x) + exp(-x) ~ exp(x)}
      if x<ln_MaxExt then cosh_x := 0.5*_exp_x(x)
      else cosh_x := _exp_x(x-ln2);
    end
    else begin
      {calculate inverse only if it is not too small}
      x := _exp_x(x);
      cosh_x := 0.5*(x + 1.0/x);
    end
  end;
end;


function sinh_x(x: extended): extended;
  {-Return the hyperbolic sine of x, accurate even for x near 0}
var
  t: extended;
const
  t0 = 23.0;     {ceil(-0.5*ln(2^-64)) = ceil(32*ln(2))}
begin
  t := abs(x);
  if t<1.0 then sinh_x := sinh_small(x)
  else begin
    {sinh(t) = 0.5*(exp(t) - exp(-t))}
    if t<=t0 then begin
      {calculate inverse only if it is not too small}
      t := _exp_x(t);
      t := 0.5*(t - 1.0/t);
    end
    else if t<ln_MaxExt then begin
      {t>t0: exp(t)-exp(-t) ~ exp(t)}
      t := 0.5*_exp_x(t)
    end
    else begin
      {exp(t) would overflow, use small margin below overflow}
      t := _exp_x(t-ln2);
    end;
    if x>0.0 then sinh_x := t
    else sinh_x := -t;
  end;
end;



function _exp_x(x: extended): extended;
begin
 if F_FastSpec = True then  _exp_x:=exp(x) else _exp_x:=exp_x(x)
end;

function _power_x(x,y: extended): extended;
begin
 if F_FastSpec = True then  _power_x:=power(x,y) else _power_x:=power_x(x,y)
end;



function sinh_small(x: extended): extended;
  {-Internal: return sinh(x) for |x| <= 1}
const
  ncs=9;
  csh: array[0..ncs-1] of THexExtW = (     {chebyshev(sinh(x)/x-1, x=-1..1, 0.1e-20)}
         ($D006,$0AFA,$F911,$B131,$3FFC),  {+0.173042194047179631675883846985    }
         ($4EEC,$D088,$9973,$B364,$3FFB),  {+0.875942219227604771549002634544e-1 }
         ($AECD,$1FE8,$437A,$8D7D,$3FF5),  {+0.107947777456713275024272706516e-2 }
         ($D1B0,$68E6,$89C6,$D5E7,$3FED),  {+0.637484926075475048156855545659e-5 }
         ($DAE7,$9306,$8CA6,$BD2E,$3FE5),  {+0.220236640492305301591904955350e-7 }
         ($E0C5,$862E,$36BE,$DB5F,$3FDC),  {+0.498794018041584931494262802066e-10}
         ($5D46,$9B41,$8645,$B389,$3FD3),  {+0.797305355411573048133738827571e-13}
         ($D7DD,$B169,$A8B3,$DA6F,$3FC9),  {+0.947315871307254445124531745434e-16}
         ($7636,$F264,$EF5C,$CD44,$3FBF)); {+0.869349205044812416919840906342e-19}
var
  css: array[0..ncs-1] of extended absolute csh;
var
  t: extended;
const
  t1 = 2.3E-10;  {~ sqrt(2^-64)}
begin
  if abs(x)<=t1 then begin
    {sinh(x) = x*(1 + 1/6*x^2 + 1/120*x^4 + O(x^6))}
    sinh_small := x;
  end
  else begin
    t := CSEvalX(2.0*sqr(x)-1.0, css, ncs);
    sinh_small := x + x*t;
  end;
end;



function CSEvalX(x: extended; const a: array of extended; n: integer): extended;
var
  b0,b1,b2: extended;
  i: integer;
begin
  {$ifdef CONST}
    {$ifdef debug}
    {.248}
    {
      if n>high(a)+1 then begin
        writeln('CSEvalX:  n > high(a)+1, n = ',n, ' vs. ', high(a)+1);
        readln;
      end;
      }
    {$endif}
    if n>high(a)+1 then n := high(a)+1;
  {$endif}
  b2 := 0.0;
  b1 := 0.0;
  b0 := 0.0;
  x  := 2.0*x;
  for i:=n-1 downto 0 do begin
    b2 := b1;
    b1 := b0;
    b0 := x*b1 - b2 + a[i];
  end;
  CSEvalX := 0.5*(b0-b2);
end;


function coshm1(x: extended): extended;
  {-Return cosh(x)-1, accurate even for x near 0}
const
  ncc=8;
  cch: array[0..ncc-1] of THexExtW = (     {chebyshev(((cosh(x)-1)/(x^2)-1/2)/x^2, x=-1..1, 0.1e-20);}
         ($7F5B,$DC92,$B00E,$AD8C,$3FFB),  {+0.847409967933085972383472682702e-1 }
         ($E233,$6D68,$4FB4,$B954,$3FF4),  {+0.706975331110557282636312722683e-3 }
         ($C401,$57F2,$9421,$D38C,$3FEC),  {+0.315232776534872632694769980081e-5 }
         ($60F4,$9F76,$CDC0,$9634,$3FE4),  {+0.874315531301413118696907944019e-8 }
         ($D666,$F7AB,$C086,$9172,$3FDB),  {+0.165355516210397415961496829471e-10}
         ($5E0D,$C9ED,$687C,$CC55,$3FD1),  {+0.226855895848535933261822520050e-13}
         ($0DEF,$8B25,$7570,$D9B9,$3FC7),  {+0.236057319781153495473072617928e-16}
         ($7B04,$4A7F,$2A3E,$B5FC,$3FBD)); {+0.192684134365813759102742156829e-19}
var
  ccs: array[0..ncc-1] of extended absolute cch;
  z: extended;
begin
  x := abs(x);
  if x<=1.0 then begin
    x := sqr(x);
    if x<eps_x then begin
      {cosh(x)-1 = 0.5*x^2*[1 + 1/12*x^2 + O(x^4)]}
      coshm1 := 0.5*x;
    end
    else begin
      z := CSEvalX(2.0*x-1.0, ccs, ncc);
      coshm1 := 0.5*x + x*x*z;
    end;
  end
  else coshm1 := cosh_x(x)-1.0;
end;



function rem_pio2_cw_x(x: extended; var z: extended): integer;
  {-Cody/Waite reduction of x: z = x - n*Pi/2, |z| <= Pi/4, result = n mod 8}
var
  i: longint;
  y: extended;
const
  HP1 : THexExtW = ($0000,$0000,$da80,$c90f,$3ffe);
  HP2 : THexExtW = ($0000,$0000,$a300,$8885,$3fe4);
  HP3 : THexExtW = ($3707,$a2e0,$3198,$8d31,$3fc8);
var
  DP1 : extended absolute HP1; {= 7.853981554508209228515625E-1;}
  DP2 : extended absolute HP2; {= 7.94662735614792836713604629E-9;}
  DP3 : extended absolute HP3; {= 3.06161699786838294306516483E-17;}
begin
  {This is my Pascal translation of the CW reduction given in}
  {sinl.c from the Cephes Math Library Release 2.7: May, 1998}
  {Copyright 1985, 1990, 1998 by Stephen L. Moshier          }
  if x=0.0 then begin
    z := 0.0;
    rem_pio2_cw_x := 0;
  end
  else begin
    y := floorx(x/Pi_4);
    i := trunc(y - 16.0*floorx(y/16.0));
    if odd(i) then begin
      inc(i);
      y := y + 1.0;
    end;
    rem_pio2_cw_x := (i shr 1) and 7;
    {Extended precision modular arithmetic}
    z := ((x - y * DP1) - y * DP2) - y * DP3;
  end;
end;





function k_rem_pio2({$ifdef CONST}const{$else}var {$endif} x: TDA02;
                    var y: TDA02; e0, nx, prec: integer): integer;
 {-Calculate y with y = x - n*Pi/2, |y| <= Pi/4 and return n mod 8}
label
  recompute;
var
  i,ih,j,jz,jx,jv,jp,jk,carry,k,m,n,q0: integer;
  t: longint; {WE: used for longint calculations, all other C-ints can be integers}
  iq: array[0..19] of longint;
  f,fq,q: array[0..19] of double;
  z,fw: double;
begin

  {initialize jk}
  jk := init_jk[prec];
  jp := jk;

  {determine jx,jv,q0, note that 3>q0}
  jx := nx-1;
  jv := (e0-3) div 24; if jv<0 then jv := 0;
  q0 := e0-24*(jv+1);

  {set up f[0] to f[jx+jk] where f[jx+jk] = ipio2[jv+jk]}
  j := jv-jx;
  m := jx+jk;
  for i:=0 to m do begin
    if j<0 then f[i] := 0.0 else f[i] := ipio2[j];
    inc(j);
  end;

  {compute q[0],q[1],...q[jk]}
  for i:=0 to jk do begin
    fw := 0.0;
    for j:=0 to jx do begin
      fw   := fw + x[j]*f[jx+i-j];
      q[i] := fw;
    end;
  end;
  jz := jk;

recompute:

  {distill q[] into iq[] reversingly}
  i := 0;
  j := jz;
  z := q[jz];
  while j>0 do begin
    fw    := trunc(twon24*z);
    iq[i] := trunc(z-two24*fw);
    z     := q[j-1]+fw;
    inc(i);
    dec(j);
  end;

  {compute n}
  z  := ldexp(z,q0);              {actual value of z}
  z  := z - 8.0*floorx(z*0.125);  {trim off integer >= 8}
  n  := trunc(z);
  z  := z - n;
  ih := 0;
  if q0>0 then begin
    {need iq[jz-1] to determine n}
    t  := (iq[jz-1] shr (24-q0));
    inc(n,t);
    dec(iq[jz-1], t shl (24-q0));
    ih := iq[jz-1] shr (23-q0);
  end
  else if q0=0 then ih := iq[jz-1] shr 23
  else if z>=0.5 then ih := 2;

  if ih>0 then begin
    {q > 0.5}
    inc(n);
    carry := 0;
    for i:=0 to jz-1 do begin
      {compute 1-q}
      t := iq[i];
      if carry=0 then begin
        if t<>0 then begin
          carry := 1;
          iq[i] := $1000000 - t;
        end
      end
      else iq[i] := $ffffff - t;
    end;
    if q0>0 then begin
      {rare case: chance is 1 in 12}
      case q0 of
        1: iq[jz-1] := iq[jz-1] and $7fffff;
        2: iq[jz-1] := iq[jz-1] and $3fffff;
      end;
    end;
    if ih=2 then begin
      z := 1.0 - z;
      if carry<>0 then  z := z - ldexp(1.0,q0);
    end;
  end;

  {check if recomputation is needed}
  if z=0.0 then begin
    t := 0;
    for i:=jz-1 downto jk do t := t or iq[i];
    if t=0 then begin
      {need recomputation}
      k := 1;
      while iq[jk-k]=0 do inc(k);   {k = no. of terms needed}
      for i:=jz+1 to jz+k do begin
        {add q[jz+1] to q[jz+k]}
        f[jx+i] := ipio2[jv+i];
        fw := 0.0;
        for j:=0 to jx do fw := fw + x[j]*f[jx+i-j];
        q[i] := fw;
      end;
      inc(jz,k);
      goto recompute;
    end;
  end;

  {chop off zero terms}
  if z=0.0 then begin
    dec(jz);
    dec(q0,24);
    while iq[jz]=0 do begin
      dec(jz);
      dec(q0,24);
    end;
  end
  else begin
    {break z into 24-bit if necessary}
    z := ldexp(z,-q0);
    if z>=two24 then begin
      fw := trunc(twon24*z);
      iq[jz] := trunc(z-two24*fw);
      inc(jz);
      inc(q0,24);
      iq[jz] := trunc(fw);
    end
    else iq[jz] := trunc(z);
  end;

  {convert integer "bit" chunk to floating-point value}
  fw := ldexp(1.0,q0);
  for i:=jz downto 0 do begin
    q[i] := fw*iq[i];
    fw := fw*twon24;
  end;

  {compute PIo2[0,...,jp]*q[jz,...,0]}
  for i:=jz downto 0 do begin
    fw :=0.0;
    k := 0;
    while (k<=jp) and (k<=jz-i) do begin
      fw := fw + double(PIo2[k])*(q[i+k]);
      fq[jz-i] := fw;
      inc(k);
    end;
  end;

  {compress fq[] into y[]}
  case prec of
     0:  begin
           fw := 0.0;
           for i:=jz downto 0 do fw := fw + fq[i];
           if ih=0 then y[0] := fw else y[0] := -fw;
         end;
     1,
     2:  begin
           fw := 0.0;
           for i:=jz downto 0 do fw := fw + fq[i];
           if ih=0 then y[0] := fw else y[0] := -fw;
           fw := fq[0]-fw;
           for i:=1 to jz do fw := fw + fq[i];
           if ih=0 then y[1] := fw else y[1] := -fw;
         end;
     3:  begin
           {painful}
           for i:=jz downto 1 do begin
             fw     := fq[i-1]+fq[i];
             fq[i]  := fq[i] + (fq[i-1]-fw);
             fq[i-1]:= fw;
           end;
           for i:=jz downto 2 do begin
             fw      := fq[i-1]+fq[i];
             fq[i]  := fq[i] + (fq[i-1]-fw);
             fq[i-1]:= fw;
           end;
           fw := 0.0;
           for i:=jz downto 2 do fw := fw + fq[i];
           if ih=0 then begin
             y[0] :=  fq[0];
             y[1] :=  fq[1];
             y[2] :=  fw;
           end
           else begin
             y[0] := -fq[0];
             y[1] := -fq[1];
             y[2] := -fw;
           end;
         end;
  end;
  k_rem_pio2 := n and 7;
end;




function rem_pio2_ph_x(x: extended; var z: extended): integer;
  {-Payne/Hanek reduction of x:  z = x - n*Pi/2, |z| <= Pi/4, result = n mod 8}
var
  ax, ay: TDA02;
  e0: integer;
begin
  z := abs(x);
  if (x=0.0) or IsNanOrInf(x) then rem_pio2_ph_x := 0
  else begin
    e0 := ilogb(z)-23;
    z  := ldexp(z,-e0);
    ax[0] := trunc(z); z := ldexp(z-ax[0],24);
    ax[1] := trunc(z);
    ax[2] := trunc(ldexp(z-ax[1],24));
    e0 := k_rem_pio2(ax,ay, e0, 3, 2);
    if x>=0 then begin
      z := ay[0]+ay[1];
      rem_pio2_ph_x := e0;
    end
    else begin
      z := -ay[0]-ay[1];
      rem_pio2_ph_x := (-e0) and 7;
    end;
  end;
end;

function rem_pio2_x(x: extended; var z: extended): integer;
  {-Argument reduction of x:  z = x - n*Pi/2, |z| <= Pi/4, result = n mod 8.}
  { Uses Payne/Hanek if |x| >= ph_cutoff, Cody/Waite otherwise}
const
  tol: extended = 5.9604644775390625e-8;  {ph_cutoff*eps_x}
begin
  z := abs(x);
  if z<Pi_4 then begin
    z := x;
    rem_pio2_x := 0;
  end
  else if z > ph_cutoff then rem_pio2_x := rem_pio2_ph_x(x,z)
  else begin
    rem_pio2_x := rem_pio2_cw_x(x,z);
    if abs(z) <= tol then begin
      {If x is close to a multiple of Pi/2, the C/W relative error may be}
      {large, e.g. 1.0389e-6 for x = 491299012647073665.0/16777216.0 with}
      {z_cw = 5.0672257144385922E-20 and z_ph = 5.0672204500144216E-20.  }
      {In this case redo the calculation with the Payne/Hanek algorithm. }
      rem_pio2_x := rem_pio2_ph_x(x,z);
    end;
  end;
end;

function PolEvalX(x: extended; const va:array of extended; n: integer): extended;
  {-Evaluate polynomial; return a[0] + a[1]*x + ... + a[n-1]*x^(n-1)}
var
  a: TExtVector absolute va;
var
  i: integer;
  s: extended;
begin
  if n<=0 then begin
    PolEvalX := 0.0;
    exit;
  end;
  s := a[n-1];
  for i:=n-2 downto 0 do s := s*x + a[i];
  PolEvalX := s;
end;






function sfc_gaminv_small(x: extended): extended;
  {- 1/gamma(x) for small argument values, |x| < 0.03125}
const
  {1/gamma(x) = x*P(x), 0 < x < 0.03125, peak relative error 4.2e-23}
  SPHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 1.000000000000000000000E+0}
           ($c7a9,$7db0,$67e3,$93c4,$3ffe),   { 5.772156649015328608253E-1}
           ($7bf6,$57d1,$a013,$a7e7,$bffe),   {-6.558780715202540684668E-1}
           ($f183,$126b,$f47d,$ac0a,$bffa),   {-4.200263503403344054473E-2}
           ($6b8d,$7515,$1905,$aa89,$3ffc),   { 1.665386113720805206758E-1}
           ($10b0,$ec17,$87dc,$acd7,$bffa),   {-4.219773360705915470089E-2}
           ($9225,$dfef,$b0e9,$9da5,$bff8),   {-9.622023360406271645744E-3}
           ($fe9a,$ceb4,$c74e,$ec9a,$3ff7),   { 7.220599478036909672331E-3}
           ($baeb,$d6d3,$25e5,$9c7e,$bff5));  {-1.193945051381510095614E-3}

  {1/gamma(-x) = x*P(x), 0 < x < 0.03125, peak relative error 5.16e-23}
  SNHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$bfff),   {-1.000000000000000000000E+0}
           ($c7aa,$7db0,$67e3,$93c4,$3ffe),   { 5.772156649015328608727E-1}
           ($5e26,$57d1,$a013,$a7e7,$3ffe),   { 6.558780715202536547116E-1}
           ($7f64,$1234,$f47d,$ac0a,$bffa),   {-4.200263503402112910504E-2}
           ($7a5b,$d76d,$1905,$aa89,$bffc),   {-1.665386113944413519335E-1}
           ($783f,$41dd,$87d1,$acd7,$bffa),   {-4.219773343731191721664E-2}
           ($2ca1,$18f0,$386f,$9da5,$3ff8),   { 9.621911155035976733706E-3}
           ($989b,$dd68,$c5f1,$ec9c,$3ff7),   { 7.220837261893170325704E-3}
           ($5dd1,$02de,$b9f7,$948d,$3ff5));  { 1.133374167243894382010E-3}
var
  SP: array[0..8] of extended absolute SPHex;
  SN: array[0..8] of extended absolute SNHex;
var
  p: extended;
begin
  {Ref: Cephes [7], function gammal/label small in file ldouble/gammal.c}
  if x=0.0 then sfc_gaminv_small := 0.0
  else begin
    if x<0.0 then begin
      x := -x;
      p := PolEvalX(x, SN, 9);
    end
    else p := PolEvalX(x, SP, 9);
    sfc_gaminv_small := x*p;
  end;
end;




{---------------------------------------------------------------------------}
function sfc_gamma_medium(x: extended): extended;
  {-Return gamma(x), |x| <= 13, x negative integer produces div by 0}
const
  {gamma(x+2) = P(x)/Q(x), 0 <= x <= 1, peak relative error = 1.83e-20}
  PHex: array[0..7] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 1.000000000000000000009E+0}
           ($29cf,$19b3,$16c8,$d67a,$3ffe),   { 8.378004301573126728826E-1}
           ($8d75,$23af,$c8e4,$b9d4,$3ffd),   { 3.629515436640239168939E-1}
           ($9549,$8eb5,$8c3a,$e3f4,$3ffb),   { 1.113062816019361559013E-1}
           ($7f43,$5196,$b166,$c368,$3ff9),   { 2.385363243461108252554E-2}
           ($be6c,$3757,$c717,$861b,$3ff7),   { 4.092666828394035500949E-3}
           ($f5aa,$e82f,$335b,$ee2e,$3ff3),   { 4.542931960608009155600E-4}
           ($434a,$3f22,$2bda,$b0b2,$3ff0));  { 4.212760487471622013093E-5}

  QHex: array[0..8] of THexExtW = (
           ($0000,$0000,$0000,$8000,$3fff),   { 9.999999999999999999908E-1}
           ($e458,$2ec7,$fd57,$d47c,$3ffd),   { 4.150160950588455434583E-1}
           ($75ef,$3ab7,$4ad3,$e5bc,$bffc),   {-2.243510905670329164562E-1}
           ($3295,$3698,$d580,$bdcd,$bffa),   {-4.633887671244534213831E-2}
           ($0417,$7989,$d7bc,$e338,$3ff9),   { 2.773706565840072979165E-2}
           ($296e,$7cb1,$5dfd,$d08f,$bff4),   {-7.955933682494738320586E-4}
           ($beed,$1853,$a691,$a23d,$bff5),   {-1.237799246653152231188E-3}
           ($334b,$c2f0,$a2dd,$f60e,$3ff2),   { 2.346584059160635244282E-4}
           ($5473,$2de8,$1268,$ea67,$bfee));  {-1.397148517476170440917E-5}
var
  PX: array[0..7] of extended absolute PHex;
  QX: array[0..8] of extended absolute QHex;
var
  z: extended;
begin
  {Ref: Cephes [7], function gammal in file ldouble/gammal.c}
  {-13 <= x <= 13. Use recurrence formula to bring argument to [2,3)}
  z := 1.0;
  while x >= 3.0 do begin
    x := x-1.0;
    z := z*x;
  end;
  while x < -0.03125 do begin
    z := z/x;
    x := x+1.0;
  end;
  {Here -0.03125 <= x < 3}
  if x <= 0.03125 then begin
    {argument near a pole, use approximation of 1/gamma}
     sfc_gamma_medium := z/sfc_gaminv_small(x);
  end
  else begin
    {finish reduction to [2,3)}
    while x < 2.0 do begin
      z := z/x;
      x := x+1.0;
    end;
    if x=2.0 then  sfc_gamma_medium := z
    else begin
      x := x-2.0;
       sfc_gamma_medium := z * PolEvalX(x,PX,8) / PolEvalX(x,QX,9);
    end;
  end;
end;





function stirf(x: extended): extended;
  {-Stirling's formula for the gamma function for x > 13.0}
  { gamma(x) = sqrt(2*Pi)*x^(x-0.5)*exp(-x)*(1 + 1/x * P(1/x))}
const
  {13 <= x <= 1024, relative peak error = 9.44E-21, relative error spread = 8.8e-4}
  SPHex: array[0..8] of THexExtW = (
           ($a1d5,$aaaa,$aaaa,$aaaa,$3ffb),  { 8.333333333333331800504E-2}
           ($c3c9,$906e,$38e3,$e38e,$3ff6),  { 3.472222222230075327854E-3}
           ($3a1c,$5ac8,$3478,$afb9,$bff6),  {-2.681327161876304418288E-3}
           ($bef3,$7023,$6a08,$f09e,$bff2),  {-2.294719747873185405699E-4}
           ($30b7,$1a21,$98b2,$cd87,$3ff4),  { 7.840334842744753003862E-4}
           ($5704,$1a39,$b11d,$9293,$3ff1),  { 6.989332260623193171870E-5}
           ($ba6f,$7c59,$5e47,$9bfb,$bff4),  {-5.950237554056330156018E-4}
           ($c395,$0295,$4443,$c64b,$bfef),  {-2.363848809501759061727E-5}
           ($6ede,$69f7,$54e3,$bb5d,$3ff4)); { 7.147391378143610789273E-4}
const
  {x > 1024, rational coefficients from the analytical expansion  }
  { exp(sum(i=1,20,x^(2*i-1)*bernfrac(2*i)/(2*i)/(2*i-1)))+O(x^7) }
  { = 1 + 1/12*x + 1/288*x^2 - 139/51840*x^3 - 571/2488320*x^4    }
  {     + 163879/209018880*x^5 + 5246819/75246796800*x^6 + O(x^7) }
  SAHex: array[0..5] of THexExtW = (
           ($aaab,$aaaa,$aaaa,$aaaa,$3ffb),  { 8.33333333333333333333E-2} {1/12}
           ($e38e,$8e38,$38e3,$e38e,$3ff6),  { 3.47222222222222222222E-3} {1/288}
           ($3df2,$d5a6,$3476,$afb9,$bff6),  {-2.68132716049382716049E-3} {-139/51840}
           ($cab1,$fd42,$7232,$f09e,$bff2),  {-2.29472093621399176955E-4} {-571/2488320}
           ($20e4,$a796,$fb43,$cd87,$3ff4),  { 7.84039221720066627474E-4} {163879/209018880}
           ($c3f2,$ce01,$0241,$923b,$3ff1)); { 6.97281375836585777429E-5} {5246819/75246796800}
const
  XMAX = 1500.0; {max. argument for direct power}
var
  SP: array[0..8] of extended absolute SPHex;
  SA: array[0..5] of extended absolute SAHex;
var
  v,w,y: extended;
begin
  {Ref: Cephes [7], function stirf in file ldouble/gammal.c}
  w := 1.0/x;
  if x > 1024.0 then w := 1.0 + w*PolEvalX(w, SA, 6)
  else w := 1.0 + w*PolEvalX(w, SP, 9);
  if x > XMAX then begin
    {Avoid overflow in power}
    y := _exp_x(-x);
    v := _power_x(x,0.5*x-0.25);
    y := v*(v*y);
  end
  else begin
    y := _exp_x(x);
    y := _power_x(x, x-0.5)/y;
  end;
  stirf := Sqrt_TwoPi*y*w;
end;





function rem_int2(x: extended; var z: extended): integer;
  {-Argument reduction of x: z*Pi = x*Pi - n*Pi/2, |z|<=1/4, result = n mod 8.}
  { Used for argument reduction in sin(Pi*x) and cos(Pi*x)}
var
  y: extended;
  i: integer;
begin
  if IsNanOrInf(x) or (abs(x)<=0.25) then begin
    rem_int2 := 0;
    z := x;
    exit;
  end;
  if frac(x)=0.0 then begin
    {Here x is an integer or abs(x) >= 2^64}
    z := 0;
    i := 0;
    {set i=2, if x is a odd}
    if (TExtRec(x).xp and $7FFF < $403F) and (frac(0.5*x)<>0.0) then i:=2;
  end
  else begin
    {Here x is not an integer. First calculate x mod 2,}
    {this leaves Pi*x = Pi*(x mod 2) mod 2*Pi invariant}
    x := 0.5*x;
    x := 2.0*(x-floorx(x));
    {then apply the Cody/Waite style range reduction}
    y := floorx(4.0*x);
    i := trunc(y - 16.0*floorx(y/16.0));
    if odd(i) then begin
      inc(i);
      y := y + 1.0;
    end;
    i := (i shr 1) and 7;
    z := x-0.25*y;
  end;
  rem_int2 := i;
end;



{---------------------------------------------------------------------------}
function ln1p(x: extended): extended;
  {-Return ln(1+x), accurate even for x near 0}
var
  y: extended;
const
  x0 = 4.0;
begin
  if x >= x0 then ln1p := ln(1.0 + x)
  else begin
    y := 1.0 + x;
    {The following formula is more accurate than Goldberg [3], Theorem 4. The}
    {Taylor series f(x) = f(x0) + (x-x0)*f'(x0) + .. for f(x) = ln(1+x) gives}
    {ln1p(x) = ln(1+x0) + (x-x0)/(1+x0) = ln(y) + (x-(y-1))/y, with y = 1+x0.}
    if y=1.0 then ln1p := x
    else ln1p := ln(y) + (x-(y-1.0))/y;
  end
end;


function lngamma_small(x,xm1,xm2: extended; useln1p: boolean): extended;
  {-Return ln(gamma(x), 0<=x<8; xm1=x-1, xm2=x-2 are supplied for increased precision}
  { useln1p should be true if xm1 is supposed to be more accurate than x-1}
var
  r,t,z: extended;
const
  Y20: THexExtW = ($0,$0,$6000,$A2C7,$3FFC); {0.158963680267333984375 = 333371/2097152}
  P20: array[0..6] of extended = (
         -0.180355685678449379109e-1,
          0.25126649619989678683e-1,
          0.494103151567532234274e-1,
          0.172491608709613993966e-1,
         -0.259453563205438108893e-3,
         -0.541009869215204396339e-3,
         -0.324588649825948492091e-4);
  Q20: array[0..7] of extended = (
          1.0,
          0.196202987197795200688e1,
          0.148019669424231326694e1,
          0.541391432071720958364e0,
          0.988504251128010129477e-1,
          0.82130967464889339326e-2,
          0.224936291922115757597e-3,
         -0.223352763208617092964e-6);
const
  Y10: THexExtW = ($0,$0,$1000,$8735,$3FFE);  {0.52815341949462890625 = 1107618/2097152}
  P10: array[0..6] of extended = (
          0.490622454069039543534e-1,
         -0.969117530159521214579e-1,
         -0.414983358359495381969e0,
         -0.406567124211938417342e0,
         -0.158413586390692192217e0,
         -0.240149820648571559892e-1,
         -0.100346687696279557415e-2);
  Q10: array[0..6] of extended = (
          1.0,
          0.302349829846463038743e1,
          0.348739585360723852576e1,
          0.191415588274426679201e1,
          0.507137738614363510846e0,
          0.577039722690451849648e-1,
          0.195768102601107189171e-2);
const
  Y15: THexExtW = ($0,$0,$D000,$E76E,$3FFD);  {0.452017307281494140625 = 947949/2097152}
  P15: array[0..5] of extended = (
         -0.292329721830270012337e-1,
          0.144216267757192309184e0,
         -0.142440390738631274135e0,
          0.542809694055053558157e-1,
         -0.850535976868336437746e-2,
          0.431171342679297331241e-3);
  Q15: array[0..6] of extended = (
          1.0,
         -0.150169356054485044494e1,
          0.846973248876495016101e0,
         -0.220095151814995745555e0,
          0.25582797155975869989e-1,
         -0.100666795539143372762e-2,
         -0.827193521891290553639e-6);
begin
  {Based on \boost\math\special_functions\detail\lgamma_small.hpp [19]}
  {Copyright John Maddock 2006, see 3rdparty.ama for Boost license}
  if x<eps_x then begin
    lngamma_small := -ln(x);
    exit;
  end;
  if (xm1=0.0) or (xm2=0.0) then begin
    lngamma_small := 0.0;
    exit;
  end;
  {The Boost rational approximations are optimized for low absolute error.}
  {As long as their absolute error is small compared to the constants Yxx }
  {then any rounding errors in their computation will get wiped out.      }
  if x>2.0 then begin
    {Use recurrence formula to bring argument to [2,3)}
    if x>=3 then begin
      z := 1.0;
      repeat
        x := x-1.0;
        z := z*x;
      until x<3.0;
      xm2 := x-2.0;
      z := ln(z);
    end
    else z := 0.0;
    {Use the following form: lngamma(x) = (x-2)*(x+1)*(Y20 + R(x-2))}
    r := PolEvalX(xm2, P20, 7) / PolEvalX(xm2, Q20, 8);
    t := xm2*(x+1.0);
    lngamma_small := z + (t*extended(Y20) + t*r);
  end
  else begin
    if x<1.0 then begin
      {bring argument into [1,2)}
      if useln1p then z := -ln1p(xm1)
      else z := -ln(x);
      xm2 := xm1;
      xm1 := x;
      x   := x+1.0;
    end
    else z := 0.0;
    {Two approximations, one for x in [1.0,1.5] and one for x in (1.5,2]}
    if x<=1.5 then begin
      {Use the following form: lngamma(x) = (x-1)*(x-2)*(Y10 + R(x-1))}
      r := PolEvalX(xm1, P10, 7) / PolEvalX(xm1, Q10, 7);
      t := xm1*xm2;
      lngamma_small := z + (t*extended(Y10) + t*r);
    end
    else begin
      {Use the following form: lngamma(x) = (2-x)*(1-x)*(Y15 + R(2-x))}
      r := PolEvalX(-xm2, P15, 6) / PolEvalX(-xm2, Q15, 7);
      t := xm1*xm2;
      lngamma_small := z + (t*extended(Y15) + t*r);
    end;
  end;
end;



{---------------------------------------------------------------------------}
function sinPi(x: extended): extended;
  {-Return sin(Pi*x), result will be 0 for abs(x) >= 2^64}
var
 t: extended;
 i: integer;
begin
  i := rem_int2(x,t) and 3;
  t := Pi*t;
  case i of
      0:  sinPi :=  system.sin(t);
      1:  sinPi :=  system.cos(t);
      2:  sinPi := -system.sin(t);
    else  sinPi := -system.cos(t);
  end;
end;


procedure _sincos(x: extended; var s,c: extended); assembler; {&Frame-} {&Uses none}
asm
  fld [x]
  fsincos
  fxch
  mov     eax, s
  fstp    tbyte ptr [eax]
  mov     eax, c
  fstp    tbyte ptr [eax]
  fwait
end;


procedure sincosPi(x: extended; var s,c: extended);
  {-Return s=sin(Pi*x), c=cos(Pi*x); (s,c)=(0,1) for abs(x) >= 2^64}
var
  t,ss,cc: extended;
  n: integer;
begin
  n := rem_int2(x,t) and 3;
  t := Pi*t;
  _sincos(t,ss,cc);
  case n of
      0: begin s:= ss; c:= cc; end;
      1: begin s:= cc; c:=-ss; end;
      2: begin s:=-ss; c:=-cc; end;
    else begin s:=-cc; c:= ss; end;
  end;
end;


function powm1(x,y: extended): extended;
  {-Return x^y - 1; special code for small x,y}
var
  p: extended;
begin
  if y=0.0 then begin
    powm1 := 0.0;
    exit;
  end;
  if (x>0.0) and ((x<2.0) or (abs(y)<2.0)) then begin
    p := y*ln(x);
    if abs(p) < 4.0 then begin
      powm1 := expm1(p);
      exit;
    end;
  end;
  powm1 := power(x,y)-1.0;
end;



function exp2(x: extended): extended; assembler; {&Frame-} {&Uses none}
  {-Return 2^x}
asm
  fld     [x]                { x                       }
  fld     st(0)              { x | x                   }
  frndint                    { int(x) | x              }
  fxch    st(1)              { x | int(x)              }
  fsub    st(0),st(1)        { frac(x) | int(x)        }
  f2xm1                      { 2^frac(x)-1 | int(x)    }
  fld1                       { 1 | 2^frac(x)-1 | int(x)}
  faddp   st(1),st           { 2^frac(x) | int(x)      }
  fscale                     { 2^z | int(x)            }
  fstp    st(1)
  fwait
end;


{---------------------------------------------------------------------------}
function exp2m1(x: extended): extended;
  {-Return 2^x-1, accurate even for x near 0}
begin
  if abs(x)<=1 then begin
    {$ifdef BASM}
      asm
        fld   [x]
        f2xm1
        {$ifdef NO_FPU_RESULT}
          fstp   [x]
        {$else}
          fstp   [@result]
        {$endif}
      end;
       {$ifdef NO_FPU_RESULT}
         exp2m1 := x;
       {$endif}
    {$else}
      exp2m1 := powm1(2.0,x);
    {$endif}
  end
  else exp2m1 := exp2(x)-1.0;
end;




{---------------------------------------------------------------------------}
function sfc_bernoulli(n: integer): extended;
  {-Return the nth Bernoulli number, 0 if n<0 or odd n >= 3}
var
  bn,p4: extended;
  m: integer;
const
  bx32: array[0..68] of THexExtw = (        {Bernoulli(32*n+128)}
          ($52FD,$BE3D,$B49C,$DA4E,$C178),  {-5.2500923086774133900E+113}
          ($FF69,$87B9,$BE6D,$ABE4,$C209),  {-1.8437723552033869727E+157}
          ($4263,$EB29,$A09A,$A2CC,$C2A3),  {-3.9876744968232207445E+203}
          ($F6AA,$C1D8,$F6B1,$FC3C,$C344),  {-1.8059559586909309014E+252}
          ($B09F,$C9C2,$93E6,$9464,$C3ED),  {-7.9502125045885252855E+302}
          ($9CD8,$18C3,$6CDC,$9557,$C49B),  {-1.9158673530031573512E+355}
          ($4195,$04C1,$073E,$A49C,$C54E),  {-1.6181135520838065925E+409}
          ($739D,$571F,$11CD,$8B25,$C606),  {-3.3538244754399336890E+464}
          ($6270,$777F,$2738,$86CE,$C6C2),  {-1.2747336393845383643E+521}
          ($7FAB,$7570,$F9E6,$EADB,$C781),  {-6.9702671752326436791E+578}
          ($4BFF,$0F88,$1349,$95D1,$C845),  {-4.4656129117005935721E+637}
          ($77B2,$1772,$C601,$EAC7,$C90B),  {-2.8113943110814931668E+697}
          ($215A,$6E86,$91AA,$C206,$C9D5),  {-1.4934073418970347178E+758}
          ($5C7F,$D540,$F299,$9400,$CAA2),  {-5.8578844571843963827E+819}
          ($9A59,$4576,$14FE,$B949,$CB71),  {-1.5084075750891085972E+882}
          ($F479,$B007,$0BB5,$AB72,$CC43),  {-2.2966905497257900647E+945}
          ($C675,$9DA9,$A628,$D590,$CD17),  {-1.8830664597658071289E+1009}
          ($0CCE,$2F96,$B80B,$A49C,$CDEE),  {-7.6426951845750635249E+1073}
          ($3EA8,$63C6,$E69D,$9180,$CEC7),  {-1.4228762140674037692E+1139}
          ($4F15,$2C2F,$CD6C,$899F,$CFA2),  {-1.1338551314597730180E+1205}
          ($3CE2,$FDE0,$5490,$82C2,$D07F),  {-3.6304766452190330209E+1271}
          ($3E25,$B6E8,$D002,$EB8A,$D15D),  {-4.4077763139644032336E+1338}
          ($C753,$F062,$8250,$BEAA,$D23E),  {-1.9238585974016076227E+1406}
          ($0C77,$16EC,$4BAC,$840D,$D321),  {-2.8737803110109338078E+1474}
          ($4BEB,$3C03,$6AAB,$9587,$D405),  {-1.4036962824375857856E+1543}
          ($55C7,$0248,$8353,$84AE,$D4EB),  {-2.1491058465822269708E+1612}
          ($E368,$D032,$65E9,$B163,$D5D2),  {-9.9151825072869898178E+1681}
          ($BDEF,$D7FE,$04EF,$AC37,$D6BB),  {-1.3287264083350159101E+1752}
          ($7A90,$3C2E,$9735,$EA9B,$D7A5),  {-4.9971936871087436659E+1822}
          ($9FC1,$AC12,$A7FA,$D91F,$D891),  {-5.1070475539922579356E+1893}
          ($C234,$DE06,$003E,$8470,$D97F),  {-1.3759818684504019193E+1965}
          ($D312,$510C,$861B,$CEFB,$DA6D),  {-9.4989226808690414704E+2036}
          ($B32D,$A0BE,$8870,$C9B7,$DB5D),  {-1.6356184195123832511E+2109}
          ($F8D5,$2E47,$4E40,$EF06,$DC4E),  {-6.8487492292184253538E+2181}
          ($4706,$AFF3,$6F23,$A81A,$DD41),  {-6.8082216303292659152E+2254}
          ($269B,$331E,$4461,$892E,$DE35),  {-1.5706145196821570477E+2328}
          ($B8F7,$3321,$712A,$FE3E,$DF29),  {-8.2289799425426414984E+2401}
          ($D598,$89BA,$1E5D,$830E,$E020),  {-9.5930887251893861974E+2475}
          ($5173,$0867,$AB5F,$9368,$E117),  {-2.4402644753692194590E+2550}
          ($E63F,$76FF,$2870,$B191,$E20F),  {-1.3295801865055646274E+2625}
          ($F314,$7CE5,$BF4B,$E10C,$E308),  {-1.5244028786126305655E+2700}
          ($C94F,$DECD,$6EC9,$9389,$E403),  {-3.6161842428643845094E+2775}
          ($5940,$BF69,$792C,$C4E9,$E4FE),  {-1.7464299021830195980E+2851}
          ($A1C0,$F220,$16A1,$83B6,$E5FB),  {-1.6907945786261035527E+2927}
          ($92B8,$8BC5,$F001,$AE03,$E6F8),  {-3.2332881196060927594E+3003}
          ($DBF1,$FA98,$71AD,$DFDC,$E7F6),  {-1.2040771662431166519E+3080}
          ($BEF9,$BB8F,$DE3B,$8A4F,$E8F6),  {-8.6141885195778356538E+3156}
          ($D7D6,$F08B,$2222,$A20A,$E9F6),  {-1.1685695524501785594E+3234}
          ($E709,$C538,$6958,$B1BD,$EAF7),  {-2.9684329116433388662E+3311}
          ($9EFA,$77E4,$442C,$B459,$EBF9),  {-1.3950642936150073193E+3389}
          ($9693,$9589,$9A1E,$A753,$ECFC),  {-1.1989884572796487307E+3467}
          ($1F33,$FD26,$C024,$8C5F,$EE00),  {-1.8635271302518619006E+3545}
          ($BFDB,$D0DD,$0E46,$D2AF,$EF04),  {-5.1817804064721174490E+3623}
          ($36A4,$A1D6,$BAE4,$8BF7,$F00A),  {-2.5511413022051873656E+3702}
          ($04E7,$CCC3,$B596,$A2FF,$F110),  {-2.2016594557163485556E+3781}
          ($0B04,$CD7E,$45F3,$A4C4,$F217),  {-3.2985559030415466710E+3860}
          ($E49A,$0504,$EDB4,$8F39,$F31F),  {-8.4995513646331021388E+3939}
          ($9F07,$D98B,$CC09,$D433,$F427),  {-3.7328650389340701819E+4019}
          ($1417,$52A2,$A454,$84CC,$F531),  {-2.7699188001914063216E+4099}
          ($3D97,$75F1,$0EF2,$8B3C,$F63B),  {-3.4434756013646314132E+4179}
          ($198B,$2EA0,$521E,$F293,$F745),  {-7.1133726432191288073E+4259}
          ($A708,$C185,$31CE,$AE2D,$F851),  {-2.4224631302940113182E+4340}
          ($955A,$427F,$7027,$CC98,$F95D),  {-1.3495915388686739922E+4421}
          ($E7A1,$2BA2,$3BF4,$C31E,$FA6A),  {-1.2208793515089649122E+4502}
          ($B45B,$AA6B,$972C,$95FC,$FB78),  {-1.7804374121649736373E+4583}
          ($CB82,$33DD,$DA5B,$B88F,$FC86),  {-4.1563772250412736023E+4664}
          ($90AA,$6E95,$658B,$B48A,$FD95),  {-1.5426825919798643530E+4746}
          ($C03F,$1D53,$D309,$8B77,$FEA5),  {-9.0434733129342411538E+4827}
          ($3D11,$CC1D,$E67F,$A912,$FFB5)); {-8.3194707606651575345E+4909}
begin
  if odd(n) or (n<0) then begin
    if n=1 then sfc_bernoulli := -0.5
    else sfc_bernoulli := 0.0;
  end
  else begin
    m := n div 2;
    if m<=MaxB2nSmall then sfc_bernoulli := extended(B2nHex[m])
    else if n>MaxBernoulli then sfc_bernoulli := PosInf_x
    else begin
      {When n is even, B(2n) = -2*(-1)^n*m!/(2Pi)^m*zeta(m) with m=2n. For }
      {large m (e.g. m>63) zeta(m) is very close to 1 and we can derive the}
      {asymptotic recursion formula B(m+1) = -m*(m+1)/(2Pi)^2 * B(m). The  }
      {avg. iteration count is <4, the max. rel. error=4.5*eps_x for n=878.}
      m  := (n - 112) div 32;
      bn := extended(bx32[m]);
      m  := 32*m + 128;
      p4 := 4.0*PiSqr;
      if n>m then begin
        while n>m do begin
          inc(m,2);
          bn := bn/p4*m*(1-m);
        end;
      end
      else begin
        while m>n do begin
          bn := bn/m/(1-m)*p4;
          dec(m,2);
        end;
      end;
      sfc_bernoulli := bn;
    end;
  end;
end;




function sfc_zetaint(n: integer): extended;
  {-Return zeta(n) for integer arguments, n<>1}
const
  znhex: array[2..63] of THexExtW = (
           ($9918,$983E,$3312,$D28D,$3FFF),  {+1.6449340668482264364}
           ($09C2,$8031,$0027,$99DD,$3FFF),  {+1.2020569031595942854}
           ($41B6,$3EC2,$9156,$8A89,$3FFF),  {+1.0823232337111381915}
           ($6DBD,$53E6,$0C76,$84BA,$3FFF),  {+1.0369277551433699263}
           ($247C,$0492,$4C26,$8238,$3FFF),  {+1.0173430619844491397}
           ($C46D,$A679,$96D0,$8111,$3FFF),  {+1.0083492773819228268}
           ($B746,$C31C,$9B57,$8085,$3FFF),  {+1.0040773561979443394}
           ($AB81,$C0B5,$CF9E,$8041,$3FFF),  {+1.0020083928260822144}
           ($CBF1,$D2DD,$9719,$8020,$3FFF),  {+1.0009945751278180854}
           ($9954,$F245,$318D,$8010,$3FFF),  {+1.0004941886041194645}
           ($9821,$D966,$1052,$8008,$3FFF),  {+1.0002460865533080483}
           ($DEF6,$E845,$0564,$8004,$3FFF),  {+1.0001227133475784892}
           ($9C2B,$5E56,$01C9,$8002,$3FFF),  {+1.0000612481350587048}
           ($11BF,$BCBF,$0097,$8001,$3FFF),  {+1.0000305882363070205}
           ($9178,$66F5,$8032,$8000,$3FFF),  {+1.0000152822594086518}
           ($A19A,$C1CD,$4010,$8000,$3FFF),  {+1.0000076371976378998}
           ($6E8A,$932A,$2005,$8000,$3FFF),  {+1.0000038172932649999}
           ($F9BA,$DB08,$1001,$8000,$3FFF),  {+1.0000019082127165539}
           ($A233,$9E2C,$0800,$8000,$3FFF),  {+1.0000009539620338727}
           ($ACA0,$34AE,$0400,$8000,$3FFF),  {+1.0000004769329867878}
           ($D9D9,$118C,$0200,$8000,$3FFF),  {+1.0000002384505027277}
           ($F138,$05D8,$0100,$8000,$3FFF),  {+1.0000001192199259653}
           ($CFFF,$01F2,$0080,$8000,$3FFF),  {+1.0000000596081890513}
           ($3A95,$00A6,$0040,$8000,$3FFF),  {+1.0000000298035035146}
           ($662E,$0037,$0020,$8000,$3FFF),  {+1.0000000149015548284}
           ($76B9,$0012,$0010,$8000,$3FFF),  {+1.0000000074507117898}
           ($2768,$0006,$0008,$8000,$3FFF),  {+1.0000000037253340248}
           ($0D18,$0002,$0004,$8000,$3FFF),  {+1.0000000018626597235}
           ($AF05,$0000,$0002,$8000,$3FFF),  {+1.0000000009313274324}
           ($3A56,$0000,$0001,$8000,$3FFF),  {+1.0000000004656629064}
           ($1372,$8000,$0000,$8000,$3FFF),  {+1.0000000002328311834}
           ($067B,$4000,$0000,$8000,$3FFF),  {+1.0000000001164155017}
           ($0229,$2000,$0000,$8000,$3FFF),  {+1.0000000000582077209}
           ($00B8,$1000,$0000,$8000,$3FFF),  {+1.0000000000291038504}
           ($003D,$0800,$0000,$8000,$3FFF),  {+1.0000000000145519218}
           ($0014,$0400,$0000,$8000,$3FFF),  {+1.0000000000072759598}
           ($0007,$0200,$0000,$8000,$3FFF),  {+1.0000000000036379796}
           ($0002,$0100,$0000,$8000,$3FFF),  {+1.0000000000018189896}
           ($0001,$0080,$0000,$8000,$3FFF),  {+1.0000000000009094948}
           ($0000,$0040,$0000,$8000,$3FFF),  {+1.0000000000004547474}
           ($0000,$0020,$0000,$8000,$3FFF),  {+1.0000000000002273737}
           ($0000,$0010,$0000,$8000,$3FFF),  {+1.0000000000001136868}
           ($0000,$0008,$0000,$8000,$3FFF),  {+1.0000000000000568434}
           ($0000,$0004,$0000,$8000,$3FFF),  {+1.0000000000000284217}
           ($0000,$0002,$0000,$8000,$3FFF),  {+1.0000000000000142108}
           ($0000,$0001,$0000,$8000,$3FFF),  {+1.0000000000000071054}
           ($8000,$0000,$0000,$8000,$3FFF),  {+1.0000000000000035527}
           ($4000,$0000,$0000,$8000,$3FFF),  {+1.0000000000000017764}
           ($2000,$0000,$0000,$8000,$3FFF),  {+1.0000000000000008882}
           ($1000,$0000,$0000,$8000,$3FFF),  {+1.0000000000000004441}
           ($0800,$0000,$0000,$8000,$3FFF),  {+1.0000000000000002220}
           ($0400,$0000,$0000,$8000,$3FFF),  {+1.0000000000000001110}
           ($0200,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000555}
           ($0100,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000278}
           ($0080,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000139}
           ($0040,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000069}
           ($0020,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000035}
           ($0010,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000017}
           ($0008,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000009}
           ($0004,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000004}
           ($0002,$0000,$0000,$8000,$3FFF),  {+1.0000000000000000002}
           ($0001,$0000,$0000,$8000,$3FFF)); {+1.0000000000000000001}
begin
  if n>63 then sfc_zetaint := 1.0
  else if n>1 then sfc_zetaint := extended(znhex[n])
  else if n<0 then sfc_zetaint := sfc_bernoulli(1-n)/(n-1)+0.0  {avoid -0}
  else if n=0 then sfc_zetaint := -0.5
  else sfc_zetaint := PosInf_x;
end;


{---------------------------------------------------------------------------}
function zetap(s,sc: extended): extended;
  {-Return the Riemann zeta function at s>0, s<>1, sc=1-s}
var
  y: extended;
{Based on boost_1_42_0\boost\math\special_functions\zeta.hpp [19]}
{Copyright John Maddock 2007, see 3rdparty.ama for Boost license}
const
  P1: array[0..5] of extended = (
        0.243392944335937499969,
       -0.496837806864865688082,
        0.0680008039723709987107,
       -0.00511620413006619942112,
        0.000455369899250053003335,
       -0.279496685273033761927e-4);
  Q1: array[0..6] of extended = (
        1.0,
       -0.30425480068225790522,
        0.050052748580371598736,
       -0.00519355671064700627862,
        0.000360623385771198350257,
       -0.159600883054550987633e-4,
        0.339770279812410586032e-6);
  P2: array[0..5] of extended = (
        0.577215664901532860605,
        0.222537368917162139445,
        0.0356286324033215682729,
        0.00304465292366350081446,
        0.000178102511649069421904,
        0.700867470265983665042e-5);
  Q2: array[0..6] of extended = (
        1.0,
        0.259385759149531030085,
        0.0373974962106091316854,
        0.00332735159183332820617,
        0.000188690420706998606469,
        0.635994377921861930071e-5,
        0.226583954978371199405e-7);
  P4: array[0..6] of extended = (
       -0.053725830002359501027,
        0.0470551187571475844778,
        0.0101339410415759517471,
        0.00100240326666092854528,
        0.685027119098122814867e-4,
        0.390972820219765942117e-5,
        0.540319769113543934483e-7);
  Q4: array[0..7] of extended = (
        1.0,
        0.286577739726542730421,
        0.0447355811517733225843,
        0.00430125107610252363302,
        0.000284956969089786662045,
        0.116188101609848411329e-4,
        0.278090318191657278204e-6,
       -0.19683620233222028478e-8);
  P7: array[0..7] of extended = (
       -2.49710190602259407065,
       -3.36664913245960625334,
       -1.77180020623777595452,
       -0.464717885249654313933,
       -0.0643694921293579472583,
       -0.00464265386202805715487,
       -0.000165556579779704340166,
       -0.252884970740994069582e-5);
  Q7: array[0..8] of extended = (
        1.0,
        1.01300131390690459085,
        0.387898115758643503827,
        0.0695071490045701135188,
        0.00586908595251442839291,
        0.000217752974064612188616,
        0.397626583349419011731e-5,
       -0.927884739284359700764e-8,
        0.119810501805618894381e-9);
  P15: array[0..8] of extended = (
        -4.78558028495135548083,
        -3.23873322238609358947,
        -0.892338582881021799922,
        -0.131326296217965913809,
        -0.0115651591773783712996,
        -0.000657728968362695775205,
        -0.252051328129449973047e-4,
        -0.626503445372641798925e-6,
        -0.815696314790853893484e-8);
  Q15: array[0..8] of extended = (
         1.0,
         0.525765665400123515036,
         0.10852641753657122787,
         0.0115669945375362045249,
         0.000732896513858274091966,
         0.30683952282420248448e-4,
         0.819649214609633126119e-6,
         0.117957556472335968146e-7,
        -0.193432300973017671137e-12);
  P42: array[0..8] of extended = (
        -10.3948950573308861781,
         -2.82646012777913950108,
         -0.342144362739570333665,
         -0.0249285145498722647472,
         -0.00122493108848097114118,
         -0.423055371192592850196e-4,
         -0.1025215577185967488e-5,
         -0.165096762663509467061e-7,
         -0.145392555873022044329e-9);
  Q42: array[0..9] of extended = (
         1.0,
         0.205135978585281988052,
         0.0192359357875879453602,
         0.00111496452029715514119,
         0.434928449016693986857e-4,
         0.116911068726610725891e-5,
         0.206704342290235237475e-7,
         0.209772836100827647474e-9,
        -0.939798249922234703384e-16,
         0.264584017421245080294e-18);
const
   x1: extended = 81487.0/65536.0;
   x4: extended = 366299.0/524288.0;
begin
  {$ifdef debug}
   {.248}
    {
    if abs((s+sc)-1.0)>5e-19*maxx(1.0,abs(s)) then begin
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      zetap := NaN_x;
      exit;
    end;
    }
  {$endif}
  if s<1.0 then begin
    y := PolEvalX(sc,P1,6)/PolEvalX(sc,Q1,7) - x1;
    zetap := (y+sc)/sc;
  end
  else if s<=2.0 then begin
    s := -sc;
    y := PolEvalX(s,P2,6)/PolEvalX(s,Q2,7);
    zetap := y - 1.0/sc;
  end
  else if s<=4.0 then begin
    s := s-2.0;
    y := PolEvalX(s,P4,7)/PolEvalX(s,Q4,8) + x4;
    zetap := y - 1.0/sc;
  end
  else if s<=7.0 then begin
    s := s-4.0;
    y := PolEvalX(s,P7,8)/PolEvalX(s,Q7,9);
    zetap := 1.0 + exp(y);
  end
  else if s<15.0 then begin
    s := s-7.0;
    y := PolEvalX(s,P15,9)/PolEvalX(s,Q15,9);
    zetap := 1.0 + exp(y);
  end
  else if s<42.0 then begin
    s := s-15.0;
    y := PolEvalX(s,P42,9)/PolEvalX(s,Q42,10);
    zetap := 1.0 + exp(y);
  end
  else if s<64.0 then zetap := 1.0 + exp2(-s)
  else zetap := 1.0;
end;


{---------------------------------------------------------------------------}
function exp3(x: extended): extended; assembler; {&Frame-} {&Uses none}
  {-Return 3^x}
const
  l32_hi: THexDblA = ($00,$00,$30,$98,$93,$30,$E4,$3F);
  l32_lo: THexDblA = ($58,$1F,$02,$A7,$F4,$D4,$C4,$3D);
  log23h: THexExtW = ($43D0,$FDEB,$0D1C,$CAE0,$3FFF);  {1.5849625007211561815}
asm
  fld     [x]                { x                                                  }
  fld     tbyte ptr [log23h] { log2(3) | x                                        }
  fmul    st,st(1)           { z = x * log2(3) | x                                }
  frndint                    { int(z) | x                                         }
  fld     qword ptr [l32_hi] { l32_hi | int(z) | x                                }
  fmul    st,st(1)           { int(z)*l32_hi | int(z) | x                         }
  fsubp   st(2),st           { int(z) | x-int(z)*l32_hi                           }
  fld     qword ptr [l32_lo] { l32_lo | int(z) | x-int(z)*l32_hi                  }
  fmul    st, st(1)          { int(z)*l32_lo | int(z) | x-int(z)*l32_hi           }
  fsubp   st(2),st           { int(z) | (x-int(z)*l32_hi)-int(z)*l32_lo           }
  fxch    st(1)              { (x-int(z)*l32_hi)-int(z)*l32_lo | int(z)           }
  fld     tbyte ptr [log23h] { log2(3) | (x-int(z)*l32_hi)-int(z)*l32_lo | int(z) }
  fmulp   st(1),st           { frac(z) | int(z)                                   }

{$ifndef use_fast_exp}
  {See the exp code for a description of these conditional lines}
  fld     st
  fabs                       { abs(frac(z)) | frac(z) | int(z)                   }
  fld1                       { 1 | abs(frac(z)) | frac(z) | int(z)               }
  fcompp
  fstsw   ax
  sahf
  jae     @@1                { frac(z) <= 1, no special action needed            }
  fld     st(1)              { int(z) | frac(z) | int(z)                         }
  fabs                       { abs(int(z)) | frac(z) | int(z)                    }
  fcomp   [ebig]
  fstsw   ax
  sahf
  jb      @@0
  fsub    st,st              { set frac=0 and scale with too large int(z)}
  jmp     @@1
@@0:
  fmul    dword ptr [half]   { frac(z)/2  | int(z)                               }
  f2xm1                      { 2^(frac(z)/2)-1 | int(z)                          }
  fld     st                 { 2^(frac(z)/2)-1 | 2^(frac(z)/2)-1 | int(z)        }
  fadd    dword ptr [two]    { 2^(frac(z)/2)+1 | 2^(frac(z)/2)-1 | int(z)        }
  fmulp   st(1),st           { 2^frac(z)-1 | int(z)                              }
  jmp     @@2
{$endif}

@@1:
  f2xm1                      { 2^frac(z)-1 | int(z)                              }

@@2:
  fld1                       { 1 | 2^frac(z)-1 | int(z)                          }
  faddp   st(1),st           { 2^frac(z) | int(z)                                }
  fscale                     { 2^z | int(z)                                      }
  fstp    st(1)              { 2^z = 3^x                                          }
  fwait
end;


{---------------------------------------------------------------------------}
function exp10(x: extended): extended; assembler; {&Frame-} {&Uses none}
  {-Return 10^x}
const
  lg2_hi: THexDblA = ($00,$00,$80,$50,$13,$44,$D3,$3F);
  lg2_lo: THexDblA = ($2B,$F1,$11,$F3,$FE,$79,$DF,$3D);
asm
  fld     [x]                { x                                                  }
  fldl2t                     { log2(10) | x                                       }
  fmul    st,st(1)           { z = x * log2(10) | x                               }
  frndint                    { int(z) | x                                         }
  fld     qword ptr [lg2_hi] { lg2_hi | int(z) | x                                }
  fmul    st,st(1)           { int(z)*lg2_hi | int(z) | x                         }
  fsubp   st(2),st           { int(z) | x-int(z)*lg2_hi                           }
  fld     qword ptr [lg2_lo] { lg2_lo | int(z) | x-int(z)*lg2_hi                  }
  fmul    st, st(1)          { int(z)*lg2_lo | int(z) | x-int(z)*lg2_hi           }
  fsubp   st(2),st           { int(z) | (x-int(z)*lg2_hi)-int(z)*lg2_lo           }
  fxch    st(1)              { (x-int(z)*lg2_hi)-int(z)*lg2_lo | int(z)           }
  fldl2t                     { log2(10) | (x-int(z)*lg2_hi)-int(z)*lg2_lo | int(z)}
  fmulp   st(1),st           { frac(z) | int(z)                                   }

{$ifndef use_fast_exp}
  {See the exp code for a description of these conditional lines}
  fld     st
  fabs                       { abs(frac(z)) | frac(z) | int(z)                   }
  fld1                       { 1 | abs(frac(z)) | frac(z) | int(z)               }
  fcompp
  fstsw   ax
  sahf
  jae     @@1                { frac(z) <= 1, no special action needed            }
  fld     st(1)              { int(z) | frac(z) | int(z)                         }
  fabs                       { abs(int(z)) | frac(z) | int(z)                    }
  fcomp   [ebig]
  fstsw   ax
  sahf
  jb      @@0
  fsub    st,st              { set frac=0 and scale with too large int(z)}
  jmp     @@1
@@0:
  fmul    dword ptr [half]   { frac(z)/2  | int(z)                               }
  f2xm1                      { 2^(frac(z)/2)-1 | int(z)                          }
  fld     st                 { 2^(frac(z)/2)-1 | 2^(frac(z)/2)-1 | int(z)        }
  fadd    dword ptr [two]    { 2^(frac(z)/2)+1 | 2^(frac(z)/2)-1 | int(z)        }
  fmulp   st(1),st           { 2^frac(z)-1 | int(z)                              }
  jmp     @@2
{$endif}

@@1:
  f2xm1                      { 2^frac(z)-1 | int(z)                              }

@@2:
  fld1                       { 1 | 2^frac(z)-1 | int(z)                          }
  faddp   st(1),st           { 2^frac(z) | int(z)                                }
  fscale                     { 2^z | int(z)                                      }
  fstp    st(1)              { 2^z = 10^x                                         }
  fwait
end;


function exp7(x: extended): extended; assembler; {&Frame-} {&Uses none}
  {-Return 7^x}
const
  l72_hi: THexDblA = ($00,$00,$00,$3A,$19,$CC,$D6,$3F);
  l72_lo: THexDblA = ($C1,$DE,$44,$9C,$36,$D5,$09,$3E);
  log27h: THexExtW = ($66CD,$A021,$B3FA,$B3AB,$4000);  {2.8073549220576041075}
asm
  fld     [x]                { x                                                  }
  fld     tbyte ptr [log27h] { log2(7) | x                                        }
  fmul    st,st(1)           { z = x * log2(7) | x                                }
  frndint                    { int(z) | x                                         }
  fld     qword ptr [l72_hi] { l72_hi | int(z) | x                                }
  fmul    st,st(1)           { int(z)*l72_hi | int(z) | x                         }
  fsubp   st(2),st           { int(z) | x-int(z)*l72_hi                           }
  fld     qword ptr [l72_lo] { l72_lo | int(z) | x-int(z)*l72_hi                  }
  fmul    st, st(1)          { int(z)*l72_lo | int(z) | x-int(z)*l72_hi           }
  fsubp   st(2),st           { int(z) | (x-int(z)*l72_hi)-int(z)*l72_lo           }
  fxch    st(1)              { (x-int(z)*l72_hi)-int(z)*l72_lo | int(z)           }
  fld     tbyte ptr [log27h] { log2(7) | (x-int(z)*l72_hi)-int(z)*l72_lo | int(z) }
  fmulp   st(1),st           { frac(z) | int(z)                                   }

{$ifndef use_fast_exp}
  {See the exp code for a description of these conditional lines}
  fld     st
  fabs                       { abs(frac(z)) | frac(z) | int(z)                   }
  fld1                       { 1 | abs(frac(z)) | frac(z) | int(z)               }
  fcompp
  fstsw   ax
  sahf
  jae     @@1                { frac(z) <= 1, no special action needed            }
  fld     st(1)              { int(z) | frac(z) | int(z)                         }
  fabs                       { abs(int(z)) | frac(z) | int(z)                    }
  fcomp   [ebig]
  fstsw   ax
  sahf
  jb      @@0
  fsub    st,st              { set frac=0 and scale with too large int(z)}
  jmp     @@1
@@0:
  fmul    dword ptr [half]   { frac(z)/2  | int(z)                               }
  f2xm1                      { 2^(frac(z)/2)-1 | int(z)                          }
  fld     st                 { 2^(frac(z)/2)-1 | 2^(frac(z)/2)-1 | int(z)        }
  fadd    dword ptr [two]    { 2^(frac(z)/2)+1 | 2^(frac(z)/2)-1 | int(z)        }
  fmulp   st(1),st           { 2^frac(z)-1 | int(z)                              }
  jmp     @@2
{$endif}

@@1:
  f2xm1                      { 2^frac(z)-1 | int(z)                              }

@@2:
  fld1                       { 1 | 2^frac(z)-1 | int(z)                          }
  faddp   st(1),st           { 2^frac(z) | int(z)                                }
  fscale                     { 2^z | int(z)                                      }
  fstp    st(1)              { 2^z = 7^x                                          }
  fwait
end;




function etam1pos(s: extended): extended;
  {-Return the eta(s)-1 = -sum((-1)^k/k^s, k=2..) for s >= -ETAEPS; internal use}

const
  dh: array[2..24] of THexExtW = (
        ($FFA5,$FFFF,$FFFF,$FFFF,$BFFE),  {-0.99999999999999999506430323408864009302707409760650   }
        ($875D,$FFFE,$FFFF,$FFFF,$3FFE),  {+0.99999999999999477309712489986985851567146936528101   }
        ($D5A9,$FEFB,$FFFF,$FFFF,$BFFE),  {-0.99999999999907570687373807736430090108503647599525   }
        ($3E9B,$B83F,$FFFF,$FFFF,$3FFE),  {+0.99999999993474107123666050197526788003473422599175   }
        ($787A,$74E2,$FFF5,$FFFF,$BFFE),  {-0.99999999754516889043092198752546995530922208300482   }
        ($9FF0,$7505,$FF0B,$FFFF,$3FFE),  {+0.99999994306292316806008385807007727156754522290285   }
        ($2930,$A5AA,$F115,$FFFF,$BFFE),  {-0.99999911097044304457819242638771628351284408679994   }
        ($6065,$D3A9,$59D2,$FFFF,$3FFE),  {+0.99999009511126280553264350727993678656849990880578   }
        ($1A12,$43A3,$9FBC,$FFFA,$BFFE),  {-0.99991796823782089316825215441770081101374648485252   }
        ($0A33,$3791,$E44C,$FFDD,$3FFE),  {+0.99947954998748770036508902917665860666132371180327   }
        ($E514,$245E,$659A,$FF56,$BFFE),  {-0.99741206181749538061964608067153326403095105573943   }
        ($5B36,$DD12,$214D,$FD5F,$3FFE),  {+0.98973282004323819299371512908106770568956690464520   }
        ($544C,$0A46,$35FD,$F78F,$BFFE),  {-0.96702897479760824696922362003099562015851811010573   }
        ($E565,$E3A5,$FCA7,$E9C9,$3FFE),  {+0.91323832606180806715735142935851714059264865842761   }
        ($983E,$D706,$464E,$CF8F,$BFFE),  {-0.81077994751742677227759487569665336999099255999308   }
        ($6535,$017E,$D18E,$A766,$3FFE),  {+0.65391263691844299666858829009021366603535356790712   }
        ($6615,$FFBF,$F03A,$EC59,$BFFD),  {-0.46162367553904352979303183031457789989618319051142   }
        ($ED27,$D227,$22A5,$8C88,$3FFD),  {+0.27447613023930715026708917962385218097998528309956   }
        ($838F,$6B07,$85EE,$870D,$BFFC),  {-0.13188752429665086110446620766901353799621544888100   }
        ($9514,$E77C,$BD13,$C796,$3FFA),  {+0.48727739891972228733661913925793276796577280986824e-1}
        ($5449,$F08D,$C274,$D36F,$BFF8),  {-0.12905063533033740943161602774867625818271608663177e-1}
        ($A6BB,$3D4D,$0397,$8E43,$3FF6),  {+0.21707423941183752637782342766808453857479577229903e-2}
        ($1D19,$2582,$5684,$B618,$BFF2)); {-0.17365939152947002110225874213446763085983661783922e-3}
const
  ln2m1h: THexExtW = ($0CA8,$5C61,$D010,$9D1B,$BFFD);  {-3.0685281944005469057E-1}
  lnpi2h: THexExtW = ($AE23,$5098,$D92D,$E735,$3FFC);  {0.2257913526447274323630976} {ln(pi/2)/2}
var
  d: array[2..24] of extended absolute dh;
  p: array[2..24] of extended;
  x,sum: extended;
  k: integer;

begin

  if s=1.0 then begin
    etam1pos := extended(ln2m1h); {ln(2)-1}
    exit;
  end
  else if abs(s) <= ETAEPS then begin
    etam1pos := extended(lnpi2h)*s - 0.5;
    exit;
  end;

  x := -s;

  {Calculate p[k] := 1/k^s but only if necessary. Prime powers are}
  {evaluated with power/exp?, otherwise products of p[k] are used.}
  p[2] := exp2(x);
  if s >= 120.0 then begin
    etam1pos := -p[2];
    exit;
  end;

  p[3] := exp3(x);
  p[4] := p[2]*p[2];
  if s >= 50.0 then begin
    etam1pos := (p[3] - p[4]) - p[2];
    exit;
  end;

  p[10] := exp10(x);
  p[ 5] := p[10]/p[2];
  p[ 6] := p[2]*p[3];
  p[ 7] := exp7(x);
  p[ 8] := p[2]*p[4];
  p[ 9] := p[3]*p[3];
  if s >= 28.0 then begin
    sum := 0.0;
    for k:=10 downto 3 do begin
      if odd(k) then sum := sum + p[k]
      else sum := sum - p[k];
    end;
    etam1pos := sum - p[2];
    exit;
  end;

  p[11] := power(11,x);
  p[12] := p[2]*p[6];
  p[13] := power(13,x);
  p[14] := p[2]*p[7];
  p[15] := p[3]*p[5];
  p[16] := p[2]*p[8];
  p[17] := power(17,x);
  p[18] := p[2]*p[9];
  p[19] := power(19,x);
  p[20] := p[2]*p[10];
  p[21] := p[3]*p[7];
  p[22] := p[2]*p[11];
  if s>=19.5 then begin
    sum := 0.0;
    for k:=22 downto 3 do begin
      if odd(k) then sum := sum + p[k]
      else sum := sum - p[k];
    end;
    etam1pos := sum - p[2];
  end
  else begin
    p[23] := power(23,x);
    p[24] := p[3]*p[8];
    {Convergence acceleration: see P. Borwein[36]}

    {The d[] are from P. Borwein's Algorithm 2, but scaled and shifted.}
    {Calculated with Maple VR4 and T_RCalc/xh using n:=23; Digits:=50; }

    { d(n,k) = n*sum((n+j-1)!*4^j/((n-j)!*(2*j)!),j=0..k); }
    { d[2]   := 1/d(n,n)-1.0;                              }
    { d[i]   := (-1)^(i-2)*(d(n,i-2)/d(n,n)-1), i=3..n+1;  }
    sum := p[24]*d[24];
    for k:=23 downto 3 do sum := sum + d[k]*p[k];
    etam1pos := sum + d[2]*p[2];
  end;
end;



function sfc_zeta(s: extended): extended;
  {-Return the Riemann zeta function at s, s<>1}
var
  sc: extended;
  a,b: extended;
  ig: integer;
begin
  {Ref: Boost [19], file zeta.hpp}
  if IsNanOrInf(s) then begin
    if s=PosInf_x then sfc_zeta := 1.0
    else sfc_zeta := NaN_x;
    exit;
  end;
  if frac(s)=0.0 then begin
    if (s > -MaxBernoulli) and  (s <= MaxInt) then begin
      sfc_zeta := sfc_zetaint(round(s));
      exit;
    end;
  end
  else if abs(s) <= 1.6e-10 then begin
    {Small s, especially useful for s<0 to avoid reflection machinery}
    sfc_zeta := (-0.5) - LnSqrt2Pi*s;
    exit;
  end;
  sc := 1.0-s;
  if s<0.0 then begin
    if frac(0.5*s)=0.0 then sfc_zeta := 0.0
    else begin
      {compute Gamma(sc)/(2Pi)^sc}
      if sc>MaxGAMX-1 then begin
        {Not very accurate but does not overflow!  For s=-1753.5 this branch}
        {gives a rel. error of 2.0e-16 vs 1.8e-17 for the non-lngamma branch}
        a := sfc_lngammas(sc,ig);
        b := 2.0*sc*LnSqrt2Pi;
        {here Zeta(sc)=1}
        a := a-b;
        if a<Ln_MaxExt then a := exp(a)
        else a := PosInf_d;
        a := ig*a;
      end
      else begin
        a := sfc_gamma(sc);
        b := power(TwoPi,-sc);
        a := a*b;
        b := zetap(sc,s);
        a := a*b;
      end;
      {here a = zeta(sc)*Gamma(sc)/(2Pi)^sc}
      b := sinPi(0.5*s);
      sfc_zeta := 2.0*a*b;
    end;
  end
  else sfc_zeta := zetap(s,sc);
end;

function sfc_zetam1(s: extended): extended;
  {-Return Riemann zeta(s)-1, s<>1}
var
  t: extended;
begin
  if s <= 2.0 then sfc_zetam1 := sfc_zeta(s) - 1.0
  else begin
    if s >= 120.0 then sfc_zetam1 := exp2(-s)
    else begin
      t := 0.5*exp2(s);
      {$ifdef VER50}
        s := etam1pos(s);
        sfc_zetam1 := (1.0 + s*t) / (t - 1.0);
      {$else}
        sfc_zetam1 := (1.0 + etam1pos(s)*t) / (t - 1.0);
      {$endif}
    end;
  end;
end;


function bernpoly_intern(n: integer; x: extended): extended;
  {-Return the Bernoulli polynomial B_n(x), internal use: for small n only}
var
  i: integer;
  s,f: extended;
  ref: boolean;
begin
  if n<0 then bernpoly_intern := 0
  else if n=0 then bernpoly_intern := 1.0
  else if n=1 then bernpoly_intern := x - 0.5
  else begin
    ref := false;
    if abs(x-1.0) <= 0.125 then begin
      {if x is near 1 use NIST[30], 24.4.3: B_n(1-x) = (-1)^n B_n(x)}
      x := 1.0 - x;
      ref := odd(n);
    end;
    {Compute sum NIST[30], 24.2.5}
    f := n;
    s := x - 0.5*n;
    for i:=2 to n do begin
      f := (n+1-i)*f/i;
      s := s*x;
      if i and 1 = 0 then s := s + sfc_bernoulli(i)*f;
    end;
    if ref then s := -s;
    bernpoly_intern := s;
  end;
end;


function hz_a1(s,a: extended; var ok: boolean): extended;
  {-Return zetah(s,a) for a close to 1, s<>1 (normally s < 0)}
var
  x,t,f,z: extended;
  n: integer;
const
  MaxIter = 100;
begin
  hz_a1 := 0.0;
  ok := false;
  {compute sum from NIST[30], 25.11.10}
  z := sfc_zeta(s);
  if IsNanOrInf(z) then exit;
  x := 1.0-a;
  if x<>0.0 then begin
    n := 0;
    f := 1;
    repeat
      f := f*(s+n);
      inc(n);
      t := s+n;
      if (n>MaxIter) or (t=1.0) then exit;
      f := f*x/n;
      t := sfc_zeta(t);
      t := t*f;
      z := z + t;
      {do not use t for exit test because zeta(s+n) may be very small}
    until (abs(f)<=eps_x*abs(z)) and (n>2);
  end;
  ok := true;
  hz_a1 := z;
end;



{---------------------------------------------------------------------------}
function hurwitz_formula(s,a: extended): extended;
  {-Compute zetah with range reduction and Hurwitz formula, a > 0, s < 0}
var
  n,k: longint;
  a2,f,h,r,t,tol,sh,z: extended;
  done: boolean;
const
  MaxIter = 10000;
begin
  r := 0.0;
  z := -s;
  f := frac(a);
  k := trunc(a);
  if f=0.0 then begin
    {Use strict Hurwitz formula with reduced a in (0,1].}
    {Empirically it is valid also for reduced frac(a)=0.}
    f := 1.0;
    k := k-1;
  end;
  a2  := 2.0*f;
  sh  := 2.0*frac(0.25*s);  {sh = s/2 mod 2}
  tol := 0.5*eps_x;
  {NIST [30],25.11.4}
  for n:=0 to k-1 do begin
    t := power(f+n,z);
    r := r + t;
  end;
  h := 0.0;
  {Perform summation only for non-trivial arguments}
  if (frac(a2)<>0.0) or (frac(sh)<>0.0) then begin
    {A. Erdelyi [50], Higher Transcendental Functions I, 1.10 (6),}
    {see also NIST [30], 25.11.9  with s <-> 1-s and cos <-> sin. }
    n := 1;
    z := s-1.0;
    done := false;
    repeat
      t := n*a2 + sh;
      if frac(t)<>0.0 then begin
        t := sinpi(t);
        f := power(n, z);
        t := t*f;
        h := h + t;
        done := (f < tol*(abs(h) + 1e-10));  {1e-10 addition to avoid exit if h ~ 0}
      end;
      inc(n);
    until done or (n>MaxIter);
    if (n>MaxIter) and (RTE_NoConvergence>0) then RunError(byte(RTE_NoConvergence));
  end;
  if h=0.0 then hurwitz_formula := -r
  else begin
    {compute = 2*h*Gamma(1-s)/(2Pi)^(1-s)}
    z := 1.0 - s;
    if (z>0.0) and (z<MAXGAMX) then begin
      f := sfc_gamma(z);
      t := power(TwoPi, -z);
      t := (t*f)*(2.0*h);
    end
    else begin
      f := sfc_lngamma(z);
      t := ln(abs(2.0*h)) + f - z*ln(TwoPi);
      if t>= ln_MaxExt then t := PosInf_x
      else t := exp(t);
      if h<0.0 then t := -t;
    end;
    hurwitz_formula := t - r;
  end
end;


function sfc_zetah(s,a: extended): extended;
  {-Return the Hurwitz zeta function zetah(s,a) = sum(1/(i+a)^s, i=0..INF), s<>1, a>0}
var
  eps,q,r,t,u,w,z: extended;
  j,k,n: integer;
  ok: boolean;
const
  MaxIter = 40;
  eps1 = 1e-4;   {Threshold for hz_a1}
  S_HF = -8;     {Threshold for Hurzwitz formula}
label
  noconv;
begin
  {Initial reference: Cephes [7], file double\zeta.c}
  if IsNanOrInf(s) or IsNanOrInf(a) or (s=1.0) or (a<=0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_zetah := NaN_x;
    exit;
  end
  else if s=0.0 then begin
    sfc_zetah := 0.5 - a;       {NIST [30], 25.11.13}
    exit;
  end
  else if a=0.5 then begin
    if s < 16384.0 then begin
      sfc_zetah := exp2m1(s)*sfc_zeta(s);   {NIST [30], 25.11.11}
      exit;
    end;
  end
  else if a=1.0 then begin
    sfc_zetah := sfc_zeta(s);   {NIST [30], 25.11.2}
    exit;
  end
  else if a=2.0 then begin
    sfc_zetah := sfc_zetam1(s); {NIST [30], 25.11.2/3}
    exit;
  end;

  if s>0.0 then begin
    {Quick checks: avoid overflow/underflow}
    t := ln(a);
    t := -s*t;
    if t>ln_MaxExt then begin
      sfc_zetah := PosInf_x;
      exit;
    end
    else if t<ln_MinExt then begin
      {here s > 1, 1/a^s ~ 0, avoid underflow in some cases:}
      {zetah ~ a^(1-s)/(s-1) + 0.5/a^s = (a/(s-1) + 0.5)/a^s}
      {      = a^(1-s)/(s-1) + 0.5*a^(1-s)/a}
      u := power(a, 1-s);
      sfc_zetah := u/(s-1.0) + 0.5*u/a;
      exit;
    end;
  end
  else begin
    if (a>1.0) then begin
      if -s*ln(a-1.0) >ln_MaxExt then begin
        sfc_zetah := PosInf_x;
        exit;
      end;
    end;
    {Test if Bernoulli polynomials or Hurwitz formula can be used}
    if (s >= S_HF) and (frac(s)=0.0) then begin
      {s small negative integer: use 'naive' Bernoulli polynomial code}
      k := 1 - round(s);
      sfc_zetah := -bernpoly_intern(k,a)/k;  {NIST [30], 25.11.14}
      exit;
    end;
    if abs(a-1.0)<eps1 then begin
      {try expansion near a=1}
      sfc_zetah := hz_a1(s,a,ok);
      if ok then exit;
    end;
    if s < S_HF then begin
      sfc_zetah := hurwitz_formula(s,a);
      exit;
    end
  end;

  {Calculate z = zetah with the Euler-Maclaurin summation formula:}
  {See e.g. http://functions.wolfram.com/10.02.06.0020.01         }
  {z = sum1(s,a,n) + (a+n)^(1-s)/(s-1) - 1/2/(a+n)^s + sum2(s,a,n)}
  {with sum1(a,n) = sum(1/(a+k)^s, k=0..n) and                    }
  {sum2 = sum(B_2k/(2k)! * s(s+1)..(s+2k)/(a+n)^(s+2k+1),k=0..INF)}

  eps := 0.5*eps_x;
  if abs(s)<0.5 then n := 8
  else n := 9;

  {$ifdef Delphi}
    {Avoid brain-damaged paranoid warning: r,w,t might not have been initialized}
    r := 0; w := 0; t := 0;
  {$endif}

  {compute z = s1(a,n) = sum(1/(a+k)^s, k=0..n)}
  z := 0.0;
  for k:=0 to n do begin
    w := a+k;
    r := power(w,-s);
    z := z+r;
    if r < eps*z then begin
      sfc_zetah := z;
      exit;
    end
  end;

  {Add the two single terms z = z + (a+n)^(1-s)/(s-1) - 1/2/(a+n)^s}
  {Here w=(a+n), r=1/(a+n)^s}
  z := z + r*w/(s-1.0);
  z := z - 0.5*r;

  {Add the terms of sum2 until a term is < e*(partial result)}
  q := 1.0;
  k := 0;
  for j:=0 to MaxIter do begin
    q := q*(s+k);
    r := r/w;
    if j>NBoF then begin
      {Some problematic (s,a) values: Check decreasing terms in}
      {asymptotic expansion and increase convergence tolerance.}
      if j=NBOF+1 then eps := 1.5*eps_x;
      u := t;
      t := sfc_bernoulli(k+2);  {TP5 fix}
      t := t/factorial[k+2];
      t := (q*t)*r;
      if abs(u)<=abs(t) then goto noconv;
    end
    else t := (q*extended(BoFHex[j]))*r;
    z := z + t;
    if abs(t) < eps*abs(z) then begin
      sfc_zetah := z;
      exit;
    end;
    q := q*(s+k+1);
    r := r/w;
    inc(k,2);
  end;

noconv:
  {No convergence}
  sfc_zetah := z;
  {$ifdef debug}
   {.248}
   // writeln('sfc_zetah - Euler-Maclaurin issue: ',s:21,' ',a:21);
  {$endif}
  if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));
end;



{---------------------------------------------------------------------------}
function sfc_lngcorr(x: extended): extended;
  {-Return lngamma correction term lngamma(x) - ((x-0.5)*ln(x) - x + ln(sqrt(2*Pi)), x>=8}
const
  { ln gamma(x) = (x-0.5)*ln(x) - x + ln(sqrt(2*Pi) + 1/x * A(1/x^2) }
  { x >= 8, peak relative error 1.51e-21}
  AHex: array[0..6] of THexExtW = (
          ($9fcc,$aaaa,$aaaa,$aaaa,$3ffb),   { 8.333333333333331447505E-2}
          ($4d88,$03a8,$60b6,$b60b,$bff6),   {-2.777777777750349603440E-3}
          ($f8f2,$30e5,$0092,$d00d,$3ff4),   { 7.936507795855070755671E-4}
          ($8b20,$9fce,$844e,$9c09,$bff4),   {-5.952345851765688514613E-4}
          ($3bdc,$aad1,$d492,$dc88,$3ff4),   { 8.412723297322498080632E-4}
          ($3d91,$0304,$3da1,$f685,$bff5),   {-1.880801938119376907179E-3}
          ($d984,$cc08,$91c2,$a012,$3ff7));  { 4.885026142432270781165E-3}
var
  PA: array[0..6] of extended absolute AHex;
const
  xbig = 4.294967296e9;
  xmax = 2.47860728199e4930;
begin
  {Ref: Cephes [7], part function lgaml in file ldouble/gammal.c}
  if x < XLGAE then begin
    {this is not really used but return 'correct' result}
    sfc_lngcorr := sfc_lngamma(x) - ((x-0.5)*ln(x) - x + LnSqrt2Pi);
  end
  else if x > xmax then sfc_lngcorr := 0.0
  else if x > xbig then sfc_lngcorr := 1.0/(12.0*x)
  else sfc_lngcorr := PolEvalX(1.0/(x*x),PA,7)/x;
end;






function sfc_lngamma(x: extended): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, NAN/RTE if x is a non-positive integer}
var
  t: extended;
begin
  if x<0.0 then begin
    {avoid infinite loop if x is (negative) NAN}
    if IsNaN(x) then sfc_lngamma := NaN_x
    else if frac(x)=0.0 then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_lngamma := NaN_x;
    end
    else begin
      {use reflection formula}
      t := x*sinPi(x);
      sfc_lngamma := ln(abs(Pi/t)) - sfc_lngamma(abs(x));
    end;
    exit;
  end;
  if x <= XLGAE then begin
    {Note: will generate error for x=0}
    sfc_lngamma := lngamma_small(x,x-1.0,x-2.0,false)
  end
  else begin
    {Stirling approximation for x > 8}
    sfc_lngamma := (x-0.5)*ln(x) - x + LnSqrt2Pi + sfc_lngcorr(x);
  end;
end;





{---------------------------------------------------------------------------}



function sfc_lngammas(x: extended; var s: integer): extended;
  {-Return ln(|gamma(x)|), |x| <= MAXLGM, s=-1,1 is the sign of gamma}
begin
  sfc_lngammas := sfc_lngamma(x);
  if (x>0.0) or (frac(0.5*floorx(x))=0.0) then s := 1 else s := -1;
end;


function sfc_signgamma(x: extended): extended;
  {-Return sign(gamma(x)), useless for 0 or negative integer}
begin
  if IsNanOrInf(x) or (x>0.0) or (frac(0.5*floorx(x))=0.0) then sfc_signgamma := 1.0
  else sfc_signgamma := -1.0;
end;




{---------------------------------------------------------------------------}
function sfc_gamma(x: extended):extended; stdcall;
  {-Return gamma(x), x <= MAXGAMX; NAN/RTE if x is a non-positive integer}
var
  i: integer;
  q,z: extended;
begin

  {Ref: Cephes [7], function gammal in file ldouble/gammal.c}
  if x=PosInf_x then begin
    sfc_gamma := x;
    exit;
  end;

  if (x<=0.0) and (frac(x)=0.0) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma := NaN_x;
    exit;
  end;

  if x>MAXGAMX then begin
    sfc_gamma := PosInf_x;
    exit;
  end;

  q := abs(x);
  if q <= 13.0 then sfc_gamma := sfc_gamma_medium(x)
  else begin
    if x < 0.0 then begin
      if q <= MAXGAMX then begin
        i := trunc(q);
        z := q-i;
        if z > 0.5 then z := q - (i+1);
        z := abs(q * sinPi(z)) * stirf(q);
        if z <= Pi/MaxExtended then begin
          sfc_gamma := copysign(PosInf_x, sfc_signgamma(x));
          exit;
        end;
        sfc_gamma := sfc_signgamma(x)*Pi/z;
      end
      else begin
        {Use lngamma, there are only very rare 'non-zero' cases, most values}
        {underflow. gamma(-1756-0.5^24) = -2.75052000244830115684312e-4930}
        z := _exp_x(sfc_lngammas(x,i));
        sfc_gamma := z*i;
      end;
    end
    else sfc_gamma := stirf(x);
  end
end;



{---------------------------------------------------------------------------}
function sfc_psi(x: extended): extended;       stdcall;
  {-Return the psi (digamma) function of x, INF if x is a non-positive integer}
const
  BNNHex: array[0..8] of THexExtW = (      {bernreal(2k+2)/(2k+2), k=0..7}
            ($AAAB, $AAAA, $AAAA, $AAAA, $3FFB),   { 8.33333333333333333333333E-2}
            ($8889, $8888, $8888, $8888, $BFF8),   {-8.33333333333333333333333E-3}
            ($8208, $0820, $2082, $8208, $3FF7),   { 3.96825396825396825396825E-3}
            ($8889, $8888, $8888, $8888, $BFF7),   {-4.16666666666666666666667E-3}
            ($3E10, $E0F8, $0F83, $F83E, $3FF7),   { 7.57575757575757575757576E-3}
            ($ACCB, $CACC, $CCAC, $ACCA, $BFF9),   {-2.10927960927960927960928E-2}
            ($AAAB, $AAAA, $AAAA, $AAAA, $3FFB),   { 8.33333333333333333333333E-2}
            ($F2F3, $F2F2, $F2F2, $E2F2, $BFFD),   {-4.43259803921568627450980E-1}
            ($3FCE, $FF37, $FCDC, $C373, $4000));  { 3.05395433027011974380395   }
var
  BNN: array[0..7] of extended absolute BNNHex;

const
  PPHex: array[0..5] of THexExtW = (
           ($5D8F, $71C5, $5E47, $F7B9, $3FFE),  {+0.967672245447621170427444761712   }
           ($93CE, $3DA8, $4992, $9274, $3FFF),  {+1.14417380943934132177443086511    }
           ($6D52, $5F65, $79EF, $F169, $3FFD),  {+0.471507845373433246720039154386   }
           ($9EF2, $DF82, $1865, $A71D, $3FFB),  {+0.815984636391829369377116423465e-1}
           ($D564, $B95D, $BF51, $B443, $3FF7),  {+0.550124017486102827456755050064e-2}
           ($05E4, $B416, $530F, $C9AD, $3FF1)); {+0.961671107604462646458076964138e-4}
  PQHex: array[0..6] of THexExtW = (
           ($0000, $0000, $0000, $8000, $3FFF),  {+1.00000000000000000000000000000    }
           ($A4B3, $CD52, $FAA6, $D1E9, $3FFF),  {+1.63995297569876643247016291140    }
           ($2D1C, $1958, $35D7, $F872, $3FFE),  {+0.970492711080937113643931469693   }
           ($E38A, $0D1E, $6FB5, $84F8, $3FFD),  {+0.259707918978674869994308187160   }
           ($A4B6, $ADD5, $3BC7, $81BF, $3FFA),  {+0.316765151172736832756607905517e-1}
           ($C902, $DFDF, $6A6A, $C67A, $3FF5),  {+0.151426838914381207406214536130e-2}
           ($B01C, $4DC7, $7496, $8E9D, $3FEF)); {+0.170010400090619970545999542634e-4}
const
  x0hHex: THexExtW = ($0000,$0000,$8000,$BB16,$3FFF);  {1.4616241455078125}
  x0lHex: THexExtW = ($A9C9,$F6E1,$6BE3,$8635,$3FEE);  {7.9994605498412626595423257213e-6}
var
  PP : array[0..5] of extended absolute PPHex;
  PQ : array[0..6] of extended absolute PQHex;
  x0h: extended absolute x0hHex;
  x0l: extended absolute x0lHex;
var
  p,q,nz,s,w,y,z: extended;
  i,n: integer;
  subneg: boolean;

begin

  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_psi := PosInf_x
    else sfc_psi := NaN_x;
    exit;
  end;

  {General algorithm layout is as in Cephes [7] double/psi.c. Coefficients }
  {of the asymptotic expansion are calculated with Pari/GP's bernreal, the }
  {Pad approximation polynomials with Maple V's "pade(Psi(x+x0),x,[6,6])".}
  {Conversions to THexExtW form are done with MPArith/t_rcalc's xh command.}

  subneg := false;
  nz := 0.0;

  if x<=0.0 then begin
    q := x;
    p := floorx(q);
    if p=q then begin
      sfc_psi := PosInf_x;
      exit;
    end;
    {Remove the zeros of tan(Pi*x) by subtracting the nearest integer from x}
    nz := q - p;
    if nz<>0.5 then begin
      if nz>0.5 then begin
        p  := p + 1.0;
        nz := q - p;
      end;
      {nz := Pi/tan(Pi*nz)}
      sincosPi(nz,s,z);
      nz := Pi*z/s;
    end
    else nz := 0.0;
    x := 1.0 - x;
    subneg := nz<>0.0;
  end;

  {Check for small positive integers}
  if (x <= 12.0) and (x=floorx(x)) then begin
    y := 0.0;
    n := trunc(x);
    for i:=n-1 downto 1 do y := y+1.0/i;
    y := y - EulerGamma;
  end
  else if abs(x-x0h)<0.2 then begin
    {Pad approximation for x in (x0-0.2, x0+0.2), abs. error < 3e-19}
    {where x0=1.46163214496836234126.., psi(x0)=0}
    x := x-x0h;
    x := x-x0l;
    y := x*(PolEvalX(x,PP,6)/PolEvalX(x,PQ,7));
  end
  else begin
    s := x;
    w := 0.0;
    while s<12.0 do begin
      w := w + 1.0/s;
      s := s + 1.0;
    end;
    {asymptotic expansion}
    if s>1e10 then y := 0.0
    else begin
      z := 1.0/(s*s);
      y := z*PolEvalX(z,BNN,8);
    end;
    y := ln(s) - (0.5/s) - y - w;
  end;
  if subneg then y := y-nz;
  sfc_psi := y;
end;


{---------------------------------------------------------------------------}
function sfc_trigamma(x: extended): extended;   stdcall;
  {-Return the trigamma function psi'(x), INF if x is a negative integer}
var
  s: extended;
begin
  {trigamma = psi(1,x) = zetah(2,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_trigamma := 0.0
    else sfc_trigamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then sfc_trigamma := 1.0/x
    else sfc_trigamma := sfc_zetah(2.0,x)
  end
  else if frac(x)=0.0 then sfc_trigamma := PosInf_x
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=1}
    {psi(1,1-x) + psi(1,x) = Pi^2*(1+cot(Pi*x)^2) = Pi^2/sin(Pi*x)^2}
    s := sinPi(x);
    sfc_trigamma := PiSqr/sqr(s) - sfc_zetah(2.0,1.0-x);
  end;
end;


{---------------------------------------------------------------------------}
function sfc_tetragamma(x: extended): extended;   stdcall;
  {-Return the tetragamma function psi''(x), NAN/RTE if x is a negative integer}
var
  c,s: extended;
const
  PP3H: THexExtW = ($E0F0,$C4EB,$DAC9,$F80C,$4004);  {2*Pi^3}
begin
  {tetragamma = psi(2,x) = -2*zetah(3,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_tetragamma := 0.0
    else sfc_tetragamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then begin
      x := 1.0/x;
      sfc_tetragamma := -x*x;
    end
    else sfc_tetragamma := -2.0*sfc_zetah(3.0,x)
  end
  else if frac(x)=0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_tetragamma := NaN_x;
  end
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=2, 2nd derivative of cot(Pi*x) with Maple}
    {psi(2,1-x) - psi(2,x) = 2*Pi^3*cot(Pi*x)*(1+cot(Pi*x)^2)}
    sincosPi(x,s,c);
    c := c/s;
    c := c*(1.0 + c*c);
    s := -2.0*sfc_zetah(3.0,1.0-x);
    sfc_tetragamma := s - extended(PP3H)*c;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_pentagamma(x: extended): extended;   stdcall;
  {-Return the pentagamma function psi'''(x), INF if x is a negative integer}
var
  c,s: extended;
const
  PP4H: THexExtW = ($2C68,$4841,$7461,$C2D1,$4006);  {2*Pi^4}
begin
  {pentagamma = psi(3,x) = 6*zetah(4,x) for x >0}
  if IsNanOrInf(x) then begin
    if x=PosInf_x then sfc_pentagamma := 0.0
    else sfc_pentagamma := NaN_x;
  end
  else if x>0.0 then begin
    if x >= 1e20 then begin
      x := 1.0/x;
      sfc_pentagamma := 2.0*x*x*x;
    end
    else sfc_pentagamma := 6.0*sfc_zetah(4.0,x);
  end
  else if frac(x)=0.0 then sfc_pentagamma := PosInf_x
  else begin
    {Reflection formula HMF [1] 6.4.7 with n=3, 3rd derivative of cot(Pi*x) with Maple}
    {psi(3,1-x) + psi(3,x) = 2*Pi^4*(1+4*cot(Pi*x)^2 + 3*cot(Pi*x)^4)}
    sincosPi(x,s,c);
    c := sqr(c/s);
    c := 1.0 + c*(4.0 + 3.0*c);
    s := 6.0*sfc_zetah(4.0,1.0-x);
    sfc_pentagamma := extended(PP4H)*c - s;
  end;
end;




{*******************************BESSEL***********************************************}
 { (C) Copyright   Wolfgang Ehrhardt   http://wolfgang-ehrhardt.de/ }


procedure h1v_large(v, x: extended; var mv,tmx: extended);
  {-Return modulus and (phase - x) of the Hankel function H1_v(x), x > 0 large}
var
  s,m,m2,y: extended;
const
  c56 = 5.0/6.0;
begin
  {Modulus Mv: asymptotic expansion from HMF[1] 9.2.28}
  y  := sqr(0.5/x);
  m  := 4.0*sqr(v);
  m2 := sqr(m);
  s  := 1.0 + 0.5*y*(m-1.0)*(1.0 + 0.75*y*(m-9.0)*(1.0 + c56*y*(m-25.0)));
  mv := sqrt(2.0*s/(Pi*x));
  {Phase theta_v - x: asymptotic expansion from HMF[1] 9.2.29}
  y  := 0.25*y;
  s  := (5.0*m*m2 - 1535.0*m2 + 54703.0*m - 37573.0)/14.0;
  s  := s*y + (m2 - 114.0*m + 1073)/5.0;
  s  := s*y + (m-25.0)/6.0;
  tmx:= (m-1.0)*(s*y + 0.5)/(4.0*x) - Pi*(0.5*v+0.25)
end;




procedure bess_m0p0(x: extended; var m0,p0: extended);
  {-Modulus and phase for J0(x) and Y0(x), x >= 9.0}
var
  y, z: extended;
const
  m0nhex: array[0..7] of THexExtW = (
             ($cb2b,$4b73,$8075,$8aa3,$3ff8),
             ($6b78,$4cc6,$25b7,$b912,$3ffb),
             ($cfe9,$74e0,$67a1,$b75e,$3ffe),
             ($b1cd,$4e5e,$2274,$b5ec,$4000),
             ($5e4b,$e3af,$59bb,$f409,$4001),
             ($b343,$2673,$4e51,$a36b,$4002),
             ($38a3,$a663,$7b91,$dbab,$4001),
             ($8559,$f552,$3a38,$ca1d,$3ffd));
  m0dhex: array[0..7] of THexExtW = (
             ($8a83,$1b80,$003e,$adc2,$3ff8),
             ($ed5a,$31cd,$b3ac,$e7f3,$3ffb),
             ($7e3f,$b8dd,$04df,$e5fd,$3ffe),
             ($775a,$1b79,$7d9c,$e475,$4000),
             ($b837,$3075,$dbc0,$99ce,$4002),
             ($5e3d,$b5f4,$9848,$d032,$4002),
             ($4498,$3d2a,$f3fb,$91df,$4002),
             ($0000,$0000,$0000,$8000,$3fff));
  p0nhex: array[0..5] of THexExtW = (
             ($4c2f,$2dd8,$79c3,$e65d,$bfe9),
             ($dc17,$325e,$8baf,$9d35,$bff1),
             ($e514,$8866,$25a9,$8309,$bff7),
             ($8d8a,$84e7,$dbd5,$9e75,$bffb),
             ($1e30,$04da,$b769,$800a,$bffe),
             ($5106,$12a6,$4dd2,$bc55,$bffe));
  p0dhex: array[0..6] of THexExtW = (
             ($4c8c,$2dd8,$79c3,$e65d,$3fec),
             ($ac5c,$4806,$8709,$9dad,$3ff4),
             ($37bf,$fcc8,$9b9f,$844b,$3ffa),
             ($6f25,$2a95,$2dc6,$a285,$3ffe),
             ($4b69,$3f87,$131f,$891f,$4001),
             ($f3e9,$b2a5,$6652,$ec17,$4001),
             ($0000,$0000,$0000,$8000,$3fff));
var
  m0n: array[0..7] of extended absolute m0nhex;
  m0d: array[0..7] of extended absolute m0dhex;
  p0n: array[0..5] of extended absolute p0nhex;
  p0d: array[0..6] of extended absolute p0dhex;
begin
  {Ref: Cephes [7], file ldouble\j0l.c}
  {See also HMF[1], sections 9.2.17 .. 9.2.31}

  {Calculate the modulus m0(x) = sqrt(J0(x)^2 + Y0(x)^2) and the}
  {phase p0(x) = arctan(Y0(x)/J0(x)) with rational approximations}
  {For x>=9: J0(x) = m0(x)*cos(p0(x)) and Y0(x) = m0(x)*sin(p0(x))}
  z  := sqr(1.0/x);
  y  := abs(x);
  p0 := y - Pi_4 + PolEvalX(z,p0n,6)/PolEvalX(z,p0d,7)/y;
  z  := 1.0/y;
  m0 := PolEvalX(z,m0n,8)/PolEvalX(z,m0d,8)/sqrt(y);
end;


function bessj_large(v, x: extended): extended;
  {-Return J_v(x) via modulus/phase asymptotic expansion, x large}
var
  mv,tv,st,ct,sx,cx: extended;
begin
  h1v_large(v,x,mv,tv);
  sincos_x(tv,st,ct);
  sincos_x(x,sx,cx);
  {J_v := mv*cos(x+tv); cos(x+tv) = cos(x)cos(tv) - sin(x)sin(tv)}
  bessj_large := mv*(cx*ct - sx*st);
end;



procedure bess_m1p1(x: extended; var m1,p1: extended);
  {-Modulus and phase for J1(x) and Y1(x), x >= 9.0}
var
  y, z: extended;
const
  m1nhex: array[0..7] of THexExtW = (
            ($a905,$05fb,$3101,$82c9,$3ff9),
            ($6de4,$8fae,$fe26,$8097,$3ffc),
            ($2cb0,$c657,$be70,$81e0,$3fff),
            ($71e6,$88a5,$0a53,$b702,$4000),
            ($e5e2,$6914,$3a08,$e582,$4001),
            ($7d55,$db8c,$e825,$a1c2,$4000),
            ($3111,$863a,$3a61,$c8a0,$3ffd),
            ($3d53,$b598,$f3bf,$a155,$c001));
  m1dhex: array[0..8] of THexExtW = (
            ($1237,$cc6c,$7356,$a3ea,$3ff9),
            ($fc82,$02c7,$17a4,$a12b,$3ffc),
            ($37ce,$79ae,$2f15,$a24c,$3fff),
            ($77b6,$34e2,$501a,$e37a,$4000),
            ($0260,$746b,$d030,$8c14,$4002),
            ($6420,$97ce,$8e44,$a208,$4000),
            ($77b5,$8f2d,$b6bf,$ebe1,$bffe),
            ($2603,$640e,$7d8d,$c775,$c001),
            ($0000,$0000,$0000,$8000,$3fff));
  p1nhex: array[0..5] of THexExtW = (
            ($540c,$c1d5,$b096,$e54f,$3feb),
            ($f74f,$be87,$7e7d,$9741,$3ff3),
            ($a830,$f4a3,$2c60,$f144,$3ff8),
            ($e907,$28b9,$7cb7,$895c,$3ffd),
            ($6050,$98aa,$3500,$cb2f,$3fff),
            ($ebc0,$5506,$512f,$80ab,$4000));
  p1dhex: array[0..6] of THexExtW = (
            ($8d72,$2be3,$cb0f,$98df,$3fed),
            ($a853,$55fb,$6c79,$ca32,$3ff4),
            ($98f8,$d610,$3c35,$a235,$3ffa),
            ($c39e,$9c8c,$5428,$bb65,$3ffe),
            ($b1f2,$e0d2,$5ab5,$9098,$4001),
            ($efe3,$292c,$0d43,$d9e6,$4001),
            ($0000,$0000,$0000,$8000,$3fff));
var
  m1n: array[0..7] of extended absolute m1nhex;
  m1d: array[0..8] of extended absolute m1dhex;
  p1n: array[0..5] of extended absolute p1nhex;
  p1d: array[0..6] of extended absolute p1dhex;
begin
  {Ref: Cephes [7], file ldouble\j1l.c}

  {Calculate the modulus m1(x) = sign(x)*sqrt(J1(x)^2 + Y1(x)^2) and }
  {the phase p1(x) = arctan(Y1(x)/J1(x)) with rational approximations}
  {For x>=9: J1(x) = m1(x)*cos(p1(x)) and Y1(x) = m1(x)*sin(p1(x))}
  z  := sqr(1.0/x);
  y  := abs(x);
  p1 := y - 3.0*Pi_4 + PolEvalX(z,p1n,6)/PolEvalX(z,p1d,7)/y;
  z  := 1.0/y;
  m1 := PolEvalX(z,m1n,8)/PolEvalX(z,m1d,9)/sqrt(y);
  if x<0.0 then m1 := -m1;
end;



function CF_jn(n: integer; x: extended): extended;
  {-Return J_(n+1)(x) / J_n(x), use only if |x| <= |n|}
var
  b,c,d,f,k,delta,tiny,tol,limit: extended;
begin
  {Evaluate continued fraction HMF[1], 9.1.73 using modified Lentz's method.}
  {Based on boost_1_42_0\boost\math\special_functions\detail\bessel_jy.hpp [19]}
  {Copyright 2006 Xiaogang Zhang, see 3rdparty.ama for Boost license}
  tol  := eps_x;
  tiny := Sqrt_MinExt;
  n := abs(n);
  c := tiny;
  f := tiny;
  d := 0.0;
  x := 0.5*x;
  k := n+1;
  limit := 3.0*(k + 20.0*sqrt(k));
  while k<=limit do begin
    b := k/x;
    c := b - 1.0/c;
    d := b - d;
    if c=0.0 then c := tiny;
    if d=0.0 then d := tiny;
    d := 1.0/d;
    delta := c*d;
    f := f*delta;
    if abs(delta - 1.0) < tol then begin
      CF_jn := -f;
      exit;
    end;
    k := k+1.0;
  end;
  CF_jn := 0.0;
end;


function sfc_jn(n: integer; x: extended): extended;
  {-Return J_n(x), the Bessel function of the 1st kind, order n; not suitable for large n or x.}
var
  curr,prev,q,temp,init,xh: extended;
  k: integer;
  neg: boolean;
const
  small = 7.143435E-1651;  {~ cbrt(succx(0))}
  lnsml:extended = ln_succx0;
begin


  {Based on boost_1_42_0\boost\math\special_functions\detail\bessel_jn.hpp [19]}
  {Copyright 2006 Xiaogang Zhang, see 3rdparty.ama for Boost license}
  init := Sqrt_MinExt;
  {Flag to negate result for |n|}
  neg := (n<0) and odd(n);
  n := abs(n);

  if n=0 then curr := sfc_j0(x)
  else if n=1 then curr := sfc_j1(x)
  else if abs(x) <= small then begin
    if (x=0.0) or (n>2) then curr := 0.0
    else curr := 0.125*sqr(x);
  end
  else begin
    xh := 0.5*x;
    if abs(x) > n then begin
      {forward recurrence}
      prev := sfc_j0(x);
      curr := sfc_j1(x);
      for k:=1 to n-1 do begin
        temp := curr*k/xh - prev;
        prev := curr;
        curr := temp;
      end;
    end
    else begin
      {Quick check if |J_n(x)| < MinExtended from HMF[1] 9.1.63}
      {solution of z*exp(sqrt(1-z^2))=1 is z = 0.39989.. }
      q := abs(x/n);
      if n<=100 then temp := 1e-40
      else if n<1000 then temp := 1e-4
      else if n<5000 then temp := 0.07
      else temp := 0.3999;
      if q < temp then begin
        {Jn(x) <= [q*exp(sqrt(1-q^2))/(1+sqrt(1-q^2))]^n}
        temp := sqrt(1.0 - q*q);
        temp := ln(q/(1.0+temp)) + temp;
        if temp < lnsml/n then begin
          sfc_jn := 0.0;
          exit;
        end;
      end;
      {set overflow threshold for iteration}
      q := 0.5*MaxExtended*q;
      {backward recurrence}
      prev := CF_jn(n,x)*init;
      curr := init;
      for k:=n downto 1 do begin
        if abs(curr) > q then begin
          {prevent overflow and set result to zero}
          sfc_jn := 0.0;
          exit;
        end;
        temp := curr*k/xh - prev;
        prev := curr;
        curr := temp;
      end;
      curr := (init/curr)*sfc_j0(x);
    end;
  end;
  if neg then sfc_jn := -curr else sfc_jn := curr;
end;




function sfc_j0(x: extended): extended;
  {-Return J0(x), the Bessel function of the 1st kind, order zero}
var
  y,z: extended;
const
  {Squares of first three roots of J0, calculated with Maple and t_rcalc/xh}
  j1h: THexExtW = ($DBB7,$315D,$DC02,$B90F,$4001);  {5.7831859629467845213}
  j2h: THexExtW = ($7778,$0EEB,$2531,$F3C5,$4003);  {3.0471262343662086400E+1}
  j3h: THexExtW = ($40F6,$0ABB,$25C1,$95C6,$4005);  {7.4887006790695183442E+1}
var
  jz1: extended absolute j1h;
  jz2: extended absolute j2h;
  jz3: extended absolute j3h;
const
  j0nhex:  array[0..7] of THexExtW = (
             ($b96c,$c486,$fb95,$9f47,$c03a),
             ($4018,$ad26,$71ba,$e643,$4034),
             ($1b0b,$6331,$7add,$8753,$c02e),
             ($943a,$69b7,$36ca,$a996,$4026),
             ($008c,$7b60,$d119,$f792,$c01d),
             ($fe10,$b608,$4829,$d503,$4014),
             ($a9a8,$e62b,$3b28,$ca73,$c00a),
             ($f759,$4208,$23d6,$a5ff,$3fff));
  j0dhex: array[0..8] of THexExtW = (
             ($00ac,$fb2b,$6f62,$804b,$4048),
             ($fdce,$a4ca,$2ed8,$88b8,$4041),
             ($3d2c,$ed55,$20e1,$9105,$4039),
             ($0841,$8cb6,$5a46,$c9e3,$4030),
             ($fed1,$086d,$3425,$cc0a,$4027),
             ($66d2,$93fe,$0762,$9b79,$401e),
             ($e1a0,$923f,$cb5c,$b1a2,$4014),
             ($bdfe,$c832,$5b9f,$8e9f,$400a),
             ($0000,$0000,$0000,$8000,$3fff));
var
  j0n: array[0..7] of extended absolute j0nhex;
  j0d: array[0..8] of extended absolute j0dhex;
begin
  {Ref: Cephes [7], file ldouble\j0l.c}
  x := abs(x);
  if x < 9.0 then begin
    {In the interval [0,9) a rational approximation of the form }
    {J0(x) = (x^2 - r^2) (x^2 - s^2) (x^2 - t^2) P7(x^2)/Q8(x^2)}
    {is used, where r, s, t are the first three zeros of J0.}
    z := sqr(x);
    y := (z - jz1)*(z - jz2)*(z - jz3);
    sfc_j0 := y*PolEvalX(z,j0n,8)/PolEvalX(z,j0d,9);
  end
  else begin
    {For x>=9 the common rational approximations to modulus}
    {and phase are used J0(x) = modulus * cos(phase).}
    if x >= 500.0 then sfc_j0 := bessj_large(0,x)
    else begin
      bess_m0p0(x,y,z);
      sfc_j0 := y*cos_x(z);
    end;
  end;
end;



{---------------------------------------------------------------------------}
function sfc_j1(x: extended): extended;
  {-Return J1(x), the Bessel function of the 1st kind, order one}
var
  y,z: extended;
const
  {Squares of first three roots of J1, calculated with Maple and t_rcalc/xh}
  j1h: THexExtW = ($5F8E,$4C11,$5A0C,$EAE9,$4002);  {1.4681970642123893257E+1}
  j2h: THexExtW = ($9093,$9521,$B303,$C4DF,$4004);  {4.9218456321694603672E+1}
  j3h: THexExtW = ($5EBF,$C2F1,$B86B,$CEFF,$4005);  {1.0349945389513658033E+2}
var
  jz1: extended absolute j1h;
  jz2: extended absolute j2h;
  jz3: extended absolute j3h;
const
  j1nhex:  array[0..8] of THexExtW = (
             ($d8d8,$7311,$a7d2,$97a4,$c039),
             ($d3c2,$f8f0,$f852,$c144,$4033),
             ($636c,$4d29,$9f71,$cebb,$c02c),
             ($038e,$bd23,$a7fa,$f49c,$4024),
             ($1ac8,$c825,$3c9c,$b0b6,$c01c),
             ($38f5,$f72b,$0a5c,$a122,$4013),
             ($29f3,$496b,$a54c,$b6d9,$c009),
             ($6dc3,$c850,$a096,$ee6b,$3ffe),
             ($f72f,$18cc,$50b2,$8a22,$bff3));
  j1dhex: array[0..8] of THexExtW = (
             ($dd67,$f5b3,$0522,$ad0f,$404a),
             ($665d,$b178,$242e,$9af7,$4043),
             ($e6c0,$a725,$3d56,$88f7,$403b),
             ($0122,$56c0,$f2ef,$9d6e,$4032),
             ($b498,$fdd5,$209e,$820e,$4029),
             ($6041,$c9fe,$6890,$a033,$401f),
             ($6a17,$e162,$4e86,$9218,$4015),
             ($baf9,$146e,$df50,$b88a,$400a),
             ($0000,$0000,$0000,$8000,$3fff));
var
  j1n: array[0..8] of extended absolute j1nhex;
  j1d: array[0..8] of extended absolute j1dhex;
begin
  {Ref: Cephes [7], file ldouble\j1l.c}
  z := abs(x);
  if z < 9.0 then begin
    z := sqr(x);
    {In the interval [0,9) a rational approximation of the form }
    {J1(x) = x*(x^2 - r^2)*(x^2 - s^2)*(x^2 - t^2)*P8(x^2)/Q8(x^2)}
    {is used, where r, s, t are the first three zeros of J1.}
    y := x*(z - jz1)*(z - jz2)*(z - jz3);
    sfc_j1 := y*PolEvalX(z,j1n,9)/PolEvalX(z,j1d,9);
  end
  else begin
    {For x>=9 the common rational approximations to modulus}
    {and phase are used J1(x) = modulus * cos(phase).}
    if z >= 500.0 then begin
      y := bessj_large(1,z);
      if x<0.0 then sfc_j1 := -y else sfc_j1 := y;
    end
    else begin
      bess_m1p1(x,y,z);
      sfc_j1 := y*cos_x(z);
    end;
  end;
end;


function bessel_j0: extended;
var
x: extended;
  {-Return J0(x), the Bessel function of the 1st kind, order zero}
begin
asm
  fstp tbyte ptr [x]
end;
  bessel_j0 := sfc_j0(x);
end;


{---------------------------------------------------------------------------}
function bessel_j1: extended;
var
x: extended;
  {-Return J1(x), the Bessel function of the 1st kind, order one}
begin
asm
  fstp tbyte ptr [x]
end;
  bessel_j1 := sfc_j1(x);
end;


{---------------------------------------------------------------------------}
function bessel_jn: extended;
  {-Return J_n(x), the Bessel function of the 1st kind, order n; not suitable for large n or x.}
var
  n: integer;
  x: extended;
  MEM: Cardinal;
begin
asm
 fistp [n]
 fstp  tbyte ptr [x]
 //mov  n,eax
end;

  bessel_jn := sfc_jn(n,x);
end;



function gamma: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  gamma := sfc_gamma(x);
end;


function psi: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  psi := sfc_psi(x);
end;


function psid1: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  psid1 := sfc_trigamma(x);
end;


function psid2: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  psid2 := sfc_tetragamma(x);
end;


function psid3: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  psid3 := sfc_pentagamma(x);
end;


function zeta: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  zeta := sfc_zeta(x);
end;





function lanczos(x: extended; egscale: boolean): extended;
  {-Return the Lanczos sum for x, exp(g) scaled if egscale=true}
const
  lsh:  array[0..12] of THexExtW = (
          ($6F92,$D992,$7F81,$A01D,$402C),  {+4.4012138428004608955E+13}
          ($CD54,$4004,$20CB,$974E,$402C),  {+4.1590453358593200516E+13}
          ($F967,$596F,$659A,$8311,$402B),  {+1.8013842787117996778E+13}
          ($71D8,$A66C,$D4A5,$899F,$4029),  {+4.7287362634753888969E+12}
          ($96F3,$2C67,$5538,$C317,$4026),  {+8.3791008362840464705E+11}
          ($817F,$4B70,$3C24,$C4AA,$4023),  {+1.0558370727342993449E+11}
          ($125B,$9FAE,$C3E8,$908F,$4020),  {+9.7013636184949994935E+9}
          ($B883,$315C,$CD76,$9C24,$401C),  {+6.5491439754820526409E+8}
          ($29BA,$97D5,$7978,$F5F5,$4017),  {+3.2238322942133565306E+7}
          ($EA59,$87B0,$11C1,$89C2,$4013),  {+1.1285142194970914380E+6}
          ($0534,$EF56,$966A,$D053,$400D),  {+2.6665793784598589447E+4}
          ($8931,$781E,$A7EE,$BEF0,$4007),  {+3.8188012486329268705E+2}
          ($2CB3,$B138,$98FF,$A06C,$4000)); {+2.5066282746310005025}
  lseh: array[0..12] of THexExtW = (
          ($EBF0,$180B,$E131,$A434,$4019),  {+8.6091529534185372176E+7}
          ($A3D3,$B6ED,$E125,$9B2B,$4019),  {+8.1354505178580112428E+7}
          ($78B7,$D786,$C498,$866A,$4018),  {+3.5236626388154619108E+7}
          ($F7DC,$EF2B,$16FC,$8D24,$4016),  {+9.2498149880244712949E+6}
          ($8D43,$C678,$81BB,$C813,$4013),  {+1.6390242166871469602E+6}
          ($D9B5,$7AB8,$B435,$C9B0,$4010),  {+2.0653081576412250327E+5}
          ($A231,$1063,$6764,$9441,$400D),  {+1.8976701935302889156E+4}
          ($059E,$8F9A,$3482,$A022,$4009),  {+1.2810689099125594799E+3}
          ($35CD,$8C87,$6555,$FC3E,$4004),  {+6.3060933434202345361E+1}
          ($950B,$1B89,$3411,$8D47,$4000),  {+2.2074709097925276381}
          ($3037,$2E58,$56F1,$D5A6,$3FFA),  {+5.2160586946135054274E-2}
          ($653E,$8220,$AD06,$C3D1,$3FF4),  {+7.4699038089154483163E-4}
          ($6259,$720D,$001B,$A486,$3FED)); {+4.9031805734598718626E-6}
type
  TX12 = array[0..12] of extended;

  function lratev((*{$ifdef CONST}const{$else}var{$endif}*)const num: TX12; z: extended): extended;
  const
    denom: TX12 = ( 0.0,
                    39916800.0,
                    120543840.0,
                    150917976.0,
                    105258076.0,
                    45995730.0,
                    13339535.0,
                    2637558.0,
                    357423.0,
                    32670.0,
                    1925.0,
                    66.0,
                    1.0);
  var
    s1,s2: extended;
    i: integer;
  begin
    if abs(z)<=1.0 then begin
      s1 := num[12];
      s2 := denom[12];
      for i:=11 downto 0 do begin
        s1 := s1*z + num[i];
        s2 := s2*z + denom[i];
      end;
    end
    else begin
      z  := 1.0/z;
      s1 := num[0];
      s2 := denom[0];
      for i:=1 to 12 do begin
        s1 := s1*z + num[i];
        s2 := s2*z + denom[i];
      end;
    end;
    lratev := s1/s2;
  end;

begin
  {Ref: Boost [19], lanczos.hpp, struct lanczos13}
  if egscale then lanczos := lratev(TX12(lseh), x)
  else lanczos := lratev(TX12(lsh), x);
end;




function sfc_gdr_pos(x,d: extended): extended;
  {-Return gamma(x)/gamma(x+d), x>0, x+d > 0, accurate even for |d| << |x|}
var
  xgh,res: extended;
begin
  {$ifdef debug}
   {.248}
  (*
    if (x <= 0.0) and (x+d <= 0.0) then begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_gdr_pos := NaN_x;
      exit;
    end;
    *)
  {$endif}
  xgh := x + lanczos_gm05;
  if abs(d) < 10.0 then begin
    res := ln1p(d/xgh);
    res := _exp_x((0.5-x)*res);
  end
  else begin
    res := _power_x(xgh/(xgh+d), x-0.5);
  end;
  res := res*_power_x(_exp_x(1.0)/(xgh + d), d);
  sfc_gdr_pos := lanczos(x,false)/lanczos(x+d,false)*res;
end;



function sfc_gamma_delta_ratio(x,d: extended): extended;
  {-Return gamma(x)/gamma(x+d), accurate even for |d| << |x|}
var
  s,t,y: extended;
begin

  if IsNanOrInf(x) or IsNanOrInf(d) then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_gamma_delta_ratio := NaN_x;
    exit;
  end;

  y := x+d;
  if (x>0.0) and (y>0.0) then sfc_gamma_delta_ratio := sfc_gdr_pos(x,d)
  else begin
    if (y<=0.0) and (frac(y)=0.0) then begin
      {y=x+d is a non-positive integer}
      if (x>0.0) or (frac(x)<>0.0) then begin
        {gamma(x+d) = inf, gamma(x)<>0, return zero}
        sfc_gamma_delta_ratio := 0.0;
        exit;
      end;
      {both x,y non-positive integers, no need to calculate sin for reflection}
      t := sfc_gdr_pos(1.0-x,-d);
      if frac(0.5*x)<>frac(0.5*y) then t := -t;
      sfc_gamma_delta_ratio := 1.0/t;
      exit;
    end;
    if (x<0.0) and (y<0.0) then begin
      {Here both x and y are negative, use reflection formula for gammas}
      {gamma(x)/gamma(x+d) = sinPi(x+d)/sinP(x)/ [gamma(1-x)/gamma(1-x-d]}
      t := sfc_gdr_pos(1.0-x,-d);
      t := sinPi(x)*t;
      y := sinPi(y);
      sfc_gamma_delta_ratio := y/t;
    end
    else begin
      {Remaing cases: use lngamma for both arguments}
      s := sfc_signgamma(x)*sfc_signgamma(y);
      y := sfc_lngamma(y);
      t := sfc_lngamma(x);
      t := _exp_x(t-y);
      sfc_gamma_delta_ratio := s*t;
    end
  end;
end;


function sfc_gamma_ratio(x,y: extended): extended;
  {-Return gamma(x)/gamma(y)}
begin
  sfc_gamma_ratio := sfc_gamma_delta_ratio(x,y-x)
end;



function legendre_plmf(l,m: integer; x,f: extended): extended;
  {-Associated Legendre polynomial P_lm(x), f=|1-x^2|^(|m|/2) calculated externally}
var
  p0, p1, pk: extended;
  k: integer;
begin
  {Ref: HMF [1], ch.8; NR [13] ch.6.8; [19] legendre.hpp/legendre_imp}
  if l<0 then l := -l-1; {negative degree: HMF[1] 8.2.1}
  if m<0 then begin
    {Negative order: P(l,-m) = s*(l-m)!/(l+m)! P(l,m), for s see below}
    k := l+m;
    if k<0 then p1 := 0
    else p1 := legendre_plmf(l, -m, x, f);
    if p1<>0.0 then begin
      {calculate Gamma(l+m+1)/Gamma(l-m+1) via sfc_fac if possible}
      l := l-m;
      if l<MAXGAMX then p0 := factorial[k]/factorial[l] //sfc_fac(k)/sfc_fac(l)
      else p0 := sfc_gamma_ratio(k+1, l+1);
      {If |x| <= 1 there is an additional factor s=(-1)^m, but no factor}
      {if |x| > 1, see NIST [30], 14.9.3 vs 14.9.13}
      if odd(m) and (abs(x)<=1.0) then p1 := -p1*p0
      else p1 := p1*p0;
    end;
    legendre_plmf := p1;
  end
  else if m>l then legendre_plmf := 0.0
  else if m=0 then legendre_plmf := sfc_legendre_p(l,x)
  else begin
    {Starting value p0 = P_mm = (-1)^m * (2m-1)!! * f,  NR[13] 6.8.8}
    p0 := factorial2F[2*m-1]*f;//sfc_dfac(2*m-1)*f;
    if odd(m) and (x<=1.0) then p0 := -p0;
    if m=l then begin
      legendre_plmf := p0;
      exit;
    end;
    p1 := x*p0*(2*m + 1);
    k := succ(m);
    while k<l do begin
      {Recurrence relation for varying degree: HMF[1] 8.5.3}
      pk := ((2*k+1)*x*p1 - (k+m)*p0)/(k+1-m);
      p0 := p1;
      p1 := pk;
      inc(k);
    end;
    legendre_plmf := p1;
  end;
end;


function sfc_legendre_p(l: integer; x: extended): extended;   stdcall;
  {-Return P_l(x), the Legendre polynomial/function P_l, degree l}
var
  p0, p1, pk: extended;
  k: integer;
begin

  if IsNanOrInf(x) then
  begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_legendre_p := NaN_x;
    exit;
  end;

  if abs(x)=1.0 then begin
    if (x=-1.0) and odd(l) then sfc_legendre_p := -1.0
    else sfc_legendre_p := 1.0;
  end
  else begin
    if l<0 then l := -l-1;            {negative degree: HMF[1] 8.2.1}
    {here l >= 0}
    if l=0 then sfc_legendre_p := 1                    {HMF[1] 8.4.1}
    else if l=1 then sfc_legendre_p := x               {HMF[1] 8.4.3}
    else if l=2 then sfc_legendre_p := 1.5*sqr(x)-0.5  {HMF[1] 8.4.5}
    else begin
      if (x=0.0) and odd(l) then sfc_legendre_p := 0.0
      else begin
        p0 := 1.0;
        p1 := x;
        {Recurrence relation for varying degree: HMF[1] 8.5.3}
        for k:=2 to l do begin
          pk := ((2*k-1)*x*p1 - (k-1)*p0)/k;
          p0 := p1;
          p1 := pk;
        end;
        sfc_legendre_p := p1;
      end;
    end;
  end;
end;



function sfc_legendre_plm(l,m: integer; x: extended): extended;   stdcall;
  {-Return the associated Legendre polynomial P_lm(x)}
var
  f: extended;
begin

  if IsNaNorInf(x) then
   begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_legendre_plm := NaN_x;
    exit;
  end;

  if m=0 then sfc_legendre_plm := sfc_legendre_p(l,x)
  else begin
    f := (1.0-x)*(1.0+x);
    if f=0.0 then sfc_legendre_plm := 0.0
    else begin
      f := _power_x(abs(f), 0.5*abs(m));
      if x > -1.0 then sfc_legendre_plm := legendre_plmf(l,m,x,f)
      else begin
        {Switch to positive degree. Although legendre_plmf handles this case, }
        {it is already done here because it changes the odd/even status below.}
        if l<0 then l := -l-1;
        {x on the branch cut (-inf,-1), choose P_lm(x + i*0)}
        {and use Lebedev [45], 7.12.24 with integer nu = l}
        f := legendre_plmf(l,m,abs(x),f);
        if odd(l) then sfc_legendre_plm := -f
        else sfc_legendre_plm := f;
      end;
    end;
  end;
end;




{---------------------------------------------------------------------------}
function sfc_chebyshev_t(n: integer; x: extended): extended;  stdcall;
  {-Return Tn(x), the Chebyshev polynomial of the first kind, degree n}
var
  t0, t1, tk, z: extended;
  k: integer;
const
  NTR = 64;
begin
  {T(k+1,x) = 2x*T(k,x) - T(k-1,x);   T(0,x) = 1;   T(1,x) = x}
  {We use the trig/hyp form of Tn to generalize to negative degree: If n<0}
  {define T(-n,x) = T(n,x) because cos(nz) and cosh(nz) are even functions}
  n := abs(n);
  z := abs(x);
  if n=0 then sfc_chebyshev_t := 1.0
  else if n=1 then sfc_chebyshev_t := x
  else if z=1.0 then begin
    if odd(n) and (x<0.0) then sfc_chebyshev_t := -1.0
    else sfc_chebyshev_t := 1.0
  end
  else if n>NTR then begin
    {use trigonometric/hyperbolic functions if n is large}
    if z<1.0 then sfc_chebyshev_t:= cos(n*arccos(x))
    else begin
      {cannot use x < -1 for arccosh, use abs and adjust sign}
      t1 := cosh(n*arccosh(z));
      if odd(n) and (x<0.0) then sfc_chebyshev_t := -t1
      else sfc_chebyshev_t := t1;
    end;
  end
  else begin
    z  := 2.0*x;
    t0 := 1.0;
    t1 := x;
    for k:=2 to n do begin
      tk := z*t1 - t0;
      t0 := t1;
      t1 := tk;
    end;
    sfc_chebyshev_t := t1;
  end;
end;


{---------------------------------------------------------------------------}
function sfc_chebyshev_u(n: integer; x: extended): extended;  stdcall;
  {-Return Un(x), the Chebyshev polynomial of the second kind, degree n}
var
  u0, u1, uk, z: extended;
  k,n1: integer;
const
  NTR = 64;
begin
  {U(k+1,x) = 2x*U(k,x) - U(k-1,x);   U(0,x) = 1;   U(1,x) = 2x}
  if n<0 then begin
    {We use the trig/hyp form of U_n to generalize to negative degree: If n<0}
    {then sin((n+1)z) = sin((-|n|+1)z) = -sin((|n|-1)z) = -sin(((|n|-2)+1)z) }
    {and a similar argument applies for sinh. So define U(n,x) = -U(-n-2,x). }
    {The only special case is n=-1. Then we have -n-2 = -1 again. Here we can}
    {use sin((n+1)z) = sin(0z) = 0, and therefore U(-1,x) = 0 for all x.}
    if n=-1 then sfc_chebyshev_u := 0.0
    else sfc_chebyshev_u := -sfc_chebyshev_u(-n-2,x);
    exit;
  end;
  n1 := n+1;
  z  := abs(x);
  if n=0 then sfc_chebyshev_u := 1.0
  else if n=1 then sfc_chebyshev_u := 2.0*x
  else if z=1.0 then begin
    if odd(n) and (x<0.0) then sfc_chebyshev_u := -n1
    else sfc_chebyshev_u := n1
  end
  else if n>NTR then begin
    {Use trigonometric/hyperbolic functions if n is large,}
    {calculate U(n, abs(x)) and adjust sign}
    if z<1.0 then begin
      z  := arccos(z);
      u0 := sin(z);
      if u0=0.0 then u1 := n1 {u0=0 -> z=0 or Pi -> x ~ 1 or -1}
      else u1 := sin(n1*z)/u0;
    end
    else begin
      z  := arccosh(z);
      u0 := sinh(z);
      if u0=0.0 then u1 := n1  {u0=0 -> arccosh(z)=0 -> x ~ +-1}
      else u1 := sinh(n1*z)/u0;
    end;
    if odd(n) and (x<0.0) then sfc_chebyshev_u := -u1
    else sfc_chebyshev_u := u1;
  end
  else begin
    z  := 2.0*x;
    u0 := 1.0;
    u1 := z;
    for k:=2 to n do begin
      uk := z*u1 - u0;
      u0 := u1;
      u1 := uk;
    end;
    sfc_chebyshev_u := u1;
  end;
end;



{---------------------------------------------------------------------------}
function sfc_hermite_h(n: integer; x: extended): extended;
  {-Return Hn(x), the nth Hermite polynomial, degree n >= 0}
var
  h0,h1,hk: extended;
  k: integer;
begin

  h0 := 1.0;
  h1 := 2.0*x;

  {Recurrence relation, HMF[1] 22.7.13:   H(0,x) = 1, H(1,x) = 2*x}
  {H(k,x) = 2*x*H(k-1,x) - 2*(k-1)*H(k-2,x), for k>1.}
  if n<=1 then begin
    if n=1 then sfc_hermite_h := h1
    else if n=0 then sfc_hermite_h := h0
    else sfc_hermite_h := 0.0;
    exit;
  end;

  for k:=2 to n do begin
    hk := x*h1 - pred(k)*h0;
    h0 := h1;
    h1 := 2.0*hk;
  end;
  sfc_hermite_h := h1;
end;




function sfc_laguerre(n: integer; x: extended): extended;
  {-Return Ln(a,x), the nth generalized Laguerre polynomial with parameter a;}
  { degree n must be >= 0. x >=0 and a > -1 are the standard ranges.}
var
  l0,l1,lk,a1,ax,a: extended;
  k: integer;
begin

a:=0;

  {Recurrence relation, HMF[1] 22.7.12:  L(0,a,x) = 1;  L(1,a,x) = -x+1+a}
  {L(k,a,x) = (2*k+a-1-x)*L(k-1,a,x)/k - (k+a-1)*L(k-2,a,x)/k  for  k > 1}

  if n<=1 then begin
    if n=1 then sfc_laguerre := a+1.0-x
    else if n=0 then sfc_laguerre := 1.0
    else sfc_laguerre := 0.0;
    exit;
  end;

  if (x<0.0) and (a>-1.0) then begin
    {Calculate result via confluent hypergeometric function M.}
    { = binom(a+n,n) * M(-n,a+1,x); see HMF[1] 22.5.54, 13.1.2}
    a1 := a+1.0;
    ax := 1.0;  {binom}
    lk := 1.0;  {kth term of M}
    l1 := lk;   {sum M}
    for k:=0 to n-1 do begin
      {here (k-n)*x > 0, and lk > 0}
      lk := (k-n)/(a1+k)*x/(k+1)*lk;
      l1 := l1 + lk;
      ax := ax*(a1+k)/(k+1);
    end;
    l1 := l1*ax;
  end
  else begin
    {Calculate via standard recurrence relation}
    a1 := a-1.0;
    ax := a1-x;
    l0 := 1.0;
    l1 := a+1.0-x;
    for k:=2 to n do begin
      lk := (2*k+ax)*l1 - (k+a1)*l0;
      l0 := l1;
      l1 := lk/k;
    end;
  end;
  sfc_laguerre := l1;
end;



function sfc_laguerre_a(n: integer; a,x: extended): extended;
  {-Return Ln(a,x), the nth generalized Laguerre polynomial with parameter a;}
  { degree n must be >= 0. x >=0 and a > -1 are the standard ranges.}
var
  l0,l1,lk,a1,ax: extended;
  k: integer;
begin

  {Recurrence relation, HMF[1] 22.7.12:  L(0,a,x) = 1;  L(1,a,x) = -x+1+a}
  {L(k,a,x) = (2*k+a-1-x)*L(k-1,a,x)/k - (k+a-1)*L(k-2,a,x)/k  for  k > 1}

  if n<=1 then begin
    if n=1 then sfc_laguerre_a := a+1.0-x
    else if n=0 then sfc_laguerre_a := 1.0
    else sfc_laguerre_a := 0.0;
    exit;
  end;

  if (x<0.0) and (a>-1.0) then begin
    {Calculate result via confluent hypergeometric function M.}
    { = binom(a+n,n) * M(-n,a+1,x); see HMF[1] 22.5.54, 13.1.2}
    a1 := a+1.0;
    ax := 1.0;  {binom}
    lk := 1.0;  {kth term of M}
    l1 := lk;   {sum M}
    for k:=0 to n-1 do begin
      {here (k-n)*x > 0, and lk > 0}
      lk := (k-n)/(a1+k)*x/(k+1)*lk;
      l1 := l1 + lk;
      ax := ax*(a1+k)/(k+1);
    end;
    l1 := l1*ax;
  end
  else begin
    {Calculate via standard recurrence relation}
    a1 := a-1.0;
    ax := a1-x;
    l0 := 1.0;
    l1 := a+1.0-x;
    for k:=2 to n do begin
      lk := (2*k+ax)*l1 - (k+a1)*l0;
      l0 := l1;
      l1 := lk/k;
    end;
  end;
  sfc_laguerre_a := l1;
end;



function bessel_y0: extended;
var
x: extended;
  {-Return Y0(x), the Bessel function of the 2 kind, order zero}
begin
asm
  fstp tbyte ptr [x]
end;
  bessel_y0 := sfc_y0(x);
end;


{---------------------------------------------------------------------------}
function bessel_y1: extended;
var
x: extended;
  {-Return Y1(x), the Bessel function of the 2 kind, order one}
begin
asm
  fstp tbyte ptr [x]
end;
  bessel_y1 := sfc_y1(x);
end;


function bessel_yn: extended;
  {-Return Y_n(x), the Bessel function of the 2 kind, order n; not suitable for large n or x.}
var
  n: integer;
  x: extended;
  MEM: Cardinal;
begin
asm
 fistp [n]
 fstp  tbyte ptr [x]
 //mov  n,eax
end;

  bessel_yn := sfc_yn(n,x);
end;





function bessy_large(v, x: extended): extended;
  {-Return Y_v(x) via modulus/phase asymptotic expansion, x large}
var
  mv,tv,st,ct,sx,cx: extended;
begin
  h1v_large(v,x,mv,tv);
  sincos_x(tv,st,ct);
  sincos_x(x,sx,cx);
  {Y_v := mv*sin(x+tv); sin(x+tv) = cos(x)sin(tv) + sin(x)cos(tv)}
  bessy_large := mv*(st*cx + ct*sx);
end;





function sfc_y0(x: extended): extended;
  {-Return Y0(x), the Bessel function of the 2nd kind, order zero; x>0}
var
  y, z: extended;
const
  {The first four roots of Y0, calculated with Maple and t_rcalc/xh}
  y1h: THexExtW = ($DE10,$C2B4,$9A6C,$FD4A,$4000);  {3.9576784193148578685}
  y2h: THexExtW = ($1734,$3908,$EE27,$E2C0,$4001);  {7.0860510603017726975}
  y3h: THexExtW = ($BC69,$23EA,$B9AD,$A38E,$4002);  {10.222345043496417019}
  y4h: THexExtW = ($4D04,$0F3A,$0E25,$D5C7,$4002);  {13.361097473872763478}
var
  y0z1: extended absolute y1h;
  y0z2: extended absolute y2h;
  y0z3: extended absolute y3h;
  y0z4: extended absolute y4h;
const
  y0nhex: array[0..7] of THexExtW = (
            ($5fbd,$0171,$135a,$8340,$c035),
            ($501f,$6264,$bdf4,$9d17,$4036),
            ($23c9,$6b29,$4244,$c4c9,$c032),
            ($b219,$37ba,$5142,$9f1f,$402d),
            ($3e3c,$b343,$46c9,$e45f,$c026),
            ($2fdd,$4b27,$ca98,$a1c3,$401f),
            ($2ec0,$7b95,$297f,$df70,$c016),
            ($126c,$20be,$647f,$f344,$400c));
  y0dhex: array[0..7] of THexExtW = (
            ($241a,$8f2b,$629a,$de4b,$4038),
            ($04d3,$a629,$d61d,$b410,$4032),
            ($6732,$8c1b,$c5ab,$9384,$402b),
            ($553b,$4dc8,$8695,$a0c3,$4023),
            ($97a4,$90fa,$a7e9,$801c,$401b),
            ($d938,$b6b2,$71d8,$98be,$4012),
            ($9057,$7f25,$59b7,$8219,$4009),
            ($0000,$0000,$0000,$8000,$3fff));
  y059nh: array[0..9] of THexExtW = (
            ($f90c,$3510,$0be9,$87e7,$c012),
            ($fd54,$b2fe,$0a23,$e37e,$c00f),
            ($8c07,$29e3,$11be,$9f10,$4012),
            ($49e2,$fb52,$02af,$be8a,$4010),
            ($e8fa,$4b44,$4a39,$dc5b,$400b),
            ($62e0,$c25b,$2cb3,$8f12,$c00b),
            ($d5a3,$f673,$4e59,$9a8c,$4005),
            ($5504,$035a,$59fa,$ca14,$4003),
            ($1207,$46ea,$c3db,$bc88,$bfff),
            ($992f,$ab45,$90b6,$c20b,$3ff9));
  y059dh: array[0..9] of THexExtW = (
            ($3b3b,$ea0b,$b8d1,$8bd7,$401d),
            ($ceb6,$3463,$5ddb,$d1b5,$401e),
            ($e26b,$76b9,$250a,$a7fb,$c01c),
            ($27ff,$ca92,$3d78,$cea1,$4019),
            ($ec8a,$4697,$ddde,$a742,$c016),
            ($e8b6,$d705,$da91,$d62c,$4012),
            ($a28c,$5563,$d19f,$c75e,$c00e),
            ($ad09,$8e6a,$a502,$8b0c,$400a),
            ($debf,$a468,$8a55,$f96b,$c004),
            ($0000,$0000,$0000,$8000,$3fff));
var
  y059n: array[0..9] of extended absolute y059nh;
  y059d: array[0..9] of extended absolute y059dh;
  y0n:   array[0..7] of extended absolute y0nhex;
  y0d:   array[0..7] of extended absolute y0dhex;
begin
  {Ref: Cephes [7], file ldouble\j0l.c}
  if x<=0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_y0 := NaN_x;
    exit;
  end;
  if x < 9.0 then begin
    z := sqr(x);
    if z < 20.25 then begin
      {In the interval [0,4.5) a rational approximation of the}
      {form Y0(x) = P7(x)/Q7(x) + 2/Pi*ln(x)*J0(x) is used.   }
      y := ln(x)*sfc_j0(x)/Pi_2;
      sfc_y0 := y + PolEvalX(z,y0n,8)/PolEvalX(z,y0d,8);
    end
    else begin
      {In the interval [4.5,9) a rational approximation of the}
      {form Y0(x) = (x - p)(x - q)(x - r)(x - s)P9(x)/Q9(x) is}
      {is used where p, q, r, s are first four zeros of Y0(x).}
      y := (x - y0z1)*(x - y0z2)*(x - y0z3)*(x - y0z4);
      sfc_y0 := y * PolEvalX(x,y059n,10)/PolEvalX(x,y059d,10);
    end;
  end
  else begin
    {For x>=9 the common rational approximations to modulus}
    {and phase are used Y0(x) = modulus * sin(phase).}
    if x >= 1600 then sfc_y0 := bessy_large(0,x)
    else begin
      bess_m0p0(x,y,z);
      sfc_y0 := y*sin_x(z);
    end;
  end;
end;



function sfc_y1(x: extended): extended;
  {-Return Y1(x), the Bessel function of the 2nd kind, order one; x>0}
var
  y, z: extended;
const
  {The first four roots of Y1, calculated with Maple and t_rcalc/xh}
  y1h: THexExtW = ($1721,$FF92,$F6A6,$8C9D,$4000); {2.1971413260310170351}
  y2h: THexExtW = ($73C0,$3D81,$F274,$ADBF,$4001); {5.4296810407941351328}
  y3h: THexExtW = ($A148,$0B4D,$3D73,$8989,$4002); {8.5960058683311689268}
  y4h: THexExtW = ($3022,$A190,$89C6,$BBFC,$4002); {11.749154830839881243}
var
  y1z1: extended absolute y1h;
  y1z2: extended absolute y2h;
  y1z3: extended absolute y3h;
  y1z4: extended absolute y4h;
const
  y1nhex: array[0..6] of THexExtW = (
            ($3a10,$0848,$5930,$9965,$c035),
            ($7f8b,$4757,$75bd,$a196,$4033),
            ($69fd,$1242,$f62d,$de75,$c02e),
            ($5633,$aa6b,$79e5,$e62c,$4028),
            ($7607,$a687,$af0a,$d892,$c021),
            ($53e4,$194c,$befa,$bd19,$4019),
            ($5b16,$f7f8,$0d7e,$fbbd,$c00f));
  y1dhex: array[0..7] of THexExtW = (
            ($7302,$b91b,$de7e,$c399,$4037),
            ($8c6a,$397e,$0963,$ad7a,$4031),
            ($aaf0,$342b,$d098,$9ca5,$402a),
            ($57e0,$1d92,$90a9,$bd99,$4022),
            ($0e86,$117b,$36d6,$a94a,$401a),
            ($298c,$29ef,$0630,$e482,$4011),
            ($dd1a,$3b8e,$ab73,$df28,$4008),
            ($0000,$0000,$0000,$8000,$3fff));
  y159nh: array[0..9] of THexExtW = (
            ($539b,$f305,$c3d8,$97f6,$4011),
            ($f62f,$d968,$8c66,$8d15,$c013),
            ($3811,$a3da,$413f,$dc24,$c013),
            ($cd43,$2f50,$1118,$d972,$c013),
            ($a33b,$8229,$1561,$f1fc,$c00f),
            ($b2bf,$4296,$65af,$a3d1,$400d),
            ($df40,$226b,$7e37,$c0d9,$400b),
            ($e917,$8486,$0ebd,$e6c3,$c008),
            ($fdf1,$41e5,$4beb,$ac44,$4004),
            ($b5e5,$bb42,$f667,$ae3f,$bffe));
  y159dh: array[0..10] of THexExtW = (
            ($a231,$6ab0,$7952,$cdb2,$4019),
            ($b3ad,$1c6d,$0f07,$8ba8,$c01d),
            ($8e0e,$e148,$5ab3,$ff44,$401e),
            ($a46a,$0273,$bc0f,$c358,$c01c),
            ($6de5,$b797,$ea1c,$e66b,$4019),
            ($e5e5,$4172,$8863,$b4a0,$c016),
            ($3c4f,$dc46,$b802,$e107,$4012),
            ($bed4,$3ad5,$2da1,$cc1d,$c00e),
            ($d0fe,$2487,$01c0,$8be3,$400a),
            ($1a6c,$1c93,$612a,$f742,$c004),
            ($0000,$0000,$0000,$8000,$3fff));
var
  y159n: array[0..9]  of extended absolute y159nh;
  y159d: array[0..10] of extended absolute y159dh;
  y1n:   array[0..6]  of extended absolute y1nhex;
  y1d:   array[0..7]  of extended absolute y1dhex;
begin
  {Ref: Cephes [7], file ldouble\j1l.c}
  if x<=0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_y1 := NaN_x;
    exit;
  end;
  if x < 9.0 then begin
    z := sqr(x);
    if z < 20.25 then begin
      {In the interval [0,4.5) a rational approximation of the form}
      {Y1(x) = x*P6(x)/Q7(x) + 2/Pi*(ln(x)*J1(x) - 1/x) is used.}
      y := (ln(x)*sfc_j1(x) - 1.0/x)/Pi_2;
      sfc_y1 := y + x*PolEvalX(z,y1n,7)/PolEvalX(z,y1d,8);
    end
    else begin
      {In the interval [4.5,9) a rational approximation of the form}
      {Y1(x) = (x - p)*(x - q)*(x - r)*(x - s)*P9(x)/Q10(x) is used}
      {where p, q, r, s are first four zeros of Y1(x).}
      y := (x - y1z1)*(x - y1z2)*(x - y1z3)*(x - y1z4);
      sfc_y1 := y * PolEvalX(x,y159n,10)/PolEvalX(x,y159d,11);
    end;
  end
  else begin
    {For x>=9 the common rational approximations to modulus}
    {and phase are used Y1(x) = modulus * sin(phase).}
    if x >= 1600 then sfc_y1 := bessy_large(1,x)
    else begin
      bess_m1p1(x,y,z);
      sfc_y1 := y*sin_x(z);
    end;
  end;
end;


function sfc_lnfac(n: longint): extended;
  {-Return ln(n!), INF if n<0}
begin
  if n>25 then sfc_lnfac := sfc_lngamma(n+1.0)
  else begin
    if n<0 then sfc_lnfac := PosInf_x
    else sfc_lnfac := ln(Factorial[n]);
  end;
end;


function sfc_yn(n: integer; x: extended): extended;
  {-Return Y_n(x), the Bessel function of the 2nd kind, order n, x>0, not suitable for large n or x}
var
  yn,yn1,t: extended;
  k: integer;
  neg: boolean;
const
  lnpi = 1.144729885849400174;
begin
  {Flag to negate result for |n|}
  neg := (n<0) and odd(n);
  n := abs(n);
  if x<=0.0 then begin
    {$ifopt R+}
      if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
    {$endif}
    sfc_yn := NaN_x;
    exit;
  end;
  if n=0 then yn := sfc_y0(x)
  else if n=1 then yn := sfc_y1(x)
  else if (n>MAXGAMX) and (x<=2.0) then yn := NegInf_x
  else begin
    if x<1.0 then begin
      {NIST[30] 10.7.4}
      t := sfc_lnfac(n-1) - n*ln(0.5*x) - lnpi;
      if t > ln_MaxExt then begin
        if neg then sfc_yn := PosInf_x else sfc_yn := NegInf_x;
        exit;
      end;
    end;
    {forward recurrence}
    yn1 := sfc_y0(x);
    yn  := sfc_y1(x);
    x := 0.5*x;
    for k:=1 to n-1 do begin
      t  := yn*k/x - yn1;
      yn1:= yn;
      yn := t;
    end;
  end;
  if neg then sfc_yn := -yn else sfc_yn := yn;
end;




{-----------------------BESSEL SPHERICAL---------------------------------------}  {shb}



function sph_bessel_jn: extended;
var
  n: integer;
  x: extended;
  MEM: Cardinal;
begin
asm
 fistp [n]
 fstp  tbyte ptr [x]
end;
  sph_bessel_jn := sfc_sph_jn(n,x);
end;


function maxx(x, y: extended): extended; {$ifdef HAS_INLINE} inline;{$endif}
  {-Return the maximum of two extendeds; x,y <> NAN}
begin
  if x>y then maxx := x
  else maxx := y;
end;

procedure sincosPi_x(x: extended; var s,c: extended);
  {-Return s=sin(Pi*x), c=cos(Pi*x); (s,c)=(0,1) for abs(x) >= 2^64}
var
  t,ss,cc: extended;
  n: integer;
begin
  n := rem_int2(x,t) and 3;
  t := Pi*t;
  sincos_x(t,ss,cc);
  case n of
      0: begin s:= ss; c:= cc; end;
      1: begin s:= cc; c:=-ss; end;
      2: begin s:=-ss; c:=-cc; end;
    else begin s:=-cc; c:= ss; end;
  end;
end;

function Jv_series(v,x: extended): extended;
  {-Power series for Bessel J_v(x), 0 <= v < MAXGAMX-1, |x|<1 or v > x^2/4}
var
  f,s,t: extended;
  n: integer;
begin
  f := 0.5*x;
  t := _power_x(f,v)/sfc_gamma(v+1.0);
  f := -f*f;
  s := t;
  n := 0;
  repeat
    inc(n);
    t := t*f/n/(v+n);
    s := s + t;
  until abs(t) <= 0.5*eps_x*abs(s);
  Jv_series := s;
end;


function Yv_series(v,x: extended): extended;
  {-Series for Yv for 'small' x < 1;  frac(v)<>0}
var
  h,k,v1,v2,t1,t2,xx,Yv: extended;
const
  hsmall: THexExtW = ($4153,$E4E4,$9836,$0A2F,$0000); {2.675476672E-4933 = ~ 1/(MaxExtended*Pi)}
begin

  {Use Yv(x) = ( Jv(v,x)cos(vx) - Jv(-v,x) )/sin(vx) }
  {and Gamma reflection to sum the two series for Jv }

  if v<MAXGAMX then h := _power_x(0.5*x, v)/sfc_gamma(v)
  else h := _exp_x(v*ln(0.5*x) - sfc_lngamma(v));

  if h <= extended(hsmall) then begin
    {1.0/(h*Pi) will overflow}
    Yv_series := NegInf_x;
    exit;
  end;

  sincosPi_x(v, v1, v2);
  t2 := 1.0/(h*Pi);
  t1 := h*(v2/(v*v1));

  Yv := t1-t2;
  xx := 0.25*x*x;
  v2 := v;
  v1 := v;
  k  := 0.0;
  repeat
    k  := k + 1.0;
    v1 := v1 + 1.0;
    v2 := v2 - 1.0;
    t1 := -t1*xx/(k*v1);
    t2 :=  t2*xx/(k*v2);
    h  := t1 - t2;
    Yv := Yv + h;
  until (abs(h)<eps_x*abs(Yv));
  Yv_series := Yv;
end;

 (*
{$ifdef CONST}
function CSEval_X(x: extended; const a: array of extended; n: integer): extended;
  {-Evaluate Chebyshev sum a[0]/2 + a[1]*T_1(x) +..+ a[n-1]*T_(n-1)(x) using Clenshaw algorithm}
{$else}
function CSEval_X(x: extended; var va{:array of extended}; n: integer): extended;
var
  a: TExtVector absolute va;
{$endif}
*)
function CSEval_X(x: extended; const a: array of extended; n: integer): extended;
var
  b0,b1,b2: extended;
  i: integer;
begin
  {$ifdef CONST}
    {$ifdef debug}
     {.248}
    {
      if n>high(a)+1 then begin
        writeln('CSEvalX:  n > high(a)+1, n = ',n, ' vs. ', high(a)+1);
        readln;
      end;
      }
    {$endif}
    if n>high(a)+1 then n := high(a)+1;
  {$endif}
  b2 := 0.0;
  b1 := 0.0;
  b0 := 0.0;
  x  := 2.0*x;
  for i:=n-1 downto 0 do begin
    b2 := b1;
    b1 := b0;
    b0 := x*b1 - b2 + a[i];
  end;
  CSEval_X := 0.5*(b0-b2);
end;


function expm1(x: extended): extended;
  {-Return exmp1(x)-1, accurate even for x near 0}
const
  nce=17;
  ceh: array[0..nce-1] of THexExtW = (     {chebyshev(((exp(x)-1)/x - 1)/x, x=-1..1, 0.1e-20);}
         ($6A89,$722B,$FAC5,$8577,$3FFF),  {+1.04272398606220957726049179176     }
         ($D006,$0AFA,$F911,$B131,$3FFC),  {+0.173042194047179631675883846985    }
         ($D2C9,$D68A,$A866,$B073,$3FF9),  {+0.215395249458237651324867090935e-1 }
         ($B98A,$635F,$18B1,$8CA8,$3FF6),  {+0.214624979834132263391667992385e-2 }
         ($6731,$7DD9,$FF11,$BAFB,$3FF2),  {+0.178322182449141937355350007488e-3 }
         ($4297,$88F5,$C8BC,$D52A,$3FEE),  {+0.127057507929428665687742064639e-4 }
         ($2C99,$3975,$4676,$D4B9,$3FEA),  {+0.792457652881593907264616203751e-6 }
         ($196B,$F18F,$09DF,$BCC1,$3FE6),  {+0.439477285666343943629044429651e-7 }
         ($A60E,$D37D,$5A0B,$96C6,$3FE2),  {+0.219406227546144888207254174079e-8 }
         ($F7F8,$EF82,$8D42,$DB05,$3FDD),  {+0.995995318266659554765505887521e-10}
         ($CB48,$CD67,$F8D6,$91D8,$3FD9),  {+0.414523660148535566445986193898e-11}
         ($9AFA,$57D1,$F72D,$B352,$3FD4),  {+0.159271781651030822298380021406e-12}
         ($560F,$44FD,$5097,$CCC2,$3FCF),  {+0.568320507930678710973086211864e-14}
         ($33CD,$BF09,$610D,$DA3C,$3FCA),  {+0.189289431283787337567205644620e-15}
         ($7E97,$7C7D,$7B0A,$DA14,$3FC5),  {+0.591107031096332196225712012488e-17}
         ($1F08,$8285,$97C9,$CD1E,$3FC0),  {+0.173742977663542892171117958472e-18}
         ($0AA6,$23C8,$CF63,$B638,$3FBB)); {+0.482337391484586252954369194368e-20}
var
  ces: array[0..nce-1] of extended absolute ceh;
  t: extended;
begin
  if abs(x)<=1.0 then begin
    t := CSEval_x(x, ces, nce);
    expm1 := x + x*x*t;
  end
  else expm1 := _exp_x(x)-1.0;
end;


function sfc_gamma1pm1(x: extended): extended;
  {-Return gamma(1+x)-1 with increased accuracy for x near 0}
begin
  if abs(x)<=eps_x then sfc_gamma1pm1 := -EulerGamma*x
  else if (x<-0.5) or (x>2.0) then sfc_gamma1pm1 := sfc_gamma(1.0+x)-1.0
  else if x>0.0 then sfc_gamma1pm1 := expm1(lngamma_small(x+1.0,x,x-1.0,true))
  else sfc_gamma1pm1 := expm1(lngamma_small(x+2.0,x+1.0,x,false) - ln1p(x));
end;



procedure CF1_j(v,x: extended; var fv: extended; var s: integer);
  {-Return J_(v+1)(x) / J_v(x), efficient only if |x| <= |v|}
var
  c,d,f,b,t,tiny,tol: extended;
  k: longint;

const
  MAXIT = longint(32000)*100;  {see note below}

begin
  s := 1;

  {Evaluate HMF [1], 9.1.73 using modified Lentz's method. s keeps track }
  {of sign changes in the denominator. Ref: NR [13] (6.7.2) and p. 244,  }
  {function bessjy and Boost [19] file bessel_jy.hpp, function CF1_jy.   }

  {Note that CF1_j needs about O(|x|) iterations if |x| > |v|. But unless}
  {there is a better implementation below the asymptotic range,  CF1_j is}
  {is used in the last resort branch of bessel_jy. If a better algorithm }
  {(like Olver/Temme uniform Airy type asymptotic expansion) is available}
  {the factor 100 in the MAXIT declaration should be removed.}

  tol  := 2.0*eps_x;
  tiny := Sqrt_MinExt;
  c := tiny;
  f := tiny;
  d := 0.0;
  for k:=1 to MAXIT do begin
    b := 2.0*(v + k)/x;
    c := b - 1.0/c;
    d := b - d;
    if c=0.0 then c := tiny;
    if d=0.0 then d := tiny;
    d := 1.0/d;
    t := c * d;
    f := f*t;
    if d<0 then s := -s;
    if abs(t-1.0) < tol then begin
      fv := -f;
      exit;
    end;
  end;
  {No convergence}
  if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));
  fv := -f;
end;






procedure temme_y(v,x: extended; var Y,Y1: extended);
  {-Calculate Y(v, x) and Y(v+1, x) by Temme's method for small |x|}
var
  k,g,h,p,q,f,c,d,s,s1,tol,a,e: extended;
  g1,g2,gp,gm,v2: extended;
const
  MAXIT = 30000;
begin

{$ifdef debug}
 {.248}
{
  if (abs(v) > 0.5) or (abs(x) > 2) then begin
    if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
  end;
  }
{$endif}

  {N.M.Temme [51], On the Numerical Evaluation of the Ordinary}
  {Bessel Function of the Second Kind, Section 2.}
  gp := sfc_gamma1pm1(v);
  gm := sfc_gamma1pm1(-v);
  a  := ln(0.5*x);
  s  := -a*v;

  if abs(v) < eps_x then begin
    e := 0.5*v*sqr(Pi);
    d := 1.0/Pi;
  end
  else begin
    e := 2.0*sqr(sinpi(0.5*v))/v;
    d := v/sinpi(v);
  end;

  if v=0.0 then g1 := -EulerGamma
  else g1 := 0.5*(gp-gm)/((1.0+gp)*(1.0+gm)*v);
  g2 := 0.5*(2.0+gp+gm)/((1.0+gp)*(1.0+gm));

  if abs(s) <= eps_x then f := 1.0 else f := sinh_x(s)/s;
  f := 2.0*(g1*cosh_x(s) - g2*a*f)*d;

  c := _power_x(0.5*x, v);
  p := d/(c*(1.0 + gm));
  q := d*c/(1.0 + gp);

  g := f + e*q;
  c := 1.0;
  s := c*g;
  s1:= c*p;

  v2 := v*v;
  d  := -0.25*x*x;

  {series summation}
  tol := 0.5*eps_x;
  {use extended k because otherwise k*k may overflow}
  k := 1.0;
  while k <= MAXIT do begin
    c := c*d/k;
    f := (k*f + p + q) / (k*k - v2);
    p := p/(k - v);
    q := q/(k + v);
    g := f + e*q;
    h := p - k*g;
    s := s  + c*g;
    s1:= s1 + c*h;
    if abs(c*g) < abs(s)*tol then begin
      Y  := -s;
      Y1 := -2.0*s1/x;
      exit;
    end;
    k := k + 1.0;
  end;
  {No convergence}
  if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));
  Y  := -s;
  Y1 := -2.0*s1/x;
end;


procedure CF2_jy(v,x: extended; var p,q: extended);
  {-Return the continued fraction p + i*q = (J' + iY') / (J + iY)}
var
  a,br,bi,cr,ci,dr,di,er,ei,fr,fi,t: extended;
  i: integer;
const
  MAXIT = 30000;
begin
  {Ref: Numerical Recipes [13], ch. 6.7, p.244, function bessjy}
  {Evaluate the continued fraction p + iq = (J' + iY') / (J + iY)}
  {NR [13] (6.7.3) using the (complex) modified Lentz's method.}

{$ifdef debug}
 {.248}
{
  if abs(x) < 1.0 then begin
    if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
  end;
  }
{$endif}

  t  := 1.0/x;
  a  := 0.25-v*v;
  fr := -0.5*t;
  fi := 1.0;
  br := 2.0*x;
  bi := 2.0;
  t  := a*t/(fr*fr+fi*fi);
  cr := br+t*fi;
  ci := bi+t*fr;
  t  := br*br+bi*bi;
  dr := +br/t;
  di := -bi/t;
  er := cr*dr-ci*di;
  ei := cr*di+ci*dr;
  t  := fr*er-fi*ei;
  fi := fr*ei+fi*er;
  fr := t;
  for i:=1 to MAXIT do begin
    a  := a + 2.0*i;
    bi := bi + 2.0;
    dr := a*dr+br;
    di := a*di+bi;
    if abs(dr)+abs(di) = 0.0 then dr := Sqrt_MinExt;
    t  := a/(cr*cr+ci*ci);
    cr := br+t*cr;
    ci := bi-t*ci;
    if abs(cr)+abs(ci) = 0.0 then cr := Sqrt_MinExt;
    t  := dr*dr+di*di;
    dr := +dr/t;
    di := -di/t;
    er := cr*dr-ci*di;
    ei := cr*di+ci*dr;
    t  := fr*er-fi*ei;
    fi := fr*ei+fi*er;
    fr := t;
    if abs(er-1.0)+abs(ei) < 8*eps_x then begin
      p := fr;
      q := fi;
      exit;
    end;
  end;
  {No convergence}
  if RTE_NoConvergence>0 then RunError(byte(RTE_NoConvergence));
  p := fr;
  q := fi;
end;



procedure bessel_jy(v,x: extended; BT: byte; var Jv,Yv: extended);
  {-Return J_v(x) and/or Y_v(x) depending on BT, x>0, |v| < MaxLongint, INF if overflow}
var
  n,k: longint;
  u, Ju, Yv1, Yu, Yu1, fv, fu: extended;
  w, p, q, g, curr, prev, next, init, t, x2: extended;
  reflect: boolean;
  s: integer;
const
  lnepsh = -44.3614195558364998;  {ln(0.5*eps_x)}

  {--------------------------------------------}
  function rec_overflow(a,b: extended): boolean;
    {-Test if a*b overflows, if yes set Yv and Jv = PosInf_x}
  begin
    if (abs(a) > 1.0) and (abs(b) >= MaxExtended/abs(a)) then begin
      rec_overflow := true;
      Jv := PosInf_x;
      Yv := PosInf_x;
    end
    else rec_overflow := false;
  end;

begin

  {Ref: Boost [19] file bessel_jy.hpp, function bessel_jy}
  {and NR [13], Ch.6.7, section Ordinary Bessel Functions}

  {For x < 0 the functions Jv and Yv are in general complex; and Yv}
  {is singular for x=0. |v| < MaxLongint is assumed, so we can work}
  {with longint in the recurrence iterative, but these routines are}
  {are not suitable for large v values anyway.}

  reflect := v < 0.0;
  if reflect then begin
    v  := -v;
    BT := BT_J + BT_Y;   {J and Y needed for reflection formula}
  end;

  x2 := 0.5*x;
  n  := round(v);
  u  := v-n;
  w  := 2.0/(Pi*x); {Wronskian}

  if x<=2.0 then begin
    if v>MAXGAMX then begin
      Yv := PosInf_x;
      if reflect then Jv := PosInf_x else Jv := 0.0;
      exit;
    end
    else begin
      {Check very 'small' x and v case with (near) overflow for Yv}
      if (x<1.0) and (u<>0.0) and (lnepsh > v*ln(0.25*sqr(x)/v)) then begin
        if BT and BT_J <> 0 then Jv := Jv_series(v, x);
        if BT and BT_Y <> 0 then Yv := Yv_series(v, x);
      end
      else begin
        temme_y(u, x, Yu,Yu1);
        if n=0 then begin
          Yv  := Yu;
          Yv1 := Yu1;
        end
        else begin
          prev := Yu;
          curr := Yu1;
          {skip last next calculation if J is not needed: it}
          {overflows in some cases like bessel_yv(1755.45,2)}
          {bessel_jv(1755.45,2) will produce an overflow!   }
          for k:=1 to n-1 do begin
            {forward recurrence for Y}
            t := (u+k)/x2;
            if rec_overflow(t,curr) then exit;
            next := t*curr - prev;
            prev := curr;
            curr := next;
          end;
          Yv  := curr;
          if BT and BT_J = 0 then Yv1 := 0.0  {keep some compilers quiet!}
          else Yv1 := ((u+n)/x2)*curr - prev;
        end;
        if BT and BT_J <> 0 then begin
          CF1_j(v, x, fv, s);
          Jv := w/(Yv*fv - Yv1);   {Wronskian relation}
        end;
      end;
    end;
  end
  else begin
    {calculate the lower limit t=t(v) for asymptotic range}
    if BT=BT_Y then t:= 1552
    else begin
      t := maxx(3.0, v*v)*121.0;
      if BT and BT_Y <> 0 then t := maxx(t,1552);
    end;
    if x>t then begin
      {Use asymptotic expansion of Hankel function H1v(x)}
      if BT and BT_J <> 0 then Jv := bessj_large(v, x);
      if BT and BT_Y <> 0 then begin
        Yu  := bessy_large(u, x);
        Yu1 := bessy_large(u + 1, x);
      end;
    end
    else begin
      CF1_j(v, x, fv, s);
      {tiny initial value to prevent overflow}
      init := Sqrt_MinExt;
      curr := s*init;
      prev := fv*curr;
      for k:=n downto 1 do begin
        {backward recurrence for J}
        t := (u+k)/x2;
        if rec_overflow(t,curr) then exit;
        next := t*curr - prev;
        prev := curr;
        curr := next;
      end;
      CF2_jy(u, x, p, q);
      fu := prev/curr;
      t  := u/x - fu;   {t = J'/J}
      g  := (p-t)/q;
      Ju := sqrt(w/(q + g*(p-t)));
      if curr<0.0 then Ju := -Ju;
      Jv := s*Ju*(init/curr); {normalization}
      Yu := g*Ju;
      Yu1:= Yu*(u/x - p - q/g);
    end;
    if BT and BT_Y <> 0 then begin
      prev := Yu;
      curr := Yu1;
      if n=0 then Yv := prev
      else begin
        for k:=1 to n-1 do begin
          {forward recurrence for Y}
          t := (u+k)/x2;
          if rec_overflow(t,curr) then exit;
          next := t*curr - prev;
          prev := curr;
          curr := next;
        end;
        Yv := curr;
      end;
    end;
  end;

  if reflect then begin
    {For negative v use reflection formula NR [13], 6.7.19}
    {J(-v,x) = cos(Pi*v)*J(v,x) - sin(Pi*v)*Y(v,x)}
    {Y(-v,x) = sin(Pi*v)*J(v,x) + cos(Pi*v)*Y(v,x)}
    sincosPi_x(v,p,q);
    t  := q*Jv - p*Yv;
    Yv := p*Jv + q*Yv;
    Jv := t;
  end;
end;



function sfc_jv(v, x: extended): extended;  stdcall;
  {-Return J_v(x), the Bessel function of the 1st kind, order v; not suitable for large v.}
var
  r,n: extended;
begin

  if IsNaNorInf(x) or IsNaNorInf(v) then begin
    sfc_jv := NaN_x;
    exit;
  end;

  n := int(v);
  if x<0.0 then begin
    if n=v then begin
      r := sfc_jv(v, -x);
      if frac(0.5*v)<>0 then r := -r;
      sfc_jv := r;
    end
    else begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_jv := NaN_x;
    end;
  end
  else if x=0.0 then begin
    if v=0.0 then sfc_jv := 1.0
    else if v>0.0 then sfc_jv := 0.0
    else begin
      {$ifopt R+}
        if RTE_ArgumentRange>0 then RunError(byte(RTE_ArgumentRange));
      {$endif}
      sfc_jv := NaN_x;
    end;
  end
  else begin
    {here x > 0}
    if n=v then begin
      {integer order}
      if abs(n)<200.0 then begin
        if x > maxx(3.0, v*v)*121.0 then begin
          r := bessj_large(abs(v),x);
          if frac(0.5*v)<0.0 then r := -r;
          sfc_jv := r;
        end
        else sfc_jv := sfc_jn(round(n),x);
        exit;
      end;
    end;
    {Here v no integer or |v| > 200}
    if ((v >= 0.0) and (v<MAXGAMX-1)) and ((x < 1.0) or (v > 0.25*x*x)) then begin
      sfc_jv := Jv_series(v, x);
    end
    else begin
      bessel_jy(v,x,BT_J,r,n);
      sfc_jv := r;
    end;
  end;
end;

function sinc(x: extended): extended;
  {-Return the cardinal sine sinc(x) = sin(x)/x}
begin
  {sin(x)/x = 1 - 1/6*x^2 + 1/120*x^4 - 1/5040*x^6 + 1/362880*x^8 + O(x^10)}
  x := abs(x);
  if x < 6.15e-4 then begin
    if x < 2.3e-10 then sinc := 1.0
    else begin
      x := sqr(x);
      if x < 2.3e-10 then sinc := 1.0 - x/6.0
      else sinc := 1.0 - (x/6.0 - sqr(x)/120.0)
    end;
  end
  else sinc := sin_x(x)/x;
end;

function sfc_sph_jn(n: integer; x: extended): extended;
  {-Return j_n(x), the spherical Bessel function of the 1st kind, order n}
var
  r,z: extended;
begin
  if IsNaNorInf(x) then begin
    sfc_sph_jn := NaN_x;
    exit;
  end;
  if n=0 then sfc_sph_jn := sinc(x)
  else begin
    if x=0.0 then sfc_sph_jn := 0.0
    else begin
      z := abs(x);
      r := sqrt(Pi_2/z)*sfc_jv(0.5+n,z);
      if (x<0.0) and odd(n) then r := -r;    {NIST 10.47.14}
      sfc_sph_jn := r;
    end;
  end;
end;


{-----------------------  ERF  --------------------------------------------------------}

function erf: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  erf := sfc_erf(x);
end;


function erfc: extended;
var
x: extended;
begin
asm
  fstp tbyte ptr [x]
end;
  erfc := sfc_erfc(x);
end;


function sfc_erf(x: extended): extended;
  {-Return the error function erf(x) = 2/sqrt(Pi)*integral((exp(-t^2), t=0..x)}
begin
  if IsInf(x) then sfc_erf := copysign(1.0,x)
  else if abs(x)<=1 then sfc_erf := erf_small(x)
  else if x > +x0e then sfc_erf := +1.0
  else if x < -x0e then sfc_erf := -1.0
  else sfc_erf := 1.0 - sfc_erfc(x);
end;


function erf_small(x: extended): extended;
  {-Return erf(x) for |x| <= 1}
const
  {erf(x) = x*T(x^2)/U(x^2),  0 <= x <= 1, Peak relative error 7.6e-23}
  THex: array[0..6] of THexExtW = (
          ($c446,$6bab,$0b2a,$86d0,$4013),  {1.104385395713178565288E6}
          ($7a56,$e45a,$a4bd,$975b,$4010),  {1.549905740900882313773E5}
          ($b954,$a987,$c60c,$bc83,$400e),  {4.825977363071025440855E4}
          ($6118,$6059,$9093,$a757,$400a),  {2.677472796799053019985E3}
          ($9517,$4e93,$540e,$8f97,$4007),  {2.871822526820825849235E2}
          ($3128,$c337,$3716,$ace5,$4001),  {5.402980370004774841217E0}
          ($fd7a,$3a1a,$705b,$e0c4,$3ffb)); {1.097496774521124996496E-1}
  UHex: array[0..6] of THexExtW = (
          ($71a7,$1cad,$012e,$eef3,$4012),  {9.787360737578177599571E5}
          ($9ad5,$1aef,$45b1,$e25e,$4011),  {4.636021778692893773576E5}
          ($481d,$445b,$c807,$c232,$400f),  {9.942956272177178491525E4}
          ($ffe8,$9cac,$3b84,$c2ac,$400c),  {1.245905812306219011252E4}
          ($71ac,$b12f,$21ca,$f2e2,$4008),  {9.715333124857259246107E2}
          ($3453,$1f8e,$f688,$b507,$4004),  {4.525777638142203713736E1}
          ($0000,$0000,$0000,$8000,$3fff)); {1.000000000000000000000E0}
var
  z: extended;
  T: array[0..6] of extended absolute THex;
  U: array[0..6] of extended absolute UHex;
begin
  {Ref: Cephes [7], file ldouble\ndtrl.c}
  if x=0.0 then erf_small := x
  else begin
    z := x*x;
    erf_small := x*PolEvalx(z,T,7)/PolEvalx(z,U,7);
  end;
end;

function sfc_erfc(x: extended): extended;
  {-Return the complementary error function erfc(x) = 1-erf(x)}
var
  p,q,a,z: extended;
begin
  {expx2 is used to suppress error amplification in computing exp(-x^2)}
  {Ref: Cephes [7], file ldouble\ndtrl.c}
  if IsInf(x) then begin
    sfc_erfc := 1.0-copysign(1.0,x);
    exit;
  end;
  a := abs(x);
  if a < 1.0 then sfc_erfc := 1.0-erf_small(x)
  else if x < -x0e then sfc_erfc := 2.0
  else if x > x1e then sfc_erfc := 0.0
  else begin
    {Here x is in one of the ranges -x0e..-1 or 1..x1e}
    {Compute accurate z = exp(-a^2)}
    z := expx2(-a);
    {Compute p/q := exp(a^2)*erfc(a)}
    sfc_erfc_iusc(a,p,q);
    z := (z*p)/q;
    if x>=0 then sfc_erfc := z
    else sfc_erfc := 2.0-z;
  end;
end;

procedure sfc_erfc_iusc(x: extended; var p,q: extended);
  {-(internal) unevaluated scaled erfc, p/q := exp(x^2)*erfc(x), 1<=x<=128}
const
  {erfc(x) = exp(-x^2)*P(1/x)/Q(1/x), 1/8<=1/x<=1, Peak relative error 5.8e-21}
  PHex: array[0..9] of THexExtW = (
          ($333b,$d9e6,$d404,$986f,$bfee),   {-9.085943037416544232472E-6}
          ($0144,$489e,$be68,$9c31,$4011),   {3.198859502299390825278E5}
          ($3b58,$3da2,$af02,$9780,$4015),   {4.964439504376477951135E6}
          ($5840,$554d,$37a3,$9239,$4018),   {3.833161455208142870198E7}
          ($4e13,$caee,$9e31,$b258,$401a),   {1.870095071120436715930E8}
          ($ada8,$356a,$4982,$94a6,$401c),   {6.234814405521647580919E8}
          ($b6d0,$c92b,$5417,$acb1,$401d),   {1.448651275892911637208E9}
          ($d025,$cfd5,$8494,$88d3,$401e),   {2.295563412811856278515E9}
          ($df23,$d843,$4032,$8881,$401e),   {2.290171954844785638925E9}
          ($4bf0,$9ad8,$7a03,$86c7,$401d));  {1.130609921802431462353E9}
  QHex: array[0..10] of THexExtW = (
          ($24fa,$96f6,$7153,$8a6c,$4012),   {5.669830829076399819566E5}
          ($3708,$33b1,$07fa,$8644,$4016),   {8.799239977351261077610E6}
          ($c4cb,$305a,$bf78,$8220,$4019),   {6.822450775590265689648E7}
          ($a75d,$436f,$30dd,$a027,$401b),   {3.358653716579278063988E8}
          ($c39d,$e415,$c43d,$87c0,$401d),   {1.138778654945478547049E9}
          ($4991,$cfda,$52f1,$a2a9,$401e),   {2.729005809811924550999E9}
          ($00e7,$7595,$cd06,$88bb,$401f),   {4.588018188918609726890E9}
          ($8eae,$8dad,$6eb4,$9aa2,$401f),   {5.188672873106859049556E9}
          ($f817,$9128,$c0f8,$d48b,$401e),   {3.565928696567031388910E9}
          ($0e43,$302d,$79ed,$86c7,$401d),   {1.130609910594093747762E9}
          ($0000,$0000,$0000,$8000,$3fff));  {1.000000000000000000000E0}
const
  {erfc(x) = exp(-x^2)*1/x*R(1/x^2)/S(1/x^2), 1/128<=1/x<1/8, Peak relative error 1.9e-21}
  RHex: array[0..4] of THexExtW = (
          ($22cf,$c711,$6c5b,$dcfb,$3ff9),   {2.697535671015506686136E-2}
          ($521d,$8527,$3435,$8dc2,$3ffe),   {5.537445669807799246891E-1}
          ($0615,$4b00,$575f,$dc7b,$4000),   {3.445028155383625172464E0}
          ($4761,$613e,$df6d,$e58e,$4001),   {7.173690522797138522298E0}
          ($260a,$ab95,$2fc7,$e7c4,$4000));  {3.621349282255624026891E0}
  SHex: array[0..5] of THexExtW = (
          ($f5af,$2fb2,$1e57,$c3d7,$3ffa),   {4.781257488046430019872E-2}
          ($3684,$3798,$b793,$80b0,$3fff),   {1.005392977603322982436E0}
          ($b611,$8f76,$f020,$d255,$4001),   {6.572990478128949439509E0}
          ($55d5,$d300,$e71e,$f564,$4002),   {1.533713447609627196926E1}
          ($5de6,$17d7,$54d6,$aba9,$4002),   {1.072884067182663823072E1}
          ($0000,$0000,$0000,$8000,$3fff));  {1.000000000000000000000E0}
var
  PP: array[0..9]  of extended absolute PHex;
  QQ: array[0..10] of extended absolute QHex;
  RR: array[0..4]  of extended absolute RHex;
  SS: array[0..5]  of extended absolute SHex;
var
  y: extended;
begin
  {Ref: Cephes [7], file ldouble\ndtrl.c}
  y := 1.0/x;
  if x<8.0 then begin
    p := PolEvalx(y,PP,10);
    q := PolEvalx(y,QQ,11);
  end
  else begin
    q := y*y;
    p := y*PolEvalx(q,RR,5);
    q := PolEvalx(q,SS,6);
  end;
end;

function IsInf(x: extended): boolean;  {$ifdef HAS_INLINE} inline;{$endif}
  {-Return true if x is +INF or -INF}
begin
  with TExtRec(x) do begin
    IsInf := (xp and $7FFF=$7FFF) and (hm=longint($80000000)) and (lm=0);
  end;
end;



function expx2(x: extended): extended;
  {-Return exp(x*|x|) with damped error amplification in computing exp of the product.}
  { Used for exp(x^2) = expx2(abs(x)) and exp(-x^2) = expx2(-abs(x))}
const
  MFAC = 32768.0;
  MINV = 3.0517578125e-5;
var
  u,u1,m,f: extended;
  neg: boolean;
begin
  if x >= 106.5669902282 then begin
    expx2 := PosInf_x;
    exit;
  end
  else if x <= -106.77 then begin
    expx2 := 0.0;
    exit;
  end;

  {Ref: Cephes [7], file ldouble\expx2l.c}
  neg := x<0;
  x := abs(x);
  if x <= 1.0 then begin
    if neg then u := -x*x else u := x*x;
    expx2 := exp(u);
  end
  else begin
    {Represent x as an exact multiple of MFAC plus a residual.}
    {MFAC is a power of 2 chosen so that exp(m * m) does not  }
    {overflow or underflow and so that |x - m| is small.      }
    m := MINV*floorx(MFAC*x + 0.5);
    f := x - m;

    {x^2 = m^2 + 2mf + f^2}
    u  := m*m;
    u1 := 2.0*m*f + f*f;

    if neg then begin
      u  := -u;
      u1 := -u1;
    end;

    if u+u1 > ln_MaxExt then expx2 := PosInf_x
    else begin
      {u is exact, u1 is small}
      expx2 := exp(u)*exp(u1);
    end;
  end;
end;

{-------------------------------------------------------------------------------}
end.
